@@ -18,13 +18,17 @@
  *   51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.              *
  ***************************************************************************/
 
+#include "config.h"
+
 #include <signal.h>
-#include <execinfo.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <libintl.h>
 
-#include "fcitx/fcitx.h"
+#if defined(ENABLE_BACKTRACE)
+#include <execinfo.h>
+#endif
+
 #include "fcitx/instance-internal.h"
 #include "fcitx/ime-internal.h"
 #include "fcitx/configfile.h"
@@ -38,6 +42,7 @@
 #endif
 
 extern FcitxInstance* instance;
+extern int selfpipe[2];
 
 void SetMyExceptionHandler(void)
 {
@@ -46,7 +51,6 @@ void SetMyExceptionHandler(void)
     for (signo = SIGHUP; signo < SIGUNUSED; signo++) {
         if (signo != SIGALRM
             && signo != SIGPIPE
-            && signo != SIGUSR1
             && signo != SIGUSR2
             && signo != SIGWINCH
         )
@@ -61,24 +65,16 @@ void OnException(int signo)
     if (signo == SIGCHLD)
         return;
 
-    FcitxLog(INFO, _("FCITX -- Get Signal No.: %d"), signo);
+    FcitxLog(INFO, "FCITX -- Get Signal No.: %d", signo);
 
-    if (signo != SIGSEGV && signo != SIGCONT) {
-        if (instance) {
-            FcitxInstanceLock(instance);
-            FcitxProfileSave(instance->profile);
-            FcitxInstanceSaveAllIM(instance);
-            FcitxInstanceUnlock(instance);
-        }
-    }
-
-    void *array[10];
+#if defined(ENABLE_BACKTRACE)
+    void *array[20];
 
     size_t size;
     char **strings = NULL;
     size_t i;
 
-    size = backtrace(array, 10);
+    size = backtrace(array, 20);
     strings = backtrace_symbols(array, size);
 
     if (strings) {
@@ -106,22 +102,29 @@ void OnException(int signo)
 
         free(strings);
     }
+#endif
 
     switch (signo) {
-
-    case SIGHUP:
+    case SIGKILL:
+        break;
+    case SIGABRT:
+    case SIGSEGV:
+        exit(1);
         break;
 
-    case SIGINT:
-
-    case SIGTERM:
-
-    case SIGPIPE:
-
-    case SIGSEGV:
-        exit(0);
-
     default:
+        {
+            if (!instance->initialized) {
+                exit(1);
+                break;
+            }
+
+            uint8_t sig = 0;
+            if (signo < 0xff)
+                sig = (uint8_t)(signo & 0xff);
+            write(selfpipe[1], &sig, 1);
+            signal(signo, OnException);
+        }
         break;
     }
 }
