@@ -78,6 +78,7 @@ struct _FcitxIMContextClass {
 
 /* functions prototype */
 static void     fcitx_im_context_class_init(FcitxIMContextClass   *klass);
+static void     fcitx_im_context_class_fini (FcitxIMContextClass *klass);
 static void     fcitx_im_context_init(FcitxIMContext        *im_context);
 static void     fcitx_im_context_finalize(GObject               *obj);
 static void     fcitx_im_context_set_client_window(GtkIMContext          *context,
@@ -163,11 +164,8 @@ _key_snooper_cb (GtkWidget   *widget,
                  GdkEventKey *event,
                  gpointer     user_data);
 
-static gboolean
-_get_boolean_env(const gchar *name,
-                 gboolean defval);
-
 static GType _fcitx_type_im_context = 0;
+static GtkIMContextClass *parent_class = NULL;
 
 static guint _signal_commit_id = 0;
 static guint _signal_preedit_changed_id = 0;
@@ -217,7 +215,7 @@ fcitx_im_context_register_type(GTypeModule *type_module)
         (GBaseInitFunc) NULL,
         (GBaseFinalizeFunc) NULL,
         (GClassInitFunc) fcitx_im_context_class_init,
-        (GClassFinalizeFunc) NULL,
+        (GClassFinalizeFunc) fcitx_im_context_class_fini,
         NULL, /* klass data */
         sizeof(FcitxIMContext),
         0,
@@ -268,6 +266,8 @@ fcitx_im_context_class_init(FcitxIMContextClass *klass)
     GtkIMContextClass *im_context_class = GTK_IM_CONTEXT_CLASS(klass);
     GObjectClass *gobject_class = G_OBJECT_CLASS(klass);
 
+    parent_class = (GtkIMContextClass *) g_type_class_peek_parent (klass);
+
     im_context_class->set_client_window = fcitx_im_context_set_client_window;
     im_context_class->filter_keypress = fcitx_im_context_filter_keypress;
     im_context_class->reset = fcitx_im_context_reset;
@@ -303,8 +303,8 @@ fcitx_im_context_class_init(FcitxIMContextClass *klass)
         g_signal_lookup("retrieve-surrounding", G_TYPE_FROM_CLASS(klass));
     g_assert(_signal_retrieve_surrounding_id != 0);
 
-    _use_key_snooper = !_get_boolean_env ("IBUS_DISABLE_SNOOPER", !(_ENABLE_SNOOPER))
-                       && !_get_boolean_env("FCITX_DISABLE_SNOOPER", !(_ENABLE_SNOOPER));
+    _use_key_snooper = !fcitx_utils_get_boolean_env ("IBUS_DISABLE_SNOOPER", !(_ENABLE_SNOOPER))
+                       && !fcitx_utils_get_boolean_env("FCITX_DISABLE_SNOOPER", !(_ENABLE_SNOOPER));
     /* env IBUS_DISABLE_SNOOPER does not exist */
     if (_use_key_snooper) {
         /* disable snooper if app is in _no_snooper_apps */
@@ -327,14 +327,24 @@ fcitx_im_context_class_init(FcitxIMContextClass *klass)
     }
 
     /* make ibus fix benefits us */
-    _use_sync_mode = _get_boolean_env("IBUS_ENABLE_SYNC_MODE", FALSE)
-                     || _get_boolean_env("FCITX_ENABLE_SYNC_MODE", FALSE);
+    _use_sync_mode = fcitx_utils_get_boolean_env("IBUS_ENABLE_SYNC_MODE", FALSE)
+                     || fcitx_utils_get_boolean_env("FCITX_ENABLE_SYNC_MODE", FALSE);
     /* always install snooper */
     if (_key_snooper_id == 0)
         _key_snooper_id = gtk_key_snooper_install (_key_snooper_cb, NULL);
 }
 
 
+static void
+fcitx_im_context_class_fini (FcitxIMContextClass *klass)
+{
+    if (_key_snooper_id != 0) {
+        gtk_key_snooper_remove (_key_snooper_id);
+        _key_snooper_id = 0;
+    }
+}
+
+
 static void
 fcitx_im_context_init(FcitxIMContext *context)
 {
@@ -399,26 +409,36 @@ fcitx_im_context_finalize(GObject *obj)
     FcitxLog(LOG_LEVEL, "fcitx_im_context_finalize");
     FcitxIMContext *context = FCITX_IM_CONTEXT(obj);
 
-    g_object_unref(context->client);
-    context->client = NULL;
+    fcitx_im_context_set_client_window(GTK_IM_CONTEXT(context), NULL);
+
+#ifndef g_signal_handlers_disconnect_by_data
+#define g_signal_handlers_disconnect_by_data(instance, data) \
+    g_signal_handlers_disconnect_matched ((instance), G_SIGNAL_MATCH_DATA, 0, 0, NULL, NULL, (data))
+#endif
+
+    if (context->client) {
+        g_signal_handlers_disconnect_by_data(context->client, context);
+        g_object_unref(context->client);
+        context->client = NULL;
+    }
 
     if (context->slave) {
+        g_signal_handlers_disconnect_by_data(context->slave, context);
         g_object_unref(context->slave);
         context->slave = NULL;
     }
 
-    if (context->preedit_string)
+    if (context->preedit_string) {
         g_free(context->preedit_string);
-    context->preedit_string = NULL;
-
-    if (context->attrlist)
-        pango_attr_list_unref(context->attrlist);
-    context->attrlist = NULL;
-
-    if (_key_snooper_id != 0) {
-        gtk_key_snooper_remove (_key_snooper_id);
-        _key_snooper_id = 0;
+        context->preedit_string = NULL;
     }
+
+    if (context->attrlist) {
+        pango_attr_list_unref(context->attrlist);
+        context->attrlist = NULL;
+    }
+
+    G_OBJECT_CLASS(parent_class)->finalize (obj);
 }
 
 
@@ -468,7 +488,7 @@ fcitx_im_context_filter_keypress(GtkIMContext *context,
 
             /* set_cursor_location_internal() will get origin from X server,
             * it blocks UI. So delay it to idle callback. */
-            g_idle_add_full(G_PRIORITY_DEFAULT_IDLE,
+            gdk_threads_add_idle_full(G_PRIORITY_DEFAULT_IDLE,
                             (GSourceFunc) _set_cursor_location_internal,
                             g_object_ref(fcitxcontext),
                             (GDestroyNotify) g_object_unref);
@@ -505,14 +525,18 @@ fcitx_im_context_filter_keypress(GtkIMContext *context,
             pks->context = fcitxcontext;
             pks->event = (GdkEventKey *)  gdk_event_copy((GdkEvent *) event);
 
-            fcitx_client_process_key(fcitxcontext->client,
-                                     _fcitx_im_context_process_key_cb,
-                                     pks,
-                                     event->keyval,
-                                     event->hardware_keycode,
-                                     event->state,
-                                     (event->type == GDK_KEY_PRESS) ? (FCITX_PRESS_KEY) : (FCITX_RELEASE_KEY),
-                                     event->time);
+            g_object_ref(fcitxcontext);
+            fcitx_client_process_key_async(fcitxcontext->client,
+                                           event->keyval,
+                                           event->hardware_keycode,
+                                           event->state,
+                                           (event->type == GDK_KEY_PRESS) ? (FCITX_PRESS_KEY) : (FCITX_RELEASE_KEY),
+                                           event->time,
+                                           -1,
+                                           NULL,
+                                           _fcitx_im_context_process_key_cb,
+                                           pks
+                                          );
             event->state |= FcitxKeyState_HandledMask;
             return TRUE;
         }
@@ -528,22 +552,16 @@ _fcitx_im_context_process_key_cb (GObject *source_object,
                                   gpointer user_data)
 {
     ProcessKeyStruct* pks = user_data;
+    FcitxIMContext* context = pks->context;
     GdkEventKey* event = pks->event;
-    GError* error = NULL;
-    int ret = -1;
-    GVariant* result = g_dbus_proxy_call_finish(G_DBUS_PROXY(source_object), res, &error);
-    if (error) {
-        g_error_free(error);
-    }
-    else if (result) {
-        g_variant_get(result, "(i)", &ret);
-    }
+    int ret = fcitx_client_process_key_finish(FCITX_CLIENT(source_object), res);
     if (ret <= 0) {
         event->state |= FcitxKeyState_IgnoredMask;
         gdk_event_put((GdkEvent *)event);
     }
     gdk_event_free((GdkEvent *)event);
     g_free(pks);
+    g_object_unref(context);
 }
 
 static void
@@ -588,6 +606,8 @@ _fcitx_im_context_update_formatted_preedit_cb(FcitxClient* im, GPtrArray* array,
             gboolean hasColor;
             GdkColor fg;
             GdkColor bg;
+            memset(&fg, 0, sizeof(GdkColor));
+            memset(&bg, 0, sizeof(GdkColor));
 
             if (context->client_window) {
                 GtkWidget *widget;
@@ -624,17 +644,21 @@ _fcitx_im_context_update_formatted_preedit_cb(FcitxClient* im, GPtrArray* array,
 
     gchar* str = g_string_free(gstr, FALSE);
 
-    context->preedit_string = g_strdup(str);
+    context->preedit_string = str;
     char* tempstr = g_strndup(str, cursor_pos);
     context->cursor_pos =  fcitx_utf8_strlen(tempstr);
     g_free(tempstr);
 
-    gboolean new_visible = false;
+    gboolean new_visible = FALSE;
 
-    if (context->preedit_string != NULL) {
-        if (strlen(context->preedit_string) != 0)
-            new_visible = true;
+    if (context->preedit_string != NULL && context->preedit_string[0] == 0) {
+        g_free(context->preedit_string);
+        context->preedit_string = NULL;
     }
+
+    if (context->preedit_string != NULL)
+        new_visible = TRUE;
+
     gboolean flag = new_visible != visible;
 
     if (new_visible) {
@@ -653,8 +677,6 @@ _fcitx_im_context_update_formatted_preedit_cb(FcitxClient* im, GPtrArray* array,
             /* do nothing */
         }
     }
-
-    g_signal_emit(context, _signal_preedit_changed_id, 0);
 }
 
 ///
@@ -685,7 +707,7 @@ fcitx_im_context_focus_in(GtkIMContext *context)
 
     /* set_cursor_location_internal() will get origin from X server,
      * it blocks UI. So delay it to idle callback. */
-    g_idle_add_full(G_PRIORITY_DEFAULT_IDLE,
+    gdk_threads_add_idle_full(G_PRIORITY_DEFAULT_IDLE,
                     (GSourceFunc) _set_cursor_location_internal,
                     g_object_ref(fcitxcontext),
                     (GDestroyNotify) g_object_unref);
@@ -1300,15 +1322,21 @@ void _fcitx_im_context_connect_cb(FcitxClient* im, void* user_data)
 {
     FcitxIMContext* context =  FCITX_IM_CONTEXT(user_data);
     _fcitx_im_context_set_capacity(context, TRUE);
-
+    if (context->has_focus && _focus_im_context == (GtkIMContext*) context && fcitx_client_is_valid(context->client))
+        fcitx_client_focus_in(context->client);
+    /* set_cursor_location_internal() will get origin from X server,
+     * it blocks UI. So delay it to idle callback. */
+    gdk_threads_add_idle_full(G_PRIORITY_DEFAULT_IDLE,
+                              (GSourceFunc) _set_cursor_location_internal,
+                              g_object_ref(context),
+                              (GDestroyNotify) g_object_unref);
 }
 
 
 static void
 _request_surrounding_text (FcitxIMContext *context)
 {
-    if (context &&
-            fcitx_client_is_valid(context->client)) {
+    if (context && fcitx_client_is_valid(context->client)) {
         gboolean return_value;
         FcitxLog(LOG_LEVEL, "requesting surrounding text");
         g_signal_emit (context, _signal_retrieve_surrounding_id, 0,
@@ -1336,6 +1364,9 @@ _key_snooper_cb (GtkWidget   *widget,
 
     FcitxIMContext *fcitxcontext = (FcitxIMContext *) _focus_im_context;
 
+    if (G_UNLIKELY(!_use_key_snooper))
+        return FALSE;
+
     if (fcitxcontext == NULL || !fcitxcontext->has_focus)
         return FALSE;
 
@@ -1350,7 +1381,15 @@ _key_snooper_cb (GtkWidget   *widget,
             break;
         }
 
+        /* according to RH#859879, something bad could happen here. */
+        g_object_add_weak_pointer ((GObject *) fcitxcontext,
+                                   (gpointer *) &fcitxcontext);
         _request_surrounding_text (fcitxcontext);
+        if (G_UNLIKELY(!fcitxcontext))
+            return FALSE;
+        else
+            g_object_remove_weak_pointer ((GObject *) fcitxcontext,
+                                          (gpointer *) &fcitxcontext);
         fcitxcontext->time = event->time;
 
         if (_use_sync_mode) {
@@ -1370,14 +1409,18 @@ _key_snooper_cb (GtkWidget   *widget,
             pks->context = fcitxcontext;
             pks->event = (GdkEventKey *)  gdk_event_copy((GdkEvent *) event);
 
-            fcitx_client_process_key(fcitxcontext->client,
-                                     _fcitx_im_context_process_key_cb,
-                                     pks,
-                                     event->keyval,
-                                     event->hardware_keycode,
-                                     event->state,
-                                     (event->type == GDK_KEY_PRESS) ? (FCITX_PRESS_KEY) : (FCITX_RELEASE_KEY),
-                                     event->time);
+            g_object_ref(fcitxcontext);
+            fcitx_client_process_key_async(fcitxcontext->client,
+                                           event->keyval,
+                                           event->hardware_keycode,
+                                           event->state,
+                                           (event->type == GDK_KEY_PRESS) ? (FCITX_PRESS_KEY) : (FCITX_RELEASE_KEY),
+                                           event->time,
+                                           -1,
+                                           NULL,
+                                           _fcitx_im_context_process_key_cb,
+                                           pks
+                                          );
             retval = TRUE;
         }
     } while(0);
@@ -1393,23 +1436,4 @@ _key_snooper_cb (GtkWidget   *widget,
     return retval;
 }
 
-static gboolean
-_get_boolean_env(const gchar *name,
-                 gboolean defval)
-{
-    const char *value = getenv(name);
-
-    if (value == NULL)
-        return defval;
-
-    if (strcmp(value, "") == 0 ||
-            strcmp(value, "0") == 0 ||
-            strcmp(value, "false") == 0 ||
-            strcmp(value, "False") == 0 ||
-            strcmp(value, "FALSE") == 0)
-        return FALSE;
-
-    return TRUE;
-}
-
 // kate: indent-mode cstyle; replace-tabs on;
