@@ -25,12 +25,13 @@
  *        again and again.
  */
 
-#include <config.h>
+#include "config.h"
 
 #include <stdlib.h>
 #include <string.h>
 #include <gtk/gtk.h>
 #include <gdk/gdkkeysyms.h>
+
 #include "fcitx/fcitx.h"
 #include "fcitximcontext.h"
 #include "fcitx-config/fcitx-config.h"
@@ -77,6 +78,7 @@ struct _FcitxIMContextClass {
 
 /* functions prototype */
 static void     fcitx_im_context_class_init(FcitxIMContextClass   *klass);
+static void     fcitx_im_context_class_fini (FcitxIMContextClass *klass);
 static void     fcitx_im_context_init(FcitxIMContext        *im_context);
 static void     fcitx_im_context_finalize(GObject               *obj);
 static void     fcitx_im_context_set_client_window(GtkIMContext          *context,
@@ -167,11 +169,8 @@ _key_snooper_cb (GtkWidget   *widget,
                  GdkEventKey *event,
                  gpointer     user_data);
 
-static gboolean
-_get_boolean_env(const gchar *name,
-                 gboolean defval);
-
 static GType _fcitx_type_im_context = 0;
+static GtkIMContextClass *parent_class = NULL;
 
 static guint _signal_commit_id = 0;
 static guint _signal_preedit_changed_id = 0;
@@ -221,7 +220,7 @@ fcitx_im_context_register_type(GTypeModule *type_module)
         (GBaseInitFunc) NULL,
         (GBaseFinalizeFunc) NULL,
         (GClassInitFunc) fcitx_im_context_class_init,
-        (GClassFinalizeFunc) NULL,
+        (GClassFinalizeFunc) fcitx_im_context_class_fini,
         NULL, /* klass data */
         sizeof(FcitxIMContext),
         0,
@@ -272,6 +271,8 @@ fcitx_im_context_class_init(FcitxIMContextClass *klass)
     GtkIMContextClass *im_context_class = GTK_IM_CONTEXT_CLASS(klass);
     GObjectClass *gobject_class = G_OBJECT_CLASS(klass);
 
+    parent_class = (GtkIMContextClass *) g_type_class_peek_parent (klass);
+
     im_context_class->set_client_window = fcitx_im_context_set_client_window;
     im_context_class->filter_keypress = fcitx_im_context_filter_keypress;
     im_context_class->reset = fcitx_im_context_reset;
@@ -307,9 +308,9 @@ fcitx_im_context_class_init(FcitxIMContextClass *klass)
         g_signal_lookup("retrieve-surrounding", G_TYPE_FROM_CLASS(klass));
     g_assert(_signal_retrieve_surrounding_id != 0);
 
-    _use_key_snooper = !_get_boolean_env ("IBUS_DISABLE_SNOOPER", !(_ENABLE_SNOOPER))
-                       && !_get_boolean_env("FCITX_DISABLE_SNOOPER", !(_ENABLE_SNOOPER));
-        /* env IBUS_DISABLE_SNOOPER does not exist */
+    _use_key_snooper = !fcitx_utils_get_boolean_env ("IBUS_DISABLE_SNOOPER", !(_ENABLE_SNOOPER))
+                       && !fcitx_utils_get_boolean_env("FCITX_DISABLE_SNOOPER", !(_ENABLE_SNOOPER));
+    /* env IBUS_DISABLE_SNOOPER does not exist */
     if (_use_key_snooper) {
         /* disable snooper if app is in _no_snooper_apps */
         const gchar * prgname = g_get_prgname ();
@@ -331,14 +332,24 @@ fcitx_im_context_class_init(FcitxIMContextClass *klass)
     }
 
     /* make ibus fix benefits us */
-    _use_sync_mode = _get_boolean_env("IBUS_ENABLE_SYNC_MODE", FALSE)
-                     || _get_boolean_env("FCITX_ENABLE_SYNC_MODE", FALSE);
+    _use_sync_mode = fcitx_utils_get_boolean_env("IBUS_ENABLE_SYNC_MODE", FALSE)
+                     || fcitx_utils_get_boolean_env("FCITX_ENABLE_SYNC_MODE", FALSE);
     /* always install snooper */
     if (_key_snooper_id == 0)
         _key_snooper_id = gtk_key_snooper_install (_key_snooper_cb, NULL);
 }
 
 
+static void
+fcitx_im_context_class_fini (FcitxIMContextClass *klass)
+{
+    if (_key_snooper_id != 0) {
+        gtk_key_snooper_remove (_key_snooper_id);
+        _key_snooper_id = 0;
+    }
+}
+
+
 static void
 fcitx_im_context_init(FcitxIMContext *context)
 {
@@ -404,18 +415,17 @@ fcitx_im_context_finalize(GObject *obj)
         context->slave = NULL;
     }
 
-    if (context->preedit_string)
+    if (context->preedit_string) {
         g_free(context->preedit_string);
-    context->preedit_string = NULL;
-
-    if (context->attrlist)
-        pango_attr_list_unref(context->attrlist);
-    context->attrlist = NULL;
-
-    if (_key_snooper_id != 0) {
-        gtk_key_snooper_remove (_key_snooper_id);
-        _key_snooper_id = 0;
+        context->preedit_string = NULL;
     }
+
+    if (context->attrlist) {
+        pango_attr_list_unref(context->attrlist);
+        context->attrlist = NULL;
+    }
+
+    G_OBJECT_CLASS(parent_class)->finalize (obj);
 }
 
 
@@ -449,6 +459,15 @@ fcitx_im_context_set_client_window(GtkIMContext          *context,
     set_ic_client_window(fcitxcontext, client_window);
 }
 
+static void
+process_key_data_free(gpointer p)
+{
+    ProcessKeyStruct* pks = p;
+    g_object_unref(pks->context);
+    gdk_event_free((GdkEvent *)pks->event);
+    g_free(p);
+}
+
 ///
 static gboolean
 fcitx_im_context_filter_keypress(GtkIMContext *context,
@@ -465,7 +484,7 @@ fcitx_im_context_filter_keypress(GtkIMContext *context,
 
             /* set_cursor_location_internal() will get origin from X server,
             * it blocks UI. So delay it to idle callback. */
-            g_idle_add_full(G_PRIORITY_DEFAULT_IDLE,
+            gdk_threads_add_idle_full(G_PRIORITY_DEFAULT_IDLE,
                             (GSourceFunc) _set_cursor_location_internal,
                             g_object_ref(fcitxcontext),
                             (GDestroyNotify) g_object_unref);
@@ -499,13 +518,13 @@ fcitx_im_context_filter_keypress(GtkIMContext *context,
             }
         } else {
             ProcessKeyStruct* pks = g_malloc0(sizeof(ProcessKeyStruct));
-            pks->context = fcitxcontext;
+            pks->context = g_object_ref(fcitxcontext);
             pks->event = (GdkEventKey *)  gdk_event_copy((GdkEvent *) event);
 
             FcitxIMClientProcessKey(fcitxcontext->client,
                                     _fcitx_im_context_process_key_cb,
                                     pks,
-                                    g_free,
+                                    process_key_data_free,
                                     event->keyval,
                                     event->hardware_keycode,
                                     event->state,
@@ -534,7 +553,6 @@ _fcitx_im_context_process_key_cb(DBusGProxy *proxy,
         event->state |= FcitxKeyState_IgnoredMask;
         gdk_event_put((GdkEvent *)event);
     }
-    gdk_event_free((GdkEvent *)event);
 }
 
 static void
@@ -584,8 +602,6 @@ _fcitx_im_context_update_preedit_cb(DBusGProxy* proxy, char* str, int cursor_pos
             /* do nothing */
         }
     }
-
-    g_signal_emit(context, _signal_preedit_changed_id, 0);
 }
 
 static void
@@ -630,6 +646,8 @@ _fcitx_im_context_update_formatted_preedit_cb(DBusGProxy* proxy, GPtrArray* arra
             gboolean hasColor;
             GdkColor fg;
             GdkColor bg;
+            memset(&fg, 0, sizeof(GdkColor));
+            memset(&bg, 0, sizeof(GdkColor));
 
             if (context->client_window) {
                 GtkWidget *widget;
@@ -666,17 +684,21 @@ _fcitx_im_context_update_formatted_preedit_cb(DBusGProxy* proxy, GPtrArray* arra
 
     gchar* str = g_string_free(gstr, FALSE);
 
-    context->preedit_string = g_strdup(str);
+    context->preedit_string = str;
     char* tempstr = g_strndup(str, cursor_pos);
     context->cursor_pos =  fcitx_utf8_strlen(tempstr);
     g_free(tempstr);
 
-    gboolean new_visible = false;
+    gboolean new_visible = FALSE;
 
-    if (context->preedit_string != NULL) {
-        if (strlen(context->preedit_string) != 0)
-            new_visible = true;
+    if (context->preedit_string != NULL && context->preedit_string[0] == 0) {
+        g_free(context->preedit_string);
+        context->preedit_string = NULL;
     }
+
+    if (context->preedit_string != NULL)
+        new_visible = TRUE;
+
     gboolean flag = new_visible != visible;
 
     if (new_visible) {
@@ -695,8 +717,6 @@ _fcitx_im_context_update_formatted_preedit_cb(DBusGProxy* proxy, GPtrArray* arra
             /* do nothing */
         }
     }
-
-    g_signal_emit(context, _signal_preedit_changed_id, 0);
 }
 
 ///
@@ -727,7 +747,7 @@ fcitx_im_context_focus_in(GtkIMContext *context)
 
     /* set_cursor_location_internal() will get origin from X server,
      * it blocks UI. So delay it to idle callback. */
-    g_idle_add_full(G_PRIORITY_DEFAULT_IDLE,
+    gdk_threads_add_idle_full(G_PRIORITY_DEFAULT_IDLE,
                     (GSourceFunc) _set_cursor_location_internal,
                     g_object_ref(fcitxcontext),
                     (GDestroyNotify) g_object_unref);
@@ -1357,6 +1377,17 @@ void _fcitx_im_context_connect_cb(FcitxIMClient* client, void* user_data)
                                    context,
                                    NULL);
         _fcitx_im_context_set_capacity(context, TRUE);
+
+        if (context->has_focus && _focus_im_context == GTK_IM_CONTEXT(context))
+            FcitxIMClientFocusIn(context->client);
+
+        /* set_cursor_location_internal() will get origin from X server,
+         * it blocks UI. So delay it to idle callback. */
+        gdk_threads_add_idle_full(G_PRIORITY_DEFAULT_IDLE,
+                                  (GSourceFunc) _set_cursor_location_internal,
+                                  g_object_ref(context),
+                                  (GDestroyNotify) g_object_unref);
+
     }
 
 }
@@ -1398,6 +1429,9 @@ _key_snooper_cb (GtkWidget   *widget,
 
     FcitxIMContext *fcitxcontext = (FcitxIMContext *) _focus_im_context;
 
+    if (G_UNLIKELY(!_use_key_snooper))
+        return FALSE;
+
     if (fcitxcontext == NULL || !fcitxcontext->has_focus)
         return FALSE;
 
@@ -1412,7 +1446,15 @@ _key_snooper_cb (GtkWidget   *widget,
             break;
         }
 
+        /* according to RH#859879, something bad could happen here. */
+        g_object_add_weak_pointer ((GObject *) fcitxcontext,
+                                   (gpointer *) &fcitxcontext);
         _request_surrounding_text (fcitxcontext);
+        if (G_UNLIKELY(!fcitxcontext))
+            return FALSE;
+        else
+            g_object_remove_weak_pointer ((GObject *) fcitxcontext,
+                                          (gpointer *) &fcitxcontext);
         fcitxcontext->time = event->time;
 
         if (_use_sync_mode) {
@@ -1429,13 +1471,13 @@ _key_snooper_cb (GtkWidget   *widget,
                 retval = TRUE;
         } else {
             ProcessKeyStruct* pks = g_malloc0(sizeof(ProcessKeyStruct));
-            pks->context = fcitxcontext;
+            pks->context = g_object_ref(fcitxcontext);
             pks->event = (GdkEventKey *)  gdk_event_copy((GdkEvent *) event);
 
             FcitxIMClientProcessKey(fcitxcontext->client,
                                     _fcitx_im_context_process_key_cb,
                                     pks,
-                                    g_free,
+                                    process_key_data_free,
                                     event->keyval,
                                     event->hardware_keycode,
                                     event->state,
@@ -1456,23 +1498,4 @@ _key_snooper_cb (GtkWidget   *widget,
     return retval;
 }
 
-static gboolean
-_get_boolean_env(const gchar *name,
-                 gboolean defval)
-{
-    const char *value = getenv(name);
-
-    if (value == NULL)
-        return defval;
-
-    if (strcmp(value, "") == 0 ||
-        strcmp(value, "0") == 0 ||
-        strcmp(value, "false") == 0 ||
-        strcmp(value, "False") == 0 ||
-        strcmp(value, "FALSE") == 0)
-        return FALSE;
-
-    return TRUE;
-}
-
 // kate: indent-mode cstyle; space-indent on; indent-width 0;
