@@ -33,6 +33,11 @@
 
 #define GetIPCIC(ic) ((FcitxIPCIC*) (ic)->privateic)
 
+typedef struct _FcitxIPCCreateICPriv {
+    DBusMessage* message;
+    DBusConnection* conn;
+} FcitxIPCCreateICPriv;
+
 typedef struct _FcitxIPCIC {
     int id;
     char path[32];
@@ -43,12 +48,15 @@ typedef struct _FcitxIPCIC {
     char* surroundingText;
     unsigned int anchor;
     unsigned int cursor;
+    boolean lastPreeditIsEmpty;
+    boolean isPriv;
 } FcitxIPCIC;
 
 typedef struct _FcitxIPCFrontend {
     int frontendid;
     int maxid;
-    DBusConnection* conn;
+    DBusConnection* _conn;
+    DBusConnection* _privconn;
     FcitxInstance* owner;
 } FcitxIPCFrontend;
 
@@ -94,9 +102,9 @@ static void IPCSetPropertyIMList(void* arg, DBusMessageIter* iter);
 static void IPCUpdateIMList(void* arg);
 static pid_t IPCGetPid(void* arg, FcitxInputContext* ic);
 
-static void FcitxDBusPropertyGet(FcitxIPCFrontend* ipc, DBusMessage* message);
-static void FcitxDBusPropertySet(FcitxIPCFrontend* ipc, DBusMessage* message);
-static void FcitxDBusPropertyGetAll(FcitxIPCFrontend* ipc, DBusMessage* message);
+static void FcitxDBusPropertyGet(FcitxIPCFrontend* ipc, DBusConnection* conn, DBusMessage* message);
+static void FcitxDBusPropertySet(FcitxIPCFrontend* ipc, DBusConnection* coon, DBusMessage* message);
+static void FcitxDBusPropertyGetAll(FcitxIPCFrontend* ipc, DBusConnection* conn, DBusMessage* message);
 
 const FcitxDBusPropertyTable propertTable[] = {
     { FCITX_IM_DBUS_INTERFACE, "IMList", "a(sssb)", IPCGetPropertyIMList, IPCSetPropertyIMList },
@@ -177,10 +185,25 @@ const char * im_introspection_xml =
     "    <method name=\"ConfigureAddon\">\n"
     "      <arg name=\"addon\" direction=\"in\" type=\"s\"/>\n"
     "    </method>\n"
+    "    <method name=\"ConfigureIM\">\n"
+    "      <arg name=\"im\" direction=\"in\" type=\"s\"/>\n"
+    "    </method>\n"
+    "    <method name=\"GetCurrentUI\">\n"
+    "      <arg name=\"addon\" direction=\"out\" type=\"s\"/>\n"
+    "    </method>\n"
     "    <method name=\"GetIMAddon\">\n"
     "      <arg name=\"im\" direction=\"in\" type=\"s\"/>\n"
     "      <arg name=\"addon\" direction=\"out\" type=\"s\"/>\n"
     "    </method>\n"
+    "    <method name=\"ActivateIM\">\n"
+    "    </method>\n"
+    "    <method name=\"InactivateIM\">\n"
+    "    </method>\n"
+    "    <method name=\"ToggleIM\">\n"
+    "    </method>\n"
+    "    <method name=\"GetCurrentState\">\n"
+    "      <arg name=\"state\" direction=\"out\" type=\"i\"/>\n"
+    "    </method>\n"
     "    <property access=\"readwrite\" type=\"a(sssb)\" name=\"IMList\">\n"
     "      <annotation name=\"org.freedesktop.DBus.Property.EmitsChangedSignal\" value=\"true\"/>"
     "    </property>\n"
@@ -275,8 +298,7 @@ const char * ic_introspection_xml =
     "  </interface>\n"
     "</node>\n";
 
-FCITX_EXPORT_API
-FcitxFrontend frontend = {
+FCITX_DEFINE_PLUGIN(fcitx_ipc, frontend, FcitxFrontend) = {
     IPCCreate,
     IPCDestroy,
     IPCCreateIC,
@@ -297,31 +319,30 @@ FcitxFrontend frontend = {
     IPCGetSurroundingText
 };
 
-FCITX_EXPORT_API
-int ABI_VERSION = FCITX_ABI_VERSION;
-
 void* IPCCreate(FcitxInstance* instance, int frontendid)
 {
     FcitxIPCFrontend* ipc = fcitx_utils_malloc0(sizeof(FcitxIPCFrontend));
     ipc->frontendid = frontendid;
     ipc->owner = instance;
 
-    FcitxModuleFunctionArg arg;
+    ipc->_conn = InvokeVaArgs(instance, FCITX_DBUS, GETCONNECTION);
+    ipc->_privconn = InvokeVaArgs(instance, FCITX_DBUS, GETPRIVCONNECTION);
 
-    ipc->conn = InvokeFunction(instance, FCITX_DBUS, GETCONNECTION, arg);
-    if (ipc->conn == NULL) {
+    if (ipc->_conn == NULL && ipc->_privconn == NULL) {
         FcitxLog(ERROR, "DBus Not initialized");
         free(ipc);
         return NULL;
     }
 
+
     DBusObjectPathVTable fcitxIPCVTable = {NULL, &IPCDBusEventHandler, NULL, NULL, NULL, NULL };
 
-    if (!dbus_connection_register_object_path(ipc->conn,  FCITX_IM_DBUS_PATH,
-            &fcitxIPCVTable, ipc)) {
-        FcitxLog(ERROR, "No memory");
-        free(ipc);
-        return NULL;
+    if (ipc->_conn) {
+        dbus_connection_register_object_path(ipc->_conn,  FCITX_IM_DBUS_PATH, &fcitxIPCVTable, ipc);
+    }
+
+    if (ipc->_privconn) {
+        dbus_connection_register_object_path(ipc->_privconn, FCITX_IM_DBUS_PATH, &fcitxIPCVTable, ipc);
     }
 
     FcitxIMEventHook hook;
@@ -340,8 +361,9 @@ boolean IPCDestroy(void* arg)
 void IPCCreateIC(void* arg, FcitxInputContext* context, void* priv)
 {
     FcitxIPCFrontend* ipc = (FcitxIPCFrontend*) arg;
-    FcitxIPCIC* ipcic = (FcitxIPCIC*) fcitx_utils_malloc0(sizeof(FcitxIPCIC));
-    DBusMessage* message = (DBusMessage*) priv;
+    FcitxIPCIC* ipcic = fcitx_utils_new(FcitxIPCIC);
+    FcitxIPCCreateICPriv* ipcpriv = (FcitxIPCCreateICPriv*) priv;
+    DBusMessage* message = ipcpriv->message;
     DBusMessage *reply = dbus_message_new_method_return(message);
     FcitxGlobalConfig* config = FcitxInstanceGetGlobalConfig(ipc->owner);
 
@@ -349,6 +371,8 @@ void IPCCreateIC(void* arg, FcitxInputContext* context, void* priv)
 
     ipcic->id = ipc->maxid;
     ipc->maxid ++;
+    ipcic->lastPreeditIsEmpty = false;
+    ipcic->isPriv = (ipcpriv->conn != ipc->_conn);
     sprintf(ipcic->path, FCITX_IC_DBUS_PATH, ipcic->id);
 
     uint32_t arg1, arg2, arg3, arg4;
@@ -427,16 +451,22 @@ void IPCCreateIC(void* arg, FcitxInputContext* context, void* priv)
                                  DBUS_TYPE_UINT32, &arg4,
                                  DBUS_TYPE_INVALID);
     }
-    dbus_connection_send(ipc->conn, reply, NULL);
+    dbus_connection_send(ipcpriv->conn, reply, NULL);
     dbus_message_unref(reply);
 
     DBusObjectPathVTable vtable = {NULL, &IPCICDBusEventHandler, NULL, NULL, NULL, NULL };
-    if (!dbus_connection_register_object_path(ipc->conn, ipcic->path, &vtable, ipc)) {
-        return ;
+    if (!ipcic->isPriv) {
+        if (ipc->_conn) {
+            dbus_connection_register_object_path(ipc->_conn, ipcic->path, &vtable, ipc);
+            dbus_connection_flush(ipc->_conn);
+        }
+    }
+    else {
+        if (ipc->_privconn) {
+            dbus_connection_register_object_path(ipc->_privconn, ipcic->path, &vtable, ipc);
+            dbus_connection_flush(ipc->_privconn);
+        }
     }
-
-    dbus_connection_flush(ipc->conn);
-
 }
 
 boolean IPCCheckIC(void* arg, FcitxInputContext* context, void* priv)
@@ -452,7 +482,14 @@ void IPCDestroyIC(void* arg, FcitxInputContext* context)
     FcitxIPCFrontend* ipc = (FcitxIPCFrontend*) arg;
     FcitxIPCIC* ipcic = GetIPCIC(context);
 
-    dbus_connection_unregister_object_path(ipc->conn, GetIPCIC(context)->path);
+    if (!ipcic->isPriv) {
+        if (ipc->_conn)
+            dbus_connection_unregister_object_path(ipc->_conn, GetIPCIC(context)->path);
+    }
+    else {
+        if (ipc->_privconn)
+            dbus_connection_unregister_object_path(ipc->_privconn, GetIPCIC(context)->path);
+    }
     if (ipcic->appname)
         free(ipcic->appname);
     if (ipcic->surroundingText)
@@ -461,57 +498,61 @@ void IPCDestroyIC(void* arg, FcitxInputContext* context)
     context->privateic = NULL;
 }
 
+void IPCSendSignal(FcitxIPCFrontend* ipc, FcitxIPCIC* ipcic, DBusMessage* msg)
+{
+    if (!ipcic || !ipcic->isPriv) {
+        if (ipc->_conn) {
+            dbus_uint32_t serial = 0; // unique number to associate replies with requests
+            dbus_connection_send(ipc->_conn, msg, &serial);
+            dbus_connection_flush(ipc->_conn);
+        }
+    }
+    if (!ipcic || ipcic->isPriv) {
+        if (ipc->_privconn) {
+            dbus_uint32_t serial = 0; // unique number to associate replies with requests
+            dbus_connection_send(ipc->_privconn, msg, &serial);
+            dbus_connection_flush(ipc->_privconn);
+        }
+    }
+    dbus_message_unref(msg);
+}
+
 void IPCEnableIM(void* arg, FcitxInputContext* ic)
 {
     FcitxIPCFrontend* ipc = (FcitxIPCFrontend*) arg;
-    dbus_uint32_t serial = 0; // unique number to associate replies with requests
     DBusMessage* msg = dbus_message_new_signal(GetIPCIC(ic)->path, // object name of the signal
                        FCITX_IC_DBUS_INTERFACE, // interface name of the signal
                        "EnableIM"); // name of the signal
 
-    if (!dbus_connection_send(ipc->conn, msg, &serial)) {
-        FcitxLog(DEBUG, "Out Of Memory!");
-    }
-    dbus_connection_flush(ipc->conn);
-    dbus_message_unref(msg);
+    IPCSendSignal(ipc, GetIPCIC(ic), msg);
 }
 
 void IPCCloseIM(void* arg, FcitxInputContext* ic)
 {
     FcitxIPCFrontend* ipc = (FcitxIPCFrontend*) arg;
-    dbus_uint32_t serial = 0; // unique number to associate replies with requests
     DBusMessage* msg = dbus_message_new_signal(GetIPCIC(ic)->path, // object name of the signal
                        FCITX_IC_DBUS_INTERFACE, // interface name of the signal
                        "CloseIM"); // name of the signal
-
-    if (!dbus_connection_send(ipc->conn, msg, &serial)) {
-        FcitxLog(DEBUG, "Out Of Memory!");
-    }
-    dbus_connection_flush(ipc->conn);
-    dbus_message_unref(msg);
+    IPCSendSignal(ipc, GetIPCIC(ic), msg);
 }
 
 void IPCCommitString(void* arg, FcitxInputContext* ic, const char* str)
 {
     FcitxIPCFrontend* ipc = (FcitxIPCFrontend*) arg;
-    dbus_uint32_t serial = 0; // unique number to associate replies with requests
+
+    if (!fcitx_utf8_check_string(str))
+        return;
     DBusMessage* msg = dbus_message_new_signal(GetIPCIC(ic)->path, // object name of the signal
                        FCITX_IC_DBUS_INTERFACE, // interface name of the signal
                        "CommitString"); // name of the signal
 
     dbus_message_append_args(msg, DBUS_TYPE_STRING, &str, DBUS_TYPE_INVALID);
-
-    if (!dbus_connection_send(ipc->conn, msg, &serial)) {
-        FcitxLog(DEBUG, "Out Of Memory!");
-    }
-    dbus_connection_flush(ipc->conn);
-    dbus_message_unref(msg);
+    IPCSendSignal(ipc, GetIPCIC(ic), msg);
 }
 
 void IPCForwardKey(void* arg, FcitxInputContext* ic, FcitxKeyEventType event, FcitxKeySym sym, unsigned int state)
 {
     FcitxIPCFrontend* ipc = (FcitxIPCFrontend*) arg;
-    dbus_uint32_t serial = 0; // unique number to associate replies with requests
     DBusMessage* msg = dbus_message_new_signal(GetIPCIC(ic)->path, // object name of the signal
                        FCITX_IC_DBUS_INTERFACE, // interface name of the signal
                        "ForwardKey"); // name of the signal
@@ -520,12 +561,7 @@ void IPCForwardKey(void* arg, FcitxInputContext* ic, FcitxKeyEventType event, Fc
     uint32_t keystate = (uint32_t) state;
     int32_t type = (int) event;
     dbus_message_append_args(msg, DBUS_TYPE_UINT32, &keyval, DBUS_TYPE_UINT32, &keystate, DBUS_TYPE_INT32, &type, DBUS_TYPE_INVALID);
-
-    if (!dbus_connection_send(ipc->conn, msg, &serial)) {
-        FcitxLog(DEBUG, "Out Of Memory!");
-    }
-    dbus_connection_flush(ipc->conn);
-    dbus_message_unref(msg);
+    IPCSendSignal(ipc, GetIPCIC(ic), msg);
 }
 
 void IPCSetWindowOffset(void* arg, FcitxInputContext* ic, int x, int y)
@@ -555,24 +591,25 @@ static DBusHandlerResult IPCDBusEventHandler(DBusConnection *connection, DBusMes
         dbus_message_unref(reply);
         return DBUS_HANDLER_RESULT_HANDLED;
     } else if (dbus_message_is_method_call(msg, DBUS_INTERFACE_PROPERTIES, "Get")) {
-        FcitxDBusPropertyGet(ipc, msg);
+        FcitxDBusPropertyGet(ipc, connection, msg);
         return DBUS_HANDLER_RESULT_HANDLED;
     } else if (dbus_message_is_method_call(msg, DBUS_INTERFACE_PROPERTIES, "Set")) {
-        FcitxDBusPropertySet(ipc, msg);
+        FcitxDBusPropertySet(ipc, connection, msg);
         return DBUS_HANDLER_RESULT_HANDLED;
     } else if (dbus_message_is_method_call(msg, DBUS_INTERFACE_PROPERTIES, "GetAll")) {
-        FcitxDBusPropertyGetAll(ipc, msg);
+        FcitxDBusPropertyGetAll(ipc, connection, msg);
         return DBUS_HANDLER_RESULT_HANDLED;
-    } else if (dbus_message_is_method_call(msg, FCITX_IM_DBUS_INTERFACE, "CreateIC")) {
-        FcitxInstanceCreateIC(ipc->owner, ipc->frontendid, msg);
-        return DBUS_HANDLER_RESULT_HANDLED;
-    } else if (dbus_message_is_method_call(msg, FCITX_IM_DBUS_INTERFACE, "CreateICv2")) {
-        FcitxInstanceCreateIC(ipc->owner, ipc->frontendid, msg);
-        return DBUS_HANDLER_RESULT_HANDLED;
-    } else if (dbus_message_is_method_call(msg, FCITX_IM_DBUS_INTERFACE, "CreateICv3")) {
-        FcitxInstanceCreateIC(ipc->owner, ipc->frontendid, msg);
+    } else if (dbus_message_is_method_call(msg, FCITX_IM_DBUS_INTERFACE, "CreateIC")
+            || dbus_message_is_method_call(msg, FCITX_IM_DBUS_INTERFACE, "CreateICv2")
+            || dbus_message_is_method_call(msg, FCITX_IM_DBUS_INTERFACE, "CreateICv3")
+            ) {
+        FcitxIPCCreateICPriv ipcpriv;
+        ipcpriv.message = msg;
+        ipcpriv.conn = connection;
+        FcitxInstanceCreateIC(ipc->owner, ipc->frontendid, &ipcpriv);
         return DBUS_HANDLER_RESULT_HANDLED;
     } else if (dbus_message_is_method_call(msg, FCITX_IM_DBUS_INTERFACE, "Exit")) {
+        FcitxLog(INFO, "Receive message ask for quit");
         DBusMessage *reply = dbus_message_new_method_return(msg);
         dbus_connection_send(connection, reply, NULL);
         dbus_message_unref(reply);
@@ -607,9 +644,9 @@ static DBusHandlerResult IPCDBusEventHandler(DBusConnection *connection, DBusMes
         dbus_error_init(&error);
         char* imname = NULL;
         if (dbus_message_get_args(msg, &error, DBUS_TYPE_STRING, &imname, DBUS_TYPE_INVALID)) {
-            FcitxIM* im = FcitxInstanceGetIMFromIMList(instance, IMAS_Enable, imname);
+            FcitxIM* im = FcitxInstanceGetIMFromIMList(instance, IMAS_Disable, imname);
             const char* name = "";
-            if (im)
+            if (im && im->owner)
                 name = im->owner->name;
             DBusMessage *reply = dbus_message_new_method_return(msg);
             dbus_message_append_args(reply, DBUS_TYPE_STRING, &name, DBUS_TYPE_INVALID);
@@ -618,12 +655,50 @@ static DBusHandlerResult IPCDBusEventHandler(DBusConnection *connection, DBusMes
         }
         dbus_error_free(&error);
         return DBUS_HANDLER_RESULT_HANDLED;
+    } else if (dbus_message_is_method_call(msg, FCITX_IM_DBUS_INTERFACE, "GetCurrentUI")) {
+        const char* name = "";
+        FcitxAddon* uiaddon  = FcitxInstanceGetCurrentUI(instance);
+        if (uiaddon) {
+            name = uiaddon->name;
+        }
+        DBusMessage *reply = dbus_message_new_method_return(msg);
+        dbus_message_append_args(reply, DBUS_TYPE_STRING, &name, DBUS_TYPE_INVALID);
+        dbus_connection_send(connection, reply, NULL);
+        dbus_message_unref(reply);
+        return DBUS_HANDLER_RESULT_HANDLED;
     } else if (dbus_message_is_method_call(msg, FCITX_IM_DBUS_INTERFACE, "Configure")) {
         DBusMessage *reply = dbus_message_new_method_return(msg);
         dbus_connection_send(connection, reply, NULL);
         dbus_message_unref(reply);
         fcitx_utils_launch_configure_tool();
         return DBUS_HANDLER_RESULT_HANDLED;
+    } else if (dbus_message_is_method_call(msg, FCITX_IM_DBUS_INTERFACE, "ActivateIM")) {
+        FcitxInstanceEnableIM(ipc->owner, FcitxInstanceGetCurrentIC(ipc->owner), false);
+        DBusMessage *reply = dbus_message_new_method_return(msg);
+        dbus_connection_send(connection, reply, NULL);
+        dbus_message_unref(reply);
+        return DBUS_HANDLER_RESULT_HANDLED;
+    } else if (dbus_message_is_method_call(msg, FCITX_IM_DBUS_INTERFACE, "InactivateIM")) {
+        FcitxInstanceCloseIM(ipc->owner, FcitxInstanceGetCurrentIC(ipc->owner));
+        DBusMessage *reply = dbus_message_new_method_return(msg);
+        dbus_connection_send(connection, reply, NULL);
+        dbus_message_unref(reply);
+        return DBUS_HANDLER_RESULT_HANDLED;
+    } else if (dbus_message_is_method_call(msg, FCITX_IM_DBUS_INTERFACE, "ToggleIM")) {
+        FcitxInstanceChangeIMState(ipc->owner, FcitxInstanceGetCurrentIC(ipc->owner));
+        DBusMessage *reply = dbus_message_new_method_return(msg);
+        dbus_connection_send(connection, reply, NULL);
+        dbus_message_unref(reply);
+        return DBUS_HANDLER_RESULT_HANDLED;
+    } else if (dbus_message_is_method_call(msg, FCITX_IM_DBUS_INTERFACE, "GetCurrentState")) {
+        int r = FcitxInstanceGetCurrentState(ipc->owner);
+        DBusMessage *reply = dbus_message_new_method_return(msg);
+        dbus_message_append_args(reply,
+                                 DBUS_TYPE_INT32, &r,
+                                 DBUS_TYPE_INVALID);
+        dbus_connection_send(connection, reply, NULL);
+        dbus_message_unref(reply);
+        return DBUS_HANDLER_RESULT_HANDLED;
     } else if (dbus_message_is_method_call(msg, FCITX_IM_DBUS_INTERFACE, "ConfigureAddon")) {
         DBusError error;
         dbus_error_init(&error);
@@ -638,6 +713,20 @@ static DBusHandlerResult IPCDBusEventHandler(DBusConnection *connection, DBusMes
         }
         dbus_error_free(&error);
         return DBUS_HANDLER_RESULT_HANDLED;
+    } else if (dbus_message_is_method_call(msg, FCITX_IM_DBUS_INTERFACE, "ConfigureIM")) {
+        DBusError error;
+        dbus_error_init(&error);
+        char* imname = NULL;
+        if (dbus_message_get_args(msg, &error, DBUS_TYPE_STRING, &imname, DBUS_TYPE_INVALID)) {
+            DBusMessage *reply = dbus_message_new_method_return(msg);
+            dbus_connection_send(connection, reply, NULL);
+            dbus_message_unref(reply);
+            if (imname) {
+                fcitx_utils_launch_configure_tool_for_addon(imname);
+            }
+        }
+        dbus_error_free(&error);
+        return DBUS_HANDLER_RESULT_HANDLED;
     } else if (dbus_message_is_method_call(msg, FCITX_IM_DBUS_INTERFACE, "ReloadConfig")) {
         DBusMessage *reply = dbus_message_new_method_return(msg);
         dbus_connection_send(connection, reply, NULL);
@@ -823,7 +912,7 @@ static int IPCProcessKey(FcitxIPCFrontend* ipc, FcitxInputContext* callic, const
     FcitxKeySym sym;
     unsigned int state;
 
-    state = originstate - (originstate & FcitxKeyState_NumLock) - (originstate & FcitxKeyState_CapsLock) - (originstate & FcitxKeyState_ScrollLock);
+    state = originstate & FcitxKeyState_SimpleMask;
     state &= FcitxKeyState_UsedMask;
     FcitxHotkeyGetKey(originsym, state, &sym, &state);
     FcitxLog(DEBUG,
@@ -930,16 +1019,28 @@ static void IPCICSetCursorRect(FcitxIPCFrontend* ipc, FcitxInputContext* ic, int
 
 void IPCUpdatePreedit(void* arg, FcitxInputContext* ic)
 {
+    FcitxIPCFrontend* ipc = (FcitxIPCFrontend*) arg;
+    FcitxInputState* input = FcitxInstanceGetInputState(ipc->owner);
+    FcitxMessages* clientPreedit = FcitxInputStateGetClientPreedit(input);
+    int i = 0;
+    for (i = 0; i < FcitxMessagesGetMessageCount(clientPreedit) ; i ++) {
+        char* str = FcitxMessagesGetMessageString(clientPreedit, i);
+        if (!fcitx_utf8_check_string(str))
+            return;
+    }
+
+    /* a small optimization, don't need to update empty preedit */
+    FcitxIPCIC* ipcic = GetIPCIC(ic);
+    if (ipcic->lastPreeditIsEmpty && FcitxMessagesGetMessageCount(clientPreedit) == 0)
+        return;
+
+    ipcic->lastPreeditIsEmpty = (FcitxMessagesGetMessageCount(clientPreedit) == 0);
+
     if (ic->contextCaps & CAPACITY_FORMATTED_PREEDIT) {
-        FcitxIPCFrontend* ipc = (FcitxIPCFrontend*) arg;
-        dbus_uint32_t serial = 0; // unique number to associate replies with requests
-        FcitxInputState* input = FcitxInstanceGetInputState(ipc->owner);
         DBusMessage* msg = dbus_message_new_signal(GetIPCIC(ic)->path, // object name of the signal
                         FCITX_IC_DBUS_INTERFACE, // interface name of the signal
                         "UpdateFormattedPreedit"); // name of the signal
 
-        FcitxMessages* clientPreedit = FcitxInputStateGetClientPreedit(input);
-
         DBusMessageIter args, array, sub;
         dbus_message_iter_init_append(msg, &args);
         dbus_message_iter_open_container(&args, DBUS_TYPE_ARRAY, "(si)", &array);
@@ -963,15 +1064,10 @@ void IPCUpdatePreedit(void* arg, FcitxInputContext* ic)
         int iCursorPos = FcitxInputStateGetClientCursorPos(input);
         dbus_message_iter_append_basic(&args, DBUS_TYPE_INT32, &iCursorPos);
 
-        if (!dbus_connection_send(ipc->conn, msg, &serial)) {
-            FcitxLog(DEBUG, "Out Of Memory!");
-        }
-        dbus_connection_flush(ipc->conn);
-        dbus_message_unref(msg);
+        IPCSendSignal(ipc, GetIPCIC(ic), msg);
     }
     else {
         FcitxIPCFrontend* ipc = (FcitxIPCFrontend*) arg;
-        dbus_uint32_t serial = 0; // unique number to associate replies with requests
         FcitxInputState* input = FcitxInstanceGetInputState(ipc->owner);
         DBusMessage* msg = dbus_message_new_signal(GetIPCIC(ic)->path, // object name of the signal
                         FCITX_IC_DBUS_INTERFACE, // interface name of the signal
@@ -988,11 +1084,7 @@ void IPCUpdatePreedit(void* arg, FcitxInputContext* ic)
 
         dbus_message_append_args(msg, DBUS_TYPE_STRING, &strPreedit, DBUS_TYPE_INT32, &iCursorPos, DBUS_TYPE_INVALID);
 
-        if (!dbus_connection_send(ipc->conn, msg, &serial)) {
-            FcitxLog(DEBUG, "Out Of Memory!");
-        }
-        dbus_connection_flush(ipc->conn);
-        dbus_message_unref(msg);
+        IPCSendSignal(ipc, GetIPCIC(ic), msg);
         free(strPreedit);
     }
 }
@@ -1000,18 +1092,13 @@ void IPCUpdatePreedit(void* arg, FcitxInputContext* ic)
 void IPCDeleteSurroundingText(void* arg, FcitxInputContext* ic, int offset, unsigned int size)
 {
     FcitxIPCFrontend* ipc = (FcitxIPCFrontend*) arg;
-    dbus_uint32_t serial = 0; // unique number to associate replies with requests
     DBusMessage* msg = dbus_message_new_signal(GetIPCIC(ic)->path, // object name of the signal
                        FCITX_IC_DBUS_INTERFACE, // interface name of the signal
                        "DeleteSurroundingText"); // name of the signal
 
     dbus_message_append_args(msg, DBUS_TYPE_INT32, &offset, DBUS_TYPE_UINT32, &size, DBUS_TYPE_INVALID);
 
-    if (!dbus_connection_send(ipc->conn, msg, &serial)) {
-        FcitxLog(DEBUG, "Out Of Memory!");
-    }
-    dbus_connection_flush(ipc->conn);
-    dbus_message_unref(msg);
+    IPCSendSignal(ipc, GetIPCIC(ic), msg);
 }
 
 
@@ -1039,7 +1126,6 @@ void IPCUpdateClientSideUI(void* arg, FcitxInputContext* ic)
 {
     FcitxIPCFrontend* ipc = (FcitxIPCFrontend*) arg;
     FcitxInputState* input = FcitxInstanceGetInputState(ipc->owner);
-    dbus_uint32_t serial = 0; // unique number to associate replies with requests
     DBusMessage* msg = dbus_message_new_signal(GetIPCIC(ic)->path, // object name of the signal
                        FCITX_IC_DBUS_INTERFACE, // interface name of the signal
                        "UpdateClientSideUI"); // name of the signal
@@ -1087,11 +1173,7 @@ void IPCUpdateClientSideUI(void* arg, FcitxInputContext* ic)
                              DBUS_TYPE_INT32, &iCursorPos,
                              DBUS_TYPE_INVALID);
 
-    if (!dbus_connection_send(ipc->conn, msg, &serial)) {
-        FcitxLog(DEBUG, "Out Of Memory!");
-    }
-    dbus_connection_flush(ipc->conn);
-    dbus_message_unref(msg);
+    IPCSendSignal(ipc, GetIPCIC(ic), msg);
     free(strAuxUp);
     free(strAuxDown);
     free(strPreedit);
@@ -1107,7 +1189,7 @@ boolean IPCCheckICFromSameApplication(void* arg, FcitxInputContext* icToCheck, F
     return strcmp(ipcicToCheck->appname, ipcic->appname) == 0;
 }
 
-void FcitxDBusPropertyGet(FcitxIPCFrontend* ipc, DBusMessage* message)
+void FcitxDBusPropertyGet(FcitxIPCFrontend* ipc, DBusConnection* conn, DBusMessage* message)
 {
     DBusError error;
     dbus_error_init(&error);
@@ -1132,11 +1214,11 @@ void FcitxDBusPropertyGet(FcitxIPCFrontend* ipc, DBusMessage* message)
             propertTable[index].getfunc(ipc, &variant);
         dbus_message_iter_close_container(&args, &variant);
     }
-    dbus_connection_send(ipc->conn, reply, NULL);
+    dbus_connection_send(conn, reply, NULL);
     dbus_message_unref(reply);
 }
 
-void FcitxDBusPropertySet(FcitxIPCFrontend* ipc, DBusMessage* message)
+void FcitxDBusPropertySet(FcitxIPCFrontend* ipc, DBusConnection* conn, DBusMessage* message)
 {
     DBusError error;
     dbus_error_init(&error);
@@ -1173,11 +1255,11 @@ void FcitxDBusPropertySet(FcitxIPCFrontend* ipc, DBusMessage* message)
         propertTable[index].setfunc(ipc, &variant);
 
 dbus_property_set_end:
-    dbus_connection_send(ipc->conn, reply, NULL);
+    dbus_connection_send(conn, reply, NULL);
     dbus_message_unref(reply);
 }
 
-void FcitxDBusPropertyGetAll(FcitxIPCFrontend* ipc, DBusMessage* message)
+void FcitxDBusPropertyGetAll(FcitxIPCFrontend* ipc, DBusConnection* conn, DBusMessage* message)
 {
     DBusError error;
     dbus_error_init(&error);
@@ -1209,7 +1291,7 @@ void FcitxDBusPropertyGetAll(FcitxIPCFrontend* ipc, DBusMessage* message)
         }
         dbus_message_iter_close_container(&args, &array);
     }
-    dbus_connection_send(ipc->conn, reply, NULL);
+    dbus_connection_send(conn, reply, NULL);
     dbus_message_unref(reply);
 }
 
@@ -1322,7 +1404,6 @@ void IPCUpdateIMList(void* arg)
                        DBUS_INTERFACE_PROPERTIES, // interface name of the signal
                        "PropertiesChanged"); // name of the signal
 
-    dbus_uint32_t serial;
     DBusMessageIter args;
     DBusMessageIter changed_properties, invalidated_properties;
     char sinterface[] = FCITX_IM_DBUS_INTERFACE;
@@ -1339,11 +1420,7 @@ void IPCUpdateIMList(void* arg)
     dbus_message_iter_append_basic(&invalidated_properties, DBUS_TYPE_STRING, &property);
     dbus_message_iter_close_container(&args, &invalidated_properties);
 
-    if (!dbus_connection_send(ipc->conn, msg, &serial)) {
-        FcitxLog(DEBUG, "Out Of Memory!");
-    }
-    dbus_connection_flush(ipc->conn);
-    dbus_message_unref(msg);
+    IPCSendSignal(ipc, NULL, msg);
 }
 
 pid_t IPCGetPid(void* arg, FcitxInputContext* ic)
