@@ -23,7 +23,6 @@
 #include <QTextCharFormat>
 
 #include <sys/time.h>
-#include <unicode/unorm.h>
 
 #include "fcitx/ui.h"
 #include "fcitx/ime.h"
@@ -59,21 +58,6 @@ static const int XKeyRelease = KeyRelease;
 #define Q_UNLIKELY(x) (x)
 #endif
 
-typedef struct _FcitxComposeTableCompact FcitxComposeTableCompact;
-struct _FcitxComposeTableCompact {
-    const quint32 *data;
-    int max_seq_len;
-    int n_index_size;
-    int n_index_stride;
-};
-
-static const FcitxComposeTableCompact fcitx_compose_table_compact = {
-    fcitx_compose_seqs_compact,
-    5,
-    23,
-    6
-};
-
 static int
 compare_seq_index(const void *key, const void *value)
 {
@@ -105,25 +89,6 @@ compare_seq(const void *key, const void *value)
     return 0;
 }
 
-static boolean
-get_boolean_env(const char *name,
-                 boolean defval)
-{
-    const char *value = getenv(name);
-
-    if (value == NULL)
-        return defval;
-
-    if (strcmp(value, "") == 0 ||
-        strcmp(value, "0") == 0 ||
-        strcmp(value, "false") == 0 ||
-        strcmp(value, "False") == 0 ||
-        strcmp(value, "FALSE") == 0)
-        return false;
-
-    return true;
-}
-
 static const uint fcitx_compose_ignore[] = {
     FcitxKey_Shift_L,
     FcitxKey_Shift_R,
@@ -148,42 +113,225 @@ typedef QInputMethodEvent::Attribute QAttribute;
 
 static bool key_filtered = false;
 
+QByteArray QFcitxInputContext::localMachineId()
+{
+#if QT_VERSION >= QT_VERSION_CHECK(4, 8, 0)
+    return QDBusConnection::localMachineId();
+#else
+    QFile file1("/var/lib/dbus/machine-id");
+    QFile file2("/etc/machine-id");
+    QFile* fileToRead = NULL;
+    if (file1.open(QIODevice::ReadOnly)) {
+        fileToRead = &file1;
+    }
+    else if (file2.open(QIODevice::ReadOnly)) {
+        fileToRead = &file2;
+    }
+    if (fileToRead) {
+        QByteArray result = fileToRead->readLine(1024);
+        fileToRead->close();
+        result = result.trimmed();
+        if (!result.isEmpty())
+            return result;
+    }
+    return "machine-id";
+#endif
+}
+
+QString
+QFcitxInputContext::socketFile()
+{
+    char* addressFile = NULL;
+
+    asprintf(&addressFile, "%s-%d", localMachineId().data(), fcitx_utils_get_display_number());
+
+    char* file = NULL;
+
+    FcitxXDGGetFileUserWithPrefix("dbus", addressFile, NULL, &file);
+
+    QString path = QString::fromUtf8(file);
+    free(file);
+    free(addressFile);
+
+    return path;
+
+}
+
+QString
+QFcitxInputContext::address()
+{
+    QString addr;
+    QByteArray addrVar = qgetenv("FCITX_DBUS_ADDRESS");
+    if (!addrVar.isNull())
+        return QString::fromLocal8Bit(addrVar);
+
+    QFile file(socketFile());
+    if (!file.open(QIODevice::ReadOnly))
+        return QString();
+
+    const int BUFSIZE = 1024;
+
+    char buffer[BUFSIZE];
+    size_t sz = file.read(buffer, BUFSIZE);
+    file.close();
+    if (sz == 0)
+        return QString();
+    char* p = buffer;
+    while(*p)
+        p++;
+    size_t addrlen = p - buffer;
+    if (sz != addrlen + 2 * sizeof(pid_t) + 1)
+        return QString();
+
+    /* skip '\0' */
+    p++;
+    pid_t *ppid = (pid_t*) p;
+    pid_t daemonpid = ppid[0];
+    pid_t fcitxpid = ppid[1];
+
+    if (!fcitx_utils_pid_exists(daemonpid)
+        || !fcitx_utils_pid_exists(fcitxpid))
+        return QString();
+
+    addr = QLatin1String(buffer);
+
+    return addr;
+
+}
+
 QFcitxInputContext::QFcitxInputContext()
-    : m_connection(QDBusConnection::sessionBus()),
-      m_dbusproxy(0),
+    : m_watcher(new QFileSystemWatcher(this)),
+      m_connection(0),
       m_improxy(0),
       m_icproxy(0),
+      m_capacity(0),
       m_id(0),
       m_path(""),
       m_has_focus(false),
       m_n_compose(0),
       m_cursorPos(0),
-      m_useSurroundingText(false)
+      m_useSurroundingText(false),
+      m_syncMode(true)
 {
     FcitxFormattedPreedit::registerMetaType();
 
     memset(m_compose_buffer, 0, sizeof(uint) * (FCITX_MAX_COMPOSE_LEN + 1));
 
-    m_dbusproxy = new org::freedesktop::DBus("org.freedesktop.DBus", "/org/freedesktop/DBus", m_connection, this);
-    connect(m_dbusproxy, SIGNAL(NameOwnerChanged(QString, QString, QString)), this, SLOT(imChanged(QString, QString, QString)));
-
     m_serviceName = QString("%1-%2").arg(FCITX_DBUS_SERVICE).arg(fcitx_utils_get_display_number());
+    m_serviceWatcher.setConnection(QDBusConnection::sessionBus());
+    m_serviceWatcher.addWatchedService(m_serviceName);
 
-    createInputContext();
+    QFileInfo info(socketFile());
+    QDir dir(info.path());
+    if (!dir.exists()) {
+        QDir rt(QDir::root());
+        rt.mkpath(info.path());
+    }
+    m_watcher.data()->addPath(info.path());
+    if (info.exists()) {
+        m_watcher.data()->addPath(info.filePath());
+    }
+
+    connect(m_watcher.data(), SIGNAL(fileChanged(QString)), this, SLOT(socketFileChanged()));
+    connect(m_watcher.data(), SIGNAL(directoryChanged(QString)), this, SLOT(socketFileChanged()));
+#if QT_VERSION < QT_VERSION_CHECK(4, 8, 0)
+    connect(qApp, SIGNAL(aboutToQuit()), m_watcher.data(), SLOT(deleteLater()));
+#endif
+    createConnection();
 }
 
 QFcitxInputContext::~QFcitxInputContext()
 {
-    delete m_dbusproxy;
-    if (m_improxy)
-        delete m_improxy;
-    if (m_icproxy) {
-        if (m_icproxy->isValid()) {
-            m_icproxy->DestroyIC();
-        }
-
-        delete m_icproxy;
+    if (m_icproxy && m_icproxy->isValid()) {
+        m_icproxy->DestroyIC();
     }
+
+    cleanUp();
+
+    if (!m_watcher.isNull())
+        delete m_watcher.data();
+}
+
+void QFcitxInputContext::socketFileChanged()
+{
+    if (m_watcher.isNull())
+        return;
+
+    QFileInfo info(socketFile());
+    if (info.exists()) {
+        if (m_watcher.data()->files().indexOf(info.filePath()) == -1)
+            m_watcher.data()->addPath(info.filePath());
+    }
+
+    QString addr = address();
+    if (addr.isNull())
+        return;
+
+    cleanUp();
+    createConnection();
+}
+
+void QFcitxInputContext::cleanUp()
+{
+    QDBusConnection::disconnectFromBus("fcitx");
+    if (m_connection) {
+        delete m_connection;
+        m_connection = NULL;
+    }
+
+    if (m_improxy) {
+        delete m_improxy;
+        m_improxy = NULL;
+    }
+
+    if (m_icproxy) {
+        delete m_icproxy;
+        m_icproxy = NULL;
+    }
+
+    reset();
+}
+
+void QFcitxInputContext::createConnection()
+{
+    m_serviceWatcher.disconnect(SIGNAL(serviceOwnerChanged(QString,QString,QString)));
+    QString addr = address();
+    // qDebug() << addr;
+    if (!addr.isNull()) {
+        QDBusConnection connection(QDBusConnection::connectToBus(addr, "fcitx"));
+        if (connection.isConnected()) {
+            // qDebug() << "create private";
+            m_connection = new QDBusConnection(connection);
+        }
+        else
+            QDBusConnection::disconnectFromBus("fcitx");
+    }
+
+    bool needCreate = true;
+    if (!m_connection) {
+        m_connection = new QDBusConnection(QDBusConnection::sessionBus());
+        connect(&m_serviceWatcher, SIGNAL(serviceOwnerChanged(QString,QString,QString)), this, SLOT(imChanged(QString,QString,QString)));
+        QDBusReply<bool> registered = m_connection->interface()->isServiceRegistered(m_serviceName);
+        if (!registered.isValid() || !registered.value())
+            needCreate = false;
+    }
+
+    m_connection->connect ("org.freedesktop.DBus.Local",
+                           "/org/freedesktop/DBus/Local",
+                           "org.freedesktop.DBus.Local",
+                           "Disconnected",
+                           this,
+                           SLOT (dbusDisconnect ()));
+
+    if (needCreate)
+        createInputContext();
+}
+
+void QFcitxInputContext::dbusDisconnect()
+{
+    cleanUp();
+    emit dbusDisconnected();
+    createConnection();
 }
 
 QString QFcitxInputContext::identifierName()
@@ -240,6 +388,16 @@ void QFcitxInputContext::update()
         }
     }
 
+    QTimer::singleShot(0, this, SLOT(updateCursor()));
+}
+
+void QFcitxInputContext::updateCursor()
+{
+    QWidget* widget = focusWidget();
+    if (widget == NULL || !isValid()) {
+        return;
+    }
+
     QRect rect = widget->inputMethodQuery(Qt::ImMicroFocus).toRect();
 
     QPoint topleft = widget->mapToGlobal(QPoint(0, 0));
@@ -251,7 +409,6 @@ void QFcitxInputContext::update()
     }
 }
 
-
 bool QFcitxInputContext::isComposing() const
 {
     return false;
@@ -259,16 +416,16 @@ bool QFcitxInputContext::isComposing() const
 
 bool QFcitxInputContext::filterEvent(const QEvent* event)
 {
-#ifndef Q_WS_X11
+#if not (defined(Q_WS_X11) && defined(ENABLE_X11))
     QWidget* keywidget = focusWidget();
 
     if (key_filtered)
         return false;
 
-    if (!keywidget->testAttribute(Qt::WA_WState_Created))
+    if (!keywidget || !keywidget->testAttribute(Qt::WA_WState_Created))
         return false;
 
-    if (!keywidget || keywidget->inputMethodHints() & (Qt::ImhExclusiveInputMask | Qt::ImhHiddenText))
+    if (keywidget->inputMethodHints() & (Qt::ImhExclusiveInputMask | Qt::ImhHiddenText))
         addCapacity(CAPACITY_PASSWORD);
     else
         removeCapacity(CAPACITY_PASSWORD);
@@ -278,18 +435,6 @@ bool QFcitxInputContext::filterEvent(const QEvent* event)
     }
 
     const QKeyEvent *key_event = static_cast<const QKeyEvent*>(event);
-    if (!m_enable) {
-        FcitxKeySym fcitxsym;
-        uint fcitxstate;
-        GetKey((FcitxKeySym) key_event->nativeVirtualKey(), key_event->nativeModifiers(), &fcitxsym, &fcitxstate);
-        if (!FcitxIsHotKey(fcitxsym, fcitxstate, m_triggerKey)) {
-            if (m_slave && m_slave->filterEvent(event))
-                return true;
-            else
-                return QInputContext::filterEvent(event);
-        }
-
-    }
     m_icproxy->FocusIn();
 
     struct timeval current_time;
@@ -300,16 +445,16 @@ bool QFcitxInputContext::filterEvent(const QEvent* event)
                                        key_event->nativeScanCode(),
                                        key_event->nativeModifiers(),
                                        (event->type() == QEvent::KeyPress) ? FCITX_PRESS_KEY : FCITX_RELEASE_KEY,
-                                       time
-                                                                       );
+                                       time);
     {
         QEventLoop loop;
         QDBusPendingCallWatcher watcher(result);
         loop.connect(&watcher, SIGNAL(finished(QDBusPendingCallWatcher*)), SLOT(quit()));
+        loop.connect(this, SIGNAL(dbusDisconnected()), SLOT(quit()));
         loop.exec(QEventLoop::ExcludeUserInputEvents | QEventLoop::WaitForMoreEvents);
     }
 
-    if (result.isError() || result.value() <= 0)
+    if (!m_connection || !result.isFinished() || result.isError() || result.value() <= 0)
         return QInputContext::filterEvent(event);
     else {
         update();
@@ -402,16 +547,17 @@ void QFcitxInputContext::setFocusWidget(QWidget* w)
 
 void QFcitxInputContext::widgetDestroyed(QWidget* w)
 {
+    if (isValid()) {
+        if (w == focusWidget())
+            m_icproxy->FocusOut();
+        update();
+    }
+
     QInputContext::widgetDestroyed(w);
-    if (!isValid())
-        return;
-    if (w == focusWidget())
-        m_icproxy->FocusOut();
-    update();
 }
 
 
-#if defined(Q_WS_X11)
+#if defined(Q_WS_X11) && defined(ENABLE_X11)
 
 bool QFcitxInputContext::x11FilterEvent(QWidget* keywidget, XEvent* event)
 {
@@ -432,15 +578,18 @@ bool QFcitxInputContext::x11FilterEvent(QWidget* keywidget, XEvent* event)
     if (Q_UNLIKELY(event->xkey.state & FcitxKeyState_IgnoredMask))
         return false;
 
-    if (Q_UNLIKELY(!isValid() || (event->type != XKeyRelease && event->type != XKeyPress))) {
-        return x11FilterEventFallback(event, 0);
-    }
+    if (Q_UNLIKELY(event->type != XKeyRelease && event->type != XKeyPress))
+        return false;
 
     KeySym sym = 0;
     char strbuf[64];
     memset(strbuf, 0, 64);
     XLookupString(&event->xkey, strbuf, 64, &sym, NULL);
 
+    if (Q_UNLIKELY(!isValid())) {
+        return x11FilterEventFallback(event, sym);
+    }
+
     QDBusPendingReply< int > result = this->m_icproxy->ProcessKeyEvent(
                                           sym,
                                           event->xkey.keycode,
@@ -452,9 +601,10 @@ bool QFcitxInputContext::x11FilterEvent(QWidget* keywidget, XEvent* event)
         QEventLoop loop;
         QDBusPendingCallWatcher watcher(result);
         loop.connect(&watcher, SIGNAL(finished(QDBusPendingCallWatcher*)), SLOT(quit()));
+        loop.connect(this, SIGNAL(dbusDisconnected()), SLOT(quit()));
         loop.exec(QEventLoop::ExcludeUserInputEvents);
 
-        if (result.isError() || result.value() <= 0) {
+        if (!m_connection || !result.isFinished() || result.isError() || result.value() <= 0) {
             QTimer::singleShot(0, this, SLOT(updateIM()));
             return x11FilterEventFallback(event, sym);
         } else {
@@ -509,28 +659,37 @@ void QFcitxInputContext::imChanged(const QString& service, const QString& oldown
 {
     if (service == m_serviceName) {
         /* old die */
-        if (oldowner.length() > 0 || newowner.length() > 0) {
-            if (m_improxy) {
-                delete m_improxy;
-                m_improxy = NULL;
-            }
-
-            if (m_icproxy) {
-                delete m_icproxy;
-                m_icproxy = NULL;
-            }
-        }
+        if (oldowner.length() > 0 || newowner.length() > 0)
+            cleanUp();
 
         /* new rise */
-        if (newowner.length() > 0)
-            createInputContext();
+        if (newowner.length() > 0) {
+            QTimer::singleShot(100, this, SLOT(newServiceAppear()));
+        }
+    }
+}
+
+void QFcitxInputContext::newServiceAppear()
+{
+    if (!isConnected()) {
+        cleanUp();
+        createConnection();
     }
 }
 
 void QFcitxInputContext::createInputContext()
 {
+    if (!m_connection)
+        return;
+
+    // qDebug() << "create Input Context" << m_connection->name();
+
     m_rect = QRect();
-    m_improxy = new org::fcitx::Fcitx::InputMethod(m_serviceName, FCITX_IM_DBUS_PATH, m_connection, this);
+    if (m_improxy) {
+        delete m_improxy;
+        m_improxy = NULL;
+    }
+    m_improxy = new org::fcitx::Fcitx::InputMethod(m_serviceName, FCITX_IM_DBUS_PATH, *m_connection, this);
 
     if (!m_improxy->isValid())
         return;
@@ -545,12 +704,22 @@ void QFcitxInputContext::createInputContext()
 void QFcitxInputContext::createInputContextFinished(QDBusPendingCallWatcher* watcher)
 {
     QDBusPendingReply< int, bool, uint, uint, uint, uint > result = *watcher;
-    if (result.isError())
-        qWarning() << result.error();
-    else {
+
+    do {
+        if (result.isError()) {
+            break;
+        }
+
+        if (!m_connection)
+            break;
+
         this->m_id = qdbus_cast<int>(result.argumentAt(0));
         this->m_path = QString(FCITX_IC_DBUS_PATH_QSTRING).arg(m_id);
-        m_icproxy = new org::fcitx::Fcitx::InputContext(m_serviceName, m_path, m_connection, this);
+        if (m_icproxy) {
+            delete m_icproxy;
+            m_icproxy = NULL;
+        }
+        m_icproxy = new org::fcitx::Fcitx::InputContext(m_serviceName, m_path, *m_connection, this);
         connect(m_icproxy, SIGNAL(CommitString(QString)), this, SLOT(commitString(QString)));
         connect(m_icproxy, SIGNAL(ForwardKey(uint, uint, int)), this, SLOT(forwardKey(uint, uint, int)));
         connect(m_icproxy, SIGNAL(UpdateFormattedPreedit(FcitxFormattedPreeditList,int)), this, SLOT(updateFormattedPreedit(FcitxFormattedPreeditList,int)));
@@ -563,14 +732,14 @@ void QFcitxInputContext::createInputContextFinished(QDBusPendingCallWatcher* wat
         flag |= CAPACITY_PREEDIT;
         flag |= CAPACITY_FORMATTED_PREEDIT;
         flag |= CAPACITY_CLIENT_UNFOCUS_COMMIT;
-        m_useSurroundingText = get_boolean_env("FCITX_QT_ENABLE_SURROUNDING_TEXT", false);
+        m_useSurroundingText = fcitx_utils_get_boolean_env("FCITX_QT_ENABLE_SURROUNDING_TEXT", false);
         if (m_useSurroundingText)
             flag |= CAPACITY_SURROUNDING_TEXT;
 
-        m_syncMode = get_boolean_env("FCITX_QT_USE_SYNC", true);
+        m_syncMode = fcitx_utils_get_boolean_env("FCITX_QT_USE_SYNC", true);
 
         addCapacity(flag, true);
-    }
+    } while(0);
     delete watcher;
 }
 
@@ -645,7 +814,7 @@ void QFcitxInputContext::forwardKey(uint keyval, uint state, int type)
     QWidget* widget = focusWidget();
     if (Q_LIKELY(widget != NULL)) {
         key_filtered = true;
-#ifdef Q_WS_X11
+#if defined(Q_WS_X11) && defined(ENABLE_X11)
         const WId window_id = widget->winId();
         Display* x11_display = QX11Info::display();
 
@@ -661,7 +830,7 @@ void QFcitxInputContext::forwardKey(uint keyval, uint state, int type)
     }
 }
 
-#ifdef Q_WS_X11
+#if defined(Q_WS_X11) && defined(ENABLE_X11)
 XEvent* QFcitxInputContext::createXEvent(Display* dpy, WId wid, uint keyval, uint state, int type)
 {
     XEvent* xevent = static_cast<XEvent*>(malloc(sizeof(XEvent)));
@@ -693,6 +862,11 @@ XEvent* QFcitxInputContext::createXEvent(Display* dpy, WId wid, uint keyval, uin
 }
 #endif
 
+bool QFcitxInputContext::isConnected()
+{
+    return m_connection != NULL && m_connection->isConnected();
+}
+
 bool QFcitxInputContext::isValid()
 {
     return m_icproxy != NULL && m_icproxy->isValid();
@@ -745,7 +919,7 @@ bool
 QFcitxInputContext::checkAlgorithmically()
 {
     int i;
-    UChar combination_buffer[FCITX_MAX_COMPOSE_LEN];
+    uint32_t combination_buffer[FCITX_MAX_COMPOSE_LEN];
 
     if (m_n_compose >= FCITX_MAX_COMPOSE_LEN)
         return false;
@@ -765,7 +939,7 @@ case FcitxKey_dead_##keysym: combination_buffer[i + 1] = unicode; break
                 CASE(grave, 0x0300);
                 CASE(acute, 0x0301);
                 CASE(circumflex, 0x0302);
-                CASE(tilde, 0x0303);  /* Also used with perispomeni, 0x342. */
+                CASE(tilde, 0x0303);   /* Also used with perispomeni, 0x342. */
                 CASE(macron, 0x0304);
                 CASE(breve, 0x0306);
                 CASE(abovedot, 0x0307);
@@ -774,25 +948,24 @@ case FcitxKey_dead_##keysym: combination_buffer[i + 1] = unicode; break
                 CASE(abovering, 0x030A);
                 CASE(doubleacute, 0x030B);
                 CASE(caron, 0x030C);
-                CASE(abovecomma, 0x0313);  /* Equivalent to psili */
-                CASE(abovereversedcomma, 0x0314);  /* Equivalent to dasia */
-                CASE(horn, 0x031B);  /* Legacy use for psili, 0x313 (or 0x343). */
+                CASE(abovecomma, 0x0313);         /* Equivalent to psili */
+                CASE(abovereversedcomma, 0x0314); /* Equivalent to dasia */
+                CASE(horn, 0x031B);    /* Legacy use for psili, 0x313 (or 0x343). */
                 CASE(belowdot, 0x0323);
                 CASE(cedilla, 0x0327);
                 CASE(ogonek, 0x0328);  /* Legacy use for dasia, 0x314.*/
                 CASE(iota, 0x0345);
-                CASE(voiced_sound, 0x3099);  /* Per Markus Kuhn keysyms.txt file. */
-                CASE(semivoiced_sound, 0x309A);  /* Per Markus Kuhn keysyms.txt file. */
+                CASE(voiced_sound, 0x3099);    /* Per Markus Kuhn keysyms.txt file. */
+                CASE(semivoiced_sound, 0x309A);    /* Per Markus Kuhn keysyms.txt file. */
+
                 /* The following cases are to be removed once xkeyboard-config,
                 * xorg are fully updated.
-                **/
-                /* Workaround for typo in 1.4.x xserver-xorg */
-            case 0xfe66:
-                combination_buffer[i + 1] = 0x314;
-                break;
-                /* CASE (dasia, 0x314); */
-                /* CASE (perispomeni, 0x342); */
-                /* CASE (psili, 0x343); */
+                */
+                    /* Workaround for typo in 1.4.x xserver-xorg */
+                case 0xfe66: combination_buffer[i+1] = 0x314; break;
+                /* CASE(dasia, 0x314); */
+                /* CASE(perispomeni, 0x342); */
+                /* CASE(psili, 0x343); */
 #undef CASE
             default:
                 combination_buffer[i + 1] = FcitxKeySymToUnicode((FcitxKeySym) m_compose_buffer[i]);
@@ -820,15 +993,14 @@ case FcitxKey_dead_##keysym: combination_buffer[i + 1] = unicode; break
             return TRUE;
         }
 #endif
-        UErrorCode state = U_ZERO_ERROR;
-        UChar result[FCITX_MAX_COMPOSE_LEN + 1];
-        i = unorm_normalize(combination_buffer, m_n_compose, UNORM_NFC, 0, result, FCITX_MAX_COMPOSE_LEN + 1, &state);
+        QString s(QString::fromUcs4(combination_buffer, m_n_compose));
+        s = s.normalized(QString::NormalizationForm_C);
 
-        // qDebug () << "combination_buffer = " << QString::fromUtf16(combination_buffer) << "m_n_compose" << m_n_compose;
-        // qDebug () << "result = " << QString::fromUtf16(result) << "i = " << i << state;
+        // qDebug () << "combination_buffer = " << QString::fromUcs4(combination_buffer, m_n_compose) << "m_n_compose" << m_n_compose;
+        // qDebug () << "result = " << s << "i = " << s.length();
 
-        if (i == 1) {
-            commitString(QString(QChar(result[0])));
+        if (s.length() == 1) {
+            commitString(QString(s[0]));
             m_compose_buffer[0] = 0;
             m_n_compose = 0;
             return true;
