@@ -20,12 +20,15 @@
 #ifndef __FCITX_INPUT_CONTEXT_H_
 #define __FCITX_INPUT_CONTEXT_H_
 
+#include "config.h"
+
 #include <QInputContext>
 #include <QList>
 #include <QDBusConnection>
 #include <QDir>
 #include <QApplication>
-#include "org.freedesktop.DBus.h"
+#include <QWeakPointer>
+
 #include "org.fcitx.Fcitx.InputMethod.h"
 #include "org.fcitx.Fcitx.InputContext.h"
 #include "fcitx-config/hotkey.h"
@@ -76,13 +79,18 @@ public:
 
     virtual void widgetDestroyed(QWidget *w);
 
-#if defined(Q_WS_X11)
+#if defined(Q_WS_X11) && defined(ENABLE_X11)
     virtual bool x11FilterEvent(QWidget *keywidget, XEvent *event);
 #endif // Q_WS_X11
     virtual bool filterEvent(const QEvent* event);
     virtual void mouseHandler(int x, QMouseEvent* event);
 
+Q_SIGNALS:
+    void dbusDisconnected();
+
 private Q_SLOTS:
+    void socketFileChanged();
+    void dbusDisconnect();
     void imChanged(const QString& service, const QString& oldowner, const QString& newowner);
     void commitString(const QString& str);
     void updateFormattedPreedit(const FcitxFormattedPreeditList& preeditList, int cursorPos);
@@ -90,20 +98,28 @@ private Q_SLOTS:
     void deleteSurroundingText(int offset, uint nchar);
     void createInputContextFinished(QDBusPendingCallWatcher* watcher);
     void updateIM();
-#if defined(Q_WS_X11)
+#if defined(Q_WS_X11) && defined(ENABLE_X11)
     void x11ProcessKeyEventCallback(QDBusPendingCallWatcher* watcher);
+    void newServiceAppear();
+    void updateCursor();
 #endif
 private:
+    static QByteArray localMachineId();
+    static QString socketFile();
+    static QString address();
+    void cleanUp();
+    void createConnection();
     void createInputContext();
     bool processCompose(uint keyval, uint state, FcitxKeyEventType event);
     bool checkAlgorithmically();
     bool checkCompactTable(const struct _FcitxComposeTableCompact *table);
-#if defined(Q_WS_X11)
-    bool x11FilterEventFallback(XEvent *event , KeySym sym);
+#if defined(Q_WS_X11) && defined(ENABLE_X11)
+    bool x11FilterEventFallback(XEvent *event, KeySym sym);
     XEvent* createXEvent(Display* dpy, WId wid, uint keyval, uint state, int type);
 #endif // Q_WS_X11
     QKeyEvent* createKeyEvent(uint keyval, uint state, int type);
     bool isValid();
+    bool isConnected();
 
     void addCapacity(QFlags<FcitxCapacityFlags> capacity, bool forceUpdage = false)
     {
@@ -126,8 +142,9 @@ private:
     void updateCapacity();
     void commitPreedit();
 
-    QDBusConnection m_connection;
-    org::freedesktop::DBus* m_dbusproxy;
+    QWeakPointer<QFileSystemWatcher> m_watcher;
+    QDBusServiceWatcher m_serviceWatcher;
+    QDBusConnection* m_connection;
     org::fcitx::Fcitx::InputMethod* m_improxy;
     org::fcitx::Fcitx::InputContext* m_icproxy;
     QFlags<FcitxCapacityFlags> m_capacity;
