@@ -81,8 +81,7 @@ static XIMStyle OnTheSpot_Styles [] = {
     0
 };
 
-FCITX_EXPORT_API
-FcitxFrontend frontend = {
+FCITX_DEFINE_PLUGIN(fcitx_xim, frontend, FcitxFrontend) = {
     XimCreate,
     XimDestroy,
     XimCreateIC,
@@ -103,9 +102,6 @@ FcitxFrontend frontend = {
     NULL
 };
 
-FCITX_EXPORT_API
-int ABI_VERSION = FCITX_ABI_VERSION;
-
 FcitxXimFrontend *ximfrontend;
 
 CONFIG_DESC_DEFINE(GetXimConfigDesc, "fcitx-xim.desc")
@@ -132,9 +128,8 @@ void* XimCreate(FcitxInstance* instance, int frontendid)
     XIMEncodings *encodings;
     char *imname = NULL;
     char *p;
-    FcitxModuleFunctionArg arg;
 
-    xim->display = InvokeFunction(instance, FCITX_X11, GETDISPLAY, arg);
+    xim->display = InvokeVaArgs(instance, FCITX_X11, GETDISPLAY);
 
     if (xim->display == NULL) {
         FcitxLog(FATAL, _("X11 not initialized"));
@@ -246,6 +241,7 @@ void* XimCreate(FcitxInstance* instance, int frontendid)
     if (xim->ims == (XIMS) NULL) {
         FcitxLog(ERROR, _("Start XIM error. Another XIM daemon named %s is running?"), imname);
         free(xim);
+        FcitxInstanceEnd(instance);
         return NULL;
     }
 
