@@ -26,9 +26,9 @@
 #include <iconv.h>
 #include <ctype.h>
 
+#if defined(ENABLE_ICU)
 #include <unicode/unorm.h>
-
-#include <enchant/enchant.h>
+#endif
 
 #include "fcitx/ime.h"
 #include "fcitx/instance.h"
@@ -36,31 +36,23 @@
 #include "fcitx/keys.h"
 #include "fcitx/candidate.h"
 #include "fcitx/module.h"
+#include "fcitx/frontend.h"
 #include "fcitx/hook.h"
 #include "fcitx-config/xdg.h"
 #include "fcitx-utils/log.h"
+#include "module/spell/spell.h"
 
 #include "xkb.h"
 #include "keyboard.h"
+#if defined(ENABLE_LIBXML2)
 #include "isocodes.h"
+#endif
 #include "fcitx-compose-data.h"
 
 #define INVALID_COMPOSE_RESULT 0xffffffff
 
-typedef struct _FcitxComposeTableCompact FcitxComposeTableCompact;
-struct _FcitxComposeTableCompact {
-    const uint32_t *data;
-    int max_seq_len;
-    int n_index_size;
-    int n_index_stride;
-};
-
-static const FcitxComposeTableCompact fcitx_compose_table_compact = {
-    fcitx_compose_seqs_compact,
-    5,
-    23,
-    6
-};
+#define PTR_TRUE ((void*)0x1)
+#define PTR_FALSE ((void*)0x0)
 
 static void* FcitxKeyboardCreate(FcitxInstance* instance);
 static boolean FcitxKeyboardInit(void *arg);
@@ -69,10 +61,6 @@ static INPUT_RETURN_VALUE FcitxKeyboardDoInput(void *arg, FcitxKeySym, unsigned
 static void  FcitxKeyboardSave(void *arg);
 static void  FcitxKeyboardReloadConfig(void *arg);
 INPUT_RETURN_VALUE FcitxKeyboardGetCandWords(void* arg);
-#ifdef PRESAGE_FOUND
-INPUT_RETURN_VALUE FcitxKeyboardGetPresageCandWord (void* arg, FcitxCandidateWord* candWord);
-#endif
-INPUT_RETURN_VALUE FcitxKeyboardGetCandWord (void* arg, FcitxCandidateWord* candWord);
 static void FcitxKeyboardLayoutCreate(FcitxKeyboard* keyboard,
                                       const char* name,
                                       const char* langCode,
@@ -88,17 +76,6 @@ static uint32_t checkCompactTable(FcitxKeyboardLayout* layout, const FcitxCompos
 static uint32_t checkAlgorithmically(FcitxKeyboardLayout* layout);
 static uint32_t processCompose(FcitxKeyboardLayout* layout, uint32_t keyval, uint32_t state);
 
-static int fcitx_keyboard_strcmp0(const char* a, const char* b)
-{
-    if (a == NULL && b == NULL)
-        return 0;
-    if (a == NULL && b)
-        return -1;
-    if (a && b == NULL)
-        return 1;
-    return strcmp(a, b);
-}
-
 static int
 compare_seq_index(const void *key, const void *value)
 {
@@ -161,16 +138,13 @@ void Ucs4ToUtf8(iconv_t conv, uint32_t ucs4, char* utf8)
 
 CONFIG_DESC_DEFINE(GetKeyboardConfigDesc, "fcitx-keyboard.desc")
 
-FCITX_EXPORT_API
-FcitxIMClass ime = {
+FCITX_DEFINE_PLUGIN(fcitx_keyboard, ime, FcitxIMClass) = {
     FcitxKeyboardCreate,
     NULL
 };
 
-FCITX_EXPORT_API
-int ABI_VERSION = FCITX_ABI_VERSION;
-
-const char* FcitxKeyboardGetPastStream   (void* arg)
+#if 0
+static const char* FcitxKeyboardGetPastStream(void* arg)
 {
     FcitxKeyboard* keyboard = arg;
     /* buggy surrounding text will cause trouble */
@@ -206,7 +180,7 @@ const char* FcitxKeyboardGetPastStream   (void* arg)
     return keyboard->buffer;
 }
 
-const char* FcitxKeyboardGetFutureStream (void* arg)
+static const char* FcitxKeyboardGetFutureStream(void* arg)
 {
     const char* result = "";
 #if 0
@@ -236,6 +210,7 @@ const char* FcitxKeyboardGetFutureStream (void* arg)
 #endif
     return result;
 }
+#endif
 
 
 void FcitxKeyboardLayoutCreate(FcitxKeyboard* keyboard,
@@ -259,16 +234,13 @@ void FcitxKeyboardLayoutCreate(FcitxKeyboard* keyboard,
 
     int iPriority = 100;
     if (strcmp(keyboard->initialLayout, layoutString) == 0
-        && fcitx_keyboard_strcmp0(keyboard->initialVariant, variantString) == 0) {
+        && fcitx_utils_strcmp0(keyboard->initialVariant, variantString) == 0) {
         iPriority = PRIORITY_MAGIC_FIRST;
     }
     else {
         boolean result = false;
-        FcitxModuleFunctionArg args;
-        args.args[0] = (void*) layoutString;
-        args.args[1] = (void*) variantString;
-        args.args[2] = &result;
-        InvokeFunction(keyboard->owner, FCITX_XKB, LAYOUTEXISTS, args);
+        InvokeVaArgs(keyboard->owner, FCITX_XKB, LAYOUTEXISTS,
+                     (void*)layoutString, (void*)variantString, &result);
         if (result)
             iPriority = 50;
     }
@@ -294,6 +266,11 @@ void FcitxKeyboardLayoutCreate(FcitxKeyboard* keyboard,
     free(uniqueName);
 }
 
+void* SimpleCopy(void* arg, void* dest, void* src)
+{
+    return src;
+}
+
 void* FcitxKeyboardCreate(FcitxInstance* instance)
 {
     FcitxKeyboard* keyboard = fcitx_utils_malloc0(sizeof(FcitxKeyboard));
@@ -316,23 +293,6 @@ void* FcitxKeyboardCreate(FcitxInstance* instance)
         keyboard->iconv = iconv_open("utf-8", "ucs-4be");
     else
         keyboard->iconv = iconv_open("utf-8", "ucs-4le");
-    keyboard->broker = enchant_broker_init();
-#ifdef PRESAGE_FOUND
-    presage_new(FcitxKeyboardGetPastStream, keyboard, FcitxKeyboardGetFutureStream, keyboard, &keyboard->presage);
-#endif
-    /* new option, since hunspell/myspell is really toooo slow */
-    switch (keyboard->config.provider) {
-        case EP_Aspell:
-            enchant_broker_set_ordering(keyboard->broker, "*", "aspell,myspell,ispell");
-            break;
-        case EP_Myspell:
-            enchant_broker_set_ordering(keyboard->broker, "*", "myspell,aspell,ispell");
-            break;
-        case EP_Default:
-        default:
-            break;
-    }
-    keyboard->enchantLanguages = fcitx_utils_new_string_list();
 
     FcitxHotkeyHook hk;
     hk.arg = keyboard;
@@ -341,21 +301,22 @@ void* FcitxKeyboardCreate(FcitxInstance* instance)
 
     FcitxInstanceRegisterHotkeyFilter(instance, hk);
 
-    FcitxModuleFunctionArg args;
-    FcitxXkbRules* rules = InvokeFunction(instance, FCITX_XKB, GETRULES, args);
+    FcitxXkbRules* rules = InvokeVaArgs(instance, FCITX_XKB, GETRULES);
     keyboard->rules = rules;
-    FcitxIsoCodes* isocodes = FcitxXkbReadIsoCodes(ISOCODES_ISO639_XML, ISOCODES_ISO3166_XML);
+#if defined(ENABLE_LIBXML2)
+#endif
 
     keyboard->initialLayout = NULL;
     keyboard->initialVariant = NULL;
 
-    args.args[0] = &keyboard->initialLayout;
-    args.args[1] = &keyboard->initialVariant;
-    InvokeFunction(instance, FCITX_XKB, GETCURRENTLAYOUT, args);
+    InvokeVaArgs(instance, FCITX_XKB, GETCURRENTLAYOUT,
+                 &keyboard->initialLayout, &keyboard->initialVariant);
     if (!keyboard->initialLayout)
         keyboard->initialLayout = strdup("us");
 
-    if (rules) {
+#if defined(ENABLE_LIBXML2)
+    if (rules && utarray_len(rules->layoutInfos)) {
+        FcitxIsoCodes* isocodes = FcitxXkbReadIsoCodes(ISOCODES_ISO639_XML, ISOCODES_ISO3166_XML);
         FcitxXkbLayoutInfo* layoutInfo;
         for (layoutInfo = (FcitxXkbLayoutInfo*) utarray_front(rules->layoutInfos);
             layoutInfo != NULL;
@@ -403,20 +364,27 @@ void* FcitxKeyboardCreate(FcitxInstance* instance)
                 free(description);
             }
         }
+        FcitxIsoCodesFree(isocodes);
     }
-    else {
+    else
+#endif
+    {
         char* description;
-        asprintf(&description, _("Keyboard - %s"),
-                dgettext("xkeyboard-config", "English (US)"));
+        asprintf(&description, _("Keyboard"));
 
+        fcitx_utils_free(keyboard->initialLayout);
+        keyboard->initialLayout = strdup("us");
+        fcitx_utils_free(keyboard->initialVariant);
+        keyboard->initialVariant = NULL;
         FcitxKeyboardLayoutCreate(keyboard, description, "en", "us", NULL);
         free(description);
     }
 
-    FcitxIsoCodesFree(isocodes);
     keyboard->lastLength = 10;
     keyboard->tempBuffer = fcitx_utils_malloc0(keyboard->lastLength);
 
+    keyboard->dataSlot = FcitxInstanceAllocDataForIC(instance, NULL, SimpleCopy, NULL, keyboard);
+
     return keyboard;
 }
 
@@ -424,57 +392,64 @@ boolean FcitxKeyboardInit(void *arg)
 {
     FcitxKeyboardLayout* layout = (FcitxKeyboardLayout*) arg;
     boolean flag = true;
-    FcitxInstanceSetContext(layout->owner->owner, CONTEXT_DISABLE_AUTOENG, &flag);
-    FcitxInstanceSetContext(layout->owner->owner, CONTEXT_DISABLE_QUICKPHRASE, &flag);
-    FcitxInstanceSetContext(layout->owner->owner, CONTEXT_DISABLE_FULLWIDTH, &flag);
+    FcitxInstanceSetContext(layout->owner->owner,
+                            CONTEXT_DISABLE_AUTOENG, &flag);
+    FcitxInstanceSetContext(layout->owner->owner,
+                            CONTEXT_DISABLE_QUICKPHRASE, &flag);
+    FcitxInstanceSetContext(layout->owner->owner,
+                            CONTEXT_DISABLE_FULLWIDTH, &flag);
     if (layout->variantString) {
         char* string;
         asprintf(&string, "%s,%s", layout->layoutString, layout->variantString);
-        FcitxInstanceSetContext(layout->owner->owner, CONTEXT_IM_KEYBOARD_LAYOUT, string);
+        FcitxInstanceSetContext(layout->owner->owner,
+                                CONTEXT_IM_KEYBOARD_LAYOUT, string);
         free(string);
-    }
-    else {
-        FcitxInstanceSetContext(layout->owner->owner, CONTEXT_IM_KEYBOARD_LAYOUT, layout->layoutString);
+    } else {
+        FcitxInstanceSetContext(layout->owner->owner,
+                                CONTEXT_IM_KEYBOARD_LAYOUT,
+                                layout->layoutString);
     }
     return true;
 }
 
 void  FcitxKeyboardResetIM(void *arg)
 {
-    FcitxKeyboardLayout* layout = (FcitxKeyboardLayout*) arg;
-    layout->owner->buffer[0] = '\0';
-    layout->owner->cursorPos = 0;
-    layout->owner->composeBuffer[0] = 0;
-    layout->owner->n_compose = 0;
+    FcitxKeyboardLayout *layout = (FcitxKeyboardLayout*) arg;
+    FcitxKeyboard *keyboard = layout->owner;
+    keyboard->buffer[0] = '\0';
+    keyboard->cursorPos = 0;
+    keyboard->composeBuffer[0] = 0;
+    keyboard->n_compose = 0;
 }
 
 boolean IsDictAvailable(FcitxKeyboard* keyboard)
 {
-    if (keyboard->dict)
-        return true;
-#ifdef PRESAGE_FOUND
-    if (keyboard->config.bUsePresage && keyboard->presage && strcmp(keyboard->dictLang, "en") == 0) {
-        return true;
-    }
-#endif
-    return false;
+    return InvokeVaArgs(keyboard->owner, FCITX_SPELL, DICT_AVAILABLE,
+                        keyboard->dictLang, NULL);
 }
 
+/* a little bit funny here LOL.... */
+static const FcitxHotkey FCITX_HYPHEN_APOS[2] = {
+    {NULL, FcitxKey_minus, FcitxKeyState_None},
+    {NULL, FcitxKey_apostrophe, FcitxKeyState_None}
+};
+
 INPUT_RETURN_VALUE FcitxKeyboardDoInput(void *arg, FcitxKeySym sym, unsigned int state)
 {
     FcitxKeyboardLayout* layout = (FcitxKeyboardLayout*) arg;
     FcitxKeyboard* keyboard = layout->owner;
     FcitxInstance* instance = keyboard->owner;
-    const char* currentLang = FcitxInstanceGetContextString(instance, CONTEXT_IM_LANGUAGE);
+    const char* currentLang = FcitxInstanceGetContextString(instance,
+                                                            CONTEXT_IM_LANGUAGE);
     if (currentLang == NULL)
         currentLang = "";
 
     if (sym == FcitxKey_Shift_L || sym == FcitxKey_Shift_R
         || sym == FcitxKey_Alt_L || sym == FcitxKey_Alt_R
-        || sym == FcitxKey_Control_L || sym == FcitxKey_Control_R)
-    {
+        || sym == FcitxKey_Control_L || sym == FcitxKey_Control_R
+        || sym == FcitxKey_Super_L || sym == FcitxKey_Super_R
+    )
         return IRV_TO_PROCESS;
-    }
 
     uint32_t result = processCompose(layout, sym, state);
     if (result == INVALID_COMPOSE_RESULT)
@@ -485,50 +460,47 @@ INPUT_RETURN_VALUE FcitxKeyboardDoInput(void *arg, FcitxKeySym sym, unsigned int
         state = 0;
     }
 
-    /* performance seems not trouble here, so we only keep one dict */
-    if (layout->owner->config.bEnableWordHint && strcmp(keyboard->dictLang, currentLang) != 0) {
-        if (keyboard->dict) {
-            enchant_broker_free_dict(keyboard->broker, keyboard->dict);
-            keyboard->dict = NULL;
-        }
-
+    FcitxInputContext* currentIC = FcitxInstanceGetCurrentIC(instance);
+    void* enableWordHint = FcitxInstanceGetICData(instance, currentIC,
+                                                  keyboard->dataSlot);
+    if (enableWordHint && strcmp(keyboard->dictLang, currentLang) != 0)
         strncpy(keyboard->dictLang, currentLang, LANGCODE_LENGTH);
-        keyboard->dict = enchant_broker_request_dict(keyboard->broker, keyboard->dictLang);
-    }
 
-    if (IsDictAvailable(keyboard) && layout->owner->config.bEnableWordHint) {
-        FcitxInputState *input = FcitxInstanceGetInputState(layout->owner->owner);
-        struct _FcitxCandidateWordList* candList = FcitxInputStateGetCandidateList(input);
+    /* dict is set and loaded as a side effect */
+    if (enableWordHint && IsDictAvailable(keyboard)) {
+        FcitxInputState *input = FcitxInstanceGetInputState(instance);
 
-        if (FcitxHotkeyIsHotKey(sym, state, layout->owner->config.hkAddToUserDict))
-        {
-            size_t len = strlen(keyboard->buffer);
-            if (len != 0) {
-                enchant_dict_add_to_personal(keyboard->dict, keyboard->buffer, len);
+        if (FcitxHotkeyIsHotKey(sym, state,
+                                keyboard->config.hkAddToUserDict)) {
+            if (InvokeVaArgs(instance, FCITX_SPELL, ADD_PERSONAL,
+                             keyboard->buffer, keyboard->dictLang))
                 return IRV_DO_NOTHING;
-            }
         }
 
-        if (IsValidChar(result) || FcitxHotkeyIsHotKeySimple(sym, state) || IsValidSym(sym, state))
-        {
-            if (IsValidChar(result) || FcitxHotkeyIsHotKeyLAZ(sym, state) || FcitxHotkeyIsHotKeyUAZ(sym, state) || IsValidSym(sym, state)) {
+        if (IsValidChar(result) || FcitxHotkeyIsHotKeySimple(sym, state) ||
+            IsValidSym(sym, state)) {
+            if (IsValidChar(result) || FcitxHotkeyIsHotKeyLAZ(sym, state) ||
+                FcitxHotkeyIsHotKeyUAZ(sym, state) || IsValidSym(sym, state) ||
+                FcitxHotkeyIsHotKey(sym, state, FCITX_HYPHEN_APOS)) {
                 char buf[UTF8_MAX_LENGTH + 1];
                 memset(buf, 0, sizeof(buf));
                 if (result)
-                    Ucs4ToUtf8(layout->owner->iconv, result, buf);
+                    Ucs4ToUtf8(keyboard->iconv, result, buf);
                 else
-                    Ucs4ToUtf8(layout->owner->iconv, FcitxKeySymToUnicode(sym), buf);
+                    Ucs4ToUtf8(keyboard->iconv, FcitxKeySymToUnicode(sym), buf);
                 size_t charlen = strlen(buf);
 
                 if (strlen(keyboard->buffer) >= FCITX_KEYBOARD_MAX_BUFFER) {
-                    FcitxInstanceCommitString(instance, FcitxInstanceGetCurrentIC(instance), keyboard->buffer);
+                    FcitxInstanceCommitString(instance, currentIC,
+                                              keyboard->buffer);
                     keyboard->cursorPos = 0;
                     keyboard->buffer[0] = '\0';
                 }
                 size_t len = strlen(keyboard->buffer);
-                if (keyboard->buffer[keyboard->cursorPos] != 0)
-                {
-                    memmove(keyboard->buffer + keyboard->cursorPos + charlen, keyboard->buffer + keyboard->cursorPos, len - keyboard->cursorPos);
+                if (keyboard->buffer[keyboard->cursorPos] != 0) {
+                    memmove(keyboard->buffer + keyboard->cursorPos + charlen,
+                            keyboard->buffer + keyboard->cursorPos,
+                            len - keyboard->cursorPos);
                 }
                 keyboard->buffer[len + charlen] = 0;
                 strncpy(&keyboard->buffer[keyboard->cursorPos], buf, charlen);
@@ -536,183 +508,123 @@ INPUT_RETURN_VALUE FcitxKeyboardDoInput(void *arg, FcitxKeySym sym, unsigned int
 
                 return IRV_DISPLAY_CANDWORDS;
             }
-            if (FcitxHotkeyIsHotKeyDigit(sym, state) && FcitxCandidateWordGetListSize(candList) != 0) {
-                return IRV_TO_PROCESS;
-            }
-        }
-        else {
-            if (FcitxHotkeyIsHotKey(sym, state, FCITX_BACKSPACE))
-            {
+        } else {
+            if (FcitxHotkeyIsHotKey(sym, state, FCITX_BACKSPACE)) {
                 size_t slen = strlen(keyboard->buffer);
-                if (slen > 0)
-                {
+                if (slen > 0) {
                     if (keyboard->cursorPos > 0) {
                         size_t len = fcitx_utf8_strlen(keyboard->buffer);
-                        char* pos = fcitx_utf8_get_nth_char(keyboard->buffer, len - 1);
+                        char *pos = fcitx_utf8_get_nth_char(keyboard->buffer,
+                                                            len - 1);
                         keyboard->cursorPos = pos - keyboard->buffer;
-                        memset(keyboard->buffer + keyboard->cursorPos, 0, sizeof(char) * (slen - keyboard->cursorPos));
+                        memset(keyboard->buffer + keyboard->cursorPos,
+                               0, sizeof(char) * (slen - keyboard->cursorPos));
                         return IRV_DISPLAY_CANDWORDS;
-                    }
-                    else
+                    } else {
                         return IRV_DO_NOTHING;
+                    }
                 }
             }
         }
 
+        FcitxCandidateWordList* candList = FcitxInputStateGetCandidateList(input);
+        if (FcitxCandidateWordGetListSize(candList)) {
+            int index = FcitxCandidateWordCheckChooseKey(candList, sym, state);
+            if (index >= 0)
+                return IRV_TO_PROCESS;
+        }
+
         if (strlen(keyboard->buffer) > 0) {
             INPUT_RETURN_VALUE irv = IRV_FLAG_FORWARD_KEY;
-            if (layout->owner->config.bUseEnterToCommit && FcitxHotkeyIsHotKey(sym, state, FCITX_ENTER)) {
+            if (keyboard->config.bUseEnterToCommit &&
+                FcitxHotkeyIsHotKey(sym, state, FCITX_ENTER)) {
                 irv = 0;
             }
-            strcpy(FcitxInputStateGetOutputString(input), keyboard->buffer);
+            FcitxInstanceCommitString(instance, currentIC, keyboard->buffer);
 
             if (result) {
-                FcitxInputState *input = FcitxInstanceGetInputState(layout->owner->owner);
                 char buf[UTF8_MAX_LENGTH + 1];
                 memset(buf, 0, sizeof(buf));
-                Ucs4ToUtf8(layout->owner->iconv, result, buf);
-                strcat(FcitxInputStateGetOutputString(input), buf);
+                Ucs4ToUtf8(keyboard->iconv, result, buf);
+                FcitxInstanceCommitString(instance, currentIC, buf);
                 irv = 0;
             }
-            irv |= IRV_COMMIT_STRING;
+            irv |= IRV_FLAG_UPDATE_INPUT_WINDOW | IRV_FLAG_RESET_INPUT;
             return irv;
         }
-    }
-    else {
+    } else {
         FcitxUICloseInputWindow(instance);
     }
     if (result) {
-        FcitxInputState *input = FcitxInstanceGetInputState(layout->owner->owner);
         char buf[UTF8_MAX_LENGTH + 1];
         memset(buf, 0, sizeof(buf));
-        Ucs4ToUtf8(layout->owner->iconv, result, buf);
-        strcpy(FcitxInputStateGetOutputString(input), buf);
-        return IRV_COMMIT_STRING;
+        Ucs4ToUtf8(keyboard->iconv, result, buf);
+        FcitxInstanceCommitString(instance, currentIC, buf);
+        return IRV_FLAG_UPDATE_INPUT_WINDOW | IRV_FLAG_RESET_INPUT;
     }
     return IRV_TO_PROCESS;
 }
 
+static INPUT_RETURN_VALUE
+FcitxKeyboardGetCandWordCb(void *arg, const char *commit)
+{
+    FcitxKeyboardLayout *layout = (FcitxKeyboardLayout*) arg;
+    FcitxKeyboard *keyboard = layout->owner;
+    FcitxInstance *instance = keyboard->owner;
+    char str[strlen(commit) + 2];
+    strcpy(str, commit);
+    if (keyboard->config.bCommitWithExtraSpace)
+        strcat(str, " ");
+    FcitxInstanceCommitString(instance,
+                              FcitxInstanceGetCurrentIC(instance), str);
+    return IRV_FLAG_RESET_INPUT;
+}
+
 INPUT_RETURN_VALUE FcitxKeyboardGetCandWords(void* arg)
 {
     FcitxKeyboardLayout* layout = (FcitxKeyboardLayout*) arg;
     FcitxKeyboard* keyboard = layout->owner;
-    FcitxInstance* instance = layout->owner->owner;
+    FcitxInstance* instance = keyboard->owner;
     FcitxInputState* input = FcitxInstanceGetInputState(instance);
-    FcitxGlobalConfig* config = FcitxInstanceGetGlobalConfig(instance);
     if (keyboard->buffer[0] == '\0')
         return IRV_CLEAN;
-    FcitxCandidateWordSetPageSize(FcitxInputStateGetCandidateList(input), config->iMaxCandWord);
-    FcitxCandidateWordSetChoose(FcitxInputStateGetCandidateList(input), DIGIT_STR_CHOOSE);
-    size_t bufferlen = strlen(keyboard->buffer);
+
+    unsigned int cmodtable[] = {FcitxKeyState_None, FcitxKeyState_Alt,
+                                FcitxKeyState_Ctrl, FcitxKeyState_Shift};
+    if (keyboard->config.chooseModifier > CM_CTRL)
+        keyboard->config.chooseModifier = CM_CTRL;
+    FcitxCandidateWordSetPageSize(FcitxInputStateGetCandidateList(input),
+                                  keyboard->config.maximumHintLength);
+    FcitxCandidateWordSetChooseAndModifier(
+        FcitxInputStateGetCandidateList(input), DIGIT_STR_CHOOSE,
+        cmodtable[keyboard->config.chooseModifier]);
+    ssize_t bufferlen = strlen(keyboard->buffer);
     strcpy(FcitxInputStateGetRawInputBuffer(input), keyboard->buffer);
     FcitxInputStateSetRawInputBufferSize(input, bufferlen);
-    FcitxMessagesAddMessageAtLast(FcitxInputStateGetClientPreedit(input), MSG_INPUT, "%s", keyboard->buffer);
-    FcitxMessagesAddMessageAtLast(FcitxInputStateGetPreedit(input), MSG_INPUT, "%s", keyboard->buffer);
+    FcitxMessagesAddMessageAtLast(FcitxInputStateGetClientPreedit(input),
+                                  MSG_INPUT, "%s", keyboard->buffer);
+    FcitxMessagesAddMessageAtLast(FcitxInputStateGetPreedit(input), MSG_INPUT,
+                                  "%s", keyboard->buffer);
     FcitxInputStateSetClientCursorPos(input, keyboard->cursorPos);
     FcitxInputStateSetCursorPos(input, keyboard->cursorPos);
 
-    char **suggestions = NULL;
+    if (bufferlen < keyboard->config.minimumHintLength)
+        return IRV_DISPLAY_CANDWORDS;
 
-#ifdef PRESAGE_FOUND
-    if (keyboard->config.bUsePresage && keyboard->presage && strcmp(keyboard->dictLang, "en") == 0) {
-        do {
-            char buf[20];
-            sprintf(buf, "%d", config->iMaxCandWord);
-            presage_config_set(keyboard->presage, "Presage.Selector.SUGGESTIONS", buf);
-            presage_error_code_t error;
-            error = presage_predict(keyboard->presage, &suggestions);
-            if (error != PRESAGE_OK) {
-                break;
-            }
-            if (suggestions) {
-                int i = 0;
-                while(suggestions[i]) {
-                    FcitxCandidateWord candWord;
-                    candWord.callback = FcitxKeyboardGetPresageCandWord;
-                    candWord.owner = layout;
-                    candWord.priv = NULL;
-                    candWord.strExtra = NULL;
-                    candWord.strWord = strdup(suggestions[i]);
-                    candWord.wordType = MSG_OTHER;
-
-                    FcitxCandidateWordAppend(FcitxInputStateGetCandidateList(input), &candWord);
-                    i++;
-                }
-                presage_free_string_array(suggestions);
-            }
-        } while(0);
-    }
-    else
-#endif
-    {
-        if (bufferlen < keyboard->config.minimumHintLength)
-            return IRV_DISPLAY_CANDWORDS;
-        size_t number = 0;
-        int count = 0;
-        suggestions = enchant_dict_suggest(keyboard->dict, keyboard->buffer, bufferlen, &number);
-        int i;
-        for (i = 0 ;i < number && count < config->iMaxCandWord; i ++)
-        {
-            FcitxCandidateWord candWord;
-            candWord.callback = FcitxKeyboardGetCandWord;
-            candWord.owner = layout;
-            candWord.priv = NULL;
-            candWord.strExtra = NULL;
-            candWord.strWord = strdup(suggestions[i]);
-            candWord.wordType = MSG_OTHER;
-            FcitxCandidateWordAppend(FcitxInputStateGetCandidateList(input), &candWord);
-            count ++;
-        }
-
-        if (suggestions && number)
-            enchant_dict_free_string_list(keyboard->dict, suggestions);
-    }
+    FcitxCandidateWordList *candList;
+    candList = InvokeVaArgs(instance, FCITX_SPELL, GET_CANDWORDS,
+                            NULL, keyboard->buffer, NULL,
+                            (void*)(long)keyboard->config.maximumHintLength,
+                            keyboard->dictLang, NULL,
+                            FcitxKeyboardGetCandWordCb, layout);
+    if (!candList)
+        return IRV_DISPLAY_CANDWORDS;
+    FcitxCandidateWordMerge(FcitxInputStateGetCandidateList(input),
+                            candList, -1);
+    FcitxCandidateWordFreeList(candList);
     return IRV_DISPLAY_CANDWORDS;
 }
 
-INPUT_RETURN_VALUE FcitxKeyboardGetCandWord (void* arg, FcitxCandidateWord* candWord)
-{
-    FcitxKeyboardLayout* layout = (FcitxKeyboardLayout*) arg;
-    FcitxInstance* instance = layout->owner->owner;
-    FcitxInputState* input = FcitxInstanceGetInputState(instance);
-    char* strGet = FcitxInputStateGetOutputString(input);
-    strncpy(strGet, candWord->strWord, MAX_USER_INPUT);
-    strGet[MAX_USER_INPUT] = '\0';
-    if (layout->owner->config.bCommitWithExtraSpace && strlen(strGet) < MAX_USER_INPUT)
-        strcat(strGet, " ");
-    return IRV_COMMIT_STRING;
-}
-
-#ifdef PRESAGE_FOUND
-INPUT_RETURN_VALUE FcitxKeyboardGetPresageCandWord(void* arg, FcitxCandidateWord* candWord)
-{
-    FcitxKeyboardLayout* layout = (FcitxKeyboardLayout*) arg;
-    FcitxKeyboard* keyboard = layout->owner;
-    char* result = NULL;
-    INPUT_RETURN_VALUE irv = IRV_DO_NOTHING;
-    do {
-        presage_error_code_t error = presage_completion(keyboard->presage, candWord->strWord, &result);
-        if (error != PRESAGE_OK)
-            break;
-        FcitxInputState* input = FcitxInstanceGetInputState(keyboard->owner);
-        char* strGet = FcitxInputStateGetOutputString(input);
-        char* nr = fcitx_utils_trim(result);
-        snprintf(strGet, MAX_USER_INPUT, "%s%s", keyboard->buffer, nr);
-        free(nr);
-        if (layout->owner->config.bCommitWithExtraSpace && strlen(strGet) < MAX_USER_INPUT)
-            strcat(strGet, " ");
-
-        irv = IRV_COMMIT_STRING;
-    } while(0);
-
-    if (result)
-        free(result);
-
-    return irv;
-}
-#endif
-
-
 void  FcitxKeyboardSave(void *arg)
 {
 }
@@ -759,20 +671,26 @@ INPUT_RETURN_VALUE FcitxKeyboardHotkeyToggleWordHint(void* arg)
 {
     FcitxKeyboard* keyboard = (FcitxKeyboard*) arg;
     FcitxIM* im = FcitxInstanceGetCurrentIM(keyboard->owner);
-    if (im && strncmp(im->uniqueName, "fcitx-keyboard", strlen("fcitx-keyboard")) == 0) {
-        keyboard->config.bEnableWordHint = !keyboard->config.bEnableWordHint;
-        if (!keyboard->config.bEnableWordHint) {
-            keyboard->dictLang[0] = 0;
-            if (keyboard->dict) {
-                enchant_broker_free_dict(keyboard->broker, keyboard->dict);
-                keyboard->dict = NULL;
-            }
-        }
-        SaveKeyboardConfig(&keyboard->config);
-        return IRV_DO_NOTHING;
-    }
-    else
+    FcitxInputContext* currentIC = FcitxInstanceGetCurrentIC(keyboard->owner);
+    if (!currentIC)
         return IRV_TO_PROCESS;
+    if (im && strncmp(im->uniqueName, "fcitx-keyboard",
+                      strlen("fcitx-keyboard")) == 0) {
+        void* enableWordHint = FcitxInstanceGetICData(keyboard->owner,
+                                                      currentIC,
+                                                      keyboard->dataSlot);
+        if (!enableWordHint) {
+            enableWordHint = PTR_TRUE;
+            IsDictAvailable(keyboard);
+        } else {
+            enableWordHint = PTR_FALSE;
+        }
+        FcitxInstanceSetICData(keyboard->owner, currentIC,
+                               keyboard->dataSlot, enableWordHint);
+        return IRV_DO_NOTHING;
+    } else {
+        return IRV_TO_PROCESS;
+    }
 }
 
 static const uint32_t validSym[] =
@@ -2683,6 +2601,7 @@ checkCompactTable(FcitxKeyboardLayout* layout, const FcitxComposeTableCompact *t
 uint32_t
 checkAlgorithmically(FcitxKeyboardLayout* layout)
 {
+#if defined(ENABLE_ICU)
     int i;
     UChar combination_buffer[FCITX_MAX_COMPOSE_LEN];
     FcitxKeyboard* keyboard = layout->owner;
@@ -2702,37 +2621,36 @@ checkAlgorithmically(FcitxKeyboardLayout* layout)
             switch (keyboard->composeBuffer[i]) {
 #define CASE(keysym, unicode) \
 case FcitxKey_dead_##keysym: combination_buffer[i + 1] = unicode; break
-                CASE(grave, 0x0300);
-                CASE(acute, 0x0301);
-                CASE(circumflex, 0x0302);
-                CASE(tilde, 0x0303);  /* Also used with perispomeni, 0x342. */
-                CASE(macron, 0x0304);
-                CASE(breve, 0x0306);
-                CASE(abovedot, 0x0307);
-                CASE(diaeresis, 0x0308);
-                CASE(hook, 0x0309);
-                CASE(abovering, 0x030A);
-                CASE(doubleacute, 0x030B);
-                CASE(caron, 0x030C);
-                CASE(abovecomma, 0x0313);  /* Equivalent to psili */
-                CASE(abovereversedcomma, 0x0314);  /* Equivalent to dasia */
-                CASE(horn, 0x031B);  /* Legacy use for psili, 0x313 (or 0x343). */
-                CASE(belowdot, 0x0323);
-                CASE(cedilla, 0x0327);
-                CASE(ogonek, 0x0328);  /* Legacy use for dasia, 0x314.*/
-                CASE(iota, 0x0345);
-                CASE(voiced_sound, 0x3099);  /* Per Markus Kuhn keysyms.txt file. */
-                CASE(semivoiced_sound, 0x309A);  /* Per Markus Kuhn keysyms.txt file. */
-                /* The following cases are to be removed once xkeyboard-config,
-                * xorg are fully updated.
-                **/
+            CASE(grave, 0x0300);
+            CASE(acute, 0x0301);
+            CASE(circumflex, 0x0302);
+            CASE(tilde, 0x0303);   /* Also used with perispomeni, 0x342. */
+            CASE(macron, 0x0304);
+            CASE(breve, 0x0306);
+            CASE(abovedot, 0x0307);
+            CASE(diaeresis, 0x0308);
+            CASE(hook, 0x0309);
+            CASE(abovering, 0x030A);
+            CASE(doubleacute, 0x030B);
+            CASE(caron, 0x030C);
+            CASE(abovecomma, 0x0313);         /* Equivalent to psili */
+            CASE(abovereversedcomma, 0x0314); /* Equivalent to dasia */
+            CASE(horn, 0x031B);    /* Legacy use for psili, 0x313 (or 0x343). */
+            CASE(belowdot, 0x0323);
+            CASE(cedilla, 0x0327);
+            CASE(ogonek, 0x0328);  /* Legacy use for dasia, 0x314.*/
+            CASE(iota, 0x0345);
+            CASE(voiced_sound, 0x3099);    /* Per Markus Kuhn keysyms.txt file. */
+            CASE(semivoiced_sound, 0x309A);    /* Per Markus Kuhn keysyms.txt file. */
+
+            /* The following cases are to be removed once xkeyboard-config,
+            * xorg are fully updated.
+            */
                 /* Workaround for typo in 1.4.x xserver-xorg */
-            case 0xfe66:
-                combination_buffer[i + 1] = 0x314;
-                break;
-                /* CASE (dasia, 0x314); */
-                /* CASE (perispomeni, 0x342); */
-                /* CASE (psili, 0x343); */
+            case 0xfe66: combination_buffer[i+1] = 0x314; break;
+            /* CASE(dasia, 0x314); */
+            /* CASE(perispomeni, 0x342); */
+            /* CASE(psili, 0x343); */
 #undef CASE
             default:
                 combination_buffer[i + 1] = FcitxKeySymToUnicode((FcitxKeySym) keyboard->composeBuffer[i]);
@@ -2773,5 +2691,6 @@ case FcitxKey_dead_##keysym: combination_buffer[i + 1] = unicode; break
             return result[0];
         }
     }
+#endif
     return 0;
 }
