@@ -24,62 +24,54 @@
 #include "config.h"
 
 #include <iconv.h>
-#include <enchant/enchant.h>
+
 #include <fcitx/ime.h>
 #include <fcitx-utils/utarray.h>
 #include <fcitx-config/fcitx-config.h>
-#ifdef PRESAGE_FOUND
-#include <presage.h>
-#endif
 #include "rules.h"
 
 #define FCITX_KEYBOARD_MAX_BUFFER 20
 #define FCITX_MAX_COMPOSE_LEN 7
 
-typedef enum _EnchantProvider {
-    EP_Default = 0,
-    EP_Aspell = 1,
-    EP_Myspell = 2
-} EnchantProvider;
+typedef enum _ChooseModifier {
+    CM_NONE,
+    CM_ALT,
+    CM_CTRL,
+    CM_SHIFT,
+} ChooseModifier;
 
 typedef struct _FcitxKeyboardConfig {
     FcitxGenericConfig gconfig;
     boolean bCommitWithExtraSpace;
-    boolean bEnableWordHint;
     boolean bUseEnterToCommit;
-    boolean bUsePresage;
     FcitxHotkey hkToggleWordHint[2];
     FcitxHotkey hkAddToUserDict[2];
     int minimumHintLength;
-    EnchantProvider provider;
+    int maximumHintLength;
+    ChooseModifier chooseModifier;
 } FcitxKeyboardConfig;
 
 typedef struct _FcitxKeyboard {
     struct _FcitxInstance* owner;
-    EnchantBroker* broker;
-    UT_array* enchantLanguages;
+    char dictLang[6];
     FcitxKeyboardConfig config;
     FcitxXkbRules* rules;
-    char dictLang[6];
-    EnchantDict* dict;
     iconv_t iconv;
-    char* initialLayout;
-    char* initialVariant;
-#ifdef PRESAGE_FOUND
-    presage_t presage;
-#endif
+    char *initialLayout;
+    char *initialVariant;
     char buffer[FCITX_KEYBOARD_MAX_BUFFER + UTF8_MAX_LENGTH + 1];
     int cursorPos;
     uint composeBuffer[FCITX_MAX_COMPOSE_LEN + 1];
     int n_compose;
-    char* tempBuffer;
+    char *tempBuffer;
     int lastLength;
+    int dataSlot;
 } FcitxKeyboard;
 
 typedef struct _FcitxKeyboardLayout {
-    FcitxKeyboard* owner;
-    char* layoutString;
-    char* variantString;
+    FcitxKeyboard *owner;
+    char *layoutString;
+    char *variantString;
 } FcitxKeyboardLayout;
 
 CONFIG_BINDING_DECLARE(FcitxKeyboardConfig);
