@@ -26,13 +26,13 @@
 #include <string.h>
 #include <stdlib.h>
 #include <libintl.h>
+#include <errno.h>
 
 #include "fcitx/fcitx.h"
 #include "fcitx/keys.h"
 #include "fcitx/ui.h"
 #include "fcitx/profile.h"
 #include "fcitx-utils/log.h"
-#include "fcitx/instance.h"
 #include "fcitx/module.h"
 #include "fcitx/frontend.h"
 #include "fcitx/candidate.h"
@@ -46,10 +46,9 @@
 
 #define MAX_TABLE_INPUT 50
 
-static void FreeTableConfig(void *v);
-const UT_icd table_icd = {sizeof(TableMetaData), NULL , NULL, FreeTableConfig};
+static void TableMetaDataFree(TableMetaData *table);
 const UT_icd tableCand_icd = {sizeof(TABLECANDWORD*), NULL, NULL, NULL };
-typedef struct _TableCandWordSortContext {
+typedef struct {
     ADJUSTORDER order;
 } TableCandWordSortContext;
 
@@ -57,42 +56,69 @@ static void *TableCreate(FcitxInstance* instance);
 static int TableCandCmp(const void* a, const void* b, void* arg);
 static INPUT_RETURN_VALUE TableKeyBlocker(void* arg, FcitxKeySym sym, unsigned int state);
 
-FCITX_EXPORT_API
-FcitxIMClass ime = {
+FCITX_DEFINE_PLUGIN(fcitx_table, ime2, FcitxIMClass2) =  {
     TableCreate,
+    NULL,
+    ReloadTableConfig,
+    NULL,
+    NULL,
+    NULL,
+    NULL,
     NULL
 };
 
-FCITX_EXPORT_API
-int ABI_VERSION = FCITX_ABI_VERSION;
+static boolean LoadTableConfig(TableConfig* config);
+static void SaveTableConfig(TableConfig* config);
+
+static inline unsigned int GetTableMod(TableMetaData* table)
+{
+    unsigned int cmodtable[] = {FcitxKeyState_None, FcitxKeyState_Alt, FcitxKeyState_Ctrl, FcitxKeyState_Shift};
+    if (table->chooseModifier > CM_CTRL)
+        table->chooseModifier = CM_CTRL;
+
+    return cmodtable[table->chooseModifier];
+}
+
+boolean LoadTableConfig(TableConfig* config)
+{
+    FcitxConfigFileDesc* configDesc = GetTableGlobalConfigDesc();
+    if (configDesc == NULL)
+        return false;
+
+    FILE *fp;
+    char *file;
+    fp = FcitxXDGGetFileUserWithPrefix("conf", "fcitx-table.config", "r", &file);
+    FcitxLog(DEBUG, "Load Config File %s", file);
+    free(file);
+    if (!fp) {
+        if (errno == ENOENT)
+            SaveTableConfig(config);
+    }
+
+    FcitxConfigFile *cfile = FcitxConfigParseConfigFileFp(fp, configDesc);
+
+    TableConfigConfigBind(config, cfile, configDesc);
+    FcitxConfigBindSync((FcitxGenericConfig*)config);
+
+    if (fp)
+        fclose(fp);
+
+    return true;
+}
+
+void SaveTableConfig(TableConfig* config)
+{
+}
 
 void *TableCreate(FcitxInstance* instance)
 {
     FcitxTableState *tbl = fcitx_utils_malloc0(sizeof(FcitxTableState));
     tbl->owner = instance;
-    LoadTableInfo(tbl);
-    TableMetaData* table;
-    for (table = (TableMetaData*) utarray_front(tbl->table);
-            table != NULL;
-            table = (TableMetaData*) utarray_next(tbl->table, table)) {
-        FcitxInstanceRegisterIM(
-            instance,
-            table,
-            (strlen(table->uniqueName) == 0) ? table->strIconName : table->uniqueName,
-            table->strName,
-            table->strIconName,
-            TableInit,
-            TableResetStatus,
-            DoTableInput,
-            TableGetCandWords,
-            TablePhraseTips,
-            SaveTableIM,
-            ReloadTableConfig,
-            TableKeyBlocker,
-            table->iPriority,
-            table->langCode
-        );
+    if (!LoadTableConfig(&tbl->config)) {
+        free(tbl);
+        return NULL;
     }
+    LoadTableInfo(tbl);
 
 
     return tbl;
@@ -108,11 +134,74 @@ void SaveTableIM(void *arg)
         SaveTableDict(table);
 }
 
+inline char* TableMetaDataGetName(TableMetaData* table)
+{
+    return (strlen(table->uniqueName) == 0) ? table->strIconName : table->uniqueName;
+}
+
+inline void TableMetaDataInsert(TableMetaData** tableSet, TableMetaData* table)
+{
+    char* name = TableMetaDataGetName(table);
+    HASH_ADD_KEYPTR(hh, *tableSet, name, strlen(name), table);
+}
+
+inline TableMetaData* TableMetaDataFind(TableMetaData* table, const char* name)
+{
+    TableMetaData* result = NULL;
+    HASH_FIND_STR(table, name, result);
+    return result;
+}
+
+inline void TableMetaDataUnlink(TableMetaData** tableSet, TableMetaData* table)
+{
+    HASH_DEL(*tableSet, table);
+}
+
+inline void TableMetaDataRemove(TableMetaData** tableSet, TableMetaData* table)
+{
+    HASH_DEL(*tableSet, table);
+    TableMetaDataFree(table);
+}
+
+inline void TableMetaDataRegister(FcitxTableState* tbl, TableMetaData* table)
+{
+    table->status = TABLE_REGISTERED;
+    FcitxInstanceRegisterIM(
+        tbl->owner,
+        table,
+        TableMetaDataGetName(table),
+        table->strName,
+        table->strIconName,
+        TableInit,
+        TableResetStatus,
+        DoTableInput,
+        TableGetCandWords,
+        TablePhraseTips,
+        SaveTableIM,
+        NULL,
+        TableKeyBlocker,
+        table->iPriority,
+        table->langCode
+    );
+}
+
+inline char* TableConfigStealTableName(FcitxConfigFile* cfile)
+{
+    FcitxConfigOption* option = FcitxConfigFileGetOption(cfile, "CodeTable", "UniqueName");
+    if (option && strlen(option->rawValue) != 0)
+        return option->rawValue;
+
+    option = FcitxConfigFileGetOption(cfile, "CodeTable", "IconName");
+    if (option)
+        return option->rawValue;
+    return NULL;
+}
 
 /*
- * 读取码表输入法的名称和文件路径
+ * Read table configuration
+ * and returns whether table im is really changed.
  */
-void LoadTableInfo(FcitxTableState *tbl)
+boolean LoadTableInfo(FcitxTableState *tbl)
 {
     char **tablePath;
     size_t len;
@@ -120,54 +209,69 @@ void LoadTableInfo(FcitxTableState *tbl)
 
     FcitxStringHashSet* sset = NULL;
     tbl->bTablePhraseTips = false;
-    tbl->iCurrentTableLoaded = -1;
-
-    if (tbl->table) {
-        utarray_free(tbl->table);
-        tbl->table = NULL;
+    if (tbl->curLoadedTable) {
+        FreeTableDict(tbl->curLoadedTable);
+        tbl->curLoadedTable = NULL;
     }
 
-    tbl->hkTableDelPhrase[0].sym = FcitxKey_7;
-    tbl->hkTableDelPhrase[0].state = FcitxKeyState_Ctrl;
-    tbl->hkTableAdjustOrder[0].sym = FcitxKey_6;
-    tbl->hkTableAdjustOrder[0].state = FcitxKeyState_Ctrl;
-    tbl->hkTableAddPhrase[0].sym = FcitxKey_8;
-    tbl->hkTableAddPhrase[0].state = FcitxKeyState_Ctrl;
-
-    tbl->iTableCount = 0;
-    utarray_new(tbl->table, &table_icd);
-
     tablePath = FcitxXDGGetPathWithPrefix(&len, "table");
     sset = FcitxXDGGetFiles("table", NULL, ".conf");
 
+    {
+        TableMetaData* titer = tbl->tables;
+        while(titer)
+        {
+            titer->status = TABLE_PENDING;
+            titer = titer->hh.next;
+        }
+    }
+
+    boolean imchanged = false;
     char **paths = fcitx_utils_malloc0(sizeof(char*) * len);
     for (i = 0; i < len ; i ++)
         paths[i] = NULL;
-    FcitxStringHashSet* string;
-    for (string = sset;
-            string != NULL;
-            string = (FcitxStringHashSet*)string->hh.next) {
+
+    HASH_FOREACH(string, sset, FcitxStringHashSet) {
         int i = 0;
         for (i = len - 1; i >= 0; i--) {
             asprintf(&paths[i], "%s/%s", tablePath[len - i - 1], string->name);
             FcitxLog(DEBUG, "Load Table Config File:%s", paths[i]);
         }
-        FcitxLog(INFO, _("Load Table Config File:%s"), string->name);
+        // FcitxLog(INFO, _("Load Table Config File:%s"), string->name);
         FcitxConfigFile* cfile = FcitxConfigParseMultiConfigFile(paths, len, GetTableConfigDesc());
         if (cfile) {
-            utarray_extend_back(tbl->table);
-            TableMetaData *t = (TableMetaData*)utarray_back(tbl->table);
-            TableMetaDataConfigBind(t, cfile, GetTableConfigDesc());
-            FcitxConfigBindSync((FcitxGenericConfig*)t);
-            FcitxLog(DEBUG, _("Table Config %s is %s"), string->name, (t->bEnabled) ? "Enabled" : "Disabled");
-            if (t->bEnabled) {
-                t->confName = strdup(string->name);
-                t->owner = tbl;
-                tbl->iTableCount ++;
-            }
-            else {
-                utarray_pop_back(tbl->table);
-            }
+            do {
+                char* tableName = TableConfigStealTableName(cfile);
+                boolean needunregister = true;
+                if (!tableName)
+                    break;
+                TableMetaData* t = TableMetaDataFind(tbl->tables, tableName);
+                if (!t) {
+                    t = fcitx_utils_new(TableMetaData);
+                    needunregister = false;
+                }
+                else {
+                    TableMetaDataUnlink(&tbl->tables, t);
+                }
+                TableMetaDataConfigBind(t, cfile, GetTableConfigDesc());
+                FcitxConfigBindSync((FcitxGenericConfig*)t);
+                if (t->bEnabled) {
+                    t->confName = strdup(string->name);
+                    t->owner = tbl;
+                    TableMetaDataInsert(&tbl->tables, t);
+                    if (needunregister)
+                        t->status = TABLE_REGISTERED;
+                    else
+                        t->status = TABLE_NEW;
+                }
+                else {
+                    if (needunregister) {
+                        FcitxInstanceUnregisterIM(tbl->owner, TableMetaDataGetName(t));
+                        imchanged = true;
+                    }
+                    TableMetaDataFree(t);
+                }
+            } while(0);
         }
 
         for (i = 0; i < len ; i ++) {
@@ -179,9 +283,39 @@ void LoadTableInfo(FcitxTableState *tbl)
     free(paths);
     FcitxXDGFreePath(tablePath);
     fcitx_utils_free_string_hash_set(sset);
+
+    {
+        TableMetaData* titer = tbl->tables;
+        while(titer)
+        {
+            /*
+             * if it's this case, the configuration file may already gone
+             * thus let's remove it
+             */
+            if (titer->status == TABLE_PENDING) {
+                TableMetaData* cur = titer;
+                FcitxInstanceUnregisterIM(tbl->owner, TableMetaDataGetName(cur));
+                TableMetaDataRemove(&tbl->tables, cur);
+                imchanged = true;
+            }
+            else {
+                if (titer->status != TABLE_REGISTERED) {
+                    // FcitxLog(INFO, "register %s", TableMetaDataGetName(titer));
+                    TableMetaDataRegister(tbl, titer);
+                    imchanged = true;
+                }
+                titer = titer->hh.next;
+            }
+        }
+    }
+
+    tbl->iTableCount = HASH_COUNT(tbl->tables);
+
+    return imchanged;
 }
 
 CONFIG_DESC_DEFINE(GetTableConfigDesc, "table.desc")
+CONFIG_DESC_DEFINE(GetTableGlobalConfigDesc, "fcitx-table.desc")
 
 boolean TableInit(void *arg)
 {
@@ -196,8 +330,6 @@ boolean TableInit(void *arg)
     }
     FcitxAddon* pyaddon = FcitxAddonsGetAddonByName(FcitxInstanceGetAddons(tbl->owner), "fcitx-pinyin");
     tbl->pyaddon = pyaddon;
-    if (pyaddon == NULL)
-        return false;
     tbl->PYBaseOrder = AD_FREQ;
 
     Table_ResetPY(tbl->pyaddon);
@@ -211,11 +343,25 @@ void TableResetStatus(void* arg)
     FcitxInputState *input = FcitxInstanceGetInputState(tbl->owner);
     tbl->bIsTableAddPhrase = false;
     tbl->bIsTableDelPhrase = false;
+    tbl->bIsTableClearFreq = false;
     tbl->bIsTableAdjustOrder = false;
     FcitxInputStateSetIsDoInputOnly(input, false);
     //bSingleHZMode = false;
 }
 
+boolean TableCheckNoMatch(TableMetaData* table, const char* code)
+{
+    FcitxInstance *instance = table->owner->owner;
+    FcitxInputState *input = FcitxInstanceGetInputState(instance);
+    FcitxCandidateWordList* candList = FcitxInputStateGetCandidateList(input);
+    if (!table->bTableExactMatch) {
+        return FcitxCandidateWordGetListSize(candList) == 0;
+    }
+    else {
+        return (FcitxCandidateWordGetListSize(candList) == 0) && TableFindFirstMatchCode(table, code, false, false) == -1;
+    }
+}
+
 INPUT_RETURN_VALUE DoTableInput(void* arg, FcitxKeySym sym, unsigned int state)
 {
     TableMetaData* table = (TableMetaData*) arg;
@@ -227,20 +373,21 @@ INPUT_RETURN_VALUE DoTableInput(void* arg, FcitxKeySym sym, unsigned int state)
     char* strCodeInput = FcitxInputStateGetRawInputBuffer(input);
     FcitxCandidateWordList* candList = FcitxInputStateGetCandidateList(input);
 
-    FcitxCandidateWordSetChoose(candList, table->strChoose);
+    FcitxCandidateWordSetChooseAndModifier(candList, table->strChoose, GetTableMod(table));
     FcitxCandidateWordSetPageSize(candList, config->iMaxCandWord);
 
-    tbl->iTableIMIndex = utarray_eltidx(tbl->table, table);
-    if (tbl->iTableIMIndex != tbl->iCurrentTableLoaded) {
-        TableMetaData* previousTable = (TableMetaData*) utarray_eltptr(tbl->table, tbl->iCurrentTableLoaded);
-        if (previousTable)
-            FreeTableDict(previousTable);
-        tbl->iCurrentTableLoaded = -1;
+    if (table != tbl->curLoadedTable && tbl->curLoadedTable) {
+        FreeTableDict(tbl->curLoadedTable);
+        tbl->curLoadedTable = NULL;
     }
 
-    if (tbl->iCurrentTableLoaded == -1) {
-        LoadTableDict(table);
-        tbl->iCurrentTableLoaded = tbl->iTableIMIndex;
+    if (tbl->curLoadedTable == NULL) {
+        if (!LoadTableDict(table)) {
+            FcitxInstanceUnregisterIM(instance, TableMetaDataGetName(table));
+            FcitxInstanceUpdateIMList(instance);
+            return IRV_DONOT_PROCESS;
+        }
+        tbl->curLoadedTable = table;
     }
 
     if (FcitxHotkeyIsHotKeyModifierCombine(sym, state))
@@ -272,7 +419,18 @@ INPUT_RETURN_VALUE DoTableInput(void* arg, FcitxKeySym sym, unsigned int state)
         FcitxInputStateSetIsInRemind(input, false);
 
         /* it's not in special state */
-        if (!tbl->bIsTableAddPhrase && !tbl->bIsTableDelPhrase && !tbl->bIsTableAdjustOrder) {
+        if (!tbl->bIsTableAddPhrase && !tbl->bIsTableDelPhrase && !tbl->bIsTableAdjustOrder && !tbl->bIsTableClearFreq) {
+            if (FcitxHotkeyCheckChooseKeyAndModifier(sym, state, table->strChoose, GetTableMod(table)) >= 0) {
+                char *buf1 = strndup(strCodeInput, FcitxInputStateGetRawInputBufferSize(input));
+                char* buf = NULL;
+                asprintf(&buf, "%s%c", buf1, (char) sym);
+                boolean result = (TableFindFirstMatchCode(table, buf, false, false) == -1);
+                free(buf);
+                free(buf1);
+                if (result)
+                    return IRV_TO_PROCESS;
+            }
+
             /* check we use Pinyin or Not */
             if (strCodeInput[0] == table->cPinyin && table->bUsePY) {
                 if (FcitxInputStateGetRawInputBufferSize(input) != (MAX_PY_LENGTH * 5 + 1)) {
@@ -286,7 +444,7 @@ INPUT_RETURN_VALUE DoTableInput(void* arg, FcitxKeySym sym, unsigned int state)
                 /* length is not too large */
                 if (((FcitxInputStateGetRawInputBufferSize(input) < table->tableDict->iCodeLength)
                     || (table->tableDict->bHasPinyin && FcitxInputStateGetRawInputBufferSize(input) < table->tableDict->iPYCodeLength)
-                    || (((FcitxCandidateWordPageCount(candList) == 0
+                    || (((TableCheckNoMatch(table, FcitxInputStateGetRawInputBuffer(input))
                         && table->bNoMatchDontCommit) || !table->bUseAutoSend)
                         && FcitxInputStateGetRawInputBufferSize(input) >= table->tableDict->iCodeLength
                     ))
@@ -320,11 +478,10 @@ INPUT_RETURN_VALUE DoTableInput(void* arg, FcitxKeySym sym, unsigned int state)
                         }
 
                         retVal = TableGetCandWords(table);
-                        FcitxModuleFunctionArg farg;
                         int key = FcitxInputStateGetRawInputBuffer(input)[0];
-                        farg.args[0] = &key;
                         if (!table->bIgnorePunc)
-                            strTemp = InvokeFunction(instance, FCITX_PUNC, GETPUNC, farg);
+                            strTemp = InvokeVaArgs(instance, FCITX_PUNC,
+                                                   GETPUNC, &key);
                         else
                             strTemp = NULL;
                         if (IsEndKey(table, sym)) {
@@ -346,7 +503,7 @@ INPUT_RETURN_VALUE DoTableInput(void* arg, FcitxKeySym sym, unsigned int state)
                                    && table->iTableAutoSendToClientWhenNone
                                    && (!(retVal & IRV_FLAG_PENDING_COMMIT_STRING))
                                    && (FcitxInputStateGetRawInputBufferSize(input) >= (table->iTableAutoSendToClientWhenNone + 1))
-                                   && FcitxCandidateWordPageCount(candList) == 0) {
+                                   && TableCheckNoMatch(table, FcitxInputStateGetRawInputBuffer(input))) {
                             if (strLastFirstCand && (lastFirstCandType != CT_AUTOPHRASE)) {
                                 FcitxInstanceCommitString(instance, FcitxInstanceGetCurrentIC(instance), strLastFirstCand);
                             } else if (table->bSendRawPreedit) {
@@ -417,7 +574,7 @@ INPUT_RETURN_VALUE DoTableInput(void* arg, FcitxKeySym sym, unsigned int state)
             }
             return IRV_DISPLAY_MESSAGE;
         }
-        if (FcitxHotkeyIsHotKey(sym, state, tbl->hkTableAddPhrase)) {
+        if (FcitxHotkeyIsHotKey(sym, state, tbl->config.hkTableAddPhrase)) {
             if (!tbl->bIsTableAddPhrase) {
                 if (table->tableDict->iHZLastInputCount < 2 || !table->tableDict->bRule) //词组最少为两个汉字
                     return IRV_DO_NOTHING;
@@ -440,7 +597,7 @@ INPUT_RETURN_VALUE DoTableInput(void* arg, FcitxKeySym sym, unsigned int state)
                 retVal = IRV_TO_PROCESS;
 
             return retVal;
-        } else if (FcitxHotkeyIsHotKey(sym, state, FCITX_CTRL_ALT_E)) {
+        } else if (FcitxHotkeyIsHotKey(sym, state, FCITX_CTRL_ALT_E) && tbl->pyaddon) {
             char            strPY[100];
 
             //如果拼音单字字库没有读入，则读入它
@@ -467,7 +624,7 @@ INPUT_RETURN_VALUE DoTableInput(void* arg, FcitxKeySym sym, unsigned int state)
             return IRV_TO_PROCESS;
 
         if (FcitxHotkeyIsHotKey(sym, state, FCITX_ESCAPE)) {
-            if (tbl->bIsTableDelPhrase || tbl->bIsTableAdjustOrder) {
+            if (tbl->bIsTableDelPhrase || tbl->bIsTableAdjustOrder || tbl->bIsTableClearFreq) {
                 TableResetStatus(table);
                 FcitxInstanceCleanInputWindowUp(instance);
                 FcitxMessagesAddMessageAtLast(FcitxInputStateGetPreedit(input), MSG_INPUT, "%s", FcitxInputStateGetRawInputBuffer(input));
@@ -492,13 +649,18 @@ INPUT_RETURN_VALUE DoTableInput(void* arg, FcitxKeySym sym, unsigned int state)
                     retVal = IRV_DISPLAY_CANDWORDS;
                 } else if (candWord->owner == table && tbl->bIsTableAdjustOrder) {
                     TableAdjustOrderByIndex(table, candWord->priv);
-                    tbl->bIsTableDelPhrase = false;
+                    tbl->bIsTableAdjustOrder = false;
                     retVal = IRV_DISPLAY_CANDWORDS;
-                } else
+                } else if (candWord->owner == table && tbl->bIsTableClearFreq) {
+                    TableClearFreqByIndex(table, candWord->priv);
+                    tbl->bIsTableClearFreq = false;
+                    retVal = IRV_DISPLAY_CANDWORDS;
+                }
+                else
                     return IRV_TO_PROCESS;
             }
-        } else if (!tbl->bIsTableDelPhrase && !tbl->bIsTableAdjustOrder) {
-            if (FcitxHotkeyIsHotKey(sym, state, tbl->hkTableAdjustOrder)) {
+        } else if (!tbl->bIsTableDelPhrase && !tbl->bIsTableAdjustOrder && !tbl->bIsTableClearFreq) {
+            if (FcitxHotkeyIsHotKey(sym, state, tbl->config.hkTableAdjustOrder)) {
                 if (FcitxCandidateWordGetListSize(candList) < 2 || FcitxInputStateGetIsInRemind(input))
                     return IRV_DO_NOTHING;
 
@@ -506,7 +668,7 @@ INPUT_RETURN_VALUE DoTableInput(void* arg, FcitxKeySym sym, unsigned int state)
                 FcitxInstanceCleanInputWindowUp(instance);
                 FcitxMessagesAddMessageAtLast(FcitxInputStateGetAuxUp(input), MSG_TIPS, _("Choose the phrase to be put in the front, Press Escape to Cancel"));
                 retVal = IRV_DISPLAY_MESSAGE;
-            } else if (FcitxHotkeyIsHotKey(sym, state, tbl->hkTableDelPhrase)) {
+            } else if (FcitxHotkeyIsHotKey(sym, state, tbl->config.hkTableDelPhrase)) {
                 if (!FcitxCandidateWordPageCount(candList) || FcitxInputStateGetIsInRemind(input))
                     return IRV_DO_NOTHING;
 
@@ -514,6 +676,14 @@ INPUT_RETURN_VALUE DoTableInput(void* arg, FcitxKeySym sym, unsigned int state)
                 FcitxInstanceCleanInputWindowUp(instance);
                 FcitxMessagesAddMessageAtLast(FcitxInputStateGetAuxUp(input), MSG_TIPS, _("Choose the phrase to be deleted, Press Escape to Cancel"));
                 retVal = IRV_DISPLAY_MESSAGE;
+            } else if (FcitxHotkeyIsHotKey(sym, state, tbl->config.hkTableClearFreq)) {
+                if (!FcitxCandidateWordPageCount(candList) || FcitxInputStateGetIsInRemind(input))
+                    return IRV_DO_NOTHING;
+
+                tbl->bIsTableClearFreq = true;
+                FcitxInstanceCleanInputWindowUp(instance);
+                FcitxMessagesAddMessageAtLast(FcitxInputStateGetAuxUp(input), MSG_TIPS, _("Choose the phrase to clear typing history, Press Escape to Cancel"));
+                retVal = IRV_DISPLAY_MESSAGE;
             } else if (FcitxHotkeyIsHotKey(sym, state, FCITX_BACKSPACE)) {
                 if (!FcitxInputStateGetRawInputBufferSize(input)) {
                     FcitxInputStateSetIsInRemind(input, false);
@@ -556,11 +726,11 @@ INPUT_RETURN_VALUE DoTableInput(void* arg, FcitxKeySym sym, unsigned int state)
     }
 
     if (!FcitxInputStateGetIsInRemind(input)) {
-        if (tbl->bIsTableDelPhrase || tbl->bIsTableAdjustOrder)
+        if (tbl->bIsTableDelPhrase || tbl->bIsTableAdjustOrder || tbl->bIsTableClearFreq)
             FcitxInputStateSetIsDoInputOnly(input, true);
     }
 
-    if (tbl->bIsTableDelPhrase || tbl->bIsTableAdjustOrder || FcitxInputStateGetIsInRemind(input))
+    if (tbl->bIsTableDelPhrase || tbl->bIsTableAdjustOrder || tbl->bIsTableClearFreq || FcitxInputStateGetIsInRemind(input))
         FcitxInputStateSetShowCursor(input, false);
     else {
         FcitxInputStateSetShowCursor(input, true);
@@ -601,9 +771,7 @@ INPUT_RETURN_VALUE _TableGetCandWord(TableMetaData* table, TABLECANDWORD* tableC
     FcitxInputState *input = FcitxInstanceGetInputState(instance);
     FcitxProfile *profile = FcitxInstanceGetProfile(instance);
 
-    if (table->strSymbol
-        && strlen(table->strSymbol) > 0
-        && strcmp(FcitxInputStateGetRawInputBuffer(input), table->strSymbol) == 0)
+    if (tableCandWord->flag == CT_FH)
         return TableGetFHCandWord(table, tableCandWord);
 
     FcitxInputStateSetIsInRemind(input, false);
@@ -680,6 +848,10 @@ INPUT_RETURN_VALUE TableGetPinyinCandWords(TableMetaData* table)
 {
     FcitxTableState* tbl = table->owner;
     FcitxInstance *instance = tbl->owner;
+
+    if (!tbl->pyaddon)
+        return IRV_DISPLAY_CANDWORDS;
+
     FcitxInputState *input = FcitxInstanceGetInputState(instance);
 
     strcpy(Table_PYGetFindString(tbl->pyaddon), FcitxInputStateGetRawInputBuffer(input) + 1);
@@ -746,10 +918,10 @@ INPUT_RETURN_VALUE TableGetCandWords(void* arg)
     if (!strcmp(FcitxInputStateGetRawInputBuffer(input), table->strSymbol))
         return TableGetFHCandWords(table);
 
-    if (FcitxInputStateGetRawInputBuffer(input)[0] == table->cPinyin && table->bUsePY)
+    if (FcitxInputStateGetRawInputBuffer(input)[0] == table->cPinyin && table->bUsePY && tbl->pyaddon)
         return TableGetPinyinCandWords(table);
 
-    if (TableFindFirstMatchCode(table, FcitxInputStateGetRawInputBuffer(input)) == -1 && !table->tableDict->iAutoPhrase) {
+    if (TableFindFirstMatchCode(table, FcitxInputStateGetRawInputBuffer(input), table->bTableExactMatch, true) == -1 && !table->tableDict->iAutoPhrase) {
         if (FcitxInputStateGetRawInputBufferSize(input)) {
             FcitxMessagesSetMessageCount(FcitxInputStateGetPreedit(input), 0);
             FcitxMessagesSetMessageCount(FcitxInputStateGetClientPreedit(input), 0);
@@ -773,7 +945,7 @@ INPUT_RETURN_VALUE TableGetCandWords(void* arg)
     while (table->tableDict->currentRecord && table->tableDict->currentRecord != table->tableDict->recordHead) {
         if (table->tableDict->currentRecord->type != RECORDTYPE_CONSTRUCT &&
             table->tableDict->currentRecord->type != RECORDTYPE_PROMPT &&
-            !TableCompareCode(table, FcitxInputStateGetRawInputBuffer(input), table->tableDict->currentRecord->strCode)) {
+            !TableCompareCode(table, FcitxInputStateGetRawInputBuffer(input), table->tableDict->currentRecord->strCode, table->bTableExactMatch)) {
             TABLECANDWORD* tableCandWord = fcitx_utils_malloc0(sizeof(TABLECANDWORD));
             TableAddCandWord(table->tableDict->currentRecord, tableCandWord);
             utarray_push_back(&candTemp, &tableCandWord);
@@ -826,8 +998,9 @@ INPUT_RETURN_VALUE TableGetCandWords(void* arg)
                         totallen += 1;
                 }
 
-                candWord.strExtra = fcitx_utils_malloc0(sizeof(char) * (totallen + 2));
-                candWord.strExtra[0] = ' ';
+                candWord.strExtra = fcitx_utils_malloc0(sizeof(char) * (totallen + 1 + 3));
+                if (codelen)
+                    strcpy(candWord.strExtra, "\xef\xbd\x9e");
                 for (i = 0; i < codelen; i ++) {
                     RECORD* rec = table->tableDict->promptCode[(uint8_t) pstr[i]];
                     if (rec) {
@@ -850,7 +1023,7 @@ INPUT_RETURN_VALUE TableGetCandWords(void* arg)
 
     if (table->tableDict->bRule && table->bAutoPhrase && FcitxInputStateGetRawInputBufferSize(input) == table->tableDict->iCodeLength) {
         for (i = table->tableDict->iAutoPhrase - 1; i >= 0; i--) {
-            if (!TableCompareCode(table, FcitxInputStateGetRawInputBuffer(input), table->tableDict->autoPhrase[i].strCode)) {
+            if (!TableCompareCode(table, FcitxInputStateGetRawInputBuffer(input), table->tableDict->autoPhrase[i].strCode, table->bTableExactMatch)) {
                 if (TableHasPhrase(table->tableDict, table->tableDict->autoPhrase[i].strCode, table->tableDict->autoPhrase[i].strHZ)) {
                     TABLECANDWORD* tableCandWord = fcitx_utils_malloc0(sizeof(TABLECANDWORD));
                     TableAddAutoCandWord(table, i, tableCandWord);
@@ -926,6 +1099,16 @@ void TableAddCandWord(RECORD * record, TABLECANDWORD* tableCandWord)
     tableCandWord->candWord.record = record;
 }
 
+void TableClearFreqByIndex(TableMetaData* table, TABLECANDWORD* tableCandWord)
+{
+    RECORD         *recTemp;
+
+    recTemp = tableCandWord->candWord.record;
+    recTemp->iHit = 0;
+
+    table->tableDict->iTableChanged++;
+}
+
 /*
  * 根据序号调整词组顺序，序号从1开始
  * 将指定的字/词调整到同样编码的最前面
@@ -1115,7 +1298,6 @@ INPUT_RETURN_VALUE TableGetFHCandWord(TableMetaData* table, TABLECANDWORD* table
     FcitxTableState* tbl = table->owner;
     FcitxInstance *instance = tbl->owner;
     FcitxInputState *input = FcitxInstanceGetInputState(instance);
-    FcitxInstanceCleanInputWindowDown(instance);
 
     strcpy(FcitxInputStateGetOutputString(input), table->tableDict->fh[tableCandWord->candWord.iFHIndex].strFH);
     return IRV_COMMIT_STRING;
@@ -1192,23 +1374,13 @@ void UpdateHZLastInput(TableMetaData* table, char *str)
         TableCreateAutoPhrase(table, (char)(fcitx_utf8_strlen(str)));
 }
 
-void FreeTableConfig(void *v)
+void TableMetaDataFree(TableMetaData *table)
 {
-    TableMetaData *table = (TableMetaData*) v;
     if (!table)
         return;
-    FcitxConfigFreeConfigFile(table->config.configFile);
-    FcitxHotkeyFree(table->hkAlternativePrevPage);
-    FcitxHotkeyFree(table->hkAlternativeNextPage);
-    free(table->strPath);
-    free(table->strSymbolFile);
-    free(table->uniqueName);
-    free(table->strName);
-    free(table->strIconName);
-    free(table->strEndCode);
-    free(table->strSymbol);
-    free(table->strChoose);
+    FcitxConfigFree(&table->config);
     fcitx_utils_free(table->confName);
+    free(table);
 }
 
 int TableCandCmp(const void* a, const void* b, void *arg)
@@ -1275,34 +1447,12 @@ INPUT_RETURN_VALUE TableKeyBlocker(void* arg, FcitxKeySym sym, unsigned int stat
     return FcitxStandardKeyBlocker(input, sym, state);
 }
 
-/* we should try not to break the existing on even if the config file is not exist anymore */
 void ReloadTableConfig(void* arg)
 {
-    TableMetaData* table = arg;
-    size_t len = 0;
-    int i = 0;
-    char** tablePath = FcitxXDGGetPathWithPrefix(&len, "table");
-
-    char **paths = fcitx_utils_malloc0(sizeof(char*) * len);
-    for (i = 0; i < len ; i ++)
-        paths[i] = NULL;
-
-    for (i = len - 1; i >= 0; i--) {
-        asprintf(&paths[i], "%s/%s", tablePath[len - i - 1], table->confName);
-        FcitxLog(DEBUG, "Load Table Config File:%s", paths[i]);
-    }
-    FcitxConfigFile* cfile = FcitxConfigParseMultiConfigFile(paths, len, GetTableConfigDesc());
-    if (cfile) {
-        TableMetaDataConfigBind(table, cfile, GetTableConfigDesc());
-        FcitxConfigBindSync((FcitxGenericConfig*)table);
-    }
-    for (i = 0; i < len ; i ++) {
-        free(paths[i]);
-        paths[i] = NULL;
-    }
-
-    free(paths);
-    FcitxXDGFreePath(tablePath);
+    FcitxTableState* tbl = arg;
+    LoadTableConfig(&tbl->config);
+    if (LoadTableInfo(tbl))
+        FcitxInstanceUpdateIMList(tbl->owner);
 }
 
 // kate: indent-mode cstyle; space-indent on; indent-width 0;
