@@ -13,6 +13,17 @@
 #define TABLE_TEMP_FILE "table_XXXXXX"
 const int iInternalVersion = INTERNAL_VERSION;
 
+void UpdateTableMetaData(TableMetaData* tableMetaData)
+{
+    if (!tableMetaData->tableDict)
+        return;
+    TableDict* tableDict = tableMetaData->tableDict;
+    if (tableMetaData->iTableAutoSendToClient == -1)
+        tableMetaData->iTableAutoSendToClient = tableDict->iCodeLength;
+    if (tableMetaData->iTableAutoSendToClientWhenNone == -1)
+        tableMetaData->iTableAutoSendToClientWhenNone = tableDict->iCodeLength;
+}
+
 boolean LoadTableDict(TableMetaData* tableMetaData)
 {
     char            strCode[MAX_CODE_LENGTH + 1];
@@ -20,159 +31,217 @@ boolean LoadTableDict(TableMetaData* tableMetaData)
     FILE           *fpDict;
     RECORD         *recTemp;
     unsigned int    i = 0;
-    unsigned int    iTemp, iTempCount;
+    uint32_t        iTemp, iTempCount;
     char            cChar = 0, cTemp;
-    char            iVersion = 1;
+    int8_t          iVersion = 1;
     int             iRecordIndex;
-    char           *pstr;
     TableDict      *tableDict;
-    tableMetaData->tableDict = fcitx_utils_malloc0(sizeof(TableDict));
-    tableDict = tableMetaData->tableDict;
-    tableDict->pool = fcitx_memory_pool_create();
 
     //读入码表
     FcitxLog(DEBUG, _("Loading Table Dict"));
 
-    fpDict = FcitxXDGGetFileWithPrefix("table", tableMetaData->strPath, "r", &pstr);
-    FcitxLog(INFO, _("Load Table Dict from %s"), pstr);
-    free(pstr);
-
-    //先读取码表的信息
-    //判断版本信息
-    fread(&iTemp, sizeof(unsigned int), 1, fpDict);
-    if (!iTemp) {
-        fread(&iVersion, sizeof(char), 1, fpDict);
-        iVersion = (iVersion < INTERNAL_VERSION);
-        fread(&iTemp, sizeof(unsigned int), 1, fpDict);
-    }
-
-    tableDict->strInputCode = (char *) realloc(tableDict->strInputCode, sizeof(char) * (iTemp + 1));
-    fread(tableDict->strInputCode, sizeof(char), iTemp + 1, fpDict);
-    /*
-     * 建立索引，加26是为了为拼音编码预留空间
-     */
-    tableDict->recordIndex = (RECORD_INDEX *) fcitx_memory_pool_alloc(tableDict->pool, (strlen(tableDict->strInputCode) + 26) * sizeof(RECORD_INDEX));
-    for (iTemp = 0; iTemp < strlen(tableDict->strInputCode) + 26; iTemp++) {
-        tableDict->recordIndex[iTemp].cCode = 0;
-        tableDict->recordIndex[iTemp].record = NULL;
-    }
-    /* ********************************************************************** */
-
-    fread(&(tableDict->iCodeLength), sizeof(unsigned char), 1, fpDict);
-    if (tableMetaData->iTableAutoSendToClient == -1)
-        tableMetaData->iTableAutoSendToClient = tableDict->iCodeLength;
-    if (tableMetaData->iTableAutoSendToClientWhenNone == -1)
-        tableMetaData->iTableAutoSendToClientWhenNone = tableDict->iCodeLength;
-
-    if (!iVersion)
-        fread(&(tableDict->iPYCodeLength), sizeof(unsigned char), 1, fpDict);
-    else
-        tableDict->iPYCodeLength = tableDict->iCodeLength;
-
-    fread(&iTemp, sizeof(unsigned int), 1, fpDict);
-    tableDict->strIgnoreChars = (char *) fcitx_memory_pool_alloc(tableDict->pool, sizeof(char) * (iTemp + 1));
-    fread(tableDict->strIgnoreChars, sizeof(char), iTemp + 1, fpDict);
-
-    fread(&(tableDict->bRule), sizeof(unsigned char), 1, fpDict);
-
-    if (tableDict->bRule) { //表示有组词规则
-        tableDict->rule = (RULE *) fcitx_memory_pool_alloc(tableDict->pool, sizeof(RULE) * (tableDict->iCodeLength - 1));
-        for (i = 0; i < tableDict->iCodeLength - 1; i++) {
-            fread(&(tableDict->rule[i].iFlag), sizeof(unsigned char), 1, fpDict);
-            fread(&(tableDict->rule[i].iWords), sizeof(unsigned char), 1, fpDict);
-            tableDict->rule[i].rule = (RULE_RULE *) fcitx_memory_pool_alloc(tableDict->pool, sizeof(RULE_RULE) * tableDict->iCodeLength);
-            for (iTemp = 0; iTemp < tableDict->iCodeLength; iTemp++) {
-                fread(&(tableDict->rule[i].rule[iTemp].iFlag), sizeof(unsigned char), 1, fpDict);
-                fread(&(tableDict->rule[i].rule[iTemp].iWhich), sizeof(unsigned char), 1, fpDict);
-                fread(&(tableDict->rule[i].rule[iTemp].iIndex), sizeof(unsigned char), 1, fpDict);
-            }
+    int reload = 0;
+    do {
+        boolean error = false;
+        if (!reload)
+            fpDict = FcitxXDGGetFileWithPrefix("table", tableMetaData->strPath, "r", NULL);
+        else {
+            char* temppath;
+            asprintf(&temppath, "table/%s", tableMetaData->strPath);
+            char* tablepath = fcitx_utils_get_fcitx_path_with_filename("pkgdatadir", temppath);
+            free(temppath);
+            fpDict = fopen(tablepath, "r");
+            free(tablepath);
         }
-    }
+        if (!fpDict)
+            return false;
 
-    tableDict->recordHead = (RECORD *) fcitx_memory_pool_alloc(tableDict->pool, sizeof(RECORD));
-    tableDict->currentRecord = tableDict->recordHead;
+        tableMetaData->tableDict = fcitx_utils_new(TableDict);
 
-    fread(&(tableDict->iRecordCount), sizeof(unsigned int), 1, fpDict);
+        tableDict = tableMetaData->tableDict;
+        tableDict->pool = fcitx_memory_pool_create();
+#define CHECK_LOAD_TABLE_ERROR(SIZE) if (size < (SIZE)) { error = true; goto table_load_error; }
 
-    for (i = 0; i < SINGLE_HZ_COUNT; i++)
-    {
-        tableDict->tableSingleHZ[i] = (RECORD *) NULL;
-        tableDict->tableSingleHZCons[i] = (RECORD *) NULL;
-    }
+        //先读取码表的信息
+        //判断版本信息
+        size_t size;
+        size = fcitx_utils_read_uint32(fpDict, &iTemp);
+        CHECK_LOAD_TABLE_ERROR(1);
 
-    iRecordIndex = 0;
-    size_t bufSize = 0;
-    for (i = 0; i < tableDict->iRecordCount; i++) {
-        fread(strCode, sizeof(char), tableDict->iPYCodeLength + 1, fpDict);
-        fread(&iTemp, sizeof(unsigned int), 1, fpDict);
-        if (iTemp > bufSize) {
-            bufSize = iTemp;
-            strHZ = realloc(strHZ, bufSize);
+        if (!iTemp) {
+            size = fread(&iVersion, sizeof(int8_t), 1, fpDict);
+            CHECK_LOAD_TABLE_ERROR(1);
+            iVersion = (iVersion < INTERNAL_VERSION);
+            size = fcitx_utils_read_uint32(fpDict, &iTemp);
+            CHECK_LOAD_TABLE_ERROR(1);
         }
-        fread(strHZ, sizeof(char), iTemp, fpDict);
-        recTemp = (RECORD *) fcitx_memory_pool_alloc(tableDict->pool, sizeof(RECORD));
-        recTemp->strCode = (char *) fcitx_memory_pool_alloc(tableDict->pool, sizeof(char) * (tableDict->iPYCodeLength + 1));
-        memset(recTemp->strCode, 0, sizeof(char) * (tableDict->iPYCodeLength + 1));
-        strcpy(recTemp->strCode, strCode);
-        recTemp->strHZ = (char *) fcitx_memory_pool_alloc(tableDict->pool, sizeof(char) * iTemp);
-        strcpy(recTemp->strHZ, strHZ);
+
+        tableDict->strInputCode = (char*)realloc(tableDict->strInputCode, sizeof(char) * (iTemp + 1));
+        size = fread(tableDict->strInputCode, sizeof(char), iTemp + 1, fpDict);
+        CHECK_LOAD_TABLE_ERROR(iTemp + 1);
+        /*
+         * 建立索引，加26是为了为拼音编码预留空间
+         */
+        tableDict->recordIndex = (RECORD_INDEX*)fcitx_memory_pool_alloc(tableDict->pool, (strlen(tableDict->strInputCode) + 26) * sizeof(RECORD_INDEX));
+        for (iTemp = 0; iTemp < strlen(tableDict->strInputCode) + 26; iTemp++) {
+            tableDict->recordIndex[iTemp].cCode = 0;
+            tableDict->recordIndex[iTemp].record = NULL;
+        }
+        /* ********************************************************************** */
+
+        size = fread(&(tableDict->iCodeLength), sizeof(uint8_t), 1, fpDict);
+        CHECK_LOAD_TABLE_ERROR(1);
+        UpdateTableMetaData(tableMetaData);
 
         if (!iVersion) {
-            fread(&cTemp, sizeof(int8_t), 1, fpDict);
-            recTemp->type = cTemp;
+            size = fread(&(tableDict->iPYCodeLength), sizeof(uint8_t), 1, fpDict);
+            CHECK_LOAD_TABLE_ERROR(1);
         }
+        else
+            tableDict->iPYCodeLength = tableDict->iCodeLength;
 
-        fread(&(recTemp->iHit), sizeof(unsigned int), 1, fpDict);
-        fread(&(recTemp->iIndex), sizeof(unsigned int), 1, fpDict);
-        if (recTemp->iIndex > tableDict->iTableIndex)
-            tableDict->iTableIndex = recTemp->iIndex;
+        size = fcitx_utils_read_uint32(fpDict, &iTemp);
+        CHECK_LOAD_TABLE_ERROR(1);
+        tableDict->strIgnoreChars = (char*)fcitx_memory_pool_alloc(tableDict->pool, sizeof(char) * (iTemp + 1));
+        size = fread(tableDict->strIgnoreChars, sizeof(char), iTemp + 1, fpDict);
+        CHECK_LOAD_TABLE_ERROR(iTemp + 1);
 
-        /* 建立索引 */
-        if (cChar != recTemp->strCode[0]) {
-            cChar = recTemp->strCode[0];
-            tableDict->recordIndex[iRecordIndex].cCode = cChar;
-            tableDict->recordIndex[iRecordIndex].record = recTemp;
-            iRecordIndex++;
-        }
-        /* **************************************************************** */
-        /** 为单字生成一个表   */
-        if (fcitx_utf8_strlen(recTemp->strHZ) == 1 && !IsIgnoreChar(tableDict, strCode[0]))
-        {
-            RECORD** tableSingleHZ = NULL;
-            if (recTemp->type == RECORDTYPE_NORMAL)
-                tableSingleHZ = tableDict->tableSingleHZ;
-            else if (recTemp->type == RECORDTYPE_CONSTRUCT)
-                tableSingleHZ = tableDict->tableSingleHZCons;
+        size = fread(&(tableDict->bRule), sizeof(unsigned char), 1, fpDict);
+        CHECK_LOAD_TABLE_ERROR(1);
 
-            if (tableSingleHZ) {
-                iTemp = CalHZIndex(recTemp->strHZ);
-                if (iTemp < SINGLE_HZ_COUNT) {
-                    if (tableSingleHZ[iTemp]) {
-                        if (strlen(strCode) > strlen(tableDict->tableSingleHZ[iTemp]->strCode))
-                            tableSingleHZ[iTemp] = recTemp;
-                    } else
-                        tableSingleHZ[iTemp] = recTemp;
+        if (tableDict->bRule) { //表示有组词规则
+            tableDict->rule = (RULE*)fcitx_memory_pool_alloc(tableDict->pool, sizeof(RULE) * (tableDict->iCodeLength - 1));
+            for (i = 0; i < tableDict->iCodeLength - 1; i++) {
+                size = fread(&(tableDict->rule[i].iFlag), sizeof(unsigned char), 1, fpDict);
+                CHECK_LOAD_TABLE_ERROR(1);
+                size = fread(&(tableDict->rule[i].iWords), sizeof(unsigned char), 1, fpDict);
+                CHECK_LOAD_TABLE_ERROR(1);
+                tableDict->rule[i].rule = (RULE_RULE*)fcitx_memory_pool_alloc(tableDict->pool, sizeof(RULE_RULE) * tableDict->iCodeLength);
+                for (iTemp = 0; iTemp < tableDict->iCodeLength; iTemp++) {
+                    size = fread(&(tableDict->rule[i].rule[iTemp].iFlag), sizeof(unsigned char), 1, fpDict);
+                    CHECK_LOAD_TABLE_ERROR(1);
+                    size = fread(&(tableDict->rule[i].rule[iTemp].iWhich), sizeof(unsigned char), 1, fpDict);
+                    CHECK_LOAD_TABLE_ERROR(1);
+                    size = fread(&(tableDict->rule[i].rule[iTemp].iIndex), sizeof(unsigned char), 1, fpDict);
+                    CHECK_LOAD_TABLE_ERROR(1);
                 }
             }
         }
 
-        if (recTemp->type == RECORDTYPE_PINYIN)
-            tableDict->bHasPinyin = true;
+        tableDict->recordHead = (RECORD*)fcitx_memory_pool_alloc(tableDict->pool, sizeof(RECORD));
+        tableDict->currentRecord = tableDict->recordHead;
 
-        if (recTemp->type == RECORDTYPE_PROMPT && strlen(recTemp->strCode) == 1)
-            tableDict->promptCode[(uint8_t) recTemp->strCode[0]] = recTemp;
+        size = fcitx_utils_read_uint32(fpDict, &tableDict->iRecordCount);
+        CHECK_LOAD_TABLE_ERROR(1);
 
-        tableDict->currentRecord->next = recTemp;
-        recTemp->prev = tableDict->currentRecord;
-        tableDict->currentRecord = recTemp;
-    }
-    if (strHZ)
-        free(strHZ);
+        for (i = 0; i < SINGLE_HZ_COUNT; i++) {
+            tableDict->tableSingleHZ[i] = (RECORD*)NULL;
+            tableDict->tableSingleHZCons[i] = (RECORD*)NULL;
+        }
 
-    tableDict->currentRecord->next = tableDict->recordHead;
-    tableDict->recordHead->prev = tableDict->currentRecord;
+        iRecordIndex = 0;
+        size_t bufSize = 0;
+        for (i = 0; i < tableDict->iRecordCount; i++) {
+            size = fread(strCode, sizeof(int8_t), tableDict->iPYCodeLength + 1, fpDict);
+            CHECK_LOAD_TABLE_ERROR(tableDict->iPYCodeLength + 1);
+            size = fcitx_utils_read_uint32(fpDict, &iTemp);
+            CHECK_LOAD_TABLE_ERROR(1);
+            /* we don't actually have such limit, but sometimes, broken table
+             * may break this, so we need to give a limitation.
+             */
+            if (iTemp > UTF8_MAX_LENGTH * 30) {
+                error = true;
+                goto table_load_error;
+            }
+            if (iTemp > bufSize) {
+                bufSize = iTemp;
+                strHZ = realloc(strHZ, bufSize);
+            }
+            size = fread(strHZ, sizeof(int8_t), iTemp, fpDict);
+            CHECK_LOAD_TABLE_ERROR(iTemp);
+            recTemp = (RECORD*)fcitx_memory_pool_alloc(tableDict->pool, sizeof(RECORD));
+            recTemp->strCode = (char*)fcitx_memory_pool_alloc(tableDict->pool, sizeof(char) * (tableDict->iPYCodeLength + 1));
+            memset(recTemp->strCode, 0, sizeof(char) * (tableDict->iPYCodeLength + 1));
+            strcpy(recTemp->strCode, strCode);
+            recTemp->strHZ = (char*)fcitx_memory_pool_alloc(tableDict->pool, sizeof(char) * iTemp);
+            strcpy(recTemp->strHZ, strHZ);
+
+            if (!iVersion) {
+                size = fread(&cTemp, sizeof(int8_t), 1, fpDict);
+                CHECK_LOAD_TABLE_ERROR(1);
+                recTemp->type = cTemp;
+            }
+
+            size = fcitx_utils_read_uint32(fpDict, &recTemp->iHit);
+            CHECK_LOAD_TABLE_ERROR(1);
+            size = fcitx_utils_read_uint32(fpDict, &recTemp->iIndex);
+            CHECK_LOAD_TABLE_ERROR(1);
+            if (recTemp->iIndex > tableDict->iTableIndex)
+                tableDict->iTableIndex = recTemp->iIndex;
+
+            /* 建立索引 */
+            if (cChar != recTemp->strCode[0]) {
+                cChar = recTemp->strCode[0];
+                tableDict->recordIndex[iRecordIndex].cCode = cChar;
+                tableDict->recordIndex[iRecordIndex].record = recTemp;
+                iRecordIndex++;
+            }
+            /* **************************************************************** */
+            /** 为单字生成一个表   */
+            if (fcitx_utf8_strlen(recTemp->strHZ) == 1 && !IsIgnoreChar(tableDict, strCode[0]))
+            {
+                RECORD** tableSingleHZ = NULL;
+                if (recTemp->type == RECORDTYPE_NORMAL)
+                    tableSingleHZ = tableDict->tableSingleHZ;
+                else if (recTemp->type == RECORDTYPE_CONSTRUCT)
+                    tableSingleHZ = tableDict->tableSingleHZCons;
+
+                if (tableSingleHZ) {
+                    iTemp = CalHZIndex(recTemp->strHZ);
+                    if (iTemp < SINGLE_HZ_COUNT) {
+                        if (tableSingleHZ[iTemp]) {
+                            if (strlen(strCode) > strlen(tableDict->tableSingleHZ[iTemp]->strCode))
+                                tableSingleHZ[iTemp] = recTemp;
+                        } else
+                            tableSingleHZ[iTemp] = recTemp;
+                    }
+                }
+            }
+
+            if (recTemp->type == RECORDTYPE_PINYIN)
+                tableDict->bHasPinyin = true;
+
+            if (recTemp->type == RECORDTYPE_PROMPT && strlen(recTemp->strCode) == 1)
+                tableDict->promptCode[(uint8_t) recTemp->strCode[0]] = recTemp;
+
+            tableDict->currentRecord->next = recTemp;
+            recTemp->prev = tableDict->currentRecord;
+            tableDict->currentRecord = recTemp;
+        }
+        if (strHZ) {
+            free(strHZ);
+            strHZ = NULL;
+        }
+
+        tableDict->currentRecord->next = tableDict->recordHead;
+        tableDict->recordHead->prev = tableDict->currentRecord;
+
+table_load_error:
+        fclose(fpDict);
+        if (error) {
+            fcitx_memory_pool_destroy(tableDict->pool);
+            tableDict->pool = NULL;
+            error = false;
+            reload++;
+        }
+        else
+            break;
+
+    } while(reload < 2);
+
+    if (!tableDict->pool)
+        return false;
 
-    fclose(fpDict);
     FcitxLog(DEBUG, _("Load Table Dict OK"));
 
     //读取相应的特殊符号表
@@ -180,24 +249,35 @@ boolean LoadTableDict(TableMetaData* tableMetaData)
 
     if (fpDict) {
         tableDict->iFH = fcitx_utils_calculate_record_number(fpDict);
-        tableDict->fh = (FH *) fcitx_memory_pool_alloc(tableDict->pool, sizeof(FH) * tableDict->iFH);
+        tableDict->fh = (FH*)fcitx_memory_pool_alloc(tableDict->pool, sizeof(FH) * tableDict->iFH);
 
+        char* strBuf = NULL;
+        size_t bufLen = 0;
         for (i = 0; i < tableDict->iFH; i++) {
-            if (EOF == fscanf(fpDict, "%s\n", tableDict->fh[i].strFH))
+            if (getline(&strBuf, &bufLen, fpDict) == -1)
                 break;
+
+            if (!fcitx_utf8_check_string(strBuf))
+                break;
+
+            if (fcitx_utf8_strlen(strBuf) > FH_MAX_LENGTH)
+                break;
+
+            strcpy(tableDict->fh[i].strFH, strBuf);
         }
+        fcitx_utils_free(strBuf);
         tableDict->iFH = i;
 
         fclose(fpDict);
     }
 
-    tableDict->strNewPhraseCode = (char *) fcitx_memory_pool_alloc(tableDict->pool, sizeof(char) * (tableDict->iCodeLength + 1));
+    tableDict->strNewPhraseCode = (char*)fcitx_memory_pool_alloc(tableDict->pool, sizeof(char) * (tableDict->iCodeLength + 1));
     tableDict->strNewPhraseCode[tableDict->iCodeLength] = '\0';
 
     tableDict->iAutoPhrase = 0;
     if (tableMetaData->bAutoPhrase) {
         //为自动词组分配空间
-        tableDict->autoPhrase = (AUTOPHRASE *) fcitx_memory_pool_alloc(tableDict->pool, sizeof(AUTOPHRASE) * AUTO_PHRASE_COUNT);
+        tableDict->autoPhrase = (AUTOPHRASE*)fcitx_memory_pool_alloc(tableDict->pool, sizeof(AUTOPHRASE) * AUTO_PHRASE_COUNT);
 
         //读取上次保存的自动词组信息
         FcitxLog(DEBUG, _("Loading Autophrase."));
@@ -208,26 +288,41 @@ boolean LoadTableDict(TableMetaData* tableMetaData)
         free(temppath);
         i = 0;
         if (fpDict) {
-            fread(&tableDict->iAutoPhrase, sizeof(unsigned int), 1, fpDict);
+            size_t size = fcitx_utils_read_int32(fpDict, &tableDict->iAutoPhrase);
+            if (size == 1) {
+                for (; i < tableDict->iAutoPhrase; i++) {
+                    tableDict->autoPhrase[i].strCode = (char*)fcitx_memory_pool_alloc(tableDict->pool, sizeof(char) * (tableDict->iCodeLength + 1));
+                    tableDict->autoPhrase[i].strHZ = (char*)fcitx_memory_pool_alloc(tableDict->pool, sizeof(char) * (PHRASE_MAX_LENGTH * UTF8_MAX_LENGTH + 1));
+                    size = fread(tableDict->autoPhrase[i].strCode, tableDict->iCodeLength + 1, 1, fpDict);
+                    if (size != 1) {
+                        tableDict->iAutoPhrase = i;
+                        break;
+                    }
+                    size = fread(tableDict->autoPhrase[i].strHZ, PHRASE_MAX_LENGTH * UTF8_MAX_LENGTH + 1, 1, fpDict);
+                    tableDict->autoPhrase[i].strHZ[PHRASE_MAX_LENGTH * UTF8_MAX_LENGTH] = 0;
+                    if (size != 1 || !fcitx_utf8_check_string(tableDict->autoPhrase[i].strHZ)) {
+                        tableDict->iAutoPhrase = i;
+                        break;
+                    }
+                    size = fcitx_utils_read_uint32(fpDict, &iTempCount);
+                    if (size != 1) {
+                        tableDict->iAutoPhrase = i;
+                        break;
+                    }
 
-            for (; i < tableDict->iAutoPhrase; i++) {
-                tableDict->autoPhrase[i].strCode = (char *) fcitx_memory_pool_alloc(tableDict->pool, sizeof(char) * (tableDict->iCodeLength + 1));
-                tableDict->autoPhrase[i].strHZ = (char *) fcitx_memory_pool_alloc(tableDict->pool, sizeof(char) * (PHRASE_MAX_LENGTH * UTF8_MAX_LENGTH + 1));
-                fread(tableDict->autoPhrase[i].strCode, tableDict->iCodeLength + 1, 1, fpDict);
-                fread(tableDict->autoPhrase[i].strHZ, PHRASE_MAX_LENGTH * UTF8_MAX_LENGTH + 1, 1, fpDict);
-                fread(&iTempCount, sizeof(unsigned int), 1, fpDict);
-                tableDict->autoPhrase[i].iSelected = iTempCount;
-                if (i == AUTO_PHRASE_COUNT - 1)
-                    tableDict->autoPhrase[i].next = &tableDict->autoPhrase[0];
-                else
-                    tableDict->autoPhrase[i].next = &tableDict->autoPhrase[i + 1];
+                    tableDict->autoPhrase[i].iSelected = iTempCount;
+                    if (i == AUTO_PHRASE_COUNT - 1)
+                        tableDict->autoPhrase[i].next = &tableDict->autoPhrase[0];
+                    else
+                        tableDict->autoPhrase[i].next = &tableDict->autoPhrase[i + 1];
+                }
             }
             fclose(fpDict);
         }
 
         for (; i < AUTO_PHRASE_COUNT; i++) {
-            tableDict->autoPhrase[i].strCode = (char *) fcitx_memory_pool_alloc(tableDict->pool, sizeof(char) * (tableDict->iCodeLength + 1));
-            tableDict->autoPhrase[i].strHZ = (char *) fcitx_memory_pool_alloc(tableDict->pool, sizeof(char) * (PHRASE_MAX_LENGTH * UTF8_MAX_LENGTH + 1));
+            tableDict->autoPhrase[i].strCode = (char*)fcitx_memory_pool_alloc(tableDict->pool, sizeof(char) * (tableDict->iCodeLength + 1));
+            tableDict->autoPhrase[i].strHZ = (char*)fcitx_memory_pool_alloc(tableDict->pool, sizeof(char) * (PHRASE_MAX_LENGTH * UTF8_MAX_LENGTH + 1));
             tableDict->autoPhrase[i].iSelected = 0;
             if (i == AUTO_PHRASE_COUNT - 1)
                 tableDict->autoPhrase[i].next = &tableDict->autoPhrase[0];
@@ -252,7 +347,7 @@ void SaveTableDict(TableMetaData *tableMetaData)
     RECORD         *recTemp;
     char           *pstr, *tempfile;
     FILE           *fpDict;
-    unsigned int    iTemp;
+    uint32_t    iTemp;
     unsigned int    i;
     int             fd;
     int8_t          cTemp;
@@ -277,17 +372,16 @@ void SaveTableDict(TableMetaData *tableMetaData)
     }
 
     // write version number
-    iTemp = 0;
-    fwrite(&iTemp, sizeof(unsigned int), 1, fpDict);
+    fcitx_utils_write_uint32(fpDict, 0);
     fwrite(&iInternalVersion, sizeof(char), 1, fpDict);
 
     iTemp = strlen(tableDict->strInputCode);
-    fwrite(&iTemp, sizeof(unsigned int), 1, fpDict);
+    fcitx_utils_write_uint32(fpDict, iTemp);
     fwrite(tableDict->strInputCode, sizeof(char), iTemp + 1, fpDict);
     fwrite(&(tableDict->iCodeLength), sizeof(char), 1, fpDict);
     fwrite(&(tableDict->iPYCodeLength), sizeof(char), 1, fpDict);
     iTemp = strlen(tableDict->strIgnoreChars);
-    fwrite(&iTemp, sizeof(unsigned int), 1, fpDict);
+    fcitx_utils_write_uint32(fpDict, iTemp);
     fwrite(tableDict->strIgnoreChars, sizeof(char), iTemp + 1, fpDict);
 
     fwrite(&(tableDict->bRule), sizeof(unsigned char), 1, fpDict);
@@ -303,19 +397,19 @@ void SaveTableDict(TableMetaData *tableMetaData)
         }
     }
 
-    fwrite(&(tableDict->iRecordCount), sizeof(unsigned int), 1, fpDict);
+    fcitx_utils_write_uint32(fpDict, tableDict->iRecordCount);
     recTemp = tableDict->recordHead->next;
     while (recTemp != tableDict->recordHead) {
         fwrite(recTemp->strCode, sizeof(char), tableDict->iPYCodeLength + 1, fpDict);
 
         iTemp = strlen(recTemp->strHZ) + 1;
-        fwrite(&iTemp, sizeof(unsigned int), 1, fpDict);
+        fcitx_utils_write_uint32(fpDict, iTemp);
         fwrite(recTemp->strHZ, sizeof(char), iTemp, fpDict);
 
         cTemp = recTemp->type;
         fwrite(&cTemp, sizeof(int8_t), 1, fpDict);
-        fwrite(&(recTemp->iHit), sizeof(unsigned int), 1, fpDict);
-        fwrite(&(recTemp->iIndex), sizeof(unsigned int), 1, fpDict);
+        fcitx_utils_write_uint32(fpDict, recTemp->iHit);
+        fcitx_utils_write_uint32(fpDict, recTemp->iIndex);
         recTemp = recTemp->next;
     }
 
@@ -343,12 +437,12 @@ void SaveTableDict(TableMetaData *tableMetaData)
             fpDict = fdopen(fd, "w");
 
         if (fpDict) {
-            fwrite(&tableDict->iAutoPhrase, sizeof(int), 1, fpDict);
+            fcitx_utils_write_uint32(fpDict, tableDict->iAutoPhrase);
             for (i = 0; i < tableDict->iAutoPhrase; i++) {
                 fwrite(tableDict->autoPhrase[i].strCode, tableDict->iCodeLength + 1, 1, fpDict);
                 fwrite(tableDict->autoPhrase[i].strHZ, PHRASE_MAX_LENGTH * UTF8_MAX_LENGTH + 1, 1, fpDict);
-                iTemp = tableDict->autoPhrase[i].iSelected;
-                fwrite(&iTemp, sizeof(unsigned int), 1, fpDict);
+                fcitx_utils_write_int32(
+                    fpDict, tableDict->autoPhrase[i].iSelected);
             }
             fclose(fpDict);
         }
@@ -426,7 +520,7 @@ void TableCreateAutoPhrase(TableMetaData* tableMetaData, char iCount)
     if (!tableDict->autoPhrase)
         return;
 
-    strHZ = (char *)fcitx_utils_malloc0((tableMetaData->iAutoPhraseLength * UTF8_MAX_LENGTH + 1) * sizeof(char));
+    strHZ = (char*)fcitx_utils_malloc0((tableMetaData->iAutoPhraseLength * UTF8_MAX_LENGTH + 1) * sizeof(char));
     /*
      * 为了提高效率，此处只重新生成新录入字构成的词组
      */
@@ -501,6 +595,7 @@ boolean TableCreatePhraseCode(TableDict* tableDict, char *strHZ)
     if (i == tableDict->iCodeLength - 1)
         return true;
 
+    int codeIdx = 0;
     for (i1 = 0; i1 < tableDict->iCodeLength; i1++) {
         int clen;
         char* ps;
@@ -527,7 +622,11 @@ boolean TableCreatePhraseCode(TableDict* tableDict, char *strHZ)
             break;
         }
 
-        tableDict->strNewPhraseCode[i1] = recTemp->strCode[tableDict->rule[i].rule[i1].iIndex - 1];
+        if (strlen(recTemp->strCode) >= tableDict->rule[i].rule[i1].iIndex) {
+            tableDict->strNewPhraseCode[codeIdx] = recTemp->strCode[tableDict->rule[i].rule[i1].iIndex - 1];
+            codeIdx++;
+        }
+
     }
 
     return bCanntFindCode;
@@ -570,11 +669,11 @@ void TableInsertPhrase(TableDict* tableDict, const char *strCode, const char *st
     if (!insertPoint)
         return;
 
-    dictNew = (RECORD *) fcitx_memory_pool_alloc(tableDict->pool, sizeof(RECORD));
-    dictNew->strCode = (char *) fcitx_memory_pool_alloc(tableDict->pool, sizeof(char) * (tableDict->iCodeLength + 1));
+    dictNew = (RECORD*)fcitx_memory_pool_alloc(tableDict->pool, sizeof(RECORD));
+    dictNew->strCode = (char*)fcitx_memory_pool_alloc(tableDict->pool, sizeof(char) * (tableDict->iCodeLength + 1));
     dictNew->type = RECORDTYPE_NORMAL;
     strcpy(dictNew->strCode, strCode);
-    dictNew->strHZ = (char *) fcitx_memory_pool_alloc(tableDict->pool, sizeof(char) * (strlen(strHZ) + 1));
+    dictNew->strHZ = (char*)fcitx_memory_pool_alloc(tableDict->pool, sizeof(char) * (strlen(strHZ) + 1));
     strcpy(dictNew->strHZ, strHZ);
     dictNew->iHit = 0;
     dictNew->iIndex = tableDict->iTableIndex;
@@ -624,7 +723,7 @@ void TableUpdateHitFrequency(TableMetaData* tableMetaData, RECORD * record)
     }
 }
 
-int TableCompareCode(const TableMetaData* tableMetaData, const char *strUser, const char *strDict)
+int TableCompareCode(const TableMetaData* tableMetaData, const char *strUser, const char *strDict, boolean exactMatch)
 {
     int             i;
 
@@ -636,7 +735,7 @@ int TableCompareCode(const TableMetaData* tableMetaData, const char *strUser, co
                 return (strUser[i] - strDict[i]);
         }
     }
-    if (tableMetaData->bTableExactMatch) {
+    if (exactMatch) {
         if (strlen(strUser) != strlen(strDict))
             return -999;    //随意的一个值
     }
@@ -644,7 +743,7 @@ int TableCompareCode(const TableMetaData* tableMetaData, const char *strUser, co
     return 0;
 }
 
-int TableFindFirstMatchCode(TableMetaData* tableMetaData, const char* strCodeInput)
+int TableFindFirstMatchCode(TableMetaData* tableMetaData, const char* strCodeInput, boolean exactMatch, boolean cacheCurrentRecord)
 {
     int             i = 0;
     TableDict      *tableDict = tableMetaData->tableDict;
@@ -661,15 +760,20 @@ int TableFindFirstMatchCode(TableMetaData* tableMetaData, const char* strCodeInp
             i++;
         }
     }
-    tableDict->currentRecord = tableDict->recordIndex[i].record;
-    if (!tableDict->currentRecord)
+    RECORD* record = NULL;
+    RECORD** pRecord = &record;
+    if (cacheCurrentRecord)
+        pRecord = &tableDict->currentRecord;
+
+    *pRecord = tableDict->recordIndex[i].record;
+    if (!*pRecord)
         return -1;
 
-    while (tableDict->currentRecord != tableDict->recordHead) {
-        if (!TableCompareCode(tableMetaData, strCodeInput, tableDict->currentRecord->strCode)) {
+    while (*pRecord != tableDict->recordHead) {
+        if (!TableCompareCode(tableMetaData, strCodeInput, (*pRecord)->strCode, exactMatch)) {
             return i;
         }
-        tableDict->currentRecord = tableDict->currentRecord->next;
+        (*pRecord) = (*pRecord)->next;
         i++;
     }
 
