@@ -59,6 +59,7 @@ KEY_LIST        keyList[] = {
     {"PGDN", FcitxKey_Page_Down},
     {"ESCAPE", FcitxKey_Escape},
     {"SPACE", FcitxKey_space},
+    {"BACKSPACE", FcitxKey_BackSpace},
     {"DELETE", FcitxKey_Delete},
     {"UP", FcitxKey_Up},
     {"DOWN", FcitxKey_Down},
@@ -343,6 +344,58 @@ KEY_LIST        keyList[] = {
     {"CANCEL",      FcitxKey_Cancel },
     {"HELP",        FcitxKey_Help },
     {"BREAK",       FcitxKey_Break },
+    {"PAUSE", FcitxKey_Pause},
+    {"SCROLLLOCK", FcitxKey_Scroll_Lock},
+    {"SYSREQ", FcitxKey_Sys_Req},
+    {"BEGIN", FcitxKey_Begin},
+    {"MODESWITCH", FcitxKey_Mode_switch},
+    {"SCRIPTSWITCH", FcitxKey_script_switch},
+    {"NUMLOCK", FcitxKey_Num_Lock},
+    {"KPSPACE", FcitxKey_KP_Space},
+    {"KPTAB", FcitxKey_KP_Tab},
+    {"KPENTER", FcitxKey_KP_Enter},
+    {"KPF1", FcitxKey_KP_F1},
+    {"KPF2", FcitxKey_KP_F2},
+    {"KPF3", FcitxKey_KP_F3},
+    {"KPF4", FcitxKey_KP_F4},
+    {"KPHOME", FcitxKey_KP_Home},
+    {"KPLEFT", FcitxKey_KP_Left},
+    {"KPUP", FcitxKey_KP_Up},
+    {"KPRIGHT", FcitxKey_KP_Right},
+    {"KPDOWN", FcitxKey_KP_Down},
+    {"KPPGUP", FcitxKey_KP_Page_Up},
+    {"KPPGDOWN", FcitxKey_KP_Page_Down},
+    {"KPEND", FcitxKey_KP_End},
+    {"KPBEGIN", FcitxKey_KP_Begin},
+    {"KPINSERT", FcitxKey_KP_Insert},
+    {"KPDELETE", FcitxKey_KP_Delete},
+    {"KPEQUAL", FcitxKey_KP_Equal},
+    {"KPMULTIPLY", FcitxKey_KP_Multiply},
+    {"KPADD", FcitxKey_KP_Add},
+    {"KPSEPARATOR", FcitxKey_KP_Separator},
+    {"KPSUBTRACT", FcitxKey_KP_Subtract},
+    {"KPDECIMAL", FcitxKey_KP_Decimal},
+    {"KPDIVIDE", FcitxKey_KP_Divide},
+    {"KP0", FcitxKey_KP_0},
+    {"KP1", FcitxKey_KP_1},
+    {"KP2", FcitxKey_KP_2},
+    {"KP3", FcitxKey_KP_3},
+    {"KP4", FcitxKey_KP_4},
+    {"KP5", FcitxKey_KP_5},
+    {"KP6", FcitxKey_KP_6},
+    {"KP7", FcitxKey_KP_7},
+    {"KP8", FcitxKey_KP_8},
+    {"KP9", FcitxKey_KP_9},
+    {"CAPSLOCK", FcitxKey_Caps_Lock},
+    {"SHIFTLOCK", FcitxKey_Shift_Lock},
+    {"LMETA", FcitxKey_Meta_L},
+    {"RMETA", FcitxKey_Meta_R},
+    {"LSUPER", FcitxKey_Super_L},
+    {"RSUPER", FcitxKey_Super_R},
+    {"LHYPER", FcitxKey_Hyper_L},
+    {"RHYPER", FcitxKey_Hyper_R},
+
+
     {"\0", 0}
 };
 
@@ -419,7 +472,7 @@ static int FcitxHotkeyGetKeyList(const char *strKey);
 static char *FcitxHotkeyGetKeyListString(int key);
 
 FCITX_EXPORT_API
-boolean FcitxHotkeyIsHotKeyDigit(FcitxKeySym sym, int state)
+boolean FcitxHotkeyIsHotKeyDigit(FcitxKeySym sym, unsigned int state)
 {
     if (state)
         return false;
@@ -431,7 +484,7 @@ boolean FcitxHotkeyIsHotKeyDigit(FcitxKeySym sym, int state)
 }
 
 FCITX_EXPORT_API
-boolean FcitxHotkeyIsHotKeyUAZ(FcitxKeySym sym, int state)
+boolean FcitxHotkeyIsHotKeyUAZ(FcitxKeySym sym, unsigned int state)
 {
     if (state)
         return false;
@@ -443,7 +496,7 @@ boolean FcitxHotkeyIsHotKeyUAZ(FcitxKeySym sym, int state)
 }
 
 FCITX_EXPORT_API
-boolean FcitxHotkeyIsHotKeySimple(FcitxKeySym sym, int state)
+boolean FcitxHotkeyIsHotKeySimple(FcitxKeySym sym, unsigned int state)
 {
     if (state)
         return false;
@@ -455,7 +508,7 @@ boolean FcitxHotkeyIsHotKeySimple(FcitxKeySym sym, int state)
 }
 
 FCITX_EXPORT_API
-boolean FcitxHotkeyIsHotKeyLAZ(FcitxKeySym sym, int state)
+boolean FcitxHotkeyIsHotKeyLAZ(FcitxKeySym sym, unsigned int state)
 {
     if (state)
         return false;
@@ -466,9 +519,15 @@ boolean FcitxHotkeyIsHotKeyLAZ(FcitxKeySym sym, int state)
     return false;
 }
 
+FCITX_EXPORT_API
+boolean FcitxHotkeyIsKey(FcitxKeySym sym, unsigned int state, FcitxKeySym symcmp, unsigned int statecmp)
+{
+    FcitxHotkey key[2] = { {0, symcmp, statecmp }, {0, 0, 0} };
+    return FcitxHotkeyIsHotKey(sym ,state, key);
+}
 
 FCITX_EXPORT_API
-boolean FcitxHotkeyIsHotKey(FcitxKeySym sym, int state, const FcitxHotkey * hotkey)
+boolean FcitxHotkeyIsHotKey(FcitxKeySym sym, unsigned int state, const FcitxHotkey * hotkey)
 {
     state &= FcitxKeyState_Ctrl_Alt_Shift | FcitxKeyState_Super;
     if (hotkey[0].sym && sym == hotkey[0].sym && (hotkey[0].state == state))
@@ -479,7 +538,7 @@ boolean FcitxHotkeyIsHotKey(FcitxKeySym sym, int state, const FcitxHotkey * hotk
 }
 
 FCITX_EXPORT_API
-boolean FcitxHotkeyIsHotkeyCursorMove(FcitxKeySym sym, int state)
+boolean FcitxHotkeyIsHotkeyCursorMove(FcitxKeySym sym, unsigned int state)
 {
     if ((
                 sym == FcitxKey_Left
@@ -504,7 +563,7 @@ boolean FcitxHotkeyIsHotkeyCursorMove(FcitxKeySym sym, int state)
 }
 
 FCITX_EXPORT_API
-boolean FcitxHotkeyIsHotKeyModifierCombine(FcitxKeySym sym, int state)
+boolean FcitxHotkeyIsHotKeyModifierCombine(FcitxKeySym sym, unsigned int state)
 {
     if (sym == FcitxKey_Control_L
             || sym == FcitxKey_Control_R
@@ -532,6 +591,9 @@ void FcitxHotkeyGetKey(FcitxKeySym keysym, unsigned int iKeyState, FcitxKeySym*
                 iKeyState = FcitxKeyState_None;
     }
 
+    if (keysym == FcitxKey_ISO_Left_Tab)
+        keysym = FcitxKey_Tab;
+
     *outk = keysym;
 
     *outs = iKeyState;
@@ -584,13 +646,8 @@ char* FcitxHotkeyGetKeyString(FcitxKeySym sym, unsigned int state)
     return str;
 }
 
-/*
- * 根据字串来判断键
- * 主要用于从设置文件中读取热键设定
- * 返回-1表示用户设置的热键不支持，一般是因为拼写错误或该热键不在列表中
- */
 FCITX_EXPORT_API
-boolean FcitxHotkeyParseKey(const char *strKey, FcitxKeySym* sym, int* state)
+boolean FcitxHotkeyParseKey(const char *strKey, FcitxKeySym* sym, unsigned int* state)
 {
     const char      *p;
     int             iKey;
@@ -682,18 +739,18 @@ char *FcitxHotkeyGetKeyListString(int key)
 }
 
 FCITX_EXPORT_API
-void FcitxHotkeySetKey(char *strKeys, FcitxHotkey * hotkey)
+void FcitxHotkeySetKey(const char *str, FcitxHotkey * hotkey)
 {
     char           *p;
     char           *strKey;
     int             i = 0, j = 0, k;
 
-    strKeys = fcitx_utils_trim(strKeys);
+    char* strKeys = fcitx_utils_trim(str);
     p = strKeys;
 
     for (k = 0; k < 2; k++) {
         FcitxKeySym sym;
-        int state;
+        unsigned int state;
         i = 0;
 
         while (p[i] != ' ' && p[i] != '\0')
@@ -727,69 +784,53 @@ void FcitxHotkeySetKey(char *strKeys, FcitxHotkey * hotkey)
     free(strKeys);
 }
 
-struct KeyPadTable {
-    FcitxKeySym keypad;
-    FcitxKeySym keymain;
-};
-
-static int key_table_cmp(const void* a, const void* b)
-{
-    const struct KeyPadTable* ka = a;
-    const struct KeyPadTable* kb = b;
-    return ka->keypad - kb->keypad;
-}
-
 FCITX_EXPORT_API
 FcitxKeySym FcitxHotkeyPadToMain(FcitxKeySym sym)
 {
-    static struct KeyPadTable keytable[] = {
-        {FcitxKey_KP_Space, FcitxKey_space},
-        {FcitxKey_KP_Tab, FcitxKey_Tab},
-        {FcitxKey_KP_Enter, FcitxKey_Return},
-        {FcitxKey_KP_F1, FcitxKey_F1},
-        {FcitxKey_KP_F2, FcitxKey_F2},
-        {FcitxKey_KP_F3, FcitxKey_F3},
-        {FcitxKey_KP_F4, FcitxKey_F4},
-        {FcitxKey_KP_Home, FcitxKey_Home},
-        {FcitxKey_KP_Left, FcitxKey_Left},
-        {FcitxKey_KP_Up, FcitxKey_Up},
-        {FcitxKey_KP_Right, FcitxKey_Right},
-        {FcitxKey_KP_Down, FcitxKey_Down},
-        {FcitxKey_KP_Prior, FcitxKey_Prior},
-        {FcitxKey_KP_Page_Up, FcitxKey_Page_Up},
-        {FcitxKey_KP_Next, FcitxKey_Next},
-        {FcitxKey_KP_Page_Down, FcitxKey_Page_Down},
-        {FcitxKey_KP_End, FcitxKey_End},
-        {FcitxKey_KP_Begin, FcitxKey_Begin},
-        {FcitxKey_KP_Insert, FcitxKey_Insert},
-        {FcitxKey_KP_Delete, FcitxKey_Delete},
-        {FcitxKey_KP_Equal, FcitxKey_equal},
-        {FcitxKey_KP_Multiply, FcitxKey_asterisk},
-        {FcitxKey_KP_Add, FcitxKey_plus},
-        {FcitxKey_KP_Separator, FcitxKey_comma},
-        {FcitxKey_KP_Subtract, FcitxKey_minus},
-        {FcitxKey_KP_Decimal, FcitxKey_period},
-        {FcitxKey_KP_Divide, FcitxKey_slash},
+#define PAD_TO_MAIN(keypad, keymain) case keypad: return keymain
+    switch (sym) {
+        PAD_TO_MAIN(FcitxKey_KP_Space, FcitxKey_space);
+        PAD_TO_MAIN(FcitxKey_KP_Tab, FcitxKey_Tab);
+        PAD_TO_MAIN(FcitxKey_KP_Enter, FcitxKey_Return);
+        PAD_TO_MAIN(FcitxKey_KP_F1, FcitxKey_F1);
+        PAD_TO_MAIN(FcitxKey_KP_F2, FcitxKey_F2);
+        PAD_TO_MAIN(FcitxKey_KP_F3, FcitxKey_F3);
+        PAD_TO_MAIN(FcitxKey_KP_F4, FcitxKey_F4);
+        PAD_TO_MAIN(FcitxKey_KP_Home, FcitxKey_Home);
+        PAD_TO_MAIN(FcitxKey_KP_Left, FcitxKey_Left);
+        PAD_TO_MAIN(FcitxKey_KP_Up, FcitxKey_Up);
+        PAD_TO_MAIN(FcitxKey_KP_Right, FcitxKey_Right);
+        PAD_TO_MAIN(FcitxKey_KP_Down, FcitxKey_Down);
+        PAD_TO_MAIN(FcitxKey_KP_Page_Up, FcitxKey_Page_Up);
+        PAD_TO_MAIN(FcitxKey_KP_Page_Down, FcitxKey_Page_Down);
+        PAD_TO_MAIN(FcitxKey_KP_End, FcitxKey_End);
+        PAD_TO_MAIN(FcitxKey_KP_Begin, FcitxKey_Begin);
+        PAD_TO_MAIN(FcitxKey_KP_Insert, FcitxKey_Insert);
+        PAD_TO_MAIN(FcitxKey_KP_Delete, FcitxKey_Delete);
+        PAD_TO_MAIN(FcitxKey_KP_Multiply, FcitxKey_asterisk);
+        PAD_TO_MAIN(FcitxKey_KP_Add, FcitxKey_plus);
+        PAD_TO_MAIN(FcitxKey_KP_Separator, FcitxKey_comma);
+        PAD_TO_MAIN(FcitxKey_KP_Subtract, FcitxKey_minus);
+        PAD_TO_MAIN(FcitxKey_KP_Decimal, FcitxKey_period);
+        PAD_TO_MAIN(FcitxKey_KP_Divide, FcitxKey_slash);
 
-        {FcitxKey_KP_0, FcitxKey_0},
-        {FcitxKey_KP_1, FcitxKey_1},
-        {FcitxKey_KP_2, FcitxKey_2},
-        {FcitxKey_KP_3, FcitxKey_3},
-        {FcitxKey_KP_4, FcitxKey_4},
-        {FcitxKey_KP_5, FcitxKey_5},
-        {FcitxKey_KP_6, FcitxKey_6},
-        {FcitxKey_KP_7, FcitxKey_7},
-        {FcitxKey_KP_8, FcitxKey_8},
-        {FcitxKey_KP_9, FcitxKey_9},
-    };
+        PAD_TO_MAIN(FcitxKey_KP_0, FcitxKey_0);
+        PAD_TO_MAIN(FcitxKey_KP_1, FcitxKey_1);
+        PAD_TO_MAIN(FcitxKey_KP_2, FcitxKey_2);
+        PAD_TO_MAIN(FcitxKey_KP_3, FcitxKey_3);
+        PAD_TO_MAIN(FcitxKey_KP_4, FcitxKey_4);
+        PAD_TO_MAIN(FcitxKey_KP_5, FcitxKey_5);
+        PAD_TO_MAIN(FcitxKey_KP_6, FcitxKey_6);
+        PAD_TO_MAIN(FcitxKey_KP_7, FcitxKey_7);
+        PAD_TO_MAIN(FcitxKey_KP_8, FcitxKey_8);
+        PAD_TO_MAIN(FcitxKey_KP_9, FcitxKey_9);
 
-    struct KeyPadTable key = { sym, FcitxKey_None };
-
-    struct KeyPadTable *result = bsearch(&key, keytable, sizeof(keytable) / sizeof(struct KeyPadTable) , sizeof(struct KeyPadTable), key_table_cmp);
-    if (result == NULL)
-        return sym;
-    else
-        return result->keymain;
+        PAD_TO_MAIN(FcitxKey_KP_Equal, FcitxKey_equal);
+    default:
+        break;
+    }
+#undef PAD_TO_MAIN
+    return sym;
 }
 
 FCITX_EXPORT_API
@@ -799,6 +840,8 @@ void FcitxHotkeyFree(FcitxHotkey* hotkey)
         free(hotkey[0].desc);
     if (hotkey[1].desc)
         free(hotkey[1].desc);
+    hotkey[0].desc = NULL;
+    hotkey[1].desc = NULL;
 }
 
 // kate: indent-mode cstyle; space-indent on; indent-width 0;
