@@ -81,12 +81,7 @@ FCITX_EXPORT_API
 FILE *FcitxXDGGetFileWithPrefix(const char* prefix, const char *fileName, const char *mode, char **retFile)
 {
     size_t len;
-    char *prefixpath;
-    asprintf(&prefixpath, "%s/%s", PACKAGE, prefix);
-    char* datadir = fcitx_utils_get_fcitx_path("datadir");
-    char ** path = FcitxXDGGetPath(&len, "XDG_CONFIG_HOME", ".config", prefixpath , datadir, prefixpath);
-    free(datadir);
-    free(prefixpath);
+    char ** path = FcitxXDGGetPathWithPrefix(&len, prefix);
 
     FILE* fp = FcitxXDGGetFile(fileName, path, mode, len, retFile);
 
@@ -116,10 +111,7 @@ FCITX_EXPORT_API
 FILE *FcitxXDGGetFileUserWithPrefix(const char* prefix, const char *fileName, const char *mode, char **retFile)
 {
     size_t len;
-    char *prefixpath;
-    asprintf(&prefixpath, "%s/%s", PACKAGE, prefix);
-    char ** path = FcitxXDGGetPath(&len, "XDG_CONFIG_HOME", ".config", prefixpath , NULL, NULL);
-    free(prefixpath);
+    char ** path = FcitxXDGGetPathUserWithPrefix(&len, prefix);
 
     FILE* fp = FcitxXDGGetFile(fileName, path, mode, len, retFile);
 
@@ -131,14 +123,14 @@ FILE *FcitxXDGGetFileUserWithPrefix(const char* prefix, const char *fileName, co
 FCITX_EXPORT_API
 FILE *FcitxXDGGetFile(const char *fileName, char **path, const char *mode, size_t len, char **retFile)
 {
-    char* buf = NULL;
     size_t i;
     FILE *fp = NULL;
 
     /* check absolute path */
 
     if (strlen(fileName) > 0 && fileName[0] == '/') {
-        fp = fopen(fileName, mode);
+        if (mode)
+            fp = fopen(fileName, mode);
 
         if (retFile)
             *retFile = strdup(fileName);
@@ -164,16 +156,16 @@ FILE *FcitxXDGGetFile(const char *fileName, char **path, const char *mode, size_
         return NULL;
     }
 
+    char* buf = NULL;
     for (i = 0; i < len; i++) {
+        fcitx_utils_free(buf);
+        buf = NULL;
         asprintf(&buf, "%s/%s", path[i], fileName);
 
         fp = fopen(buf, mode);
 
         if (fp)
             break;
-
-        free(buf);
-        buf = NULL;
     }
 
     if (!fp) {
@@ -266,6 +258,16 @@ char **FcitxXDGGetPath(
     return dirsArray;
 }
 
+FCITX_EXPORT_API
+char** FcitxXDGGetPathUserWithPrefix(size_t* len, const char* prefix)
+{
+    char *prefixpath;
+    asprintf(&prefixpath, "%s/%s", PACKAGE, prefix);
+    char** xdgPath = FcitxXDGGetPath(len, "XDG_CONFIG_HOME", ".config" , prefixpath , NULL , NULL);
+    free(prefixpath);
+    return xdgPath;
+}
+
 FCITX_EXPORT_API
 char** FcitxXDGGetPathWithPrefix(size_t* len, const char* prefix)
 {
@@ -296,12 +298,7 @@ FcitxStringHashSet* FcitxXDGGetFiles(
 
     FcitxStringHashSet* sset = NULL;
 
-    char *prefixpath;
-    asprintf(&prefixpath, "%s/%s", PACKAGE, path);
-    char* datadir = fcitx_utils_get_fcitx_path("datadir");
-    xdgPath = FcitxXDGGetPath(&len, "XDG_CONFIG_HOME", ".config" , prefixpath , datadir , prefixpath);
-    free(datadir);
-    free(prefixpath);
+    xdgPath = FcitxXDGGetPathWithPrefix(&len, path);
 
     for (i = 0; i < len; i++) {
         asprintf(&pathBuf, "%s", xdgPath[i]);
@@ -310,10 +307,10 @@ FcitxStringHashSet* FcitxXDGGetFiles(
         free(pathBuf);
         if (dir == NULL)
             continue;
-        
+
         size_t suffixlen = 0;
         size_t prefixlen = 0;
-        
+
         if (suffix) suffixlen = strlen(suffix);
         if (prefix) prefixlen = strlen(prefix);
 
