@@ -19,12 +19,25 @@
 
 #include <stdlib.h>
 #include <unistd.h>
+#include <dbus/dbus.h>
 #include "module/dbus/dbusstuff.h"
 #include "frontend/ipc/ipc.h"
 #include "fcitx/fcitx.h"
 #include "fcitxclient.h"
 #include "marshall.h"
 
+#define fcitx_gclient_debug(...) g_log ("fcitx-client",       \
+                                      G_LOG_LEVEL_DEBUG,    \
+                                      __VA_ARGS__)
+
+typedef struct _ProcessKeyStruct ProcessKeyStruct;
+
+struct _ProcessKeyStruct {
+    FcitxClient* self;
+    GAsyncReadyCallback callback;
+    void* user_data;
+};
+
 struct _FcitxClientPrivate {
     GDBusProxy* improxy;
     GDBusProxy* icproxy;
@@ -32,7 +45,10 @@ struct _FcitxClientPrivate {
     char icname[64];
     int id;
     guint watch_id;
+    GFileMonitor* monitor;
     GCancellable* cancellable;
+    GDBusConnection* connection;
+    gboolean connection_is_bus;
 };
 
 static const gchar introspection_xml[] =
@@ -55,6 +71,10 @@ static const gchar introspection_xml[] =
 static const gchar ic_introspection_xml[] =
     "<node>\n"
     "  <interface name=\"" FCITX_IC_DBUS_INTERFACE "\">\n"
+    "    <method name=\"EnableIC\">\n"
+    "    </method>\n"
+    "    <method name=\"CloseIC\">\n"
+    "    </method>\n"
     "    <method name=\"FocusIn\">\n"
     "    </method>\n"
     "    <method name=\"FocusOut\">\n"
@@ -100,6 +120,14 @@ static const gchar ic_introspection_xml[] =
     "      <arg name=\"str\" type=\"a(si)\"/>\n"
     "      <arg name=\"cursorpos\" type=\"i\"/>\n"
     "    </signal>\n"
+    "    <signal name=\"UpdateClientSideUI\">\n"
+    "      <arg name=\"auxup\" type=\"s\"/>\n"
+    "      <arg name=\"auxdown\" type=\"s\"/>\n"
+    "      <arg name=\"preedit\" type=\"s\"/>\n"
+    "      <arg name=\"candidateword\" type=\"s\"/>\n"
+    "      <arg name=\"imname\" type=\"s\"/>\n"
+    "      <arg name=\"cursorpos\" type=\"i\"/>\n"
+    "    </signal>\n"
     "    <signal name=\"ForwardKey\">\n"
     "      <arg name=\"keyval\" type=\"u\"/>\n"
     "      <arg name=\"state\" type=\"u\"/>\n"
@@ -116,13 +144,15 @@ G_DEFINE_TYPE(FcitxClient, fcitx_client, G_TYPE_OBJECT);
     (G_TYPE_INSTANCE_GET_PRIVATE((obj), FCITX_TYPE_CLIENT, FcitxClientPrivate))
 
 enum {
-    CONNTECTED_SIGNAL,
+    CONNECTED_SIGNAL,
     ENABLE_IM_SIGNAL,
     CLOSE_IM_SIGNAL,
     FORWARD_KEY_SIGNAL,
     COMMIT_STRING_SIGNAL,
     DELETE_SURROUNDING_TEXT_SIGNAL,
     UPDATED_FORMATTED_PREEDIT_SIGNAL,
+    DISCONNECTED_SIGNAL,
+    UPDATE_CLIENT_SIDE_UI_SIGNAL,
     LAST_SIGNAL
 };
 
@@ -130,15 +160,23 @@ static guint signals[LAST_SIGNAL] = {0};
 
 static GDBusInterfaceInfo *_fcitx_client_get_interface_info(void);
 static GDBusInterfaceInfo *_fcitx_client_get_clientic_info(void);
-static void _fcitx_client_create_ic(FcitxClient* self);
+static void _fcitx_client_create_ic(FcitxClient* self, gboolean use_session_bus);
 static void _fcitx_client_create_ic_phase1_finished(GObject* source_object, GAsyncResult* res, gpointer user_data);
 static void _fcitx_client_create_ic_cb(GObject *source_object, GAsyncResult *res, gpointer user_data);
 static void _fcitx_client_create_ic_phase2_finished(GObject *source_object, GAsyncResult *res, gpointer user_data);
 static void _fcitx_client_g_signal(GDBusProxy *proxy, gchar *sender_name, gchar *signal_name, GVariant *parameters, gpointer user_data);
 static void fcitx_client_init(FcitxClient *self);
 static void fcitx_client_finalize(GObject *object);
+static void fcitx_client_dispose(GObject *object);
 static void _fcitx_client_appear (GDBusConnection *connection, const gchar *name, const gchar *name_owner, gpointer user_data);
 static void _fcitx_client_vanish (GDBusConnection *connection, const gchar *name, gpointer user_data);
+static void _fcitx_client_socket_file_changed_cb (GFileMonitor *monitor, GFile *file, GFile *other_file, GFileMonitorEvent event_type, gpointer            user_data);
+static gchar* _fcitx_get_address ();
+static void _fcitx_client_create_ic_phase0_bus_finished(GObject *source_object, GAsyncResult *res, gpointer user_data);
+static void _fcitx_client_create_ic_phase0_connection_finished(GObject *source_object, GAsyncResult *res, gpointer user_data);
+static void _fcitx_client_clean_up(FcitxClient* self, gboolean dont_emit_disconn);
+static void _fcitx_client_unwatch(FcitxClient* self);
+static void _fcitx_client_watch(FcitxClient* self);
 
 static void fcitx_client_class_init(FcitxClientClass *klass);
 
@@ -174,24 +212,49 @@ _fcitx_client_get_clientic_info(void)
 
 static void
 fcitx_client_finalize(GObject *object)
+{
+    if (G_OBJECT_CLASS(fcitx_client_parent_class)->finalize != NULL)
+        G_OBJECT_CLASS(fcitx_client_parent_class)->finalize(object);
+}
+
+static void
+fcitx_client_dispose(GObject *object)
 {
     FcitxClient *self = FCITX_CLIENT(object);
 
+    if (self->priv->monitor) {
+        g_signal_handlers_disconnect_by_func(self->priv->monitor,
+                                             G_CALLBACK(_fcitx_client_socket_file_changed_cb),
+                                             self);
+        g_object_unref(self->priv->monitor);
+        self->priv->monitor= NULL;
+    }
+
     if (self->priv->icproxy) {
         g_dbus_proxy_call(self->priv->icproxy, "DestroyIC", NULL, G_DBUS_CALL_FLAGS_NONE, -1, NULL, NULL, NULL);
     }
-    g_bus_unwatch_name(self->priv->watch_id);
-    GDBusProxy* icproxy = self->priv->icproxy;
-    GDBusProxy* improxy = self->priv->improxy;
-    self->priv->icproxy = NULL;
-    self->priv->improxy = NULL;
-    if (icproxy)
-        g_object_unref(icproxy);
-    if (improxy)
-        g_object_unref(improxy);
+    _fcitx_client_unwatch(self);
 
-    if (G_OBJECT_CLASS(fcitx_client_parent_class)->finalize != NULL)
-        G_OBJECT_CLASS(fcitx_client_parent_class)->finalize(object);
+    _fcitx_client_clean_up(self, TRUE);
+
+    if (G_OBJECT_CLASS(fcitx_client_parent_class)->dispose != NULL)
+        G_OBJECT_CLASS(fcitx_client_parent_class)->dispose(object);
+}
+
+FCITX_EXPORT_API
+void fcitx_client_enable_ic(FcitxClient* self)
+{
+    if (self->priv->icproxy) {
+        g_dbus_proxy_call(self->priv->icproxy, "EnableIC", NULL, G_DBUS_CALL_FLAGS_NONE, -1, NULL, NULL, NULL);
+    }
+}
+
+FCITX_EXPORT_API
+void fcitx_client_close_ic(FcitxClient* self)
+{
+    if (self->priv->icproxy) {
+        g_dbus_proxy_call(self->priv->icproxy, "CloseIC", NULL, G_DBUS_CALL_FLAGS_NONE, -1, NULL, NULL, NULL);
+    }
 }
 
 FCITX_EXPORT_API
@@ -229,6 +292,12 @@ void fcitx_client_set_capacity(FcitxClient* self, guint flags)
 
 FCITX_EXPORT_API
 void fcitx_client_set_cusor_rect(FcitxClient* self, int x, int y, int w, int h)
+{
+    fcitx_client_set_cursor_rect(self, x, y, w, h);
+}
+
+FCITX_EXPORT_API
+void fcitx_client_set_cursor_rect(FcitxClient* self, int x, int y, int w, int h)
 {
     if (self->priv->icproxy) {
         g_dbus_proxy_call(self->priv->icproxy, "SetCursorRect", g_variant_new("(iiii)", x, y, w, h), G_DBUS_CALL_FLAGS_NONE, -1, NULL, NULL, NULL);
@@ -258,11 +327,72 @@ void fcitx_client_process_key(FcitxClient* self, GAsyncReadyCallback cb, gpointe
     }
 }
 
+FCITX_EXPORT_API
+gint fcitx_client_process_key_finish(FcitxClient* self, GAsyncResult* res)
+{
+    gint ret = -1;
+    if (!self->priv->icproxy)
+        return -1;
+
+    GVariant* result = g_dbus_proxy_call_finish(self->priv->icproxy, res, NULL);
+    if (result) {
+        g_variant_get(result, "(i)", &ret);
+        g_variant_unref(result);
+    }
+    return ret;
+}
+
+void _process_key_data_free(ProcessKeyStruct* pk)
+{
+    g_object_unref(pk->self);
+    g_free(pk);
+}
+
+void
+_fcitx_client_process_key_cb(GObject *source_object,
+                             GAsyncResult *res,
+                             gpointer user_data)
+{
+    ProcessKeyStruct* pk = user_data;
+    pk->callback(G_OBJECT(pk->self), res, pk->user_data);
+    _process_key_data_free(pk);
+}
+
+void _fcitx_client_process_key_cancelled(GCancellable* cancellable, gpointer user_data)
+{
+    ProcessKeyStruct* pk = user_data;
+    _process_key_data_free(pk);
+}
+
+FCITX_EXPORT_API
+void fcitx_client_process_key_async(FcitxClient* self,
+                                    guint32 keyval, guint32 keycode, guint32 state, gint type, guint32 t,
+                                    gint timeout_msec,
+                                    GCancellable *cancellable,
+                                    GAsyncReadyCallback callback,
+                                    gpointer user_data)
+{
+    int itype = type;
+    if (self->priv->icproxy) {
+        ProcessKeyStruct* pk = g_new(ProcessKeyStruct, 1);
+        pk->self = g_object_ref(self);
+        pk->callback = callback;
+        pk->user_data = user_data;
+        g_dbus_proxy_call(self->priv->icproxy,
+                          "ProcessKeyEvent",
+                          g_variant_new("(uuuiu)", keyval, keycode, state, itype, t),
+                          G_DBUS_CALL_FLAGS_NONE,
+                          timeout_msec,
+                          cancellable,
+                          _fcitx_client_process_key_cb,
+                          pk);
+    }
+}
+
 FCITX_EXPORT_API
 int fcitx_client_process_key_sync(FcitxClient* self, guint32 keyval, guint32 keycode, guint32 state, gint type, guint32 t)
 {
     int itype = type;
-    GError *error = NULL;
     int ret = -1;
     if (self->priv->icproxy) {
         GVariant* result =  g_dbus_proxy_call_sync(self->priv->icproxy,
@@ -270,20 +400,25 @@ int fcitx_client_process_key_sync(FcitxClient* self, guint32 keyval, guint32 key
                             g_variant_new("(uuuiu)", keyval, keycode, state, itype, t),
                             G_DBUS_CALL_FLAGS_NONE,
                             -1, NULL,
-                            &error);
+                            NULL);
 
-        if (error)
-        {
-            g_error_free(error);
-        }
-        else {
+        if (result) {
             g_variant_get(result, "(i)", &ret);
+            g_variant_unref(result);
         }
     }
 
     return ret;
 }
 
+static gboolean
+_fcitx_client_new_service_appear(gpointer user_data)
+{
+    FcitxClient* self = (FcitxClient*) user_data;
+    if (!self->priv->connection || g_dbus_connection_is_closed(self->priv->connection))
+        _fcitx_client_create_ic(self, FALSE);
+    return FALSE;
+}
 
 static void
 _fcitx_client_appear (GDBusConnection *connection,
@@ -294,17 +429,11 @@ _fcitx_client_appear (GDBusConnection *connection,
     FcitxClient* self = (FcitxClient*) user_data;
     gboolean new_owner_good = name_owner && (name_owner[0] != '\0');
     if (new_owner_good) {
-        if (self->priv->improxy) {
-            g_object_unref(self->priv->improxy);
-            self->priv->improxy = NULL;
-        }
-
-        if (self->priv->icproxy) {
-            g_object_unref(self->priv->icproxy);
-            self->priv->icproxy = NULL;
-        }
-
-        _fcitx_client_create_ic(self);
+        g_timeout_add_full(G_PRIORITY_DEFAULT,
+                           100,
+                           _fcitx_client_new_service_appear,
+                           g_object_ref(self),
+                           g_object_unref);
     }
 }
 
@@ -314,15 +443,24 @@ _fcitx_client_vanish (GDBusConnection *connection,
                       gpointer         user_data)
 {
     FcitxClient* self = (FcitxClient*) user_data;
-    if (self->priv->improxy) {
-        g_object_unref(self->priv->improxy);
-        self->priv->improxy = NULL;
-    }
+    _fcitx_client_clean_up(self, FALSE);
+}
 
-    if (self->priv->icproxy) {
-        g_object_unref(self->priv->icproxy);
-        self->priv->icproxy = NULL;
-    }
+static gchar*
+_fcitx_get_socket_path()
+{
+    char* machineId = dbus_get_local_machine_id();
+    gchar* path;
+    gchar* addressFile = g_strdup_printf("%s-%d", machineId, fcitx_utils_get_display_number());
+    dbus_free(machineId);
+
+    path = g_build_filename (g_get_user_config_dir (),
+            "fcitx",
+            "dbus",
+            addressFile,
+            NULL);
+    g_free(addressFile);
+    return path;
 }
 
 static void
@@ -332,6 +470,119 @@ fcitx_client_init(FcitxClient *self)
 
     sprintf(self->priv->servicename, "%s-%d", FCITX_DBUS_SERVICE, fcitx_utils_get_display_number());
 
+    self->priv->connection = NULL;
+    self->priv->cancellable = NULL;
+    self->priv->improxy = NULL;
+    self->priv->icproxy = NULL;
+    self->priv->watch_id = 0;
+    self->priv->connection_is_bus = FALSE;
+
+    gchar* path = _fcitx_get_socket_path();
+    GFile* file = g_file_new_for_path(path);
+    self->priv->monitor = g_file_monitor_file(file, 0, NULL, NULL);
+
+    g_signal_connect (self->priv->monitor, "changed", (GCallback) _fcitx_client_socket_file_changed_cb, self);
+
+    g_object_unref(file);
+    g_free(path);
+
+    _fcitx_client_create_ic(self, FALSE);
+}
+
+static void
+_fcitx_client_create_ic(FcitxClient *self, gboolean use_session_bus)
+{
+    fcitx_gclient_debug("_fcitx_client_create_ic");
+    _fcitx_client_unwatch(self);
+    _fcitx_client_clean_up(self, FALSE);
+    self->priv->cancellable = g_cancellable_new ();
+
+    g_object_ref(self);
+    if (!use_session_bus) {
+        gchar* address = _fcitx_get_address();
+        /* a trick for prevent cancellable being called after finalize */
+        if (address) {
+            /* since we have a possible valid address here */
+            g_dbus_connection_new_for_address(address,
+                    G_DBUS_CONNECTION_FLAGS_AUTHENTICATION_CLIENT | G_DBUS_CONNECTION_FLAGS_MESSAGE_BUS_CONNECTION,
+                    NULL,
+                    self->priv->cancellable,
+                    _fcitx_client_create_ic_phase0_connection_finished,
+                    self);
+            g_free(address);
+            return;
+        }
+    }
+
+    _fcitx_client_watch(self);
+    g_bus_get(G_BUS_TYPE_SESSION,
+              self->priv->cancellable,
+              _fcitx_client_create_ic_phase0_bus_finished,
+              self
+            );
+};
+
+static void
+_fcitx_client_connection_closed(GDBusConnection *connection,
+                                gboolean         remote_peer_vanished,
+                                GError          *error,
+                                gpointer         user_data)
+{
+    fcitx_gclient_debug("_fcitx_client_connection_closed");
+    FcitxClient* self = (FcitxClient*) user_data;
+    _fcitx_client_clean_up(self, FALSE);
+
+    _fcitx_client_watch(self);
+}
+
+static void
+_fcitx_client_create_ic_phase0_bus_finished(GObject *source_object,
+                                            GAsyncResult *res,
+                                            gpointer user_data)
+{
+    fcitx_gclient_debug("_fcitx_client_create_ic_phase0_bus_finished");
+    g_return_if_fail (user_data != NULL);
+    g_return_if_fail (FCITX_IS_CLIENT(user_data));
+
+    FcitxClient* self = (FcitxClient*) user_data;
+    if (self->priv->cancellable) {
+        g_object_unref (self->priv->cancellable);
+        self->priv->cancellable = NULL;
+    }
+
+    GDBusConnection* connection = g_bus_get_finish(res, NULL);
+
+    if (!connection) {
+        /* unref for _fcitx_client_create_ic */
+        g_object_unref(self);
+        return;
+    }
+
+    _fcitx_client_clean_up(self, FALSE);
+    self->priv->connection = connection;
+    self->priv->connection_is_bus = TRUE;
+    g_signal_connect(connection, "closed", G_CALLBACK(_fcitx_client_connection_closed), self);
+    self->priv->cancellable = g_cancellable_new ();
+    g_dbus_proxy_new(
+        connection,
+        G_DBUS_PROXY_FLAGS_DO_NOT_LOAD_PROPERTIES,
+        _fcitx_client_get_interface_info(),
+        self->priv->servicename,
+        FCITX_IM_DBUS_PATH,
+        FCITX_IM_DBUS_INTERFACE,
+        self->priv->cancellable,
+        _fcitx_client_create_ic_phase1_finished,
+        self
+    );
+}
+
+static void
+_fcitx_client_watch(FcitxClient* self)
+{
+    if (self->priv->watch_id)
+        return;
+    fcitx_gclient_debug("_fcitx_client_watch");
+
     self->priv->watch_id = g_bus_watch_name(
                        G_BUS_TYPE_SESSION,
                        self->priv->servicename,
@@ -341,30 +592,63 @@ fcitx_client_init(FcitxClient *self)
                        self,
                        NULL
                    );
-    self->priv->improxy = NULL;
-    self->priv->icproxy = NULL;
 }
 
 static void
-_fcitx_client_create_ic(FcitxClient *self)
+_fcitx_client_unwatch(FcitxClient* self)
 {
+    if (self->priv->watch_id)
+        g_bus_unwatch_name(self->priv->watch_id);
+    self->priv->watch_id = 0;
+}
+
+static void
+_fcitx_client_create_ic_phase0_connection_finished(GObject *source_object,
+                                                   GAsyncResult *res,
+                                                   gpointer user_data)
+
+{
+    fcitx_gclient_debug("_fcitx_client_create_ic_phase0_connection_finished");
+    g_return_if_fail (user_data != NULL);
+    g_return_if_fail (FCITX_IS_CLIENT(user_data));
+    FcitxClient* self = (FcitxClient*) user_data;
     if (self->priv->cancellable) {
-        g_cancellable_cancel (self->priv->cancellable);
         g_object_unref (self->priv->cancellable);
         self->priv->cancellable = NULL;
     }
-    if (self->priv->improxy) {
-        g_object_unref(self->priv->improxy);
-        self->priv->improxy = NULL;
+
+    GError* error = NULL;
+    GDBusConnection* connection = g_dbus_connection_new_for_address_finish(res, &error);
+
+    gboolean is_cancel = FALSE;
+    if (error) {
+        if (g_error_matches(error, G_IO_ERROR, G_IO_ERROR_CANCELLED))
+            is_cancel = TRUE;
+        g_error_free(error);
     }
 
-    if (self->priv->icproxy) {
-        g_object_unref(self->priv->icproxy);
-        self->priv->icproxy = NULL;
+    /* hey! if we failed here. we'd try traditional dbus way */
+    if (!connection || g_dbus_connection_is_closed(connection)) {
+        if (connection)
+            g_object_unref(connection);
+
+        if (!is_cancel)
+            _fcitx_client_create_ic(self, TRUE);
+        /* unref for _fcitx_client_create_ic */
+        g_object_unref(self);
+        return;
     }
+
+    g_dbus_connection_set_exit_on_close(connection, FALSE);
+
+    _fcitx_client_clean_up(self, FALSE);
+    self->priv->connection = connection;
+    self->priv->connection_is_bus = FALSE;
+    g_signal_connect(self->priv->connection, "closed", G_CALLBACK(_fcitx_client_connection_closed), self);
+
     self->priv->cancellable = g_cancellable_new ();
-    g_dbus_proxy_new_for_bus(
-        G_BUS_TYPE_SESSION,
+    g_dbus_proxy_new(
+        connection,
         G_DBUS_PROXY_FLAGS_DO_NOT_LOAD_PROPERTIES,
         _fcitx_client_get_interface_info(),
         self->priv->servicename,
@@ -374,37 +658,45 @@ _fcitx_client_create_ic(FcitxClient *self)
         _fcitx_client_create_ic_phase1_finished,
         self
     );
-};
+}
 
 static void
 _fcitx_client_create_ic_phase1_finished(GObject *source_object,
                                         GAsyncResult *res,
                                         gpointer user_data)
 {
+    fcitx_gclient_debug("_fcitx_client_create_ic_phase1_finished");
+    g_return_if_fail (user_data != NULL);
+    g_return_if_fail (FCITX_IS_CLIENT(user_data));
     FcitxClient* self = (FcitxClient*) user_data;
-    GError* error = NULL;
     if (self->priv->cancellable) {
         g_object_unref (self->priv->cancellable);
         self->priv->cancellable = NULL;
     }
     if (self->priv->improxy)
         g_object_unref(self->priv->improxy);
-    self->priv->improxy = g_dbus_proxy_new_for_bus_finish(res, &error);
-    if (error) {
-        g_warning ("Create fcitx input method proxy failed: %s.", error->message);
-        g_error_free(error);
-    }
-    if (!self->priv->improxy)
-        return;
+    self->priv->improxy = g_dbus_proxy_new_finish(res, NULL);
 
-    gchar* owner_name = g_dbus_proxy_get_name_owner(self->priv->improxy);
+    do {
+        if (!self->priv->improxy) {
+            break;
+        }
 
-    if (!owner_name) {
-        g_object_unref(self->priv->improxy);
-        self->priv->improxy = NULL;
+        gchar* owner_name = g_dbus_proxy_get_name_owner(self->priv->improxy);
+
+        if (!owner_name) {
+            g_object_unref(self->priv->improxy);
+            self->priv->improxy = NULL;
+            break;
+        }
+        g_free(owner_name);
+    } while(0);
+
+    if (!self->priv->improxy) {
+        /* unref for phase0 finish */
+        g_object_unref(self);
         return;
     }
-    g_free(owner_name);
 
     self->priv->cancellable = g_cancellable_new ();
     char* appname = fcitx_utils_get_process_name();
@@ -429,27 +721,28 @@ _fcitx_client_create_ic_cb(GObject *source_object,
                            gpointer user_data)
 {
     FcitxClient* self = (FcitxClient*) user_data;
-    GError* error = NULL;
     if (self->priv->cancellable) {
         g_object_unref (self->priv->cancellable);
         self->priv->cancellable = NULL;
     }
-    GVariant* result = g_dbus_proxy_call_finish(G_DBUS_PROXY(source_object), res, &error);
+    GVariant* result = g_dbus_proxy_call_finish(G_DBUS_PROXY(source_object), res, NULL);
 
-    if (error) {
-        g_error_free(error);
+    if (!result) {
+        /* unref for _fcitx_client_phase1_finish */
+        g_object_unref(self);
         return;
     }
 
     gboolean enable;
     guint32 key1, state1, key2, state2;
     g_variant_get(result, "(ibuuuu)", &self->priv->id, &enable, &key1, &state1, &key2, &state2);
+    g_variant_unref(result);
 
     sprintf(self->priv->icname, FCITX_IC_DBUS_PATH, self->priv->id);
 
     self->priv->cancellable = g_cancellable_new ();
-    g_dbus_proxy_new_for_bus(
-        G_BUS_TYPE_SESSION,
+    g_dbus_proxy_new(
+        self->priv->connection,
         G_DBUS_PROXY_FLAGS_NONE,
         _fcitx_client_get_clientic_info(),
         self->priv->servicename,
@@ -467,34 +760,38 @@ _fcitx_client_create_ic_phase2_finished(GObject *source_object,
                                         GAsyncResult *res,
                                         gpointer user_data)
 {
+    g_return_if_fail (user_data != NULL);
+    g_return_if_fail (FCITX_IS_CLIENT(user_data));
     FcitxClient* self = (FcitxClient*) user_data;
-    GError* error = NULL;
     if (self->priv->cancellable) {
         g_object_unref (self->priv->cancellable);
         self->priv->cancellable = NULL;
     }
     if (self->priv->icproxy)
         g_object_unref(self->priv->icproxy);
-    self->priv->icproxy = g_dbus_proxy_new_for_bus_finish(res, &error);
+    self->priv->icproxy = g_dbus_proxy_new_finish(res, NULL);
 
-    if (error) {
-        g_error_free(error);
+    do {
+        if (!self->priv->icproxy)
+            break;
+
+        gchar* owner_name = g_dbus_proxy_get_name_owner(self->priv->icproxy);
+
+        if (!owner_name) {
+            g_object_unref(self->priv->icproxy);
+            self->priv->icproxy = NULL;
+            break;
+        }
+        g_free(owner_name);
+    } while(0);
+
+    if (self->priv->icproxy) {
+        g_signal_connect(self->priv->icproxy, "g-signal", G_CALLBACK(_fcitx_client_g_signal), self);
+        g_signal_emit(user_data, signals[CONNECTED_SIGNAL], 0);
     }
 
-    if (!self->priv->icproxy)
-        return;
-
-    gchar* owner_name = g_dbus_proxy_get_name_owner(self->priv->icproxy);
-
-    if (!owner_name) {
-        g_object_unref(self->priv->icproxy);
-        self->priv->icproxy = NULL;
-        return;
-    }
-    g_free(owner_name);
-
-    g_signal_connect(self->priv->icproxy, "g-signal", G_CALLBACK(_fcitx_client_g_signal), self);
-    g_signal_emit(user_data, signals[CONNTECTED_SIGNAL], 0);
+    /* unref for _fcitx_client_create_ic_cb */
+    g_object_unref(self);
 }
 
 static void
@@ -537,6 +834,12 @@ _fcitx_client_g_signal(GDBusProxy *proxy,
         g_variant_get(parameters, "(iu)", &offset, &nchar);
         g_signal_emit(user_data, signals[DELETE_SURROUNDING_TEXT_SIGNAL], 0, offset, nchar);
     }
+    else if (strcmp(signal_name, "UpdateClientSideUI") == 0) {
+        const gchar* auxup, *auxdown, *preedit, *candidate, *imname;
+        int cursor;
+        g_variant_get(parameters, "(sssssi)", &auxup, &auxdown, &preedit, &candidate, &imname, &cursor);
+        g_signal_emit(user_data, signals[UPDATE_CLIENT_SIDE_UI_SIGNAL], 0, auxup, auxdown, preedit, candidate, imname, cursor);
+    }
     else if (strcmp(signal_name, "UpdateFormattedPreedit") == 0) {
         int cursor_pos;
         GPtrArray* array = g_ptr_array_new_with_free_func(_item_free);
@@ -564,6 +867,7 @@ fcitx_client_class_init(FcitxClientClass *klass)
     GObjectClass *gobject_class;
 
     gobject_class = G_OBJECT_CLASS(klass);
+    gobject_class->dispose = fcitx_client_dispose;
     gobject_class->finalize = fcitx_client_finalize;
 
     g_type_class_add_private (klass, sizeof (FcitxClientPrivate));
@@ -574,9 +878,9 @@ fcitx_client_class_init(FcitxClientClass *klass)
      *
      * @client: A FcitxClient
      *
-     * Emit when connected to fcitx
+     * Emit when connected to fcitx and created ic
      */
-    signals[CONNTECTED_SIGNAL] = g_signal_new(
+    signals[CONNECTED_SIGNAL] = g_signal_new(
                                      "connected",
                                      FCITX_TYPE_CLIENT,
                                      G_SIGNAL_RUN_LAST,
@@ -587,6 +891,24 @@ fcitx_client_class_init(FcitxClientClass *klass)
                                      G_TYPE_NONE,
                                      0);
 
+    /**
+     * FcitxClient::disconnected
+     *
+     * @client: A FcitxClient
+     *
+     * Emit when disconnected from fcitx
+     */
+    signals[DISCONNECTED_SIGNAL] = g_signal_new(
+                                     "disconnected",
+                                     FCITX_TYPE_CLIENT,
+                                     G_SIGNAL_RUN_LAST,
+                                     0,
+                                     NULL,
+                                     NULL,
+                                     g_cclosure_marshal_VOID__VOID,
+                                     G_TYPE_NONE,
+                                     0);
+
     /**
      * FcitxClient::enable-im
      *
@@ -686,6 +1008,31 @@ fcitx_client_class_init(FcitxClientClass *klass)
                 G_TYPE_INT, G_TYPE_UINT
             );
 
+    /**
+     * FcitxClient::update-client-side-ui
+     *
+     * @client: A FcitxClient
+     * @auxup: (transfer none): Aux up string
+     * @auxdown: (transfer none): Aux down string
+     * @preedit: (transfer none): preedit string
+     * @candidateword: (transfer none): candidateword in one line
+     * @imname: (transfer none): input method name
+     * @cursor_pos: cursor position
+     *
+     * Emit when input method need to update client side ui
+     */
+    signals[UPDATE_CLIENT_SIDE_UI_SIGNAL] = g_signal_new(
+            "update-client-side-ui",
+            FCITX_TYPE_CLIENT,
+            G_SIGNAL_RUN_LAST,
+            0,
+            NULL,
+            NULL,
+            fcitx_marshall_VOID__STRING_STRING_STRING_STRING_STRING_INT,
+            G_TYPE_NONE,
+            6,
+            G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_INT);
+
     /**
      * FcitxClient::update-formatted-preedit
      *
@@ -730,5 +1077,101 @@ fcitx_client_is_valid(FcitxClient* self)
     return self->priv->icproxy != NULL;
 }
 
-// kate: indent-mode cstyle; replace-tabs on;
+static void
+_fcitx_client_socket_file_changed_cb (GFileMonitor       *monitor,
+                                      GFile              *file,
+                                      GFile              *other_file,
+                                      GFileMonitorEvent   event_type,
+                                      gpointer            user_data)
+{
+    FcitxClient* client = user_data;
+    if (event_type != G_FILE_MONITOR_EVENT_CHANGED &&
+        event_type != G_FILE_MONITOR_EVENT_CREATED &&
+        event_type != G_FILE_MONITOR_EVENT_DELETED)
+        return;
 
+    _fcitx_client_create_ic(client, FALSE);
+}
+
+static gchar*
+_fcitx_get_address ()
+{
+    gchar* address = NULL;
+    address = g_strdup (g_getenv("FCITX_DBUS_ADDRESS"));
+    if (address)
+        return address;
+
+    gchar* path = _fcitx_get_socket_path();
+    FILE* fp = fopen(path, "r");
+    g_free(path);
+
+    if (!fp)
+        return NULL;
+
+    const int BUFSIZE = 1024;
+
+    char buffer[BUFSIZE];
+    size_t sz = fread(buffer, sizeof(char), BUFSIZE, fp);
+    fclose(fp);
+    if (sz == 0)
+        return NULL;
+    char* p = buffer;
+    while(*p)
+        p++;
+    size_t addrlen = p - buffer;
+    if (sz != addrlen + 2 * sizeof(pid_t) + 1)
+        return NULL;
+
+    /* skip '\0' */
+    p++;
+    pid_t *ppid = (pid_t*) p;
+    pid_t daemonpid = ppid[0];
+    pid_t fcitxpid = ppid[1];
+
+    if (!fcitx_utils_pid_exists(daemonpid)
+        || !fcitx_utils_pid_exists(fcitxpid))
+        return NULL;
+
+    address = g_strdup(buffer);
+
+    return address;
+}
+
+static void
+_fcitx_client_clean_up(FcitxClient* self, gboolean dont_emit_disconn)
+{
+    if (self->priv->connection) {
+        g_signal_handlers_disconnect_by_func(self->priv->connection,
+                                             G_CALLBACK(_fcitx_client_connection_closed),
+                                             self);
+        if (!self->priv->connection_is_bus) {
+            g_dbus_connection_close_sync(self->priv->connection, NULL, NULL);
+        }
+        g_object_unref(self->priv->connection);
+        self->priv->connection = NULL;
+    }
+
+    if (self->priv->cancellable) {
+        g_cancellable_cancel (self->priv->cancellable);
+        g_object_unref (self->priv->cancellable);
+        self->priv->cancellable = NULL;
+    }
+
+    if (self->priv->improxy) {
+        g_object_unref(self->priv->improxy);
+        self->priv->improxy = NULL;
+    }
+
+    if (self->priv->icproxy) {
+        g_signal_handlers_disconnect_by_func(self->priv->icproxy,
+                                             G_CALLBACK(_fcitx_client_g_signal),
+                                             self);
+        g_object_unref(self->priv->icproxy);
+        self->priv->icproxy = NULL;
+        if (!dont_emit_disconn)
+            g_signal_emit(self, signals[DISCONNECTED_SIGNAL], 0);
+    }
+
+}
+
+// kate: indent-mode cstyle; replace-tabs on;
