@@ -23,46 +23,60 @@
 #include <string.h>
 #include <stdio.h>
 #include <stdarg.h>
+#include <wchar.h>
 
 #include "config.h"
 #include "fcitx/fcitx.h"
+#include "utf8.h"
 #include "log.h"
 #include "utils.h"
 
 static iconv_t iconvW = NULL;
+static int init = 0;
+static int is_utf8 = 0;
 
 FCITX_EXPORT_API
 void FcitxLogFunc(ErrorLevel e, const char* filename, const int line, const char* fmt, ...)
 {
+    if (!init) {
+        init = 1;
+        is_utf8 = fcitx_utils_current_locale_is_utf8();
+    }
+
 #ifndef _DEBUG
     if (e == DEBUG)
         return;
 #endif
     switch (e) {
     case INFO:
-        fprintf(stderr, "[INFO]");
+        fprintf(stderr, "(INFO-");
         break;
     case ERROR:
-        fprintf(stderr, "[ERROR]");
+        fprintf(stderr, "(ERROR-");
         break;
     case DEBUG:
-        fprintf(stderr, "[DEBUG]");
+        fprintf(stderr, "(DEBUG-");
         break;
     case WARNING:
-        fprintf(stderr, "[WARN]");
+        fprintf(stderr, "(WARN-");
         break;
     case FATAL:
-        fprintf(stderr, "[FATAL]");
+        fprintf(stderr, "(FATAL-");
         break;
     }
 
-    char *buffer;
+    char *buffer = NULL;
     va_list ap;
-    fprintf(stderr, " %s:%u-", filename, line);
+    fprintf(stderr, "%d %s:%u) ", getpid(), filename, line);
     va_start(ap, fmt);
     vasprintf(&buffer, fmt, ap);
     va_end(ap);
 
+    if (is_utf8) {
+        fprintf(stderr, "%s\n", buffer);
+        return;
+    }
+
     if (iconvW == NULL)
         iconvW = iconv_open("WCHAR_T", "utf-8");
 
@@ -70,9 +84,9 @@ void FcitxLogFunc(ErrorLevel e, const char* filename, const int line, const char
         fprintf(stderr, "%s\n", buffer);
     } else {
         size_t len = strlen(buffer);
-        wchar_t *wmessage;
-        size_t wlen = (len + 1) * sizeof(wchar_t);
-        wmessage = (wchar_t *) fcitx_utils_malloc0((len + 1) * sizeof(wchar_t));
+        wchar_t *wmessage = NULL;
+        size_t wlen = (len) * sizeof(wchar_t);
+        wmessage = (wchar_t *) fcitx_utils_malloc0((len + 10) * sizeof(wchar_t));
 
         IconvStr inp = buffer;
         char *outp = (char*) wmessage;
