@@ -66,6 +66,7 @@ FcitxMemoryPool* fcitx_memory_pool_create();
  **/
 void* fcitx_memory_pool_alloc(FcitxMemoryPool* pool, size_t size);
 
+void* fcitx_memory_pool_alloc_align(FcitxMemoryPool* pool, size_t size, int align);
 /**
  * free memory pool and free all the memory inside the pool
  *
@@ -74,6 +75,15 @@ void* fcitx_memory_pool_alloc(FcitxMemoryPool* pool, size_t size);
  **/
 void fcitx_memory_pool_destroy(FcitxMemoryPool* pool);
 
+/**
+ * free all the memory inside the pool but keep the pool itself
+ *
+ * @param pool memory ppol
+ * @return void
+ * @since 4.2.6
+ **/
+void fcitx_memory_pool_clear(FcitxMemoryPool* pool);
+
 #endif
 
 /**
