@@ -35,6 +35,7 @@
 
 
 #include <stdlib.h>
+#include <stdint.h>
 
 #ifdef __cplusplus
 extern "C" {
@@ -47,7 +48,7 @@ extern "C" {
 #define ISUTF8_CB(c)  (((c)&0xc0) == 0x80)
 
 /**
- * Get utf8 string lenght
+ * Get utf8 string length
  *
  * @param s string
  * @return length
@@ -61,7 +62,7 @@ size_t fcitx_utf8_strlen(const char *s);
  * @param chr return unicode
  * @return next char pointer
  **/
-char*  fcitx_utf8_get_char(const char *in, int *chr);
+char*  fcitx_utf8_get_char(const char *in, uint32_t *chr);
 
 /**
  * compare utf8 string, with utf8 string length n
@@ -91,7 +92,7 @@ int    fcitx_utf8_char_len(const char *in);
  * @param n index
  * @return next n character pointer
  **/
-char*  fcitx_utf8_get_nth_char(char* s, unsigned int n);
+char*  fcitx_utf8_get_nth_char(char* s, uint32_t n);
 
 /**
  * check utf8 string is valid or not, valid is 1, invalid is 0
@@ -143,6 +144,71 @@ void fcitx_utf8_strncpy(char* str, const char* s, size_t byte);
  **/
 size_t fcitx_utf8_strnlen(const char* str, size_t byte);
 
+/**
+ * @brief get ucs4 char length
+ *
+ * @param c ucs4 char
+ * @return int
+ *
+ * @since 4.2.5
+ **/
+int fcitx_ucs4_char_len(uint32_t c);
+
+
+/**
+ * @brief convert ucs4 char to utf8
+ *
+ * @param c ucs4 char
+ * @param output output string, need to reserve enough space
+ * @return int
+ *
+ * @since 4.2.5
+ **/
+int fcitx_ucs4_to_utf8(uint32_t c, char* output);
+
+/**
+ * @brief get the ascii part at the end of a utf8 string
+ *
+ * @param string a utf8 string
+ * @return string pointer to the ascii part
+ *
+ * @since 4.2.6
+ **/
+char *fcitx_utils_get_ascii_part(char *string);
+
+/**
+ * @brief get the ascii part at the end of a utf8 string (with a given size)
+ *
+ * @param string a utf8 string
+ * @param len the length of the string
+ * @return string pointer to the ascii part
+ *
+ * @since 4.2.6
+ **/
+char *fcitx_utils_get_ascii_partn(char *string, size_t len);
+
+/**
+ * @brief get the position of the first non-ascii character in a string (with a size limit)
+ *
+ * @param string a utf8 string
+ * @param len the length of the string
+ * @return string pointer to the position of the first non-ascii character or the end of string
+ *
+ * @since 4.2.6
+ **/
+char *fcitx_utils_get_ascii_endn(const char *string, size_t len);
+
+/**
+ * @brief get the position of the first non-ascii character in a string
+ *
+ * @param string a utf8 string
+ * @param len the length of the string
+ * @return string pointer to the position of the first non-ascii character or the end of string
+ *
+ * @since 4.2.6
+ **/
+char *fcitx_utils_get_ascii_end(const char *string);
+
 #ifdef __cplusplus
 }
 #endif
