@@ -28,9 +28,10 @@
 
 #define FCITX_USE_INTERNAL_PATH
 
+#include "config.h"
+
 #include <pthread.h>
 #include <stdio.h>
-#include <libintl.h>
 #include <signal.h>
 #include <unistd.h>
 #include <sys/wait.h>
@@ -39,8 +40,12 @@
 #include <libgen.h>
 #include <ctype.h>
 #include <stdarg.h>
+#include <errno.h>
+
+#if defined(ENABLE_BACKTRACE)
+#include <execinfo.h>
+#endif
 
-#include "config.h"
 #include "fcitx/fcitx.h"
 #include "utils.h"
 #include "utf8.h"
@@ -54,6 +59,16 @@
 #include <sys/user.h>
 #endif
 
+#if defined(__linux__)
+#include <sys/prctl.h>
+#endif
+
+#if defined(__linux__)
+#include <endian.h>
+#else
+#include <sys/endian.h>
+#endif
+
 FCITX_EXPORT_API
 int fcitx_utils_calculate_record_number(FILE* fpDict)
 {
@@ -103,6 +118,8 @@ void *fcitx_utils_custom_bsearch(const void *key, const void *base,
     }
 }
 
+typedef void (*_fcitx_sighandler_t) (int);
+
 FCITX_EXPORT_API
 void fcitx_utils_init_as_daemon()
 {
@@ -112,16 +129,24 @@ void fcitx_utils_init_as_daemon()
         exit(0);
     }
     setsid();
-    signal(SIGINT, SIG_IGN);
-    signal(SIGHUP, SIG_IGN);
-    signal(SIGQUIT, SIG_IGN);
-    signal(SIGPIPE, SIG_IGN);
-    signal(SIGTTOU, SIG_IGN);
-    signal(SIGTTIN, SIG_IGN);
-    signal(SIGCHLD, SIG_IGN);
+    _fcitx_sighandler_t oldint = signal(SIGINT, SIG_IGN);
+    _fcitx_sighandler_t oldhup  =signal(SIGHUP, SIG_IGN);
+    _fcitx_sighandler_t oldquit = signal(SIGQUIT, SIG_IGN);
+    _fcitx_sighandler_t oldpipe = signal(SIGPIPE, SIG_IGN);
+    _fcitx_sighandler_t oldttou = signal(SIGTTOU, SIG_IGN);
+    _fcitx_sighandler_t oldttin = signal(SIGTTIN, SIG_IGN);
+    _fcitx_sighandler_t oldchld = signal(SIGCHLD, SIG_IGN);
     if (fork() > 0)
         exit(0);
     chdir("/");
+
+    signal(SIGINT, oldint);
+    signal(SIGHUP, oldhup);
+    signal(SIGQUIT, oldquit);
+    signal(SIGPIPE, oldpipe);
+    signal(SIGTTOU, oldttou);
+    signal(SIGTTIN, oldttin);
+    signal(SIGCHLD, oldchld);
 }
 
 FCITX_EXPORT_API
@@ -200,6 +225,20 @@ char* fcitx_utils_join_string_list(UT_array* list, char delm)
     return result;
 }
 
+FCITX_EXPORT_API
+int fcitx_utils_string_list_contains(UT_array* list, const char* scmp)
+{
+    char** str;
+    for (str = (char**) utarray_front(list);
+         str != NULL;
+         str = (char**) utarray_next(list, str))
+    {
+        if (strcmp(scmp, *str) == 0)
+            return 1;
+    }
+    return 0;
+}
+
 FCITX_EXPORT_API
 void fcitx_utils_free_string_list(UT_array* list)
 {
@@ -271,14 +310,32 @@ int fcitx_utils_get_display_number()
             *p = '\0';
         }
 
-        if (strDisplayNumber)
-            sscanf(strDisplayNumber, "%d", &displayNumber);
+        if (strDisplayNumber) {
+            displayNumber = atoi(strDisplayNumber);
+        }
 
         free(display);
     }
     return displayNumber;
 }
 
+FCITX_EXPORT_API
+int fcitx_utils_current_locale_is_utf8()
+{
+    const char* p;
+    p = getenv("LC_CTYPE");
+    if (!p) {
+        p = getenv("LC_ALL");
+        if (!p)
+            p = getenv("LANG");
+    }
+    if (p) {
+        if (strcasestr(p, "utf8") || strcasestr(p, "utf-8"))
+            return 1;
+    }
+    return 0;
+}
+
 FCITX_EXPORT_API
 char* fcitx_utils_get_current_langcode()
 {
@@ -306,71 +363,45 @@ char* fcitx_utils_get_current_langcode()
 FCITX_EXPORT_API
 int fcitx_utils_pid_exists(pid_t pid)
 {
-#if defined(__linux__)
-    // verify if a process is running
-    char path[256];
-    snprintf(path, 255, "/proc/%d", pid);
-    struct stat buf;
-    if (stat(path, &buf) != 0) {
+    if (pid <= 0)
         return 0;
-    }
-    if (!S_ISDIR(buf.st_mode)) {
-        return 0;
-    }
-    return 1;
-#elif defined(LIBKVM_FOUND)
-    kvm_t *vm = kvm_open(0, "/dev/null", 0, O_RDONLY, NULL);
-    if (vm == 0) // ignore all error
-        return 1;
-
-    int result = 1;
-    do {
-        int cnt;
-        struct kinfo_proc * kp = kvm_getprocs(vm, KERN_PROC_PID, pid, &cnt);
-        if (kp == 0) {
-            result = 1;
-            break;
-        }
-        int i;
-        for (i = 0; i < cnt; i++)
-            if (kp->ki_pid == pid)
-                break;
-        if (i != cnt)
-            result = 1;
-        else
-            result = 0;
-    } while (0);
-    kvm_close(vm);
-    return result;
-#else
-    return 1;
-#endif
+    return !(kill(pid, 0) && (errno == ESRCH));
 }
 
 FCITX_EXPORT_API
 char* fcitx_utils_get_process_name()
 {
 #if defined(__linux__)
-    do {
-        FILE* fp = fopen("/proc/self/stat", "r");
-        if (!fp)
-            break;
+    #define _PR_GET_NAME_MAX 16
+    char name[_PR_GET_NAME_MAX + 1];
+    if (prctl(PR_GET_NAME, (unsigned long)name, 0, 0, 0))
+        return strdup("");
+    name[_PR_GET_NAME_MAX] = '\0';
+    return strdup(name);
+    /**
+     * Keep the old code here in case we want to get the name of
+     * another process with known pid sometime.
+     **/
+    /* do { */
+    /*     FILE* fp = fopen("/proc/self/stat", "r"); */
+    /*     if (!fp) */
+    /*         break; */
 
-        const size_t bufsize = 1024;
-        char buf[bufsize];
-        fgets(buf, bufsize, fp);
-        fclose(fp);
+    /*     const size_t bufsize = 1024; */
+    /*     char buf[bufsize]; */
+    /*     fgets(buf, bufsize, fp); */
+    /*     fclose(fp); */
 
-        char* S = strchr(buf, '(');
-        if (!S)
-            break;
-        char* E = strchr(S, ')');
-        if (!E)
-            break;
+    /*     char* S = strchr(buf, '('); */
+    /*     if (!S) */
+    /*         break; */
+    /*     char* E = strchr(S, ')'); */
+    /*     if (!E) */
+    /*         break; */
 
-        return strndup(S+1, E-S-1);
-    } while(0);
-    return strdup("");
+    /*     return strndup(S+1, E-S-1); */
+    /* } while(0); */
+    /* return strdup(""); */
 #elif defined(LIBKVM_FOUND)
     kvm_t *vm = kvm_open(0, "/dev/null", 0, O_RDONLY, NULL);
     if (vm == 0)
@@ -467,46 +498,198 @@ void fcitx_utils_string_swap(char** obj, const char* str)
 }
 
 FCITX_EXPORT_API
-void fcitx_utils_launch_configure_tool()
+void fcitx_utils_launch_tool(const char* name, const char* arg)
 {
-    char* command = fcitx_utils_get_fcitx_path_with_filename("bindir", "/fcitx-configtool &");
-    FILE* p = popen(command, "r");
+    char* command = fcitx_utils_get_fcitx_path_with_filename("bindir", name);
+    char* darg = NULL;
+    if (arg)
+        darg = strdup(arg);
+    char* args[] = {
+        command,
+        darg,
+        0
+    };
+    fcitx_utils_start_process(args);
+    fcitx_utils_free(darg);
     free(command);
-    if (p)
-        pclose(p);
-    else
-        FcitxLog(ERROR, _("Unable to create process"));
 }
 
+FCITX_EXPORT_API
+void fcitx_utils_launch_configure_tool()
+{
+    fcitx_utils_launch_tool("fcitx-configtool", NULL);
+}
 
 FCITX_EXPORT_API
 void fcitx_utils_launch_configure_tool_for_addon(const char* imaddon)
 {
-    char* command = fcitx_utils_get_fcitx_path_with_filename("bindir", "/fcitx-configtool");
-    char* commandf;
-    asprintf(&commandf, "%s %s &", command, imaddon);
-    FILE* p = popen(commandf, "r");
-    free(command);
-    free(commandf);
-    if (p)
-        pclose(p);
-    else
-        FcitxLog(ERROR, _("Unable to create process"));
+    fcitx_utils_launch_tool("fcitx-configtool", imaddon);
 }
 
 FCITX_EXPORT_API
 void fcitx_utils_launch_restart()
 {
-    char* command = fcitx_utils_get_fcitx_path_with_filename("bindir", "/fcitx");
-    char* commandf;
-    asprintf(&commandf, "%s -r &", command);
-    FILE* p = popen(commandf, "r");
-    free(command);
-    free(commandf);
-    if (p)
-        pclose(p);
-    else
-        FcitxLog(ERROR, _("Unable to create process"));
+    fcitx_utils_launch_tool("fcitx", "-r");
+}
+
+FCITX_EXPORT_API
+void fcitx_utils_start_process(char** args)
+{
+    /* exec command */
+    pid_t child_pid;
+
+    child_pid = fork();
+    if (child_pid < 0) {
+        perror("fork");
+    } else if (child_pid == 0) {         /* child process  */
+        pid_t grandchild_pid;
+
+        grandchild_pid = fork();
+        if (grandchild_pid < 0) {
+            perror("fork");
+            _exit(1);
+        } else if (grandchild_pid == 0) { /* grandchild process  */
+            execvp(args[0], args);
+            perror("execvp");
+            _exit(1);
+        } else {
+            _exit(0);
+        }
+    } else {                              /* parent process */
+        int status;
+        waitpid(child_pid, &status, 0);
+    }
+}
+
+FCITX_EXPORT_API
+void
+fcitx_utils_backtrace()
+{
+#if defined(ENABLE_BACKTRACE)
+    void *array[20];
+
+    size_t size;
+    char **strings = NULL;
+    size_t i;
+
+    size = backtrace(array, 20);
+    strings = backtrace_symbols(array, size);
+
+    if (strings) {
+        fprintf(stderr, "Obtained %zd stack frames.\n", size);
+
+        for (i = 0; i < size; i++) {
+            fprintf(stderr, "%s\n", strings[i]);
+        }
+
+        free(strings);
+    }
+#endif
+}
+
+FCITX_EXPORT_API
+int
+fcitx_utils_get_boolean_env(const char *name,
+                            int defval)
+{
+    const char *value = getenv(name);
+
+    if (value == NULL)
+        return defval;
+
+    if ((!*value) ||
+        strcmp(value, "0") == 0 ||
+        strcasecmp(value, "false") == 0)
+        return 0;
+
+    return 1;
+}
+
+FCITX_EXPORT_API
+size_t
+fcitx_utils_read_uint32(FILE *fp, uint32_t *p)
+{
+    uint32_t res = 0;
+    size_t size;
+    size = fread(&res, sizeof(uint32_t), 1, fp);
+    *p = le32toh(res);
+    return size;
+}
+
+FCITX_EXPORT_API
+size_t
+fcitx_utils_write_uint32(FILE *fp, uint32_t i)
+{
+    i = htole32(i);
+    return fwrite(&i, sizeof(uint32_t), 1, fp);
+}
+
+FCITX_EXPORT_API
+size_t
+fcitx_utils_read_uint16(FILE *fp, uint16_t *p)
+{
+    uint16_t res = 0;
+    size_t size;
+    size = fread(&res, sizeof(uint16_t), 1, fp);
+    *p = le16toh(res);
+    return size;
+}
+
+FCITX_EXPORT_API
+size_t
+fcitx_utils_write_uint16(FILE *fp, uint16_t i)
+{
+    i = htole16(i);
+    return fwrite(&i, sizeof(uint16_t), 1, fp);
+}
+
+FCITX_EXPORT_API
+size_t
+fcitx_utils_read_uint64(FILE *fp, uint64_t *p)
+{
+    uint64_t res = 0;
+    size_t size;
+    size = fread(&res, sizeof(uint64_t), 1, fp);
+    *p = le64toh(res);
+    return size;
+}
+
+FCITX_EXPORT_API
+size_t
+fcitx_utils_write_uint64(FILE *fp, uint64_t i)
+{
+    i = htole64(i);
+    return fwrite(&i, sizeof(uint64_t), 1, fp);
+}
+
+FCITX_EXPORT_API
+int
+fcitx_utils_strcmp0(const char* a, const char* b)
+{
+    if (a == NULL && b == NULL)
+        return 0;
+    if (a == NULL && b)
+        return -1;
+    if (a && b == NULL)
+        return 1;
+    return strcmp(a, b);
+
+}
+
+FCITX_EXPORT_API
+int
+fcitx_utils_strcmp_empty(const char* a, const char* b)
+{
+    int isemptya = (a == NULL || (*a) == 0);
+    int isemptyb = (b == NULL || (*b) == 0);
+    if (isemptya && isemptyb)
+        return 0;
+    if (isemptya && !isemptyb)
+        return -1;
+    if (!isemptya && isemptyb)
+        return 1;
+    return strcmp(a, b);
+
 }
 
 // kate: indent-mode cstyle; space-indent on; indent-width 0;
