@@ -68,16 +68,43 @@ void FcitxAddonsInit(UT_array* addons)
     utarray_init(addons, &addon_icd);
 }
 
+void* FcitxGetSymbol(void* handle, const char* addonName, const char* symbolName)
+{
+    char* escapedAddonName;
+    asprintf(&escapedAddonName, "%s_%s", addonName, symbolName);
+    char* p = escapedAddonName;
+    while (*p) {
+        if (*p == '-') {
+            *p = '_';
+        }
+        p++;
+    }
+    void* result = dlsym(handle, escapedAddonName);
+    if (!result)
+        result = dlsym(handle, symbolName);
+
+    return result;
+}
+
 /**
  * Load Addon Info
  */
 FCITX_EXPORT_API
 void FcitxAddonsLoad(UT_array* addons)
+{
+    FcitxAddonsLoadInternal(addons, false);
+}
+
+FcitxAddon* FcitxAddonsLoadInternal(UT_array* addons, boolean reloadIM)
 {
     char **addonPath;
     size_t len;
-    size_t i = 0;
-    utarray_clear(addons);
+    size_t k = 0;
+    size_t start;
+    if (!reloadIM)
+        utarray_clear(addons);
+
+    start = utarray_len(addons);
 
     FcitxStringHashSet* sset = FcitxXDGGetFiles(
                               "addon",
@@ -86,18 +113,16 @@ void FcitxAddonsLoad(UT_array* addons)
                           );
     addonPath = FcitxXDGGetPathWithPrefix(&len, "addon");
     char **paths = malloc(sizeof(char*) * len);
-    for (i = 0; i < len ; i ++)
-        paths[i] = NULL;
-    FcitxStringHashSet* string;
-    for (string = sset;
-            string != NULL;
-            string = (FcitxStringHashSet*)string->hh.next) {
+    for (k = 0; k < len ; k ++)
+        paths[k] = NULL;
+
+
+    HASH_FOREACH(string, sset, FcitxStringHashSet) {
         int i = 0;
         for (i = len - 1; i >= 0; i--) {
             asprintf(&paths[i], "%s/%s", addonPath[len - i - 1], string->name);
             FcitxLog(DEBUG, "Load Addon Config File:%s", paths[i]);
         }
-        FcitxLog(INFO, _("Load Addon Config File:%s"), string->name);
         FcitxConfigFile* cfile = FcitxConfigParseMultiConfigFile(paths, len, FcitxAddonGetConfigDesc());
         if (cfile) {
             utarray_extend_back(addons);
@@ -106,6 +131,19 @@ void FcitxAddonsLoad(UT_array* addons)
             FcitxAddonConfigBind(a, cfile, FcitxAddonGetConfigDesc());
             FcitxConfigBindSync((FcitxGenericConfig*)a);
             FcitxLog(DEBUG, _("Addon Config %s is %s"), string->name, (a->bEnabled) ? "Enabled" : "Disabled");
+            boolean error = false;
+            if (reloadIM) {
+                if (a->category !=  AC_INPUTMETHOD)
+                    error = true;
+            }
+            /* if loaded, don't touch the old one */
+            if (FcitxAddonsGetAddonByNameInternal(addons, a->name, true) != a)
+                error = true;
+
+            if (error)
+                utarray_pop_back(addons);
+            else
+                FcitxLog(INFO, _("Load Addon Config File:%s"), string->name);
         }
 
         for (i = 0; i < len ; i ++) {
@@ -119,73 +157,112 @@ void FcitxAddonsLoad(UT_array* addons)
 
     fcitx_utils_free_string_hash_set(sset);
 
-    utarray_sort(addons, AddonPriorityCmp);
+    size_t to = utarray_len(addons);
+    utarray_sort_range(addons, AddonPriorityCmp, start, to);
+
+    return (FcitxAddon*) utarray_eltptr(addons, start);
+}
+
+void FcitxInstanceFillAddonOwner(FcitxInstance* instance, FcitxAddon* addonHead)
+{
+    /* FIXME: a walkaround for not have instance in function FcitxModuleInvokeFunction */
+    FcitxAddon* addon;
+    if (addonHead)
+        addon = addonHead;
+    else
+        addon = (FcitxAddon *) utarray_front(&instance->addons);
+    for (; addon != NULL; addon = (FcitxAddon *) utarray_next(&instance->addons, addon)) {
+        addon->owner = instance;
+    }
 }
 
 FCITX_EXPORT_API
 void FcitxInstanceResolveAddonDependency(FcitxInstance* instance)
+{
+    FcitxInstanceResolveAddonDependencyInternal(instance, NULL);
+}
+
+void FcitxInstanceResolveAddonDependencyInternal(FcitxInstance* instance, FcitxAddon* startAddon)
 {
     UT_array* addons = &instance->addons;
     boolean remove = true;
     FcitxAddon *addon;
     FcitxAddon *uiaddon = NULL, *uifallbackaddon = NULL;
+    boolean reloadIM = true;
 
-    /* choose ui */
-    for (addon = (FcitxAddon *) utarray_front(addons);
-            addon != NULL;
-            addon = (FcitxAddon *) utarray_next(addons, addon)) {
-        if (addon->category == AC_UI) {
-            if (instance->uiname == NULL) {
-                if (addon->bEnabled) {
-                    uiaddon = addon;
-                    break;
+    if (!startAddon) {
+        startAddon = (FcitxAddon*) utarray_front(addons);
+        reloadIM = false;
+    }
+
+    /* override the enable and disable option */
+    for (addon = startAddon;
+         addon != NULL;
+         addon = (FcitxAddon *) utarray_next(addons, addon)) {
+        if (instance->disableList && fcitx_utils_string_list_contains(instance->disableList, addon->name))
+            addon->bEnabled = false;
+        else if (instance->enableList && fcitx_utils_string_list_contains(instance->enableList, addon->name))
+            addon->bEnabled = true;
+    }
+
+    if (!reloadIM) {
+        /* choose ui */
+        for (addon = startAddon;
+             addon != NULL;
+             addon = (FcitxAddon *) utarray_next(addons, addon)) {
+            if (addon->category == AC_UI) {
+                if (instance->uiname == NULL) {
+                    if (addon->bEnabled) {
+                        uiaddon = addon;
+                        break;
+                    }
+                } else {
+                    if (strcmp(instance->uiname, addon->name) == 0) {
+                        addon->bEnabled = true;
+                        uiaddon = addon;
+                        break;
+                    }
                 }
-            } else {
-                if (strcmp(instance->uiname, addon->name) == 0) {
-                    addon->bEnabled = true;
-                    uiaddon = addon;
+            }
+        }
+
+        if (uiaddon && uiaddon->uifallback) {
+            for (addon = startAddon;
+                 addon != NULL;
+                 addon = (FcitxAddon *) utarray_next(addons, addon)) {
+                if (addon->category == AC_UI && addon->bEnabled && strcmp(uiaddon->uifallback, addon->name) == 0) {
+                    FcitxAddon temp;
+                    int uiidx = utarray_eltidx(addons, uiaddon);
+                    int fallbackidx = utarray_eltidx(addons, addon);
+                    if (fallbackidx < uiidx) {
+                        temp = *uiaddon;
+                        *uiaddon = *addon;
+                        *addon = temp;
+
+                        /* they swapped, addon is normal ui, and ui addon is fallback */
+                        uifallbackaddon = uiaddon;
+                        uiaddon = addon;
+                    }
+                    else {
+                        uifallbackaddon = addon;
+                    }
                     break;
                 }
             }
         }
-    }
 
-    if (uiaddon && uiaddon->uifallback) {
-        for (addon = (FcitxAddon *) utarray_front(addons);
+        for (addon = startAddon;
                 addon != NULL;
                 addon = (FcitxAddon *) utarray_next(addons, addon)) {
-            if (addon->category == AC_UI && addon->bEnabled && strcmp(uiaddon->uifallback, addon->name) == 0) {
-                FcitxAddon temp;
-                int uiidx = utarray_eltidx(addons, uiaddon);
-                int fallbackidx = utarray_eltidx(addons, addon);
-                if (fallbackidx < uiidx) {
-                    temp = *uiaddon;
-                    *uiaddon = *addon;
-                    *addon = temp;
-
-                    /* they swapped, addon is normal ui, and ui addon is fallback */
-                    uifallbackaddon = uiaddon;
-                    uiaddon = addon;
-                }
-                else {
-                    uifallbackaddon = addon;
-                }
-                break;
+            if (addon->category == AC_UI && addon != uiaddon && addon != uifallbackaddon) {
+                addon->bEnabled = false;
             }
         }
     }
 
-    for (addon = (FcitxAddon *) utarray_front(addons);
-            addon != NULL;
-            addon = (FcitxAddon *) utarray_next(addons, addon)) {
-        if (addon->category == AC_UI && addon != uiaddon && addon != uifallbackaddon) {
-            addon->bEnabled = false;
-        }
-    }
-
     while (remove) {
         remove = false;
-        for (addon = (FcitxAddon *) utarray_front(addons);
+        for (addon = startAddon;
                 addon != NULL;
                 addon = (FcitxAddon *) utarray_next(addons, addon)) {
             if (!addon->bEnabled)
@@ -225,18 +302,25 @@ boolean FcitxAddonsIsAddonAvailable(UT_array* addons, const char* name)
 }
 
 FCITX_EXPORT_API
-FcitxAddon* FcitxAddonsGetAddonByName(UT_array* addons, const char* name)
+FcitxAddon* FcitxAddonsGetAddonByNameInternal(UT_array* addons, const char* name, boolean checkDisabled)
 {
     FcitxAddon *addon;
     for (addon = (FcitxAddon *) utarray_front(addons);
             addon != NULL;
             addon = (FcitxAddon *) utarray_next(addons, addon)) {
-        if (addon->bEnabled && strcmp(name, addon->name) == 0)
+        if ((checkDisabled || addon->bEnabled) && strcmp(name, addon->name) == 0)
             return addon;
     }
     return NULL;
 }
 
+
+FCITX_EXPORT_API
+FcitxAddon* FcitxAddonsGetAddonByName(UT_array* addons, const char* name)
+{
+    return FcitxAddonsGetAddonByNameInternal(addons, name, false);
+}
+
 /**
  * Load addon.desc file
  *
@@ -260,9 +344,9 @@ void FcitxAddonFree(void* v)
     free(addon->subconfig);
 }
 
-boolean CheckABIVersion(void* handle)
+boolean FcitxCheckABIVersion(void* handle, const char* addonName)
 {
-    int* version = (int*) dlsym(handle, "ABI_VERSION");
+    int* version = (int*) FcitxGetSymbol(handle, addonName, "ABI_VERSION");
     if (!version)
         return false;
     if (*version < FCITX_ABI_VERSION)
