@@ -44,6 +44,8 @@ CONFIG_BINDING_REGISTER("Output", "SendTextWhenSwitchEng", bSendTextWhenSwitchEn
 CONFIG_BINDING_REGISTER("Output", "CandidateWordNumber", iMaxCandWord)
 CONFIG_BINDING_REGISTER("Output", "PhraseTips", bPhraseTips)
 CONFIG_BINDING_REGISTER("Output", "DontCommitPreeditWhenUnfocus", bDontCommitPreeditWhenUnfocus)
+CONFIG_BINDING_REGISTER("Appearance", "ShowInputWindowOnlyWhenActive", bShowInputWindowOnlyWhenActive)
+CONFIG_BINDING_REGISTER("Appearance", "ShowInputWindowWhenFocusIn", bShowInputWindowWhenFocusIn)
 CONFIG_BINDING_REGISTER("Appearance", "ShowInputWindowAfterTriggering", bShowInputWindowTriggering)
 CONFIG_BINDING_REGISTER("Appearance", "ShowInputSpeed", bShowUserSpeed)
 CONFIG_BINDING_REGISTER("Appearance", "ShowVersion", bShowVersion)
@@ -52,7 +54,9 @@ CONFIG_BINDING_REGISTER("Appearance", "HideInputWindowWhenOnlyOneCandidate", bHi
 CONFIG_BINDING_REGISTER("Hotkey", "TriggerKey", hkTrigger)
 CONFIG_BINDING_REGISTER("Hotkey", "ActivateKey", hkActivate)
 CONFIG_BINDING_REGISTER("Hotkey", "InactivateKey", hkInactivate)
+CONFIG_BINDING_REGISTER("Hotkey", "UseExtraTriggerKeyOnlyWhenUseItToInactivate", bUseExtraTriggerKeyOnlyWhenUseItToInactivate)
 CONFIG_BINDING_REGISTER("Hotkey", "IMSwitchKey", bIMSwitchKey)
+CONFIG_BINDING_REGISTER("Hotkey", "IMSwitchIncludeInactive", bIMSwitchIncludeInactive)
 CONFIG_BINDING_REGISTER("Hotkey", "IMSwitchHotkey", iIMSwitchKey)
 CONFIG_BINDING_REGISTER("Hotkey", "SwitchKey", iSwitchKey)
 CONFIG_BINDING_REGISTER("Hotkey", "DoubleSwitchKey", bDoubleSwitchKey)
