@@ -65,6 +65,17 @@ extern "C" {
 /** fcitx addon ABI version, need to be used with addon */
 #define FCITX_ABI_VERSION 5
 
+#define FCITX_DEFINE_PLUGIN(name, category, type) \
+FCITX_EXPORT_API int name##_ABI_VERSION = FCITX_ABI_VERSION; \
+FCITX_EXPORT_API type name##_##category 
+
+#if defined(__GNUC__) && ((__GNUC__ * 100 + __GNUC_MINOR__) >= 301)
+# define FCITX_DEPRECATED  __attribute__((deprecated))
+#else /* not gcc >= 3.1 */
+# define FCITX_DEPRECATED
+#endif
+
+
 #ifdef __cplusplus
 }
 #endif
