@@ -26,6 +26,8 @@
 #ifndef _FCITX_FRONTEND_H_
 #define _FCITX_FRONTEND_H_
 
+#include <stdint.h>
+
 #include <fcitx-utils/utarray.h>
 #include <fcitx-config/fcitx-config.h>
 #include <fcitx-config/hotkey.h>
@@ -36,6 +38,9 @@ extern "C" {
 #endif
 
     struct _FcitxInstance;
+    typedef void* (*FcitxICDataAllocCallback)(void* arg);
+    typedef void (*FcitxICDataFreeCallback)(void* arg, void* data);
+    typedef void* (*FcitxICDataCopyCallback)(void* arg, void* data, void* src);
 
     /**
      * Input Method State
@@ -83,7 +88,9 @@ extern "C" {
     typedef struct _FcitxInputContext2 {
         FcitxInputContext inputContext;
         char* imname;
-        void* padding[16];
+        intptr_t switchBySwitchKey;
+        UT_array* data;
+        void* padding[14];
     } FcitxInputContext2;
 
     /**
@@ -278,6 +285,59 @@ extern "C" {
      **/
     void FcitxInstanceSetICStateFromSameApplication(struct _FcitxInstance* instance, int frontendid, FcitxInputContext *ic);
 
+    /**
+     * get a per-ic data
+     *
+     * @param instance fcitx instnace
+     * @param ic input context
+     * @param data slot id
+     *
+     * @return data
+     *
+     * @since 4.2.5
+     **/
+    void* FcitxInstanceGetICData(struct _FcitxInstance* instance, FcitxInputContext* ic, int icdataid);
+
+
+    /**
+     * get a per-ic data
+     *
+     * @param instance fcitx instnace
+     * @param ic input context
+     * @param data slot id
+     * @param newdata new data
+     *
+     * @return void
+     *
+     * @since 4.2.5
+     **/
+    void FcitxInstanceSetICData(struct _FcitxInstance* instance, FcitxInputContext* ic, int icdataid, void* newdata);
+
+    /**
+     * alloc a per-ic data slot
+     *
+     * @param instance fcitx instnace
+     * @param allocCallback alloc callback (set pointer to null, when callback is null)
+     * @param copyCallback copy callback (used when need to share state)
+     * @param freeCallback free callback
+     * @param arg closure data
+     * @return data slot id
+     *
+     * @since 4.2.5
+     **/
+    int FcitxInstanceAllocDataForIC(struct _FcitxInstance* instance, FcitxICDataAllocCallback allocCallback, FcitxICDataCopyCallback copyCallback, FcitxICDataFreeCallback freeCallback, void* arg);
+
+    /**
+     * check current ic support preedit or not
+     *
+     * @param instance fcitx instnace
+     * @param ic input context
+     * @return result
+     *
+     * @since 4.2.5
+     **/
+    boolean FcitxInstanceICSupportPreedit(struct _FcitxInstance* instance, FcitxInputContext* ic);
+
 #ifdef __cplusplus
 }
 #endif
