@@ -62,6 +62,9 @@ static const FcitxHotkey* switchKey1[] = {
     FCITX_ALT_RSHIFT,
     FCITX_RCTRL,
     FCITX_RSHIFT,
+    FCITX_LALT,
+    FCITX_RALT,
+    FCITX_LALT,
     FCITX_NONE_KEY,
 };
 
@@ -74,6 +77,9 @@ static const FcitxHotkey* switchKey2[] = {
     FCITX_NONE_KEY,
     FCITX_LCTRL,
     FCITX_LSHIFT,
+    FCITX_NONE_KEY,
+    FCITX_NONE_KEY,
+    FCITX_RALT,
     FCITX_NONE_KEY
 };
 
@@ -112,9 +118,9 @@ static boolean IMMenuAction(FcitxUIMenu* menu, int index);
 static void UpdateIMMenuItem(FcitxUIMenu *menu);
 static void FcitxInstanceEnableIMInternal(FcitxInstance* instance, FcitxInputContext* ic, boolean keepState);
 static void FcitxInstanceCloseIMInternal(FcitxInstance* instance, FcitxInputContext* ic);
-static void FcitxInstanceChangeIMStateInternal(FcitxInstance* instance, FcitxInputContext* ic, FcitxContextState objectState);
+static void FcitxInstanceChangeIMStateWithKey(FcitxInstance* instance, FcitxInputContext* ic, boolean withSwitchKey);
+static void FcitxInstanceChangeIMStateInternal(FcitxInstance* instance, FcitxInputContext* ic, FcitxContextState objectState, boolean withSwitchKey);
 static void FreeIMEntry(FcitxIMEntry* entry);
-static boolean FcitxInstanceCheckICFromSameApplication (FcitxInstance* instance, FcitxInputContext* rec, FcitxInputContext* ic);
 
 FCITX_GETTER_VALUE(FcitxInputState, IsInRemind, bIsInRemind, boolean)
 FCITX_SETTER(FcitxInputState, IsInRemind, bIsInRemind, boolean)
@@ -233,8 +239,7 @@ void FcitxInstanceSaveAllIM(FcitxInstance* instance)
     }
 }
 
-
-static boolean FcitxInstanceCheckICFromSameApplication (FcitxInstance* instance, FcitxInputContext* rec, FcitxInputContext* ic) {
+boolean FcitxInstanceCheckICFromSameApplication (FcitxInstance* instance, FcitxInputContext* rec, FcitxInputContext* ic) {
     if (rec->frontendid != ic->frontendid)
         return false;
     if (rec == ic)
@@ -264,13 +269,13 @@ void FcitxInstanceLoadIM(FcitxInstance* instance, FcitxAddon* addon)
         char* modulePath;
         FILE *fp = FcitxXDGGetLibFile(addon->library, "r", &modulePath);
         void *handle;
-        FcitxIMClass * imclass;
+        FcitxIMClass2 * imclass2;
         if (!fp) {
             free(modulePath);
             return;
         }
         fclose(fp);
-        handle = dlopen(modulePath, RTLD_LAZY | RTLD_GLOBAL);
+        handle = dlopen(modulePath, RTLD_NOW | RTLD_GLOBAL);
         if (!handle) {
             FcitxLog(ERROR, _("IM: open %s fail %s") , modulePath , dlerror());
             free(modulePath);
@@ -278,43 +283,48 @@ void FcitxInstanceLoadIM(FcitxInstance* instance, FcitxAddon* addon)
         }
         free(modulePath);
 
-        if (!CheckABIVersion(handle)) {
+        if (!FcitxCheckABIVersion(handle, addon->name)) {
             FcitxLog(ERROR, "%s ABI Version Error", addon->name);
             dlclose(handle);
             return;
         }
 
-        imclass = dlsym(handle, "ime");
-        if (!imclass || !imclass->Create) {
+        boolean isIMClass2 = false;
+        imclass2 = FcitxGetSymbol(handle, addon->name, "ime2");
+        if (!imclass2)
+            imclass2 = FcitxGetSymbol(handle, addon->name, "ime");
+        else
+            isIMClass2 = true;
+        if (!imclass2 || !imclass2->Create) {
             FcitxLog(ERROR, _("IM: bad im %s"), addon->name);
             dlclose(handle);
             return;
         }
 
+        addon->isIMClass2 = isIMClass2;
         instance->currentIMAddon = addon;
-        if ((addon->addonInstance = imclass->Create(instance)) == NULL) {
+        if ((addon->addonInstance = imclass2->Create(instance)) == NULL) {
             dlclose(handle);
             return;
         }
         instance->currentIMAddon = NULL;
-        addon->imclass = imclass;
+        addon->imclass2 = imclass2;
         utarray_push_back(&instance->imeclasses, &addon);
     }
 }
 
-void FcitxRegisterEmptyEntry(FcitxInstance *instance,
-                             const char* name,
-                             const char* uniqueName,
-                             const char* iconName,
-                             int priority,
-                             const char *langCode,
-                             FcitxAddon* addon
-                            )
+void FcitxInstanceRegisterEmptyIMEntry(FcitxInstance *instance,
+                                       const char* name,
+                                       const char* uniqueName,
+                                       const char* iconName,
+                                       int priority,
+                                       const char *langCode,
+                                       FcitxAddon* addon
+                                      )
 {
     UT_array* imes = &instance->availimes ;
     FcitxIM* entry = FcitxInstanceGetIMFromIMList(instance, IMAS_Disable, uniqueName);
     if (entry) {
-        FcitxLog(ERROR, "%s already exists", uniqueName);
         return;
     } else {
         utarray_extend_back(imes);
@@ -384,7 +394,7 @@ void FcitxInstanceRegisterIMv2(FcitxInstance *instance,
                        const char *langCode
                       )
 {
-    if (priority < 0)
+    if (priority <= 0)
         return ;
 
     if (priority == PRIORITY_MAGIC_FIRST)
@@ -456,7 +466,7 @@ boolean FcitxInstanceLoadAllIM(FcitxInstance* instance)
                 && addon->category == AC_INPUTMETHOD
                 && addon->registerMethod == IMRM_CONFIGFILE
                ) {
-                FcitxRegisterEmptyEntry(instance,
+                FcitxInstanceRegisterEmptyIMEntry(instance,
                                         entry->name,
                                         entry->uniqueName,
                                         entry->iconName,
@@ -477,7 +487,11 @@ boolean FcitxInstanceLoadAllIM(FcitxInstance* instance)
     for (addon = (FcitxAddon *) utarray_front(addons);
             addon != NULL;
             addon = (FcitxAddon *) utarray_next(addons, addon)) {
-        if (addon->bEnabled && addon->category == AC_INPUTMETHOD) {
+        /*
+         * since we want to use this function to do some "reload"
+         * we need to check addon->addonInstance
+         */
+        if (addon->bEnabled && addon->category == AC_INPUTMETHOD && !addon->addonInstance) {
             switch (addon->type) {
             case AT_SHAREDLIBRARY: {
                 if (addon->registerMethod == IMRM_SELF)
@@ -523,12 +537,21 @@ INPUT_RETURN_VALUE FcitxInstanceProcessKey(
     if (instance->CurrentIC == NULL)
         return IRV_TO_PROCESS;
 
+    if (event == FCITX_PRESS_KEY
+        && !FcitxHotkeyIsHotKeyModifierCombine(sym, state))
+    {
+        if (FcitxInstanceRemoveTimeoutByFunc(instance, HideInputSpeed))
+            HideInputSpeed(instance);
+    }
+
     if (instance->CurrentIC->contextCaps & CAPACITY_PASSWORD)
         return IRV_TO_PROCESS;
 
     if (currentIM == NULL)
         return IRV_TO_PROCESS;
 
+#define HAVE_IM (utarray_len(&instance->imes) > 1)
+
     /*
      * for following reason, we cannot just process switch key, 2nd, 3rd key as other simple hotkey
      * because ctrl, shift, alt are compose key, so hotkey like ctrl+a will also produce a key
@@ -537,37 +560,46 @@ INPUT_RETURN_VALUE FcitxInstanceProcessKey(
      */
 
     /* process keyrelease event for switch key and 2nd, 3rd key */
-    if (event == FCITX_RELEASE_KEY) {
-        if (FcitxInstanceGetCurrentState(instance) != IS_CLOSED) {
-            if ((timestamp - input->lastKeyPressedTime) < 500 && (!input->bIsDoInputOnly)) {
-                if ((input->bIsInRemind || FcitxCandidateWordGetPageSize(input->candList) != 0)
-                    && FcitxHotkeyIsHotKey(sym, state, fc->i2ndSelectKey)
-                    && input->keyReleased == KR_2ND_SELECTKEY) {
-                    if (!input->bIsInRemind) {
-                        retVal = FcitxCandidateWordChooseByIndex(input->candList, 1);
-                    } else {
-                        strcpy(FcitxInputStateGetOutputString(input), " ");
-                        retVal = IRV_COMMIT_STRING;
-                    }
-                    input->keyReleased = KR_OTHER;
-                } else if ((input->bIsInRemind || FcitxCandidateWordGetPageSize(input->candList) != 0)
-                            && FcitxHotkeyIsHotKey(sym, state, fc->i3rdSelectKey)
-                            && input->keyReleased == KR_3RD_SELECTKEY) {
-                    if (!input->bIsInRemind) {
-                        retVal = FcitxCandidateWordChooseByIndex(input->candList, 2);
-                    } else {
-                        strcpy(FcitxInputStateGetOutputString(input), "\xe3\x80\x80");
-                        retVal = IRV_COMMIT_STRING;
-                    }
+    if (event == FCITX_RELEASE_KEY
+        && FcitxInstanceGetCurrentState(instance) != IS_CLOSED
+        && (timestamp - input->lastKeyPressedTime) < 500
+        && (!input->bIsDoInputOnly)) {
+        if ((input->bIsInRemind || FcitxCandidateWordGetPageSize(input->candList) != 0)
+            && FcitxHotkeyIsHotKey(sym, state, fc->i2ndSelectKey)
+            && input->keyReleased == KR_2ND_SELECTKEY) {
+            if (!input->bIsInRemind) {
+                retVal = FcitxCandidateWordChooseByIndex(input->candList, 1);
+            } else {
+                strcpy(FcitxInputStateGetOutputString(input), " ");
+                retVal = IRV_COMMIT_STRING;
+            }
+            input->keyReleased = KR_OTHER;
+        } else if ((input->bIsInRemind || FcitxCandidateWordGetPageSize(input->candList) != 0)
+                    && FcitxHotkeyIsHotKey(sym, state, fc->i3rdSelectKey)
+                    && input->keyReleased == KR_3RD_SELECTKEY) {
+            if (!input->bIsInRemind) {
+                retVal = FcitxCandidateWordChooseByIndex(input->candList, 2);
+            } else {
+                strcpy(FcitxInputStateGetOutputString(input), "\xe3\x80\x80");
+                retVal = IRV_COMMIT_STRING;
+            }
 
-                    input->keyReleased = KR_OTHER;
-                } else if (fc->bIMSwitchKey && (FcitxHotkeyIsHotKey(sym, state, imSWNextKey1[fc->iIMSwitchKey]) || FcitxHotkeyIsHotKey(sym, state, imSWNextKey2[fc->iIMSwitchKey]))) {
+            input->keyReleased = KR_OTHER;
+        } else {
+            if (HAVE_IM) {
+                if (fc->bIMSwitchKey
+                    && (fc->bIMSwitchIncludeInactive || FcitxInstanceGetCurrentState(instance) == IS_ACTIVE)
+                    && (FcitxHotkeyIsHotKey(sym, state, imSWNextKey1[fc->iIMSwitchKey]) || FcitxHotkeyIsHotKey(sym, state, imSWNextKey2[fc->iIMSwitchKey]))
+                       ) {
                     if (input->keyReleased == KR_SWITCH_IM) {
-                        FcitxInstanceSwitchIMByIndex(instance, -1);
+                        FcitxInstanceSwitchIMByIndex(instance, fc->bIMSwitchIncludeInactive ? -1 : -3);
                     }
-                } else if (fc->bIMSwitchKey && (FcitxHotkeyIsHotKey(sym, state, imSWPrevKey1[fc->iIMSwitchKey]) || FcitxHotkeyIsHotKey(sym, state, imSWPrevKey2[fc->iIMSwitchKey]))) {
+                } else if (fc->bIMSwitchKey
+                           && (fc->bIMSwitchIncludeInactive || FcitxInstanceGetCurrentState(instance) == IS_ACTIVE)
+                           && (FcitxHotkeyIsHotKey(sym, state, imSWPrevKey1[fc->iIMSwitchKey]) || FcitxHotkeyIsHotKey(sym, state, imSWPrevKey2[fc->iIMSwitchKey]))
+                        ) {
                     if (input->keyReleased == KR_SWITCH_IM_REVERSE) {
-                        FcitxInstanceSwitchIMByIndex(instance, -2);
+                        FcitxInstanceSwitchIMByIndex(instance, fc->bIMSwitchIncludeInactive ? -2 : -4);
                     }
                 } else if ((FcitxHotkeyIsHotKey(sym, state, switchKey1[fc->iSwitchKey]) || FcitxHotkeyIsHotKey(sym, state, switchKey2[fc->iSwitchKey])) && input->keyReleased == KR_SWITCH && !fc->bDoubleSwitchKey) {
                     retVal = IRV_DONOT_PROCESS;
@@ -578,9 +610,8 @@ INPUT_RETURN_VALUE FcitxInstanceProcessKey(
                         }
                     }
                     input->keyReleased = KR_OTHER;
-                    if (FcitxInstanceGetCurrentState(instance) == IS_INACTIVE)
-                        FcitxInstanceShowInputSpeed(instance);
-                    FcitxInstanceChangeIMState(instance, instance->CurrentIC);
+                    FcitxInstanceChangeIMStateWithKey(instance, instance->CurrentIC, true);
+                    FcitxInstanceShowInputSpeed(instance);
                 }
             }
         }
@@ -610,17 +641,20 @@ INPUT_RETURN_VALUE FcitxInstanceProcessKey(
                 }
             }
             if (!(FcitxHotkeyIsHotKey(sym, state, switchKey1[fc->iSwitchKey]) || FcitxHotkeyIsHotKey(sym, state, switchKey2[fc->iSwitchKey])))
-                    input->keyReleased = KR_OTHER;
+                input->keyReleased = KR_OTHER;
             else {
                 if ((input->keyReleased == KR_SWITCH)
                         && (timestamp - input->lastKeyPressedTime < fc->iTimeInterval)
                         && fc->bDoubleSwitchKey) {
                     FcitxInstanceCommitString(instance, instance->CurrentIC, FcitxInputStateGetRawInputBuffer(input));
-                    FcitxInstanceChangeIMState(instance, instance->CurrentIC);
+                    FcitxInstanceChangeIMStateWithKey(instance, instance->CurrentIC, true);
                 }
             }
 
-            if (FcitxHotkeyIsHotKey(sym, state, switchKey1[fc->iSwitchKey]) || FcitxHotkeyIsHotKey(sym, state, switchKey2[fc->iSwitchKey])) {
+            FcitxInputContext2* currentIC2 = (FcitxInputContext2*) instance->CurrentIC;
+
+            if ((instance->CurrentIC->state == IS_ACTIVE || !fc->bUseExtraTriggerKeyOnlyWhenUseItToInactivate || currentIC2->switchBySwitchKey) &&
+                (FcitxHotkeyIsHotKey(sym, state, switchKey1[fc->iSwitchKey]) || FcitxHotkeyIsHotKey(sym, state, switchKey2[fc->iSwitchKey]))) {
                 input->keyReleased = KR_SWITCH;
                 retVal = IRV_DONOT_PROCESS;
             } else if (fc->bIMSwitchKey && (FcitxHotkeyIsHotKey(sym, state, imSWNextKey1[fc->iIMSwitchKey]) || FcitxHotkeyIsHotKey(sym, state, imSWNextKey2[fc->iIMSwitchKey]))) {
@@ -629,24 +663,26 @@ INPUT_RETURN_VALUE FcitxInstanceProcessKey(
             } else if (fc->bIMSwitchKey && (FcitxHotkeyIsHotKey(sym, state, imSWPrevKey1[fc->iIMSwitchKey]) || FcitxHotkeyIsHotKey(sym, state, imSWPrevKey2[fc->iIMSwitchKey]))) {
                 input->keyReleased = KR_SWITCH_IM_REVERSE;
                 retVal = IRV_DONOT_PROCESS;
-            } else if (FcitxHotkeyIsHotKey(sym, state, fc->hkTrigger)) {
-                /* trigger key has the highest priority, so we check it first */
-                if (FcitxInstanceGetCurrentState(instance) == IS_INACTIVE) {
-                    FcitxInstanceChangeIMState(instance, instance->CurrentIC);
-                    FcitxInstanceShowInputSpeed(instance);
-                } else
-                    FcitxInstanceCloseIM(instance, instance->CurrentIC);
+            } else if (HAVE_IM) {
+                if (FcitxHotkeyIsHotKey(sym, state, fc->hkTrigger)) {
+                    /* trigger key has the highest priority, so we check it first */
+                    if (FcitxInstanceGetCurrentState(instance) == IS_INACTIVE) {
+                        FcitxInstanceChangeIMState(instance, instance->CurrentIC);
+                        FcitxInstanceShowInputSpeed(instance);
+                    } else
+                        FcitxInstanceCloseIM(instance, instance->CurrentIC);
 
-                retVal = IRV_DO_NOTHING;
-            } else if (FcitxHotkeyIsHotKey(sym, state, fc->hkActivate)) {
-                if (FcitxInstanceGetCurrentState(instance) != IS_ACTIVE) {
-                    FcitxInstanceEnableIM(instance, instance->CurrentIC, false);
-                    retVal = IRV_DO_NOTHING;
-                }
-            } else if (FcitxHotkeyIsHotKey(sym, state, fc->hkInactivate)) {
-                if (FcitxInstanceGetCurrentState(instance) == IS_ACTIVE) {
-                    FcitxInstanceCloseIM(instance, instance->CurrentIC);
                     retVal = IRV_DO_NOTHING;
+                } else if (FcitxHotkeyIsHotKey(sym, state, fc->hkActivate)) {
+                    if (FcitxInstanceGetCurrentState(instance) != IS_ACTIVE) {
+                        FcitxInstanceEnableIM(instance, instance->CurrentIC, false);
+                        retVal = IRV_DO_NOTHING;
+                    }
+                } else if (FcitxHotkeyIsHotKey(sym, state, fc->hkInactivate)) {
+                    if (FcitxInstanceGetCurrentState(instance) == IS_ACTIVE) {
+                        FcitxInstanceCloseIM(instance, instance->CurrentIC);
+                        retVal = IRV_DO_NOTHING;
+                    }
                 }
             }
         }
@@ -678,12 +714,11 @@ INPUT_RETURN_VALUE FcitxInstanceProcessKey(
 
         /* check choose key first, because it might trigger update candidates */
         if (!input->bIsDoInputOnly && retVal == IRV_TO_PROCESS) {
-            int index = FcitxHotkeyCheckChooseKeyAndModifier(sym, state,
-                                                  FcitxCandidateWordGetChoose(input->candList),
-                                                  FcitxCandidateWordGetModifier(input->candList)
-                                                 );
+            int index = FcitxCandidateWordCheckChooseKey(input->candList,
+                                                         sym, state);
             if (index >= 0)
-                retVal = FcitxCandidateWordChooseByIndex(input->candList, index);
+                retVal = FcitxCandidateWordChooseByIndex(input->candList,
+                                                         index);
         }
     }
 
@@ -838,10 +873,7 @@ void FcitxInstanceForwardKey(FcitxInstance* instance, FcitxInputContext *ic, Fci
 FCITX_EXPORT_API
 void FcitxInstanceSwitchIM(FcitxInstance* instance, int index)
 {
-    FcitxInputContext* ic = FcitxInstanceGetCurrentIC(instance);
-    if (ic)
-        FcitxInstanceSetLocalIMName(instance, ic, NULL);
-    FcitxInstanceSwitchIMInternal(instance, index, true, true);
+    FcitxInstanceSwitchIMByIndex(instance, index);
 }
 
 FCITX_EXPORT_API
@@ -850,14 +882,9 @@ void FcitxInstanceSwitchIMByName(FcitxInstance* instance, const char* name)
     FcitxIM* im = FcitxInstanceGetIMFromIMList(instance, IMAS_Enable, name);
     if (im) {
         int idx = FcitxInstanceGetIMIndexByName(instance, name);
-        if (idx == 0)
-            FcitxInstanceCloseIM(instance, FcitxInstanceGetCurrentIC(instance));
-        else {
-            FcitxInstanceSwitchIM(instance, idx);
-            if (FcitxInstanceGetCurrentState(instance) != IS_ACTIVE) {
-                FcitxInstanceEnableIM(instance, FcitxInstanceGetCurrentIC(instance), false);
-            }
-        }
+        if (idx < 0)
+            return;
+        FcitxInstanceSwitchIMByIndex(instance, idx);
     }
 }
 
@@ -871,29 +898,50 @@ void FcitxInstanceSwitchIMByIndex(FcitxInstance* instance, int index)
      * -1 scroll forward
      * 0~positive select
      */
-    if (index < -2 || index >= iIMCount)
+    if (index < -4 || index >= iIMCount)
         return;
-    else if (index == -2) {
-        if (instance->iIMIndex > 0)
+
+    boolean skipZero = (index == -3 || index == -4);
+    if (index == -2 || index == -4) {
+        if (instance->iIMIndex > 0) {
             index = instance->iIMIndex -1;
+            if (index == 0 && skipZero)
+                index = iIMCount - 1;
+        }
         else
             index = iIMCount - 1;
-    } else if (index == -1) {
+    } else if (index == -1 || index == -3) {
         if (instance->iIMIndex >= (iIMCount - 1))
-            index = 0;
+            index = skipZero ? 1 : 0;
         else
             index = instance->iIMIndex + 1;
     }
+    FcitxInputContext* ic = FcitxInstanceGetCurrentIC(instance);
     if (index == 0)
-        FcitxInstanceCloseIM(instance, FcitxInstanceGetCurrentIC(instance));
+        FcitxInstanceCloseIM(instance, ic);
     else {
-        FcitxInstanceSwitchIM(instance, index);
+        if (ic)
+            FcitxInstanceSetLocalIMName(instance, ic, NULL);
+        FcitxInstanceSwitchIMInternal(instance, index, true, true);
+        FcitxInstanceShowInputSpeed(instance);
+
         if (FcitxInstanceGetCurrentState(instance) != IS_ACTIVE) {
             FcitxInstanceEnableIM(instance, FcitxInstanceGetCurrentIC(instance), false);
         }
     }
 }
 
+FCITX_EXPORT_API
+void FcitxInstanceUnregisterIM(FcitxInstance* instance, const char* name)
+{
+    FcitxIM* im = FcitxInstanceGetIMFromIMList(instance, IMAS_Disable, name);
+    if (!im)
+        return;
+
+    int index = utarray_eltidx(&instance->availimes, im);
+    utarray_erase(&instance->availimes, index, 1);
+}
+
 void FcitxInstanceSwitchIMInternal(FcitxInstance* instance, int index, boolean skipZero, boolean updateGlobal)
 {
     UT_array* imes = &instance->imes;
@@ -905,36 +953,30 @@ void FcitxInstanceSwitchIMInternal(FcitxInstance* instance, int index, boolean s
 
     FcitxIM* lastIM, *newIM;
 
+    /* set lastIM */
     if (instance->iIMIndex >= iIMCount || instance->iIMIndex < 0)
         lastIM = NULL;
     else {
         lastIM = (FcitxIM*) utarray_eltptr(imes, instance->iIMIndex);
     }
 
+    /* update instance->iIMIndex start */
     if (index >= iIMCount)
         instance->iIMIndex = iIMCount - 1;
-    else if (index < -2)
+    else if (index < 0)
         instance->iIMIndex = 0;
-    else if (index == -2) {
-        if (instance->iIMIndex > 0)
-            instance->iIMIndex--;
-        else
-            instance->iIMIndex = iIMCount - 1;
-    } else if (index == -1) {
-        if (instance->iIMIndex >= (iIMCount - 1))
-            instance->iIMIndex = 0;
-        else
-            instance->iIMIndex++;
-    } else if (index >= 0)
+    else if (index >= 0)
         instance->iIMIndex = index;
 
     if (skipZero && instance->iIMIndex == 0) {
-        if (index == -2)
-            instance->iIMIndex = iIMCount -1;
-        else
-            instance->iIMIndex = 1;
+        instance->iIMIndex = 1;
     }
+    if (iIMCount <= 1) {
+        instance->iIMIndex = 0;
+    }
+    /* update instance->iIMIndex end */
 
+    /* set newIM */
     if (instance->iIMIndex >= iIMCount || instance->iIMIndex < 0)
         newIM = NULL;
     else {
@@ -968,9 +1010,11 @@ void FcitxInstanceSwitchIMInternal(FcitxInstance* instance, int index, boolean s
     }
 
     if (newIM && updateGlobal) {
-        fcitx_utils_free(instance->globalIMName);
-        instance->globalIMName = strdup(newIM->uniqueName);
-        FcitxProfileSave(instance->profile);
+        if (fcitx_utils_strcmp0(instance->globalIMName, newIM->uniqueName) != 0) {
+            fcitx_utils_free(instance->globalIMName);
+            instance->globalIMName = strdup(newIM->uniqueName);
+            FcitxProfileSave(instance->profile);
+        }
     }
 
     FcitxInstanceResetInput(instance);
@@ -1082,8 +1126,8 @@ void FcitxInstanceReloadConfig(FcitxInstance *instance)
 
     FcitxAddon *addon;
     for (addon = (FcitxAddon *) utarray_front(addons);
-            addon != NULL;
-            addon = (FcitxAddon *) utarray_next(addons, addon)) {
+         addon != NULL;
+         addon = (FcitxAddon *) utarray_next(addons, addon)) {
         if (addon->category == AC_MODULE &&
                 addon->bEnabled &&
                 addon->addonInstance) {
@@ -1093,8 +1137,8 @@ void FcitxInstanceReloadConfig(FcitxInstance *instance)
     }
 
     for (addon = (FcitxAddon *) utarray_front(addons);
-            addon != NULL;
-            addon = (FcitxAddon *) utarray_next(addons, addon)) {
+         addon != NULL;
+         addon = (FcitxAddon *) utarray_next(addons, addon)) {
         if (addon->category == AC_FRONTEND &&
                 addon->bEnabled &&
                 addon->addonInstance) {
@@ -1103,6 +1147,18 @@ void FcitxInstanceReloadConfig(FcitxInstance *instance)
         }
     }
 
+    for (addon = (FcitxAddon *) utarray_front(addons);
+         addon != NULL;
+         addon = (FcitxAddon *) utarray_next(addons, addon)) {
+        if (addon->category == AC_INPUTMETHOD &&
+                addon->bEnabled &&
+                addon->addonInstance &&
+                addon->isIMClass2) {
+            if (addon->imclass2->ReloadConfig)
+                addon->imclass2->ReloadConfig(addon->addonInstance);
+        }
+    }
+
 
     UT_array* imes = &instance->imes;
     FcitxIM* pim;
@@ -1115,6 +1171,11 @@ void FcitxInstanceReloadConfig(FcitxInstance *instance)
 
     if (instance->ui && instance->ui->ui->ReloadConfig)
         instance->ui->ui->ReloadConfig(instance->ui->addonInstance);
+
+    FcitxAddon* addonHead = FcitxAddonsLoadInternal(&instance->addons, true);
+    FcitxInstanceFillAddonOwner(instance, addonHead);
+    FcitxInstanceResolveAddonDependencyInternal(instance, addonHead);
+    FcitxInstanceLoadAllIM(instance);
 }
 
 FCITX_EXPORT_API
@@ -1125,6 +1186,14 @@ FcitxIM* FcitxInstanceGetCurrentIM(FcitxInstance* instance)
     return pcurrentIM;
 }
 
+static inline void FcitxInstanceSetICStatus(FcitxInstance* instance, FcitxInputContext* ic, FcitxContextState state)
+{
+    if (ic->state != state) {
+        ic->state = state;
+        FcitxInstanceProcessICStateChangedHook(instance, ic);
+    }
+}
+
 FCITX_EXPORT_API
 void FcitxInstanceSetLocalIMName(FcitxInstance* instance, FcitxInputContext* ic, const char* imname)
 {
@@ -1141,10 +1210,10 @@ void FcitxInstanceSetLocalIMName(FcitxInstance* instance, FcitxInputContext* ic,
         FcitxInstanceUpdateCurrentIM(instance, false);
 }
 
-void FcitxInstanceUpdateCurrentIM(FcitxInstance* instance, boolean force) {
+boolean FcitxInstanceUpdateCurrentIM(FcitxInstance* instance, boolean force) {
     FcitxInputContext* ic = FcitxInstanceGetCurrentIC(instance);
     if (!ic && !force)
-        return;
+        return false;
     FcitxInputContext2* ic2 = (FcitxInputContext2*) ic;
     int globalIndex = FcitxInstanceGetIMIndexByName(instance, instance->globalIMName);
     boolean forceSwtich = force;
@@ -1180,8 +1249,12 @@ void FcitxInstanceUpdateCurrentIM(FcitxInstance* instance, boolean force) {
         skipZero = true;
     }
 
-    if (forceSwtich || targetIMIndex != instance->iIMIndex)
+    if (forceSwtich || targetIMIndex != instance->iIMIndex) {
         FcitxInstanceSwitchIMInternal(instance, targetIMIndex, skipZero, forceSwtich);
+        return true;
+    }
+    else
+        return false;
 }
 
 FCITX_EXPORT_API
@@ -1228,7 +1301,7 @@ void FcitxInstanceEnableIMInternal(FcitxInstance* instance, FcitxInputContext* i
         return;
     FcitxFrontend* frontend = (*pfrontend)->frontend;
     FcitxContextState oldstate = ic->state;
-    ic->state = IS_ACTIVE;
+    FcitxInstanceSetICStatus(instance, ic, IS_ACTIVE);
     if (oldstate == IS_CLOSED)
         frontend->EnableIM((*pfrontend)->addonInstance, ic);
 
@@ -1286,7 +1359,7 @@ void FcitxInstanceCloseIMInternal(FcitxInstance* instance, FcitxInputContext* ic
     if (pfrontend == NULL)
         return;
     FcitxFrontend* frontend = (*pfrontend)->frontend;
-    ic->state = IS_CLOSED;
+    FcitxInstanceSetICStatus(instance, ic, IS_CLOSED);
     frontend->CloseIM((*pfrontend)->addonInstance, ic);
 
     if (ic == instance->CurrentIC) {
@@ -1296,13 +1369,7 @@ void FcitxInstanceCloseIMInternal(FcitxInstance* instance, FcitxInputContext* ic
     }
 }
 
-/**
- * 更改输入法状态
- *
- * @param _connect_id
- */
-FCITX_EXPORT_API
-void FcitxInstanceChangeIMState(FcitxInstance* instance, FcitxInputContext* ic)
+void FcitxInstanceChangeIMStateWithKey(FcitxInstance* instance, FcitxInputContext* ic, boolean withSwitchKey)
 {
     if (ic == NULL)
         return;
@@ -1326,26 +1393,39 @@ void FcitxInstanceChangeIMState(FcitxInstance* instance, FcitxInputContext* ic)
             }
 
             if (flag && (rec == ic || !(rec->contextCaps & CAPACITY_CLIENT_SIDE_CONTROL_STATE)))
-                FcitxInstanceChangeIMStateInternal(instance, rec, objectState);
+                FcitxInstanceChangeIMStateInternal(instance, rec, objectState, withSwitchKey);
             rec = rec->next;
         }
     }
     break;
     case ShareState_None:
-        FcitxInstanceChangeIMStateInternal(instance, ic, objectState);
+        FcitxInstanceChangeIMStateInternal(instance, ic, objectState, withSwitchKey);
         break;
     }
 
     FcitxInstanceUpdateCurrentIM(instance, false);
 }
 
-void FcitxInstanceChangeIMStateInternal(FcitxInstance* instance, FcitxInputContext* ic, FcitxContextState objectState)
+/**
+ * 更改输入法状态
+ *
+ * @param _connect_id
+ */
+FCITX_EXPORT_API
+void FcitxInstanceChangeIMState(FcitxInstance* instance, FcitxInputContext* ic)
+{
+    FcitxInstanceChangeIMStateWithKey(instance, ic, false);
+}
+
+void FcitxInstanceChangeIMStateInternal(FcitxInstance* instance, FcitxInputContext* ic, FcitxContextState objectState, boolean withSwitchKey)
 {
     if (!ic)
         return;
     if (ic->state == objectState)
         return;
-    ic->state = objectState;
+    FcitxInstanceSetICStatus(instance, ic, objectState);
+    FcitxInputContext2* ic2 = (FcitxInputContext2*) ic;
+    ic2->switchBySwitchKey = withSwitchKey;
     if (ic == instance->CurrentIC) {
         if (objectState == IS_ACTIVE) {
             FcitxInstanceResetInput(instance);
@@ -1372,14 +1452,7 @@ boolean IMMenuAction(FcitxUIMenu *menu, int index)
 {
     FcitxInstance* instance = (FcitxInstance*) menu->priv;
 
-    if (index == 0)
-        FcitxInstanceCloseIM(instance, instance->CurrentIC);
-    else {
-        FcitxInstanceSwitchIM(instance, index);
-        if (FcitxInstanceGetCurrentState(instance) != IS_ACTIVE) {
-            FcitxInstanceEnableIM(instance, instance->CurrentIC, false);
-        }
-    }
+    FcitxInstanceSwitchIMByIndex(instance, index);
     return true;
 }
 
@@ -1398,13 +1471,39 @@ void UpdateIMMenuItem(FcitxUIMenu *menu)
     menu->mark = instance->iIMIndex;
 }
 
+void HideInputSpeed(void* arg)
+{
+    FcitxInstance *instance = arg;
+    FcitxInputState* input = instance->input;
+    if (FcitxMessagesIsMessageChanged(input->msgAuxUp)
+        || FcitxMessagesIsMessageChanged(input->msgAuxDown)
+        || FcitxMessagesGetMessageCount(input->msgPreedit)
+        || FcitxMessagesGetMessageCount(input->msgClientPreedit)
+        || FcitxCandidateWordGetListSize(input->candList))
+        return;
+    FcitxUICloseInputWindow(instance);
+}
+
 void FcitxInstanceShowInputSpeed(FcitxInstance* instance)
 {
     FcitxInputState* input = instance->input;
 
+    if (!instance->initialized)
+        return;
+
     if (!instance->config->bShowInputWindowTriggering)
         return;
 
+    if (FcitxInstanceGetCurrentState(instance) != IS_ACTIVE && instance->config->bShowInputWindowOnlyWhenActive)
+        return;
+
+    if (FcitxMessagesGetMessageCount(input->msgAuxUp)
+        || FcitxMessagesGetMessageCount(input->msgAuxDown)
+        || FcitxMessagesGetMessageCount(input->msgPreedit)
+        || FcitxMessagesGetMessageCount(input->msgClientPreedit)
+        || FcitxCandidateWordGetListSize(input->candList))
+        return;
+
     input->bShowCursor = false;
 
     FcitxInstanceCleanInputWindow(instance);
@@ -1414,10 +1513,10 @@ void FcitxInstanceShowInputSpeed(FcitxInstance* instance)
         return;
 
     if (instance->config->bShowVersion) {
-        FcitxMessagesAddMessageAtLast(input->msgAuxUp, MSG_TIPS, "FCITX " VERSION);
-        if (im) {
-            FcitxMessagesAddMessageAtLast(input->msgAuxUp, MSG_TIPS, " %s", im->strName);
-        }
+        FcitxMessagesAddMessageAtLast(input->msgAuxUp, MSG_TIPS, "FCITX " VERSION " ");
+    }
+    if (im) {
+        FcitxMessagesAddMessageAtLast(input->msgAuxUp, MSG_TIPS, "%s", im->strName);
     }
 
     //显示打字速度
@@ -1439,8 +1538,12 @@ void FcitxInstanceShowInputSpeed(FcitxInstance* instance)
     }
 
     FcitxUIUpdateInputWindow(instance);
+
+    if (!FcitxInstanceCheckTimeoutByFunc(instance, HideInputSpeed))
+        FcitxInstanceAddTimeout(instance, 1000, HideInputSpeed, instance);
 }
 
+
 INPUT_RETURN_VALUE ImProcessSaveAll(void *arg)
 {
     FcitxInstance *instance = (FcitxInstance*) arg;
@@ -1483,28 +1586,19 @@ void FcitxInstanceCleanInputWindowDown(FcitxInstance* instance)
 }
 
 FCITX_EXPORT_API
-int FcitxHotkeyCheckChooseKey(FcitxKeySym sym, int state, const char* strChoose)
+int FcitxHotkeyCheckChooseKey(FcitxKeySym sym, unsigned int state, const char* strChoose)
 {
     return FcitxHotkeyCheckChooseKeyAndModifier(sym, state, strChoose, FcitxKeyState_None);
 }
 
 FCITX_EXPORT_API
-int FcitxHotkeyCheckChooseKeyAndModifier(FcitxKeySym sym, int state, const char* strChoose, int candState)
+int FcitxHotkeyCheckChooseKeyAndModifier(FcitxKeySym sym, unsigned int state, const char* strChoose, int candState)
 {
     if (state != candState)
         return -1;
 
-    sym = FcitxHotkeyPadToMain(sym);
-
-    int i = 0;
-
-    while (strChoose[i]) {
-        if (sym == strChoose[i])
-            return i;
-        i++;
-    }
-
-    return -1;
+    char *p = strchr(strChoose, FcitxHotkeyPadToMain(sym));
+    return p ? p - strChoose : -1;
 }
 
 FCITX_EXPORT_API
@@ -1512,7 +1606,7 @@ int FcitxInstanceGetIMIndexByName(FcitxInstance* instance, const char* imName)
 {
     UT_array* imes = &instance->imes;
     FcitxIM *pim;
-    int index = 0, i = 0;
+    int index = -1, i = 0;
     for (pim = (FcitxIM *) utarray_front(imes);
             pim != NULL;
             pim = (FcitxIM *) utarray_next(imes, pim)) {
@@ -1660,7 +1754,12 @@ FCITX_EXPORT_API
 INPUT_RETURN_VALUE FcitxStandardKeyBlocker(FcitxInputState* input, FcitxKeySym key, unsigned int state)
 {
     if (FcitxInputStateGetRawInputBufferSize(input) != 0
-        && (FcitxHotkeyIsHotKeySimple(key, state) || FcitxHotkeyIsHotkeyCursorMove(key, state)))
+        && (FcitxHotkeyIsHotKeySimple(key, state)
+        || FcitxHotkeyIsHotkeyCursorMove(key, state)
+        || FcitxHotkeyIsHotKey(key, state, FCITX_SHIFT_SPACE)
+        || FcitxHotkeyIsHotKey(key, state, FCITX_TAB)
+        || FcitxHotkeyIsHotKey(key, state, FCITX_SHIFT_ENTER)
+        ))
         return IRV_DO_NOTHING;
     else
         return IRV_TO_PROCESS;
