@@ -135,6 +135,22 @@ extern "C" {
         void  (*Destroy)(void *arg); /**< interface for destroy all input method created by this class */
     } FcitxIMClass;
 
+    /**
+     * Fcitx Input Method class, it can register more than one input
+     *        method in create function
+     **/
+    typedef struct _FcitxIMClass2 {
+        void* (*Create)(struct _FcitxInstance* instance); /**< interface for create a input method */
+        void  (*Destroy)(void *arg); /**< interface for destroy all input method created by this class */
+        void  (*ReloadConfig)(void *arg); /**< interface for destroy all input method created by this class */
+        void  (*padding1)(void *arg); /**< padding */
+        void  (*padding2)(void *arg); /**< padding */
+        void  (*padding3)(void *arg); /**< padding */
+        void  (*padding4)(void *arg); /**< padding */
+        void  (*padding5)(void *arg); /**< padding */
+    } FcitxIMClass2;
+
+
     typedef boolean(*FcitxIMInit)(void *arg); /**< FcitxIMInit */
     typedef void (*FcitxIMResetIM)(void *arg); /**< FcitxIMResetIM */
     typedef INPUT_RETURN_VALUE(*FcitxIMDoInput)(void *arg, FcitxKeySym, unsigned int); /**< FcitxIMDoInput */
@@ -507,7 +523,7 @@ extern "C" {
      * @param index input method index
      * @return void
      **/
-    void FcitxInstanceSwitchIM(struct _FcitxInstance* instance, int index);
+    FCITX_DEPRECATED void FcitxInstanceSwitchIM(struct _FcitxInstance* instance, int index);
 
 
     /**
@@ -525,7 +541,8 @@ extern "C" {
     /**
      * switch to a input method by index, index need to be valid, otherwise it have no effect
      * And if the object index is zero, the state will automatically change to inactive
-     * -1 means scroll forward, and -2 means scroll back.
+     * -1 means scroll forward, and -2 means scroll backward.
+     * -3 means scroll forward without first one, -4 mean scroll backward without first one.
      *
      *
      * @param instance fcitx instance
@@ -544,7 +561,7 @@ extern "C" {
      * @param strChoose choose key string
      * @return int
      **/
-    int FcitxHotkeyCheckChooseKey(FcitxKeySym sym, int state, const char* strChoose);
+    int FcitxHotkeyCheckChooseKey(FcitxKeySym sym, unsigned int state, const char* strChoose);
 
     /**
      * check is choose key or not, if so, return the choose index
@@ -555,7 +572,7 @@ extern "C" {
      * @param candState candidate keystate
      * @return int
      **/
-    int FcitxHotkeyCheckChooseKeyAndModifier(FcitxKeySym sym, int state, const char* strChoose, int candState);
+    int FcitxHotkeyCheckChooseKeyAndModifier(FcitxKeySym sym, unsigned int state, const char* strChoose, int candState);
 
     /**
      * get im index by im name
@@ -826,12 +843,24 @@ extern "C" {
     /**
      * set local input method name
      *
-     * @param ic name
+     * @param instance Fcitx Instance
+     * @param ic input context
      * @param imname im name
      * @return void
      **/
     void FcitxInstanceSetLocalIMName(struct _FcitxInstance* instance, struct _FcitxInputContext* ic, const char* imname);
 
+    /**
+     * unregister an im entry
+     *
+     * @param instance Fcitx Instance
+     * @param name imname
+     * @return void
+     *
+     * @since 4.2.6
+     */
+    void FcitxInstanceUnregisterIM(struct _FcitxInstance* instance, const char* name);
+
 #ifdef __cplusplus
 }
 #endif
