@@ -42,6 +42,21 @@ typedef struct _UnusedIMItem {
     UT_hash_handle hh;
 } UnusedIMItem;
 
+typedef struct _TimeoutItem {
+    FcitxTimeoutCallback callback;
+    void* arg;
+    long int milli;
+    uint64_t idx;
+    uint64_t time;
+} TimeoutItem;
+
+typedef struct _FcitxICDataInfo {
+    FcitxICDataAllocCallback allocCallback;
+    FcitxICDataCopyCallback copyCallback;
+    FcitxICDataFreeCallback freeCallback;
+    void* arg;
+} FcitxICDataInfo;
+
 struct _FcitxInstance {
     pthread_mutex_t fcitxMutex;
     UT_array uistats;
@@ -86,6 +101,7 @@ struct _FcitxInstance {
     struct _HookStack* hookUpdateCandidateWordHook;
     struct _HookStack* hookUpdateIMListHook;
     struct _HookStack* hookCommitFilter;
+    struct _HookStack* hookICStateChangedHook;
 
     FcitxUIFlag uiflag;
 
@@ -116,6 +132,16 @@ struct _FcitxInstance {
     FcitxAddon* currentIMAddon;
 
     UnusedIMItem* unusedItem;
+    UT_array* disableList;
+    UT_array* enableList;
+
+    UT_array timeout;
+    boolean initialized;
+    uint64_t timeoutIdx;
+
+    UT_array icdata;
+    boolean loadingFatalError;
+    int fd;
 };
 
 #endif
