@@ -27,6 +27,9 @@
 #include <pthread.h>
 #include <semaphore.h>
 #include <getopt.h>
+#include <sys/time.h>
+#include <signal.h>
+#include <fcntl.h>
 
 #include "instance.h"
 #include "fcitx-utils/log.h"
@@ -41,6 +44,7 @@
 #include "fcitx-internal.h"
 #include "instance-internal.h"
 #include "module-internal.h"
+#include "addon-internal.h"
 
 #define CHECK_ENV(env, value, icase) (!getenv(env) \
                                       || (icase ? \
@@ -56,6 +60,7 @@ FCITX_GETTER_REF(FcitxInstance, AvailIMEs, availimes, UT_array)
 FCITX_GETTER_REF(FcitxInstance, ReadFDSet, rfds, fd_set)
 FCITX_GETTER_REF(FcitxInstance, WriteFDSet, wfds, fd_set)
 FCITX_GETTER_REF(FcitxInstance, ExceptFDSet, efds, fd_set)
+FCITX_GETTER_VALUE(FcitxInstance, CurrentUI, ui, FcitxAddon*)
 FCITX_GETTER_VALUE(FcitxInstance, MaxFD, maxfd, int)
 FCITX_SETTER(FcitxInstance, MaxFD, maxfd, int)
 FCITX_GETTER_VALUE(FcitxInstance, GlobalConfig, config, FcitxGlobalConfig*)
@@ -65,6 +70,8 @@ FCITX_GETTER_VALUE(FcitxInstance, InputState, input, FcitxInputState*)
 const UT_icd stat_icd = {sizeof(FcitxUIStatus), 0, 0, 0};
 const UT_icd compstat_icd = {sizeof(FcitxUIComplexStatus), 0, 0, 0};
 const UT_icd menup_icd = {sizeof(FcitxUIMenu*), 0, 0, 0};
+const UT_icd timeout_icd = {sizeof(TimeoutItem), 0, 0, 0};
+const UT_icd icdata_icd = {sizeof(FcitxICDataInfo), NULL, NULL, NULL };
 static void FcitxInitThread(FcitxInstance* inst);
 static void ToggleRemindState(void* arg);
 static boolean GetRemindEnabled(void* arg);
@@ -87,6 +94,8 @@ void Usage()
            "\t-s[sleep time]\t\toverride delay start time in config file, 0 for immediate start\n"
            "\t-v\t\t\tdisplay the version information and exit\n"
            "\t-u, --ui\t\tspecify the user interface to use\n"
+           "\t--enable\t\tspecify a comma separated list for addon that will override the enable option\n"
+           "\t--disable\t\tspecify a comma separated list for addon that will explicitly disabled, priority is higher than --enable\n"
            "\t-h, --help\t\tdisplay this help and exit\n");
 }
 
@@ -100,6 +109,12 @@ void Version()
 
 FCITX_EXPORT_API
 FcitxInstance* FcitxInstanceCreate(sem_t *sem, int argc, char* argv[])
+{
+    return FcitxInstanceCreateWithFD(sem, argc, argv, 0);
+}
+
+FCITX_EXPORT_API
+FcitxInstance* FcitxInstanceCreateWithFD(sem_t *sem, int argc, char* argv[], int fd)
 {
     FcitxInstance* instance = fcitx_utils_malloc0(sizeof(FcitxInstance));
     FcitxAddonsInit(&instance->addons);
@@ -110,11 +125,17 @@ FcitxInstance* FcitxInstanceCreate(sem_t *sem, int argc, char* argv[])
     utarray_init(&instance->uistats, &stat_icd);
     utarray_init(&instance->uicompstats, &compstat_icd);
     utarray_init(&instance->uimenus, &menup_icd);
+    utarray_init(&instance->timeout, &timeout_icd);
+    utarray_init(&instance->icdata, &icdata_icd);
     instance->input = FcitxInputStateCreate();
     instance->sem = sem;
     instance->config = fcitx_utils_malloc0(sizeof(FcitxGlobalConfig));
     instance->profile = fcitx_utils_malloc0(sizeof(FcitxProfile));
     instance->globalIMName = strdup("");
+    if (fd > 0) {
+        fcntl(fd, F_SETFL, O_NONBLOCK);
+        instance->fd = fd;
+    }
 
     if (!FcitxGlobalConfigLoad(instance->config))
         goto error_exit;
@@ -137,18 +158,13 @@ FcitxInstance* FcitxInstanceCreate(sem_t *sem, int argc, char* argv[])
         goto error_exit;
 
     FcitxAddonsLoad(&instance->addons);
-
-    /* FIXME: a walkaround for not have instance in function FcitxModuleInvokeFunction */
-    FcitxAddon* addon;
-    for (addon = (FcitxAddon *) utarray_front(&instance->addons);
-            addon != NULL;
-            addon = (FcitxAddon *) utarray_next(&instance->addons, addon)) {
-        addon->owner = instance;
-    }
+    FcitxInstanceFillAddonOwner(instance, NULL);
     FcitxInstanceResolveAddonDependency(instance);
     FcitxInstanceInitBuiltInHotkey(instance);
     FcitxInstanceInitBuiltContext(instance);
     FcitxModuleLoad(instance);
+    if (instance->loadingFatalError)
+        return instance;
     if (!FcitxInstanceLoadAllIM(instance)) {
         FcitxInstanceEnd(instance);
         return instance;
@@ -165,7 +181,7 @@ FcitxInstance* FcitxInstanceCreate(sem_t *sem, int argc, char* argv[])
 
     instance->iIMIndex = FcitxInstanceGetIMIndexByName(instance, instance->profile->imName);
 
-    FcitxInstanceSwitchIM(instance, instance->iIMIndex);
+    FcitxInstanceSwitchIMByIndex(instance, instance->iIMIndex);
 
     if (!FcitxInstanceLoadFrontend(instance)) {
         FcitxInstanceEnd(instance);
@@ -192,8 +208,19 @@ void FcitxInstanceSetRecheckEvent(FcitxInstance* instance)
 void* RunInstance(void* arg)
 {
     FcitxInstance* instance = (FcitxInstance*) arg;
+    instance->initialized = true;
+    int64_t curtime = 0;
     while (1) {
         FcitxAddon** pmodule;
+        uint8_t signo = 0;
+        while (read(instance->fd, &signo, sizeof(char)) > 0) {
+            if (signo == SIGINT || signo == SIGTERM || signo == SIGQUIT)
+                FcitxInstanceEnd(instance);
+            else if (signo == SIGHUP)
+                fcitx_utils_launch_restart();
+            else if (signo == SIGUSR1)
+                FcitxInstanceReloadConfig(instance);
+        }
         do {
             instance->uiflag = UI_NONE;
             for (pmodule = (FcitxAddon**) utarray_front(&instance->eventmodules);
@@ -202,6 +229,30 @@ void* RunInstance(void* arg)
                 FcitxModule* module = (*pmodule)->module;
                 module->ProcessEvent((*pmodule)->addonInstance);
             }
+            struct timeval current_time;
+            gettimeofday(&current_time, NULL);
+            curtime = (current_time.tv_sec * 1000LL) + (current_time.tv_usec / 1000LL);
+
+            int idx = 0;
+            while(idx < utarray_len(&instance->timeout))
+            {
+                TimeoutItem* ti = (TimeoutItem*) utarray_eltptr(&instance->timeout, idx);
+                uint64_t id = ti->idx;
+                if (ti->time + ti->milli <= curtime) {
+                    ti->callback(ti->arg);
+                    ti = (TimeoutItem*) utarray_eltptr(&instance->timeout, idx);
+                    /* faster remove */
+                    if (ti && ti->idx == id)
+                        utarray_remove_quick(&instance->timeout, idx);
+                    else {
+                        FcitxInstanceRemoveTimeoutById(instance, id);
+                        idx = 0;
+                    }
+                }
+                else {
+                    idx++;
+                }
+            }
 
             if (instance->uiflag & UI_MOVE)
                 FcitxUIMoveInputWindowReal(instance);
@@ -215,6 +266,10 @@ void* RunInstance(void* arg)
         FD_ZERO(&instance->efds);
 
         instance->maxfd = 0;
+        if (instance->fd > 0) {
+            instance->maxfd = instance->fd;
+            FD_SET(instance->fd, &instance->rfds);
+        }
         for (pmodule = (FcitxAddon**) utarray_front(&instance->eventmodules);
                 pmodule != NULL;
                 pmodule = (FcitxAddon**) utarray_next(&instance->eventmodules, pmodule)) {
@@ -223,7 +278,24 @@ void* RunInstance(void* arg)
         }
         if (instance->maxfd == 0)
             break;
-        select(instance->maxfd + 1, &instance->rfds, &instance->wfds, &instance->efds, NULL);
+        struct timeval tval;
+        struct timeval* ptval = NULL;
+        if (utarray_len(&instance->timeout) != 0) {
+            long int min_time = LONG_MAX;
+            TimeoutItem* ti;
+            for (ti = (TimeoutItem*) utarray_front(&instance->timeout);
+                 ti != NULL;
+                 ti = (TimeoutItem*) utarray_next(&instance->timeout, ti))
+            {
+                if (ti->time + ti->milli - curtime < min_time) {
+                    min_time = ti->time + ti->milli - curtime;
+                }
+            }
+            tval.tv_usec = (min_time % 1000) * 1000;
+            tval.tv_sec = min_time / 1000;
+            ptval = &tval;
+        }
+        select(instance->maxfd + 1, &instance->rfds, &instance->wfds, &instance->efds, ptval);
     }
     return NULL;
 }
@@ -231,6 +303,15 @@ void* RunInstance(void* arg)
 FCITX_EXPORT_API
 void FcitxInstanceEnd(FcitxInstance* instance)
 {
+    if (!instance->initialized) {
+        if (!instance->loadingFatalError) {
+            FcitxLog(ERROR, "Exiting.");
+            instance->loadingFatalError = true;
+            sem_post(instance->sem);
+        }
+        return;
+    }
+
     FcitxProfileSave(instance->profile);
     FcitxInstanceSaveAllIM(instance);
 
@@ -293,6 +374,13 @@ void FcitxInstanceEnd(FcitxInstance* instance)
     }
 
     sem_post(instance->sem);
+
+    /* don't return to main loop, wait for exit */
+    int countDown = 5;
+    while(countDown--) {
+        sleep(1000);
+    }
+    exit(0);
 }
 
 void FcitxInitThread(FcitxInstance* inst)
@@ -340,7 +428,10 @@ boolean ProcessOption(FcitxInstance* instance, int argc, char* argv[])
     struct option longOptions[] = {
         {"ui", 1, 0, 0},
         {"replace", 0, 0, 0},
-        {"help", 0, 0, 0}
+        {"enable", 1, 0, 0},
+        {"disable", 1, 0, 0},
+        {"help", 0, 0, 0},
+        {NULL, 0, 0, 0}
     };
 
     int optionIndex = 0;
@@ -358,6 +449,20 @@ boolean ProcessOption(FcitxInstance* instance, int argc, char* argv[])
             case 1:
                 instance->tryReplace = true;
                 break;
+            case 2:
+                {
+                    if (instance->enableList)
+                        fcitx_utils_free_string_list(instance->enableList);
+                    instance->enableList = fcitx_utils_split_string(optarg, ',');
+                }
+                break;
+            case 3:
+                {
+                    if (instance->disableList)
+                        fcitx_utils_free_string_list(instance->disableList);
+                    instance->disableList = fcitx_utils_split_string(optarg, ',');
+                }
+                break;
             default:
                 Usage();
                 return false;
@@ -466,6 +571,74 @@ void FcitxInstanceResetTryReplace(FcitxInstance* instance)
     instance->tryReplace = false;
 }
 
+FCITX_EXPORT_API
+uint64_t FcitxInstanceAddTimeout(FcitxInstance* instance, long int milli, FcitxTimeoutCallback callback , void* arg)
+{
+    if (milli < 0)
+        return 0;
+
+    struct timeval current_time;
+    gettimeofday(&current_time, NULL);
+    TimeoutItem item;
+    item.arg = arg;
+    item.callback =callback;
+    item.milli = milli;
+    item.idx = ++instance->timeoutIdx;
+    item.time = (current_time.tv_sec * 1000LL) + (current_time.tv_usec / 1000LL);
+    utarray_push_back(&instance->timeout, &item);
+
+    return item.idx;
+}
+
+FCITX_EXPORT_API
+boolean FcitxInstanceCheckTimeoutByFunc(FcitxInstance* instance, FcitxTimeoutCallback callback)
+{
+    TimeoutItem* ti;
+    for (ti = (TimeoutItem*) utarray_front(&instance->timeout);
+         ti != NULL;
+         ti = (TimeoutItem*) utarray_next(&instance->timeout, ti))
+    {
+        if (ti->callback == callback)
+            return true;
+    }
+    return false;
+}
+
+FCITX_EXPORT_API
+boolean FcitxInstanceRemoveTimeoutByFunc(FcitxInstance* instance, FcitxTimeoutCallback callback)
+{
+    TimeoutItem* ti;
+    for (ti = (TimeoutItem*) utarray_front(&instance->timeout);
+         ti != NULL;
+         ti = (TimeoutItem*) utarray_next(&instance->timeout, ti))
+    {
+        if (ti->callback == callback) {
+            int idx = utarray_eltidx(&instance->timeout, ti);
+            utarray_remove_quick(&instance->timeout, idx);
+            return true;
+        }
+    }
+    return false;
+}
+
+FCITX_EXPORT_API
+boolean FcitxInstanceRemoveTimeoutById(FcitxInstance* instance, uint64_t id)
+{
+    if (id == 0)
+        return false;
+    TimeoutItem* ti;
+    for (ti = (TimeoutItem*) utarray_front(&instance->timeout);
+         ti != NULL;
+         ti = (TimeoutItem*) utarray_next(&instance->timeout, ti))
+    {
+        if (ti->idx == id) {
+            int idx = utarray_eltidx(&instance->timeout, ti);
+            utarray_remove_quick(&instance->timeout, idx);
+            return true;
+        }
+    }
+    return false;
+}
 
 
 // kate: indent-mode cstyle; space-indent on; indent-width 0;
