@@ -47,6 +47,8 @@ extern "C" {
      **/
     typedef struct _FcitxInstance FcitxInstance;
 
+    typedef void (*FcitxTimeoutCallback)(void* arg);
+
     /**
      * create new fcitx instance
      *
@@ -57,6 +59,22 @@ extern "C" {
      **/
     FcitxInstance* FcitxInstanceCreate(sem_t *sem, int argc, char* argv[]);
 
+
+    /**
+     * create new fcitx instance
+     *
+     * @param sem semaphore to notify the instance is end
+     * @param argc argc
+     * @param argv argv
+     * @param signal fd
+     * @return FcitxInstance*
+     *
+     * @see FcitxInstanceCreate
+     *
+     * @since 4.2.5
+     **/
+    FcitxInstance* FcitxInstanceCreateWithFD(sem_t *sem, int argc, char* argv[], int fd);
+
     /**
      * replace existing fcitx instance
      *
@@ -147,6 +165,8 @@ extern "C" {
 
     fd_set* FcitxInstanceGetExceptFDSet(FcitxInstance* instance);
 
+    struct _FcitxAddon* FcitxInstanceGetCurrentUI(FcitxInstance* instance);
+
     int FcitxInstanceGetMaxFD(FcitxInstance* instance);
 
     void FcitxInstanceSetMaxFD(FcitxInstance* instance, int maxfd);
@@ -157,6 +177,13 @@ extern "C" {
 
     FcitxInputState* FcitxInstanceGetInputState(FcitxInstance* instance);
 
+    uint64_t FcitxInstanceAddTimeout(FcitxInstance* instance, long int milli, FcitxTimeoutCallback callback , void* arg);
+
+    boolean FcitxInstanceCheckTimeoutByFunc(FcitxInstance* instance, FcitxTimeoutCallback callback);
+
+    boolean FcitxInstanceRemoveTimeoutByFunc(FcitxInstance* instance, FcitxTimeoutCallback callback);
+
+    boolean FcitxInstanceRemoveTimeoutById(FcitxInstance* instance, uint64_t id);
 #ifdef __cplusplus
 }
 #endif
