@@ -164,6 +164,12 @@ FcitxHotkey FCITX_SPACE[2] = {
     {NULL, 0, 0},
 };
 
+FCITX_EXPORT_API
+FcitxHotkey FCITX_TAB[2] = {
+    {NULL, FcitxKey_Tab, FcitxKeyState_None},
+    {NULL, 0, 0},
+};
+
 FCITX_EXPORT_API
 FcitxHotkey FCITX_CTRL_5[2] = {
     {NULL, FcitxKey_5, FcitxKeyState_Ctrl},
@@ -182,6 +188,18 @@ FcitxHotkey FCITX_LCTRL[2] = {
     {NULL, FcitxKey_Control_L, FcitxKeyState_Ctrl},
 };
 
+FCITX_EXPORT_API
+FcitxHotkey FCITX_LALT[2] = {
+    {NULL, FcitxKey_Alt_L, FcitxKeyState_None},
+    {NULL, FcitxKey_Alt_L, FcitxKeyState_Alt},
+};
+
+FCITX_EXPORT_API
+FcitxHotkey FCITX_RALT[2] = {
+    {NULL, FcitxKey_Alt_L, FcitxKeyState_None},
+    {NULL, FcitxKey_Alt_R, FcitxKeyState_Alt},
+};
+
 FCITX_EXPORT_API
 FcitxHotkey FCITX_LSHIFT[2] = {
     {NULL, FcitxKey_Shift_L, FcitxKeyState_None},
@@ -261,5 +279,17 @@ FcitxHotkey FCITX_RALT_RSUPER2[2] = {
     {NULL, 0, 0},
 };
 
+FCITX_EXPORT_API
+FcitxHotkey FCITX_SHIFT_SPACE[2] = {
+    {NULL, FcitxKey_space, FcitxKeyState_Shift},
+    {NULL, 0, 0},
+};
+
+FCITX_EXPORT_API
+FcitxHotkey FCITX_SHIFT_ENTER[2] = {
+    {NULL, FcitxKey_Return, FcitxKeyState_Shift},
+    {NULL, FcitxKey_KP_Enter, FcitxKeyState_Shift},
+};
+
 
 // kate: indent-mode cstyle; space-indent on; indent-width 0;
