@@ -59,19 +59,19 @@ void FcitxModuleLoad(FcitxInstance* instance)
                 if (!fp)
                     break;
                 fclose(fp);
-                handle = dlopen(modulePath, RTLD_LAZY | RTLD_GLOBAL);
+                handle = dlopen(modulePath, RTLD_NOW | RTLD_GLOBAL);
                 if (!handle) {
                     FcitxLog(ERROR, _("Module: open %s fail %s") , modulePath , dlerror());
                     break;
                 }
 
-                if (!CheckABIVersion(handle)) {
+                if (!FcitxCheckABIVersion(handle, addon->name)) {
                     FcitxLog(ERROR, "%s ABI Version Error", addon->name);
                     dlclose(handle);
                     break;
                 }
 
-                module = dlsym(handle, "module");
+                module = FcitxGetSymbol(handle, addon->name, "module");
                 if (!module || !module->Create) {
                     FcitxLog(ERROR, _("Module: bad module"));
                     dlclose(handle);
@@ -81,6 +81,8 @@ void FcitxModuleLoad(FcitxInstance* instance)
                     dlclose(handle);
                     break;
                 }
+                if (instance->loadingFatalError)
+                    return;
                 addon->module = module;
                 addon->addonInstance = moduleinstance;
                 if (module->ProcessEvent && module->SetFD)
@@ -119,8 +121,10 @@ void* FcitxModuleInvokeFunction(FcitxAddon* addon, int functionId, FcitxModuleFu
                 break;
             }
         }
-        if (!flag)
+        if (!flag && !addon->addonInstance) {
             FcitxInstanceLoadIM(addon->owner, addon);
+            FcitxInstanceUpdateIMList(addon->owner);
+        }
     }
 
     FcitxModuleFunction* func = (FcitxModuleFunction*) utarray_eltptr(&addon->functionList, functionId);
