@@ -89,6 +89,9 @@ extern "C" {
      * @return void*
      **/
     void* FcitxModuleInvokeFunction(struct _FcitxAddon* addon, int functionId, FcitxModuleFunctionArg args);
+#define FcitxModuleInvokeVaArgs(addon, functionId, ARGV...)             \
+    (FcitxModuleInvokeFunction(addon, functionId,                       \
+                               (FcitxModuleFunctionArg){ .args = {ARGV} }))
 
     /**
      * invoke inter module function with addon name, returns NULL when fails (the function itself can also return NULL)
@@ -100,11 +103,24 @@ extern "C" {
      * @return void*
      **/
     void* FcitxModuleInvokeFunctionByName(struct _FcitxInstance* instance, const char* name, int functionId, FcitxModuleFunctionArg args);
+#define FcitxModuleInvokeVaArgsByName(instance, name, functionId, ARGV...) \
+    (FcitxModuleInvokeFunctionByName(instance, name, functionId,        \
+                               (FcitxModuleFunctionArg){ .args = {ARGV} }))
 
 /** call a function provides by other addon */
 #define InvokeFunction(INST, MODULE, FUNC, ARG)  \
     ((MODULE##_##FUNC##_RETURNTYPE) FcitxModuleInvokeFunctionByName(INST, MODULE##_NAME, MODULE##_##FUNC, ARG))
 
+#define InvokeVaArgs(INST, MODULE, FUNC, ARGV...)                       \
+    ((MODULE##_##FUNC##_RETURNTYPE)FcitxModuleInvokeFunctionByName(     \
+        INST, MODULE##_NAME, MODULE##_##FUNC,                           \
+        (FcitxModuleFunctionArg){ .args = {ARGV} }))
+/**
+ * NOTE: (int)(sizeof((void*[]){ARGV}) / sizeof(void*)) can be used to get
+ * the number of arguements. (can be useful for removing the argument number
+ * limit) as well as directly call the function without a FunctionArg wrapper.
+ **/
+
 /** add a function to current addon */
 #define AddFunction(ADDON, Realname) \
     do { \
