@@ -183,20 +183,20 @@ boolean FcitxUILoadInternal(FcitxInstance* instance, FcitxAddon* addon)
 
         fclose(fp);
 
-        handle = dlopen(modulePath, RTLD_LAZY | RTLD_GLOBAL);
+        handle = dlopen(modulePath, RTLD_NOW | RTLD_GLOBAL);
 
         if (!handle) {
             FcitxLog(ERROR, _("UI: open %s fail %s") , modulePath , dlerror());
             break;
         }
 
-        if (!CheckABIVersion(handle)) {
+        if (!FcitxCheckABIVersion(handle, addon->name)) {
             FcitxLog(ERROR, "%s ABI Version Error", addon->name);
             dlclose(handle);
             break;
         }
 
-        addon->ui = dlsym(handle, "ui");
+        addon->ui = FcitxGetSymbol(handle, addon->name, "ui");
 
         if (!addon->ui || !addon->ui->Create) {
             FcitxLog(ERROR, _("UI: bad ui"));
@@ -226,8 +226,8 @@ boolean FcitxUILoadInternal(FcitxInstance* instance, FcitxAddon* addon)
             FcitxUIComplexStatus *status;
 
             for (status = (FcitxUIComplexStatus *) utarray_front(uicompstats);
-                    status != NULL;
-                    status = (FcitxUIComplexStatus *) utarray_next(uicompstats, status))
+                 status != NULL;
+                 status = (FcitxUIComplexStatus *) utarray_next(uicompstats, status))
                 addon->ui->RegisterComplexStatus(addon->addonInstance, status);
         }
 
@@ -321,9 +321,6 @@ FCITX_EXPORT_API
 void FcitxUIUpdateInputWindow(FcitxInstance *instance)
 {
     instance->uiflag |= UI_UPDATE;
-
-    if (FcitxMessagesIsMessageChanged(instance->input->msgClientPreedit))
-        FcitxInstanceUpdatePreedit(instance, FcitxInstanceGetCurrentIC(instance));
 }
 
 void FcitxUIShowInputWindow(FcitxInstance* instance)
@@ -586,9 +583,12 @@ void FcitxUIOnInputFocus(FcitxInstance* instance)
 
     FcitxInstanceResetInput(instance);
 
-    FcitxInstanceUpdateCurrentIM(instance, false);
+    boolean changed = FcitxInstanceUpdateCurrentIM(instance, false);
 
-    FcitxUICloseInputWindow(instance);
+    if (instance->config->bShowInputWindowWhenFocusIn && changed)
+        FcitxInstanceShowInputSpeed(instance);
+    else
+        FcitxUICloseInputWindow(instance);
 }
 
 FCITX_EXPORT_API
@@ -601,7 +601,7 @@ void FcitxUIOnInputUnFocus(struct _FcitxInstance* instance)
 }
 
 FCITX_EXPORT_API
-void FcitxUICommitPreedit(struct _FcitxInstance* instance)
+void FcitxUICommitPreedit(FcitxInstance* instance)
 {
     if (instance->CurrentIC
         && !instance->config->bDontCommitPreeditWhenUnfocus
@@ -719,7 +719,17 @@ int FcitxUINewMessageToOldStyleMessage(FcitxInstance* instance, FcitxMessages* m
         if (instance->config->bPointAfterNumber)
             strTemp[1] = '.';
 
-        FcitxMessagesAddMessageAtLast(msgDown, MSG_INDEX, "%s", strTemp);
+        if (candWord->strWord == NULL)
+            continue;
+
+        unsigned int mod = FcitxCandidateWordGetModifier(input->candList);
+
+        FcitxMessagesAddMessageAtLast(msgDown, MSG_INDEX, "%s%s%s%s%s",
+                                      (mod & FcitxKeyState_Super) ? "M-" : "",
+                                      (mod & FcitxKeyState_Ctrl) ? "C-" : "",
+                                      (mod & FcitxKeyState_Alt) ? "A-" : "",
+                                      (mod & FcitxKeyState_Shift) ? "S-" : "",
+                                      strTemp);
 
         FcitxMessageType type = candWord->wordType;
 
@@ -746,14 +756,21 @@ char* FcitxUIMessagesToCString(FcitxMessages* messages)
 {
     int length = 0;
     int i = 0;
+    int count = FcitxMessagesGetMessageCount(messages);
+    char *message_strs[count];
+    int msg_count = 0;
 
-    for (i = 0; i < FcitxMessagesGetMessageCount(messages) ; i ++)
-        length += strlen(FcitxMessagesGetMessageString(messages, i));
+    for (i = 0;i < count;i++) {
+        char *msg_str = FcitxMessagesGetMessageString(messages, i);
+        message_strs[msg_count++] = msg_str;
+        length += strlen(msg_str);
+    }
 
-    char* str = fcitx_utils_malloc0(sizeof(char) * (length + 1));
+    char *str = fcitx_utils_malloc0(sizeof(char) * (length + 1));
 
-    for (i = 0; i < FcitxMessagesGetMessageCount(messages) ; i ++)
-        strcat(str, FcitxMessagesGetMessageString(messages, i));
+    for (i = 0;i < msg_count;i++) {
+        strcat(str, message_strs[i]);
+    }
 
     return str;
 }
@@ -762,17 +779,23 @@ char* FcitxUIMessagesToCStringForCommit(FcitxMessages* messages)
 {
     int length = 0;
     int i = 0;
+    int count = FcitxMessagesGetMessageCount(messages);
+    char *message_strs[count];
+    int msg_count = 0;
 
-    for (i = 0; i < FcitxMessagesGetMessageCount(messages) ; i ++) {
-        if ((FcitxMessagesGetClientMessageType(messages, i) & MSG_DONOT_COMMIT_WHEN_UNFOCUS) == 0)
-            length += strlen(FcitxMessagesGetMessageString(messages, i));
+    for (i = 0;i < count;i++) {
+        if ((FcitxMessagesGetClientMessageType(messages, i) &
+             MSG_DONOT_COMMIT_WHEN_UNFOCUS) == 0) {
+            char *msg_str = FcitxMessagesGetMessageString(messages, i);
+            message_strs[msg_count++] = msg_str;
+            length += strlen(msg_str);
+        }
     }
 
     char* str = fcitx_utils_malloc0(sizeof(char) * (length + 1));
 
-    for (i = 0; i < FcitxMessagesGetMessageCount(messages) ; i ++) {
-        if ((FcitxMessagesGetClientMessageType(messages, i) & MSG_DONOT_COMMIT_WHEN_UNFOCUS) == 0)
-            strcat(str, FcitxMessagesGetMessageString(messages, i));
+    for (i = 0;i < msg_count;i++) {
+        strcat(str, message_strs[i]);
     }
 
     return str;
@@ -840,10 +863,12 @@ void FcitxUIUpdateInputWindowReal(FcitxInstance *instance)
         return;
     }
 
+    FcitxInstanceUpdatePreedit(instance, ic);
+
     boolean toshow = false;
 
     if (FcitxMessagesGetMessageCount(input->msgAuxUp) != 0
-            || FcitxMessagesGetMessageCount(input->msgAuxDown) != 0)
+        || FcitxMessagesGetMessageCount(input->msgAuxDown) != 0)
         toshow = true;
 
     if (FcitxCandidateWordGetListSize(input->candList) > 1)
@@ -859,11 +884,15 @@ void FcitxUIUpdateInputWindowReal(FcitxInstance *instance)
         toshow = true;
 
     if (!toshow) {
-        FcitxInstanceUpdatePreedit(instance, ic);
         if (UI_FUNC_IS_VALID(CloseInputWindow))
             instance->ui->ui->CloseInputWindow(instance->ui->addonInstance);
     } else
         FcitxUIShowInputWindow(instance);
+
+    FcitxMessagesSetMessageChanged(input->msgAuxUp, false);
+    FcitxMessagesSetMessageChanged(input->msgAuxDown, false);
+    FcitxMessagesSetMessageChanged(input->msgPreedit, false);
+    FcitxMessagesSetMessageChanged(input->msgClientPreedit, false);
 }
 
 void FcitxUIMoveInputWindowReal(FcitxInstance *instance)
