@@ -90,8 +90,7 @@ CONFIG_BINDING_REGISTER("TraditionalChinese", "Enabled", enabled)
 CONFIG_BINDING_REGISTER("TraditionalChinese", "Hotkey", hkToggle)
 CONFIG_BINDING_END()
 
-FCITX_EXPORT_API
-FcitxModule module = {
+FCITX_DEFINE_PLUGIN(fcitx_chttrans, module, FcitxModule) = {
     ChttransCreate,
     NULL,
     NULL,
@@ -99,9 +98,6 @@ FcitxModule module = {
     ReloadChttrans
 };
 
-FCITX_EXPORT_API
-int ABI_VERSION = FCITX_ABI_VERSION;
-
 void* ChttransCreate(FcitxInstance* instance)
 {
     FcitxChttrans* transState = fcitx_utils_malloc0(sizeof(FcitxChttrans));
@@ -262,7 +258,7 @@ char *ConvertGBKSimple2Tradition(FcitxChttrans* transState, const char *strHZ)
             while (getline(&strBuf, &bufLen, fp) != -1) {
                 simple2trad_t *s2t;
                 char *ps;
-                int wc;
+                uint32_t wc;
 
                 ps = fcitx_utf8_get_char(strBuf, &wc);
                 s2t = (simple2trad_t*) malloc(sizeof(simple2trad_t));
@@ -284,7 +280,7 @@ char *ConvertGBKSimple2Tradition(FcitxChttrans* transState, const char *strHZ)
         ps = strHZ;
         ret[0] = '\0';
         for (; i < len; ++i) {
-            int wc;
+            uint32_t wc;
             simple2trad_t *s2t = NULL;
             int chr_len = fcitx_utf8_char_len(ps);
             char *nps;
@@ -364,7 +360,7 @@ char *ConvertGBKTradition2Simple(FcitxChttrans* transState, const char *strHZ)
             while (getline(&strBuf, &bufLen, fp) != -1) {
                 simple2trad_t *t2s;
                 char *ps;
-                int wc;
+                uint32_t wc;
 
                 ps = fcitx_utf8_get_char(strBuf, &wc);
                 t2s = (simple2trad_t*) malloc(sizeof(simple2trad_t));
@@ -388,7 +384,7 @@ char *ConvertGBKTradition2Simple(FcitxChttrans* transState, const char *strHZ)
         ps = strHZ;
         ret[0] = '\0';
         for (; i < len; ++i) {
-            int wc;
+            uint32_t wc;
             simple2trad_t *t2s = NULL;
             int chr_len = fcitx_utf8_char_len(ps);
             char *nps;
@@ -439,7 +435,8 @@ boolean LoadChttransConfig(FcitxChttrans* transState)
 
     if (newconfig) {
         char *p = fcitx_utils_get_current_langcode();
-        if (strcmp(p, "zh_TW") == 0 || strcmp(p, "zh_HK") == 0) {
+        /* also check en_HK for default setting */
+        if (strcmp(p, "zh_TW") == 0 || strcmp(p, "en_HK") == 0 || strcmp(p, "zh_HK") == 0) {
             transState->enabled = true;
             SaveChttransConfig(transState);
         }
