@@ -79,8 +79,7 @@ typedef struct _FcitxFullWidthChar {
     FcitxInstance* owner;
 } FcitxFullWidthChar;
 
-FCITX_EXPORT_API
-FcitxModule module = {
+FCITX_DEFINE_PLUGIN(fcitx_fullwidth_char, module, FcitxModule) = {
     FullWidthCharCreate,
     NULL,
     NULL,
@@ -88,9 +87,6 @@ FcitxModule module = {
     NULL
 };
 
-FCITX_EXPORT_API
-int ABI_VERSION = FCITX_ABI_VERSION;
-
 void* FullWidthCharCreate(FcitxInstance* instance)
 {
     FcitxFullWidthChar* fwchar = fcitx_utils_malloc0(sizeof(FcitxFullWidthChar));
@@ -163,7 +159,7 @@ char* ProcessFullWidthChar(void* arg, const char* str)
         const char* ps = str;
         ret[0] = '\0';
         for (; i < len; ++i) {
-            int wc;
+            uint32_t wc;
             int chr_len = fcitx_utf8_char_len(ps);
             char *nps;
             nps = fcitx_utf8_get_char(ps , &wc);
