@@ -236,15 +236,13 @@ static void LuaResultItemCopy(void *_dst, const void *_src) {
     LuaResultItem *src = (LuaResultItem *)_src;
     dst->result = src->result ? strdup(src->result) : NULL;
     dst->help = src->help ? strdup(src->help) : NULL;
+    dst->tip = src->tip ? strdup(src->tip) : NULL;
 }
 static void LuaResultItemDtor(void *_elt) {
     LuaResultItem *elt = (LuaResultItem *)_elt;
-    if (elt->result) {
-        free(elt->result);
-    }
-    if (elt->help) {
-        free(elt->help);
-    }
+    fcitx_utils_free(elt->result);
+    fcitx_utils_free(elt->help);
+    fcitx_utils_free(elt->tip);
 }
 
 static void FreeTrigger(TriggerItem **triggers, LuaExtension *extension) {
@@ -304,8 +302,7 @@ static void FreeConverter(ConverterItem **converters, LuaExtension *extension) {
 
 static void UpdateShortestInputTriggerKeyLength(LuaModule *module) {
     size_t length = UINT_MAX;
-    TriggerItem *trigger;
-    for (trigger = module->input_triggers; trigger; trigger = trigger->hh.next) {
+    HASH_FOREACH(trigger, module->input_triggers, TriggerItem) {
         size_t keylen = strlen(trigger->key);
         if (keylen < length) {
             length = keylen;
@@ -412,8 +409,7 @@ static LuaExtension * FindExtension(lua_State *lua) {
         FcitxLog(ERROR, "LuaModule not found");
         return NULL;
     }
-    LuaExtension *e;
-    for (e = module->extensions; e != NULL; e = e->hh.next) {
+    HASH_FOREACH(e, module->extensions, LuaExtension) {
         if (e->lua == lua) {
             return e;
         }
@@ -636,7 +632,8 @@ static UT_array * LuaCallFunction(lua_State *lua,
         const char *str = lua_tostring(lua, -1);
         if (str) {
             utarray_new(result, &LuaResultItem_icd);
-            LuaResultItem r = {.result = (char *)str, .help = NULL};
+            /* str is ok for now and it will be copied by utarray */
+            LuaResultItem r = {.result = (char *)str, .help = NULL, .tip =  NULL};
             utarray_push_back(result, &r);
         } else {
             FcitxLog(WARNING, "lua function return return null");
@@ -649,31 +646,57 @@ static UT_array * LuaCallFunction(lua_State *lua,
         utarray_new(result, &LuaResultItem_icd);
         for (i = 1; i <= len; ++i) {
             lua_pushinteger(lua, i);
+            /* stack, table, integer */
             lua_gettable(lua, -2);
             char istable = 0;
             if (lua_type(lua, -1) == LUA_TTABLE) {
                 istable = 1;
+                /* stack, table, result-item */
                 lua_pushstring(lua, "help");
+                /* stack, table, result-item, "help" */
                 lua_gettable(lua, -2);
             }
+            LuaResultItem r = {NULL, NULL, NULL};
             const char *str = lua_tostring(lua, -1);
             if (str == NULL) {
-                FcitxLog(WARNING, "function %s() result[%d] is not string", i);
+                FcitxLog(WARNING, "function %s() result[%d] is not string", function_name, i);
             } else {
-                LuaResultItem r;
-                r.result = (char *)str;
+                r.result = strdup(str);
+            }
+            /* stack, table, result-item, string */
+            lua_pop(lua, 1);
+            /* stack, table, result-item */
+
+            if (r.result) {
                 if (istable) {
+                    const char* p;
+                    /* stack, table, result-item, "suggest" */
                     lua_pushstring(lua, "suggest");
-                    lua_gettable(lua, -3);
-                    r.help = (char *)lua_tostring(lua, -1);
+                    lua_gettable(lua, -2);
+                    /* stack, table, result-item, suggest-string */
+                    p = lua_tostring(lua, -1);
+                    if (p)
+                        r.help = strdup(p);
+                    lua_pop(lua, 1);
+                    /* stack, table, result-item */
+
+                    /* stack, table, result-item, "tip" */
+                    lua_pushstring(lua, "tip");
+                    lua_gettable(lua, -2);
+                    /* stack, table, result-item, tip string */
+                    p = lua_tostring(lua, -1);
+                    if (p)
+                        r.tip = strdup(p);
+                    lua_pop(lua, 1);
+                    /* stack, table, result-item */
                 } else {
                     r.help = NULL;
+                    r.tip = NULL;
                 }
                 utarray_push_back(result, &r);
             }
+            LuaResultItemDtor(&r);
             if (istable) {
-                lua_pop(lua, 3);
-            } else {
                 lua_pop(lua, 1);
             }
         }
@@ -740,4 +763,4 @@ void UnloadAllExtension(LuaModule* luamodule)
 {
     while(luamodule->extensions)
         UnloadExtensionByName(luamodule, luamodule->extensions->name);
-}
\ No newline at end of file
+}
