@@ -131,8 +131,7 @@ static void SelectVK(FcitxVKState* vkstate, int vkidx);
 
 static FcitxConfigColor blackColor = {0, 0, 0};
 
-FCITX_EXPORT_API
-FcitxModule module = {
+FCITX_DEFINE_PLUGIN(fcitx_vk, module, FcitxModule) = {
     VKCreate,
     NULL,
     NULL,
@@ -140,9 +139,6 @@ FcitxModule module = {
     ReloadVK
 };
 
-FCITX_EXPORT_API
-int ABI_VERSION = FCITX_ABI_VERSION;
-
 void *VKCreate(FcitxInstance* instance)
 {
     FcitxVKState *vkstate = fcitx_utils_malloc0(sizeof(FcitxVKState));
@@ -276,15 +272,14 @@ VKWindow* CreateVKWindow(FcitxVKState* vkstate)
     Colormap cmap;
     Visual * vs;
     int depth;
-    FcitxModuleFunctionArg arg;
-    VKWindow* vkWindow = fcitx_utils_malloc0(sizeof(VKWindow));
+    VKWindow* vkWindow = fcitx_utils_new(VKWindow);
     vkWindow->owner = vkstate;
 
     LoadVKImage(vkWindow);
 
     vs = VKFindARGBVisual(vkstate);
     VKInitWindowAttribute(vkstate, &vs, &cmap, &attrib, &attribmask, &depth);
-    vkWindow->dpy = InvokeFunction(vkstate->owner, FCITX_X11, GETDISPLAY, arg);
+    vkWindow->dpy = InvokeVaArgs(vkstate->owner, FCITX_X11, GETDISPLAY);
 
     vkWindow->fontSize = 12;
     vkWindow->defaultFont = strdup("sans");
@@ -306,10 +301,8 @@ VKWindow* CreateVKWindow(FcitxVKState* vkstate)
 
     VKSetWindowProperty(vkstate, vkWindow->window, FCITX_WINDOW_DOCK, strWindowName);
 
-    FcitxModuleFunctionArg arg2;
-    arg2.args[0] = VKWindowEventHandler;
-    arg2.args[1] = vkWindow;
-    InvokeFunction(vkstate->owner, FCITX_X11, ADDXEVENTHANDLER, arg2);
+    InvokeVaArgs(vkstate->owner, FCITX_X11, ADDXEVENTHANDLER,
+                 VKWindowEventHandler, vkWindow);
 
     return vkWindow;
 }
@@ -345,12 +338,10 @@ boolean VKWindowEventHandler(void* arg, XEvent* event)
 cairo_surface_t* LoadVKImage(VKWindow* vkWindow)
 {
     FcitxVKState* vkstate = vkWindow->owner;
-    FcitxModuleFunctionArg arg;
     boolean fallback = true;
     char vkimage[] = "keyboard.png";
-    arg.args[0] = vkimage;
-    arg.args[1] = &fallback;
-    cairo_surface_t* image = InvokeFunction(vkstate->owner, FCITX_CLASSIC_UI, LOADIMAGE, arg);
+    cairo_surface_t *image = InvokeVaArgs(vkstate->owner, FCITX_CLASSIC_UI,
+                                          LOADIMAGE, vkimage, &fallback);
 
     if (image)
         return image;
@@ -382,9 +373,9 @@ void DrawVKWindow(VKWindow* vkWindow)
     FcitxVKState *vkstate = vkWindow->owner;
     VKS *vks = vkstate->vks;
 
-    FcitxModuleFunctionArg arg;
-    FcitxConfigColor* fontColor = InvokeFunction(vkstate->owner, FCITX_CLASSIC_UI, GETKEYBOARDFONTCOLOR, arg);
-    char** font = InvokeFunction(vkstate->owner, FCITX_CLASSIC_UI, GETFONT, arg);
+    FcitxConfigColor *fontColor = InvokeVaArgs(vkstate->owner, FCITX_CLASSIC_UI,
+                                               GETKEYBOARDFONTCOLOR);
+    char **font = InvokeVaArgs(vkstate->owner, FCITX_CLASSIC_UI, GETFONT);
 
     if (!fontColor || !font) {
         fontColor = &blackColor;
@@ -397,39 +388,39 @@ void DrawVKWindow(VKWindow* vkWindow)
     cairo_paint(cr);
     /* 显示字符 */
     /* 名称 */
-    OutputString(cr, vks[vkstate->iCurrentVK].strName, *font, vkWindow->fontSize , (VK_WINDOW_WIDTH - StringWidth(vks[vkstate->iCurrentVK].strName, *font, vkWindow->fontSize)) / 2, 6, fontColor);
+    OutputString(cr, vks[vkstate->iCurrentVK].strName, *font, vkWindow->fontSize, false , (VK_WINDOW_WIDTH - StringWidth(vks[vkstate->iCurrentVK].strName, *font, vkWindow->fontSize, false)) / 2, 6, fontColor);
 
     /* 第一排 */
     iPos = 13;
     for (i = 0; i < 13; i++) {
-        OutputString(cr, vks[vkstate->iCurrentVK].strSymbol[i][1], *font, vkWindow->fontSize, iPos, 27, fontColor);
-        OutputString(cr, vks[vkstate->iCurrentVK].strSymbol[i][0], *font, vkWindow->fontSize, iPos - 5, 40, fontColor);
+        OutputString(cr, vks[vkstate->iCurrentVK].strSymbol[i][1], *font, vkWindow->fontSize, false, iPos, 27, fontColor);
+        OutputString(cr, vks[vkstate->iCurrentVK].strSymbol[i][0], *font, vkWindow->fontSize, false, iPos - 5, 40, fontColor);
         iPos += 24;
     }
     /* 第二排 */
     iPos = 48;
     for (i = 13; i < 26; i++) {
-        OutputString(cr, vks[vkstate->iCurrentVK].strSymbol[i][1], *font, vkWindow->fontSize, iPos, 55, fontColor);
-        OutputString(cr, vks[vkstate->iCurrentVK].strSymbol[i][0], *font, vkWindow->fontSize, iPos - 5, 68, fontColor);
+        OutputString(cr, vks[vkstate->iCurrentVK].strSymbol[i][1], *font, vkWindow->fontSize, false, iPos, 55, fontColor);
+        OutputString(cr, vks[vkstate->iCurrentVK].strSymbol[i][0], *font, vkWindow->fontSize, false, iPos - 5, 68, fontColor);
         iPos += 24;
     }
     /* 第三排 */
     iPos = 55;
     for (i = 26; i < 37; i++) {
-        OutputString(cr, vks[vkstate->iCurrentVK].strSymbol[i][1], *font, vkWindow->fontSize, iPos, 83, fontColor);
-        OutputString(cr, vks[vkstate->iCurrentVK].strSymbol[i][0], *font, vkWindow->fontSize, iPos - 5, 96, fontColor);
+        OutputString(cr, vks[vkstate->iCurrentVK].strSymbol[i][1], *font, vkWindow->fontSize, false, iPos, 83, fontColor);
+        OutputString(cr, vks[vkstate->iCurrentVK].strSymbol[i][0], *font, vkWindow->fontSize, false, iPos - 5, 96, fontColor);
         iPos += 24;
     }
 
     /* 第四排 */
     iPos = 72;
     for (i = 37; i < 47; i++) {
-        OutputString(cr, vks[vkstate->iCurrentVK].strSymbol[i][1], *font, vkWindow->fontSize, iPos, 111, fontColor);
-        OutputString(cr, vks[vkstate->iCurrentVK].strSymbol[i][0], *font, vkWindow->fontSize, iPos - 5, 124, fontColor);
+        OutputString(cr, vks[vkstate->iCurrentVK].strSymbol[i][1], *font, vkWindow->fontSize, false, iPos, 111, fontColor);
+        OutputString(cr, vks[vkstate->iCurrentVK].strSymbol[i][0], *font, vkWindow->fontSize, false, iPos - 5, 124, fontColor);
         iPos += 24;
     }
 
-    cairo_destroy(cr);
+    _CAIRO_DESTROY(cr);
 }
 
 /*
@@ -721,10 +712,8 @@ void SwitchVK(FcitxVKState *vkstate)
     if (vkstate->bVK) {
         int             x, y;
         int dwidth, dheight;
-        FcitxModuleFunctionArg arg;
-        arg.args[0] = &dwidth;
-        arg.args[1] = &dheight;
-        InvokeFunction(vkstate->owner, FCITX_X11, GETSCREENSIZE, arg);
+        InvokeVaArgs(vkstate->owner, FCITX_X11, GETSCREENSIZE,
+                     &dwidth, &dheight);
 
         if (!FcitxUISupportMainWindow(instance)) {
             x = dwidth / 2 - VK_WINDOW_WIDTH / 2;
@@ -777,41 +766,26 @@ VKInitWindowAttribute(FcitxVKState* vkstate, Visual ** vs, Colormap * cmap,
                       XSetWindowAttributes * attrib,
                       unsigned long *attribmask, int *depth)
 {
-    FcitxModuleFunctionArg arg;
-    arg.args[0] = vs;
-    arg.args[1] = cmap;
-    arg.args[2] = attrib;
-    arg.args[3] = attribmask;
-    arg.args[4] = depth;
-    InvokeFunction(vkstate->owner, FCITX_X11, INITWINDOWATTR, arg);
+    InvokeVaArgs(vkstate->owner, FCITX_X11, INITWINDOWATTR,
+                 vs, cmap, attrib, attribmask, depth);
 }
 
 Visual * VKFindARGBVisual(FcitxVKState* vkstate)
 {
-    FcitxModuleFunctionArg arg;
-    return InvokeFunction(vkstate->owner, FCITX_X11, FINDARGBVISUAL, arg);
+    return InvokeVaArgs(vkstate->owner, FCITX_X11, FINDARGBVISUAL);
 }
 
 void VKSetWindowProperty(FcitxVKState* vkstate, Window window, FcitxXWindowType type, char *windowTitle)
 {
-    FcitxModuleFunctionArg arg;
-    arg.args[0] = &window;
-    arg.args[1] = &type;
-    arg.args[2] = windowTitle;
-    InvokeFunction(vkstate->owner, FCITX_X11, SETWINDOWPROP, arg);
+    InvokeVaArgs(vkstate->owner, FCITX_X11, SETWINDOWPROP,
+                 &window, &type, windowTitle);
 }
 
 boolean
 VKMouseClick(FcitxVKState* vkstate, Window window, int *x, int *y)
 {
-    boolean            bMoved = false;
-    FcitxModuleFunctionArg arg;
-    arg.args[0] = &window;
-    arg.args[1] = x;
-    arg.args[2] = y;
-    arg.args[3] = &bMoved;
-    InvokeFunction(vkstate->owner, FCITX_X11, MOUSECLICK, arg);
-
+    boolean bMoved = false;
+    InvokeVaArgs(vkstate->owner, FCITX_X11, MOUSECLICK, &window, x, y, &bMoved);
     return bMoved;
 }
 
