@@ -41,6 +41,7 @@
 
 #include "config.h"
 #include "xkb.h"
+#include "xkb-internal.h"
 #include "rules.h"
 
 #ifndef XKB_RULES_XML_FILE
@@ -52,9 +53,12 @@
 
 static void* FcitxXkbGetRules(void* arg, FcitxModuleFunctionArg args);
 static void* FcitxXkbLayoutExists(void* arg, FcitxModuleFunctionArg args);
+static void FcitxXkbGetCurrentLayoutInternal(FcitxXkb *xkb,
+                                             char **layout, char **variant);
 static void* FcitxXkbGetCurrentLayout(void* arg, FcitxModuleFunctionArg args);
 static void* FcitxXkbGetLayoutOverride(void* arg, FcitxModuleFunctionArg args);
 static void* FcitxXkbSetLayoutOverride(void* arg, FcitxModuleFunctionArg args);
+static void* FcitxXkbSetDefaultLayout(void* arg, FcitxModuleFunctionArg args);
 static void* FcitxXkbCreate(struct _FcitxInstance* instance);
 static void FcitxXkbDestroy(void*);
 static void FcitxXkbReloadConfig(void*);
@@ -67,11 +71,18 @@ static Bool FcitxXkbUpdateProperties (FcitxXkb* xkb,
                      const char *rules_file, const char *model,
                      const char *all_layouts, const char *all_variants,
                      const char *all_options);
+static void FcitxXkbApplyCustomScript(FcitxXkb* xkb);
+
 #if 0
 static char* FcitxXkbGetCurrentLayout (FcitxXkb* xkb);
 static char* FcitxXkbGetCurrentModel (FcitxXkb* xkb);
 static char* FcitxXkbGetCurrentOption (FcitxXkb* xkb);
 #endif
+static boolean
+FcitxXkbSetLayoutByName(FcitxXkb *xkb,
+                        const char *layout, const char *variant);
+static void
+FcitxXkbRetrieveCloseGroup(FcitxXkb *xkb);
 static boolean FcitxXkbSetLayout  (FcitxXkb* xkb,
                const char *layouts,
                const char *variants,
@@ -96,8 +107,7 @@ typedef struct _LayoutOverride {
     UT_hash_handle hh;
 } LayoutOverride;
 
-FCITX_EXPORT_API
-FcitxModule module = {
+FCITX_DEFINE_PLUGIN(fcitx_xkb, module, FcitxModule) = {
     FcitxXkbCreate,
     NULL,
     NULL,
@@ -105,9 +115,6 @@ FcitxModule module = {
     FcitxXkbReloadConfig
 };
 
-FCITX_EXPORT_API
-int ABI_VERSION = FCITX_ABI_VERSION;
-
 boolean FcitxXkbSupported(FcitxXkb* xkb, int* xkbOpcode)
 {
     // Verify the Xlib has matching XKB extension.
@@ -139,7 +146,7 @@ boolean FcitxXkbSupported(FcitxXkb* xkb, int* xkbOpcode)
 }
 
 
-char* FcitxXkbGetRulesName(FcitxXkb* xkb)
+static char* FcitxXkbGetRulesName(FcitxXkb* xkb)
 {
     XkbRF_VarDefsRec vd;
     char *tmp = NULL;
@@ -158,48 +165,52 @@ static char* FcitxXkbFindXkbRulesFile(FcitxXkb* xkb)
     char* rulesName = FcitxXkbGetRulesName(xkb);
 
     if ( rulesName != NULL ) {
-        char* xkbParentDir = NULL;
-
-        const char* base = XLIBDIR;
-
-        int count = 0, i = 0;
-        while(base[i]) {
-            if (base[i] == '/')
-                count++;
-            i ++;
+        if (rulesName[0] == '/') {
+            asprintf(&rulesFile, "%s.xml", rulesName);
         }
+        else {
+            char* xkbParentDir = NULL;
 
-        if( count >= 3 ) {
-            // .../usr/lib/X11 -> /usr/share/X11/xkb vs .../usr/X11/lib -> /usr/X11/share/X11/xkb
-            const char* delta = StringEndsWith(base, "X11") ? "/../../share/X11" : "/../share/X11";
-            char* dirPath;
-            asprintf(&dirPath, "%s%s", base, delta);
+            const char* base = XLIBDIR;
 
-            DIR* dir = opendir(dirPath);
-            if( dir ) {
-                closedir(dir);
-                xkbParentDir = realpath(dirPath, NULL);
-                free(dirPath);
+            int count = 0, i = 0;
+            while(base[i]) {
+                if (base[i] == '/')
+                    count++;
+                i ++;
             }
-            else {
-                free(dirPath);
-                asprintf(&dirPath, "%s/X11", base);
+
+            if( count >= 3 ) {
+                // .../usr/lib/X11 -> /usr/share/X11/xkb vs .../usr/X11/lib -> /usr/X11/share/X11/xkb
+                const char* delta = StringEndsWith(base, "X11") ? "/../../share/X11" : "/../share/X11";
+                char* dirPath;
+                asprintf(&dirPath, "%s%s", base, delta);
+
                 DIR* dir = opendir(dirPath);
                 if( dir ) {
                     closedir(dir);
                     xkbParentDir = realpath(dirPath, NULL);
+                    free(dirPath);
+                }
+                else {
+                    free(dirPath);
+                    asprintf(&dirPath, "%s/X11", base);
+                    DIR* dir = opendir(dirPath);
+                    if( dir ) {
+                        closedir(dir);
+                        xkbParentDir = realpath(dirPath, NULL);
+                    }
+                    free(dirPath);
                 }
-                free(dirPath);
             }
-        }
 
-        if( xkbParentDir == NULL || strlen(xkbParentDir) == 0 ) {
-            xkbParentDir = strdup("/usr/share/X11");
-        }
+            if( xkbParentDir == NULL || strlen(xkbParentDir) == 0 ) {
+                xkbParentDir = strdup("/usr/share/X11");
+            }
 
-        rulesFile = fcitx_utils_malloc0(sizeof(char) * (1 + strlen(xkbParentDir) + strlen(rulesName) + strlen("/xkb/rules/")));
-        sprintf(rulesFile, "%s/xkb/rules/%s.xml", xkbParentDir, rulesName);
-        fcitx_utils_free(xkbParentDir);
+            asprintf(&rulesFile, "%s/xkb/rules/%s.xml", xkbParentDir, rulesName);
+            fcitx_utils_free(xkbParentDir);
+        }
     }
 
     fcitx_utils_free(rulesName);
@@ -263,7 +274,13 @@ FcitxXkbSetRules (FcitxXkb* xkb,
     XkbComponentNamesRec rnames;
     XkbDescPtr xkbDesc;
 
-    rulesPath = strdup("./rules/evdev");
+    if (!rules_file)
+        return False;
+
+    if (rules_file[0] != '/')
+        asprintf(&rulesPath, "./rules/%s", rules_file);
+    else
+        rulesPath = strdup(rules_file);
     rules = XkbRF_Load (rulesPath, "C", True, True);
     if (rules == NULL) {
         free (rulesPath);
@@ -293,18 +310,24 @@ FcitxXkbSetRules (FcitxXkb* xkb,
                                 XkbGBN_AllComponentsMask,
                                 XkbGBN_AllComponentsMask &
                                 (~XkbGBN_GeometryMask), True);
+
+    Bool result = True;
     if (!xkbDesc) {
         FcitxLog (WARNING, "Cannot load new keyboard description.");
-        return False;
+        result = False;
+    }
+    else {
+       char* tempstr = strdup(rules_file);
+       XkbRF_SetNamesProp (dpy, tempstr, &rdefs);
+       free (tempstr);
     }
-    XkbRF_SetNamesProp (dpy, rulesPath, &rdefs);
     free (rulesPath);
     free (rdefs.model);
     free (rdefs.layout);
     free (rdefs.variant);
     free (rdefs.options);
 
-    return True;
+    return result;
 }
 
 static Bool
@@ -376,29 +399,31 @@ FcitxXkbUpdateProperties (FcitxXkb* xkb,
     return True;
 }
 
-void*
-FcitxXkbGetCurrentLayout (void* arg, FcitxModuleFunctionArg args)
+static void
+FcitxXkbGetCurrentLayoutInternal(FcitxXkb *xkb, char **layout, char **variant)
 {
-    FcitxXkb* xkb = arg;
-    char** layout = args.args[0];
-    char** variant = args.args[1];
     char* const * layoutName = (char* const*) utarray_eltptr(xkb->defaultLayouts, FcitxXkbGetCurrentGroup(xkb));
     char* const * pVariantName = (char* const*) utarray_eltptr(xkb->defaultVariants, FcitxXkbGetCurrentGroup(xkb));
-    if (layoutName) {
+    if (layoutName)
         *layout = strdup(*layoutName);
-    }
     else
         *layout = NULL;
     if (pVariantName && strlen(*pVariantName) != 0)
         *variant = strdup(*pVariantName);
     else
         *variant = NULL;
+}
 
+static void*
+FcitxXkbGetCurrentLayout (void* arg, FcitxModuleFunctionArg args)
+{
+    FcitxXkbGetCurrentLayoutInternal((FcitxXkb*)arg,
+                                     args.args[0], args.args[1]);
     return NULL;
 }
 #if 0
 
-char *
+static char *
 FcitxXkbGetCurrentModel (FcitxXkb* xkb)
 {
     if (xkb->defaultModels == NULL) {
@@ -407,7 +432,7 @@ FcitxXkbGetCurrentModel (FcitxXkb* xkb)
     return fcitx_utils_join_string_list(xkb->defaultModels, ',');
 }
 
-char *
+static char *
 FcitxXkbGetCurrentOption (FcitxXkb* xkb)
 {
     if (xkb->defaultOptions == NULL) {
@@ -417,7 +442,8 @@ FcitxXkbGetCurrentOption (FcitxXkb* xkb)
 }
 #endif
 
-boolean
+/* This SHOULD be _SetLayouts_ .... */
+static boolean
 FcitxXkbSetLayout  (FcitxXkb* xkb,
                const char *layouts,
                const char *variants,
@@ -452,10 +478,10 @@ FcitxXkbSetLayout  (FcitxXkb* xkb,
     model_line = fcitx_utils_join_string_list(xkb->defaultModels, ',');
 
     retval = FcitxXkbSetRules(xkb,
-                              "evdev", model_line,
+                              FcitxXkbGetRulesName(xkb), model_line,
                               layouts_line, variants_line, options_line);
     FcitxXkbUpdateProperties(xkb,
-                             "evdev", model_line,
+                             FcitxXkbGetRulesName(xkb), model_line,
                              layouts_line, variants_line, options_line);
     free (layouts_line);
     free (variants_line);
@@ -465,7 +491,7 @@ FcitxXkbSetLayout  (FcitxXkb* xkb,
     return retval;
 }
 
-int
+static int
 FcitxXkbGetCurrentGroup (FcitxXkb* xkb)
 {
     Display *dpy = xkb->dpy;
@@ -484,8 +510,11 @@ FcitxXkbGetCurrentGroup (FcitxXkb* xkb)
     return state.group;
 }
 
-void FcitxXkbAddNewLayout(FcitxXkb* xkb, const char* layoutString, const char* variantString)
+static void FcitxXkbAddNewLayout(FcitxXkb* xkb, const char* layoutString, const char* variantString)
 {
+    if (!layoutString)
+        return;
+
     while (utarray_len(xkb->defaultVariants) < utarray_len(xkb->defaultLayouts)) {
         const char* dummy = "";
         utarray_push_back(xkb->defaultVariants, &dummy);
@@ -504,23 +533,60 @@ void FcitxXkbAddNewLayout(FcitxXkb* xkb, const char* layoutString, const char* v
     FcitxXkbSetLayout(xkb, NULL, NULL, NULL);
 }
 
-void FcitxXkbIMKeyboardLayoutChanged(void* arg, const void* value)
+static int
+FcitxXkbFindOrAddLayout(FcitxXkb *xkb, const char *layout, const char *variant)
 {
-    /* fcitx-keyboard will be useless, if fcitx-xkb don't change the layout for them */
+    int index;
+    if (layout == NULL)
+        return -1;
+    index = FcitxXkbFindLayoutIndex(xkb, layout, variant);
+    if (index >= 0)
+        return index;
+    if (!xkb->config.bOverrideSystemXKBSettings)
+        return -1;
+    FcitxXkbAddNewLayout(xkb, layout, variant);
+    FcitxXkbInitDefaultLayout(xkb);
+    return FcitxXkbFindLayoutIndex(xkb, layout, variant);
+}
+
+static boolean
+FcitxXkbSetLayoutByName(FcitxXkb *xkb, const char *layout, const char *variant)
+{
+    int index;
+    index = FcitxXkbFindOrAddLayout(xkb, layout, variant);
+    if (index < 0) {
+        return false;
+    }
+    XkbLockGroup(xkb->dpy, XkbUseCoreKbd, index);
+    return true;
+}
+
+static void
+FcitxXkbRetrieveCloseGroup(FcitxXkb *xkb)
+{
+    LayoutOverride* item = NULL;
+    HASH_FIND_STR(xkb->layoutOverride, "default", item);
+    if (item)
+        FcitxXkbSetLayoutByName(xkb, item->layout, item->variant);
+    else
+        FcitxXkbSetLayoutByName(xkb, xkb->closeLayout, xkb->closeVariant);
+}
+
+static void FcitxXkbIMKeyboardLayoutChanged(void* arg, const void* value)
+{
+    /* fcitx-keyboard will be useless, if fcitx-xkb doesn't change the layout for them */
     FcitxXkb* xkb = (FcitxXkb*) arg;
     FcitxIM* currentIM = FcitxInstanceGetCurrentIM(xkb->owner);
     if (xkb->config.bIgnoreInputMethodLayoutRequest
-        && (!currentIM || strncmp(currentIM->uniqueName, "fcitx-keyboard", strlen("fcitx-keyboard")) != 0))
+        && (!currentIM || strncmp(currentIM->uniqueName, "fcitx-keyboard",
+                                  strlen("fcitx-keyboard")) != 0))
     {
-        XkbLockGroup(xkb->dpy, XkbUseCoreKbd, xkb->closeGroup);
+        FcitxXkbRetrieveCloseGroup(xkb);
         return;
     }
 
     /* active means im will be take care */
     if (FcitxInstanceGetCurrentStatev2(xkb->owner) == IS_ACTIVE) {
-        if (value == NULL)
-            return;
-
         char* layoutString = NULL;
         char* variantString = NULL;
         LayoutOverride* item = NULL;
@@ -533,51 +599,29 @@ void FcitxXkbIMKeyboardLayoutChanged(void* arg, const void* value)
         }
         else {
             const char* layout = (const char*) value;
-            UT_array* s = fcitx_utils_split_string(layout, ',');
-            char** pLayoutString = (char**) utarray_eltptr(s, 0);
-            char** pVariantString = (char**) utarray_eltptr(s, 1);
-            layoutString = (pLayoutString)? *pLayoutString: NULL;
-            variantString = (pVariantString)? *pVariantString: NULL;
-        }
-        int idx = FcitxXkbFindLayoutIndex(xkb, layoutString, variantString);
-        if (idx >= 0) {
-            XkbLockGroup(xkb->dpy, XkbUseCoreKbd, idx);
-        }
-        else {
-
-            if (xkb->config.bOverrideSystemXKBSettings) {
-                FcitxXkbAddNewLayout(xkb, layoutString, variantString);
-                FcitxXkbInitDefaultLayout(xkb);
+            if (layout) {
+                UT_array* s = fcitx_utils_split_string(layout, ',');
+                char** pLayoutString = (char**) utarray_eltptr(s, 0);
+                char** pVariantString = (char**) utarray_eltptr(s, 1);
+                layoutString = (pLayoutString)? *pLayoutString: NULL;
+                variantString = (pVariantString)? *pVariantString: NULL;
+            }
+            else {
+                layoutString = NULL;
+                variantString = NULL;
             }
-
-            int idx = FcitxXkbFindLayoutIndex(xkb, layoutString, variantString);
-            if (idx >= 0)
-                XkbLockGroup(xkb->dpy, XkbUseCoreKbd, idx);
-            else
-                XkbLockGroup(xkb->dpy, XkbUseCoreKbd, xkb->closeGroup);
         }
+        if (!FcitxXkbSetLayoutByName(xkb, layoutString, variantString))
+            FcitxXkbRetrieveCloseGroup(xkb);
         if (s)
             fcitx_utils_free_string_list(s);
     }
     else {
-        XkbLockGroup(xkb->dpy, XkbUseCoreKbd, xkb->closeGroup);
+        FcitxXkbRetrieveCloseGroup(xkb);
     }
 }
 
-boolean strcmp0(const char* a, const char* b) {
-    boolean isemptya = false, isemptyb = false;
-    if (a == NULL || strlen(a) == 0)
-        isemptya = true;
-    if (b == NULL || strlen(b) == 0)
-        isemptyb = true;
-    if (isemptya && isemptyb)
-        return true;
-    if (isemptya ^ isemptyb)
-        return false;
-    return strcmp(a, b) == 0;
-}
-
-int FcitxXkbFindLayoutIndex(FcitxXkb* xkb, const char* layout, const char* variant)
+static int FcitxXkbFindLayoutIndex(FcitxXkb* xkb, const char* layout, const char* variant)
 {
     char** layoutName;
     char* variantName, **pVariantName;
@@ -590,26 +634,42 @@ int FcitxXkbFindLayoutIndex(FcitxXkb* xkb, const char* layout, const char* varia
         layoutName = (char**) utarray_eltptr(xkb->defaultLayouts, i);
         pVariantName = (char**) utarray_eltptr(xkb->defaultVariants, i);
         variantName = pVariantName ? *pVariantName : NULL;
-        if (strcmp(*layoutName, layout) == 0 && strcmp0(variantName, variant)) {
+        if (strcmp(*layoutName, layout) == 0 && fcitx_utils_strcmp_empty(variantName, variant) == 0) {
             return i;
         }
     }
     return -1;
 }
 
-void* FcitxXkbCreate(FcitxInstance* instance)
+static void
+FcitxXkbSaveCloseGroup(FcitxXkb *xkb)
+{
+    char *tmplayout;
+    char *tmpvariant;
+    FcitxXkbGetCurrentLayoutInternal(xkb, &tmplayout, &tmpvariant);
+    if (!tmplayout) {
+        if (tmpvariant)
+            free(tmpvariant);
+        return;
+    }
+    if (xkb->closeLayout)
+        free(xkb->closeLayout);
+    xkb->closeLayout = tmplayout;
+    if (xkb->closeVariant)
+        free(xkb->closeVariant);
+    xkb->closeVariant = tmpvariant;
+}
+
+static void* FcitxXkbCreate(FcitxInstance* instance)
 {
     FcitxAddon* addon = FcitxAddonsGetAddonByName(FcitxInstanceGetAddons(instance), "fcitx-xkb");
     FcitxXkb* xkb = (FcitxXkb*) fcitx_utils_malloc0(sizeof(FcitxXkb));
     xkb->owner = instance;
     do {
-
-        FcitxModuleFunctionArg args;
-        xkb->dpy = InvokeFunction(xkb->owner, FCITX_X11, GETDISPLAY, args);
+        xkb->dpy = InvokeVaArgs(xkb->owner, FCITX_X11, GETDISPLAY);
         if (!xkb->dpy)
             break;
 
-
         if (!FcitxXkbSupported(xkb, &xkb->xkbOpcode))
             break;
 
@@ -622,16 +682,13 @@ void* FcitxXkbCreate(FcitxInstance* instance)
         free(rulesPath);
 
         FcitxXkbInitDefaultLayout(xkb);
-
-        xkb->closeGroup = FcitxXkbGetCurrentGroup(xkb);
+        FcitxXkbSaveCloseGroup(xkb);
 
         int eventMask = XkbNewKeyboardNotifyMask | XkbStateNotifyMask;
         XkbSelectEvents(xkb->dpy, XkbUseCoreKbd, eventMask, eventMask);
 
-        FcitxModuleFunctionArg arg2;
-        arg2.args[0] = FcitxXkbEventHandler;
-        arg2.args[1] = xkb;
-        InvokeFunction(xkb->owner, FCITX_X11, ADDXEVENTHANDLER, arg2);
+        InvokeVaArgs(xkb->owner, FCITX_X11, ADDXEVENTHANDLER,
+                     FcitxXkbEventHandler, xkb);
 
         FcitxInstanceWatchContext(instance, CONTEXT_IM_KEYBOARD_LAYOUT, FcitxXkbIMKeyboardLayoutChanged, xkb);
 
@@ -651,6 +708,10 @@ void* FcitxXkbCreate(FcitxInstance* instance)
         AddFunction(addon, FcitxXkbLayoutExists);
         AddFunction(addon, FcitxXkbGetLayoutOverride);
         AddFunction(addon, FcitxXkbSetLayoutOverride);
+        AddFunction(addon, FcitxXkbSetDefaultLayout);
+
+        if (xkb->config.bOverrideSystemXKBSettings)
+            FcitxXkbSetLayout(xkb, NULL, NULL, NULL);
 
         return xkb;
     } while (0);
@@ -660,7 +721,7 @@ void* FcitxXkbCreate(FcitxInstance* instance)
     return NULL;
 }
 
-boolean FcitxXkbEventHandler(void* arg, XEvent* event)
+static boolean FcitxXkbEventHandler(void* arg, XEvent* event)
 {
     FcitxXkb* xkb = (FcitxXkb*) arg;
 
@@ -671,36 +732,43 @@ boolean FcitxXkbEventHandler(void* arg, XEvent* event)
             FcitxIM* currentIM = FcitxInstanceGetCurrentIM(xkb->owner);
             if (FcitxInstanceGetCurrentStatev2(xkb->owner) != IS_ACTIVE
                 || (xkb->config.bIgnoreInputMethodLayoutRequest
-                    && (!currentIM || strncmp(currentIM->uniqueName, "fcitx-keyboard", strlen("fcitx-keyboard")) != 0)))
-                xkb->closeGroup = FcitxXkbGetCurrentGroup(xkb);
+                    && (!currentIM ||
+                        strncmp(currentIM->uniqueName, "fcitx-keyboard",
+                                strlen("fcitx-keyboard")) != 0)))
+                FcitxXkbSaveCloseGroup(xkb);
         }
 
         if (xkbEvent->any.xkb_type == XkbNewKeyboardNotify) {
+            XSync(xkb->dpy, False);
+            FcitxUIUpdateInputWindow(xkb->owner);
             FcitxXkbInitDefaultLayout(xkb);
+            FcitxXkbApplyCustomScript(xkb);
         }
         return true;
     }
     return false;
 }
 
-void FcitxXkbCurrentStateChanged(void* arg)
+static void FcitxXkbCurrentStateChanged(void* arg)
 {
     FcitxXkb* xkb = (FcitxXkb*) arg;
     const char* layout = FcitxInstanceGetContextString(xkb->owner, CONTEXT_IM_KEYBOARD_LAYOUT);
     FcitxXkbIMKeyboardLayoutChanged(xkb, layout);
 }
 
-void FcitxXkbCurrentStateChangedTriggerOn(void* arg)
+static void FcitxXkbCurrentStateChangedTriggerOn(void* arg)
 {
     FcitxXkb* xkb = (FcitxXkb*) arg;
-    xkb->closeGroup = FcitxXkbGetCurrentGroup(xkb);
+    FcitxXkbSaveCloseGroup(xkb);
     FcitxXkbCurrentStateChanged(arg);
 }
 
-void FcitxXkbDestroy(void* arg)
+static void FcitxXkbDestroy(void* arg)
 {
     FcitxXkb* xkb = (FcitxXkb*) arg;
-    XkbLockGroup(xkb->dpy, XkbUseCoreKbd, xkb->closeGroup);
+    if (xkb->config.bOverrideSystemXKBSettings)
+        FcitxXkbSetLayout(xkb, NULL, NULL, NULL);
+    FcitxXkbRetrieveCloseGroup(xkb);
     XSync(xkb->dpy, False);
     FcitxXkbRulesFree(xkb->rules);
     fcitx_utils_free_string_list (xkb->defaultVariants);
@@ -714,21 +782,23 @@ void FcitxXkbDestroy(void* arg)
     free(xkb);
 }
 
-void FcitxXkbReloadConfig(void* arg)
+static void FcitxXkbReloadConfig(void* arg)
 {
     FcitxXkb* xkb = (FcitxXkb*) arg;
     LoadXkbConfig(xkb);
     FcitxXkbCurrentStateChanged(xkb);
+    if (xkb->config.bOverrideSystemXKBSettings)
+        FcitxXkbSetLayout(xkb, NULL, NULL, NULL);
 }
 
 
-void* FcitxXkbGetRules(void* arg, FcitxModuleFunctionArg args)
+static void* FcitxXkbGetRules(void* arg, FcitxModuleFunctionArg args)
 {
     FcitxXkb* xkb = (FcitxXkb*) arg;
     return xkb->rules;
 }
 
-void* FcitxXkbLayoutExists(void* arg, FcitxModuleFunctionArg args)
+static void* FcitxXkbLayoutExists(void* arg, FcitxModuleFunctionArg args)
 {
     FcitxXkb* xkb = (FcitxXkb*) arg;
     int idx = FcitxXkbFindLayoutIndex(xkb, args.args[0], args.args[1]);
@@ -737,6 +807,27 @@ void* FcitxXkbLayoutExists(void* arg, FcitxModuleFunctionArg args)
     return NULL;
 }
 
+static void FcitxXkbApplyCustomScript(FcitxXkb* xkb)
+{
+    if (!xkb->config.bOverrideSystemXKBSettings)
+        return;
+
+    if (!(xkb->config.xmodmapCommand && xkb->config.xmodmapCommand[0]
+        && xkb->config.customXModmapScript && xkb->config.customXModmapScript[0]))
+        return;
+
+    char* customXModmapScript = NULL;
+    FcitxXDGGetFileUserWithPrefix("data", xkb->config.customXModmapScript, NULL, &customXModmapScript);
+
+    char* args[] = {
+        xkb->config.xmodmapCommand,
+        customXModmapScript,
+        0
+    };
+    fcitx_utils_start_process(args);
+    free(customXModmapScript);
+}
+
 static inline void LayoutOverrideFree(LayoutOverride* item) {
     fcitx_utils_free(item->im);
     fcitx_utils_free(item->layout);
@@ -779,7 +870,8 @@ void LoadLayoutOverride(FcitxXkb* xkb)
         LayoutOverride* override = NULL;
         HASH_FIND_STR(xkb->layoutOverride, imString, override);
         /* if rule exists or the rule is for fcitx-keyboard, skip it */
-        if (override || strncmp(imString, "fcitx-keyboard", strlen("fcitx-keyboard")) == 0) {
+        if (override || strncmp(imString, "fcitx-keyboard",
+                                strlen("fcitx-keyboard")) == 0) {
             continue;
         }
 
@@ -833,9 +925,9 @@ void SaveLayoutOverride(FcitxXkb* xkb)
     LayoutOverride* item = xkb->layoutOverride;
     while(item) {
         if (item->variant)
-            fprintf(fp, "%s,%s,%s", item->im, item->layout, item->variant);
+            fprintf(fp, "%s,%s,%s\n", item->im, item->layout, item->variant);
         else
-            fprintf(fp, "%s,%s", item->im, item->layout);
+            fprintf(fp, "%s,%s\n", item->im, item->layout);
         item = item->hh.next;
     }
 
@@ -887,7 +979,8 @@ void* FcitxXkbSetLayoutOverride(void* arg, FcitxModuleFunctionArg args)
         LayoutOverrideFree(item);
     }
 
-    if (layout && layout[0] != '\0' && strncmp(imname, "fcitx-keyboard", strlen("fcitx-keyboard")) != 0) {
+    if (layout && layout[0] != '\0' &&
+        strncmp(imname, "fcitx-keyboard", strlen("fcitx-keyboard")) != 0) {
         item = fcitx_utils_new(LayoutOverride);
         item->im = strdup(imname);
         item->layout = strdup(layout);
@@ -900,4 +993,31 @@ void* FcitxXkbSetLayoutOverride(void* arg, FcitxModuleFunctionArg args)
     FcitxXkbCurrentStateChanged(xkb);
 
     return NULL;
-}
\ No newline at end of file
+}
+
+void* FcitxXkbSetDefaultLayout(void* arg, FcitxModuleFunctionArg args)
+{
+    FcitxXkb* xkb = arg;
+    const char* layout = args.args[0];
+    const char* variant = args.args[1];
+    LayoutOverride* item = NULL;
+    HASH_FIND_STR(xkb->layoutOverride, "default", item);
+    if (item) {
+        HASH_DEL(xkb->layoutOverride, item);
+        LayoutOverrideFree(item);
+    }
+
+    if (layout && layout[0] != '\0') {
+        item = fcitx_utils_new(LayoutOverride);
+        item->im = strdup("default");
+        item->layout = strdup(layout);
+        item->variant= (variant && variant[0] != '\0')? strdup(variant) : NULL;
+
+        HASH_ADD_KEYPTR(hh, xkb->layoutOverride, item->im, strlen(item->im), item);
+    }
+
+    SaveLayoutOverride(xkb);
+    FcitxXkbCurrentStateChanged(xkb);
+
+    return NULL;
+}
