@@ -24,6 +24,7 @@
 
 #include "libintl.h"
 #include "module/xkb/xkb.h"
+#include "module/xkb/rules.h"
 #include "module/dbus/dbusstuff.h"
 #include "fcitx-utils/log.h"
 #include "fcitx/module.h"
@@ -36,8 +37,7 @@ typedef struct _FcitxXkbDBus {
     FcitxIsoCodes* isocodes;
 } FcitxXkbDBus;
 
-FCITX_EXPORT_API
-FcitxModule module = {
+FCITX_DEFINE_PLUGIN(fcitx_xkbdbus, module, FcitxModule) = {
     FcitxXkbDBusCreate,
     NULL,
     NULL,
@@ -45,9 +45,6 @@ FcitxModule module = {
     NULL
 };
 
-FCITX_EXPORT_API
-int ABI_VERSION = FCITX_ABI_VERSION;
-
 static DBusHandlerResult FcitxXkbDBusEventHandler (DBusConnection  *connection,
                                                    DBusMessage     *message,
                                                    void            *user_data);
@@ -62,6 +59,10 @@ const char *introspection_xml =
     "    </method>\n"
     "  </interface>\n"
     "  <interface name=\"" FCITX_XKB_INTERFACE "\">\n"
+    "    <method name=\"SetDefaultLayout\">\n"
+    "        <arg name=\"layout\" direction=\"in\" type=\"s\"/>"
+    "        <arg name=\"variant\" direction=\"in\" type=\"s\"/>"
+    "    </method>\n"
     "    <method name=\"GetLayouts\">\n"
     "      <arg name=\"layouts\" direction=\"out\" type=\"a(ssss)\"/>\n"
     "    </method>\n"
@@ -83,8 +84,8 @@ void* FcitxXkbDBusCreate(FcitxInstance* instance)
     FcitxXkbDBus* xkbdbus = fcitx_utils_new(FcitxXkbDBus);
     xkbdbus->owner = instance;
     do {
-        FcitxModuleFunctionArg arg;
-        DBusConnection* conn = InvokeFunction(instance, FCITX_DBUS, GETCONNECTION, arg);
+        DBusConnection *conn = InvokeVaArgs(instance, FCITX_DBUS,
+                                            GETCONNECTION);
         if (conn == NULL) {
             FcitxLog(ERROR, "DBus Not initialized");
             break;
@@ -97,8 +98,7 @@ void* FcitxXkbDBusCreate(FcitxInstance* instance)
             FcitxLog(ERROR, "No memory");
             break;
         }
-        FcitxModuleFunctionArg args;
-        FcitxXkbRules* rules = InvokeFunction(instance, FCITX_XKB, GETRULES, args);
+        FcitxXkbRules *rules = InvokeVaArgs(instance, FCITX_XKB, GETRULES);
 
         if (!rules)
             break;
@@ -156,7 +156,7 @@ void FcitxXkbDBusGetLayouts(FcitxXkbDBus* xkbdbus, DBusMessage* message)
             layoutInfo = (FcitxXkbLayoutInfo*) utarray_next(rules->layoutInfos, layoutInfo))
         {
             char* description;
-            asprintf(&description, _("%s"),
+            asprintf(&description, "%s",
                      dgettext("xkeyboard-config", layoutInfo->description));
 
             GET_LANG;
@@ -213,27 +213,34 @@ DBusHandlerResult FcitxXkbDBusEventHandler (DBusConnection  *connection,
         dbus_error_init(&error);
         char* im, *layout, *variant;
         if (dbus_message_get_args(message, &error, DBUS_TYPE_STRING, &im, DBUS_TYPE_STRING, &layout, DBUS_TYPE_STRING, &variant, DBUS_TYPE_INVALID)) {
-            FcitxModuleFunctionArg args;
-            args.args[0] = im;
-            args.args[1] = layout;
-            args.args[2] = variant;
-            InvokeFunction(xkbdbus->owner, FCITX_XKB, SETLAYOUTOVERRIDE, args);
+            InvokeVaArgs(xkbdbus->owner, FCITX_XKB, SETLAYOUTOVERRIDE,
+                         im, layout, variant);
         }
         DBusMessage *reply = dbus_message_new_method_return(message);
         dbus_connection_send(connection, reply, NULL);
         dbus_message_unref(reply);
         dbus_connection_flush(connection);
         return DBUS_HANDLER_RESULT_HANDLED;
-    } else if (dbus_message_is_method_call(message, FCITX_XKB_INTERFACE, "GetLayoutForIM")) {
+    } else if (dbus_message_is_method_call(message, FCITX_XKB_INTERFACE, "SetDefaultLayout")) {
+        DBusError error;
+        dbus_error_init(&error);
+        char *layout, *variant;
+        if (dbus_message_get_args(message, &error, DBUS_TYPE_STRING, &layout, DBUS_TYPE_STRING, &variant, DBUS_TYPE_INVALID)) {
+            InvokeVaArgs(xkbdbus->owner, FCITX_XKB, SETDEFAULTLAYOUT,
+                           layout, variant);
+        }
+        DBusMessage *reply = dbus_message_new_method_return(message);
+        dbus_connection_send(connection, reply, NULL);
+        dbus_message_unref(reply);
+        dbus_connection_flush(connection);
+        return DBUS_HANDLER_RESULT_HANDLED;
+    }  else if (dbus_message_is_method_call(message, FCITX_XKB_INTERFACE, "GetLayoutForIM")) {
         DBusError error;
         dbus_error_init(&error);
         char* im = NULL, *layout = NULL, *variant = NULL;
         if (dbus_message_get_args(message, &error, DBUS_TYPE_STRING, &im, DBUS_TYPE_INVALID)) {
-            FcitxModuleFunctionArg args;
-            args.args[0] = im;
-            args.args[1] = &layout;
-            args.args[2] = &variant;
-            InvokeFunction(xkbdbus->owner, FCITX_XKB, GETLAYOUTOVERRIDE, args);
+            InvokeVaArgs(xkbdbus->owner, FCITX_XKB, GETLAYOUTOVERRIDE,
+                         im, &layout, &variant);
 
             if (!layout)
                 layout = "";
@@ -249,4 +256,4 @@ DBusHandlerResult FcitxXkbDBusEventHandler (DBusConnection  *connection,
         return DBUS_HANDLER_RESULT_HANDLED;
     }
     return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
-}
\ No newline at end of file
+}
