@@ -119,14 +119,10 @@ MainWindow* CreateMainWindow(FcitxClassicUI* classicui)
     mainWindow->owner = classicui;
     InitMainWindow(mainWindow);
 
-    FcitxModuleFunctionArg arg;
-    arg.args[0] = MainWindowEventHandler;
-    arg.args[1] = mainWindow;
-    InvokeFunction(classicui->owner, FCITX_X11, ADDXEVENTHANDLER, arg);
-
-    arg.args[0] = ReloadMainWindow;
-    arg.args[1] = mainWindow;
-    InvokeFunction(classicui->owner, FCITX_X11, ADDCOMPOSITEHANDLER, arg);
+    InvokeVaArgs(classicui->owner, FCITX_X11, ADDXEVENTHANDLER,
+                 MainWindowEventHandler, mainWindow);
+    InvokeVaArgs(classicui->owner, FCITX_X11, ADDCOMPOSITEHANDLER,
+                 ReloadMainWindow, mainWindow);
     return mainWindow;
 }
 
@@ -216,7 +212,7 @@ void DrawMainWindow(MainWindow* mainWindow)
                     if (FcitxInstanceGetCurrentStatev2(instance) != IS_ACTIVE || FcitxInstanceGetCurrentIM(instance) == NULL)
                         imicon = LoadImage(sc, sc->skinMainBar.eng, false);
                     else {
-                        imicon = GetIMIcon(instance, sc, sc->skinMainBar.active, 3, false);
+                        imicon = GetIMIcon(classicui, sc, sc->skinMainBar.active, 3, false);
                     }
 
                     if (imicon) {
@@ -285,7 +281,7 @@ void DrawMainWindow(MainWindow* mainWindow)
             if (FcitxInstanceGetCurrentStatev2(instance) != IS_ACTIVE || FcitxInstanceGetCurrentIM(instance) == NULL)
                 imicon = LoadImage(sc, sc->skinMainBar.eng, false);
             else {
-                imicon = GetIMIcon(instance, sc, sc->skinMainBar.active, 3, false);
+                imicon = GetIMIcon(classicui, sc, sc->skinMainBar.active, 3, false);
             }
 
             if (imicon) {
@@ -306,13 +302,25 @@ void DrawMainWindow(MainWindow* mainWindow)
                     continue;
                 const char* icon = compstatus->getIconName(compstatus->arg);
                 char* path;
-                if (icon[0] != '/')
-                    asprintf(&path, "%s.png", compstatus->name);
+                if (icon[0] != '/') {
+                    if (icon[0] == '\0')
+                        asprintf(&path, "%s", compstatus->shortDescription);
+                    else
+                        asprintf(&path, "%s.png", icon);
+                }
                 else
                     path = strdup(icon);
-                SkinImage* statusicon = LoadImage(sc, path, false);
-                if (statusicon == NULL) {
-                    continue;
+                SkinImage* statusicon = NULL;
+                if (icon[0] != '\0')
+                    statusicon = LoadImage(sc, path, false);
+                if (statusicon == NULL || statusicon->textIcon) {
+                    if (activeIcon && icon[0] == '\0') {
+                        statusicon = LoadImageWithText(classicui, sc, path, compstatus->shortDescription,
+                                                       cairo_image_surface_get_width(activeIcon->image),
+                                                       cairo_image_surface_get_height(activeIcon->image),
+                                                       true
+                                                      );
+                    }
                 }
                 else {
                     if (icon[0] == '/' && activeIcon) {
@@ -322,6 +330,10 @@ void DrawMainWindow(MainWindow* mainWindow)
                     }
                 }
                 free(path);
+
+                if (statusicon == NULL)
+                    continue;
+
                 currentX += cairo_image_surface_get_width(statusicon->image);
                 imageheight = cairo_image_surface_get_height(statusicon->image);
                 if (imageheight > height)
@@ -400,8 +412,12 @@ void DrawMainWindow(MainWindow* mainWindow)
                     continue;
                 const char* icon = compstatus->getIconName(compstatus->arg);
                 char* path;
-                if (icon[0] != '/')
-                    asprintf(&path, "%s.png", compstatus->name);
+                if (icon[0] != '/') {
+                    if (icon[0] == '\0')
+                        asprintf(&path, "%s", compstatus->shortDescription);
+                    else
+                        asprintf(&path, "%s.png", compstatus->name);
+                }
                 else
                     path = strdup(icon);
                 SkinImage* statusicon = LoadImage(sc, path, false);
@@ -441,8 +457,8 @@ void DrawMainWindow(MainWindow* mainWindow)
             }
         }
 
-        cairo_destroy(c);
-        cairo_xlib_surface_set_size(mainWindow->cs_x_main_bar,
+        _CAIRO_DESTROY(c);
+        _CAIRO_SETSIZE(mainWindow->cs_x_main_bar,
                                     width,
                                     height);
 
@@ -452,7 +468,7 @@ void DrawMainWindow(MainWindow* mainWindow)
         cairo_rectangle(c, 0, 0, width, height);
         cairo_clip(c);
         cairo_paint(c);
-        cairo_destroy(c);
+        _CAIRO_DESTROY(c);
 
         XMapRaised(mainWindow->dpy, mainWindow->window);
     } else
@@ -607,11 +623,7 @@ boolean MainWindowEventHandler(void *arg, XEvent* event)
                     SaveClassicUIConfig(classicui);
                 } else if (IsInRspArea(event->xbutton.x, event->xbutton.y, &mainWindow->imiconstat)) {
                     mouse = &mainWindow->imiconstat.mouse;
-                    /* if close, we enable it, if enabled, we scroll it */
-                    if (FcitxInstanceGetCurrentState(instance) != IS_ACTIVE)
-                        FcitxInstanceEnableIM(instance, FcitxInstanceGetCurrentIC(instance), false);
-                    else
-                        FcitxInstanceSwitchIM(instance, -1);
+                    FcitxInstanceSwitchIMByIndex(instance, -1);
                 } else {
                     FcitxUIComplexStatus *compstatus;
                     UT_array* uicompstats = FcitxInstanceGetUIComplexStats(instance);
