@@ -63,10 +63,8 @@ MessageWindow* CreateMessageWindow(FcitxClassicUI * classicui)
     InitMessageWindowProperty(messageWindow);
     XSelectInput(dpy, messageWindow->window, ExposureMask | ButtonPressMask | ButtonReleaseMask  | PointerMotionMask);
 
-    FcitxModuleFunctionArg arg;
-    arg.args[0] = MessageWindowEventHandler;
-    arg.args[1] = messageWindow;
-    InvokeFunction(classicui->owner, FCITX_X11, ADDXEVENTHANDLER, arg);
+    InvokeVaArgs(classicui->owner, FCITX_X11, ADDXEVENTHANDLER,
+                 MessageWindowEventHandler, messageWindow);
 
     return messageWindow;
 }
@@ -169,20 +167,20 @@ void DrawMessageWindow(MessageWindow* messageWindow, char *title, char **msg, in
     messageWindow->width = 0;
 
     for (i = 0; i < length ; i ++) {
-        int width = StringWidth(msg[i], classicui->font, messageWindow->fontSize);
+        int width = StringWidth(msg[i], classicui->font, messageWindow->fontSize, false);
         if (width > messageWindow->width)
             messageWindow->width = width;
     }
 
     messageWindow->width += MESSAGE_WINDOW_MARGIN * 2;
     XResizeWindow(dpy, messageWindow->window, messageWindow->width, messageWindow->height);
-    cairo_xlib_surface_set_size(messageWindow->surface, messageWindow->width, messageWindow->height);
+    _CAIRO_SETSIZE(messageWindow->surface, messageWindow->width, messageWindow->height);
 
     cairo_t *c = cairo_create(messageWindow->surface);
     cairo_set_source_rgb(c, messageWindow->color.r, messageWindow->color.g, messageWindow->color.b);
     cairo_set_operator(c, CAIRO_OPERATOR_SOURCE);
 
-    SetFontContext(c, classicui->font, messageWindow->fontSize);
+    SetFontContext(c, classicui->font, messageWindow->fontSize, 0);
 
     cairo_paint(c);
 
@@ -192,12 +190,12 @@ void DrawMessageWindow(MessageWindow* messageWindow, char *title, char **msg, in
     x = MESSAGE_WINDOW_MARGIN;
     y = MESSAGE_WINDOW_MARGIN;
     for (i = 0; i < length ; i ++) {
-        OutputStringWithContext(c, msg[i], x, y);
+        OutputStringWithContext(c, 0, msg[i], x, y);
         y += messageWindow->fontSize + MESSAGE_WINDOW_LINESPACE;
     }
 
     ResetFontContext();
-    cairo_destroy(c);
+    _CAIRO_DESTROY(c);
 
     ActivateWindow(dpy, classicui->iScreen, messageWindow->window);
 }
