@@ -77,8 +77,7 @@ static void* ClassicUILoadImage(void *arg, FcitxModuleFunctionArg args);
 static void* ClassicUIGetKeyBoardFontColor(void* arg, FcitxModuleFunctionArg args);
 static void* ClassicUIGetFont(void *arg, FcitxModuleFunctionArg args);
 
-FCITX_EXPORT_API
-FcitxUI ui = {
+FCITX_DEFINE_PLUGIN(fcitx_classic_ui, ui, FcitxUI) = {
     ClassicUICreate,
     ClassicUICloseInputWindow,
     ClassicUIShowInputWindow,
@@ -101,13 +100,9 @@ FcitxUI ui = {
     NULL
 };
 
-FCITX_EXPORT_API
-int ABI_VERSION = FCITX_ABI_VERSION;
-
 void* ClassicUICreate(FcitxInstance* instance)
 {
     FcitxAddon* classicuiaddon = FcitxAddonsGetAddonByName(FcitxInstanceGetAddons(instance), FCITX_CLASSIC_UI_NAME);
-    FcitxModuleFunctionArg arg;
     FcitxClassicUI* classicui = fcitx_utils_malloc0(sizeof(FcitxClassicUI));
     classicui->owner = instance;
     if (!LoadClassicUIConfig(classicui)) {
@@ -118,12 +113,16 @@ void* ClassicUICreate(FcitxInstance* instance)
         free(classicui);
         return NULL;
     }
-    classicui->dpy = InvokeFunction(instance, FCITX_X11, GETDISPLAY, arg);
+    classicui->dpy = InvokeVaArgs(instance, FCITX_X11, GETDISPLAY);
     if (classicui->dpy == NULL) {
         free(classicui);
         return NULL;
     }
 
+    InvokeVaArgs(instance, FCITX_X11, GETDPI, &classicui->dpi, NULL);
+    if (classicui->dpi <= 0)
+        classicui->dpi = 96;
+
     if (LoadSkinConfig(&classicui->skin, &classicui->skinType)) {
         free(classicui);
         return NULL;
@@ -170,11 +169,8 @@ void* ClassicUICreate(FcitxInstance* instance)
 
 void ClassicUISetWindowProperty(FcitxClassicUI* classicui, Window window, FcitxXWindowType type, char *windowTitle)
 {
-    FcitxModuleFunctionArg arg;
-    arg.args[0] = &window;
-    arg.args[1] = &type;
-    arg.args[2] = windowTitle;
-    InvokeFunction(classicui->owner, FCITX_X11, SETWINDOWPROP, arg);
+    InvokeVaArgs(classicui->owner, FCITX_X11, SETWINDOWPROP,
+                 &window, &type, windowTitle);
 }
 
 static void ClassicUIInputReset(void *arg)
@@ -303,21 +299,14 @@ void ActivateWindow(Display *dpy, int iScreen, Window window)
 
 void GetScreenSize(FcitxClassicUI* classicui, int* width, int* height)
 {
-    FcitxModuleFunctionArg arg;
-    arg.args[0] = width;
-    arg.args[1] = height;
-    InvokeFunction(classicui->owner, FCITX_X11, GETSCREENSIZE, arg);
+    InvokeVaArgs(classicui->owner, FCITX_X11, GETSCREENSIZE, width, height);
 }
 
 FcitxRect GetScreenGeometry(FcitxClassicUI* classicui, int x, int y)
 {
-    FcitxModuleFunctionArg arg;
     FcitxRect result = { 0, 0 , 0 , 0 };
-    arg.args[0] = &x;
-    arg.args[1] = &y;
-    arg.args[2] = &result;
-    InvokeFunction(classicui->owner, FCITX_X11, GETSCREENGEOMETRY, arg);
-
+    InvokeVaArgs(classicui->owner, FCITX_X11, GETSCREENGEOMETRY,
+                 &x, &y, &result);
     return result;
 }
 
@@ -368,14 +357,9 @@ boolean IsInRspArea(int x0, int y0, FcitxClassicUIStatus* status)
 boolean
 ClassicUIMouseClick(FcitxClassicUI* classicui, Window window, int *x, int *y)
 {
-    boolean            bMoved = false;
-    FcitxModuleFunctionArg arg;
-    arg.args[0] = &window;
-    arg.args[1] = x;
-    arg.args[2] = y;
-    arg.args[3] = &bMoved;
-    InvokeFunction(classicui->owner, FCITX_X11, MOUSECLICK, arg);
-
+    boolean bMoved = false;
+    InvokeVaArgs(classicui->owner, FCITX_X11, MOUSECLICK,
+                 &window, x, y, &bMoved);
     return bMoved;
 }
 
@@ -475,6 +459,7 @@ static void UpdateMainMenu(FcitxUIMenu* menu)
     FcitxMenuAddMenuItem(menu, NULL, MENUTYPE_DIVLINE, NULL);
     FcitxMenuAddMenuItem(menu, _("Configure Current Input Method"), MENUTYPE_SIMPLE, NULL);
     FcitxMenuAddMenuItem(menu, _("Configure"), MENUTYPE_SIMPLE, NULL);
+    FcitxMenuAddMenuItem(menu, _("Restart"), MENUTYPE_SIMPLE, NULL);
     FcitxMenuAddMenuItem(menu, _("Exit"), MENUTYPE_SIMPLE, NULL);
 }
 
@@ -484,19 +469,25 @@ boolean MainMenuAction(FcitxUIMenu* menu, int index)
     FcitxInstance* instance = classicui->owner;
     int length = utarray_len(&menu->shell);
     if (index == 0) {
-        FILE* p = popen("xdg-open http://fcitx-im.org/ &", "r");
-        if (p)
-            pclose(p);
-        else
-            FcitxLog(ERROR, _("Unable to create process"));
+        char* args[] = {
+            "xdg-open",
+            "http://fcitx-im.org/",
+            0
+        };
+        fcitx_utils_start_process(args);
     } else if (index == length - 1) { /* Exit */
         FcitxInstanceEnd(classicui->owner);
-    } else if (index == length - 2) { /* Configuration */
-        fcitx_utils_launch_configure_tool();
+    } else if (index == length - 2) { /* Restart */
+        fcitx_utils_launch_restart();
     } else if (index == length - 3) { /* Configuration */
+        fcitx_utils_launch_configure_tool();
+    } else if (index == length - 4) { /* Configuration */
         FcitxIM* im = FcitxInstanceGetCurrentIM(classicui->owner);
         if (im && im->owner) {
-            fcitx_utils_launch_configure_tool_for_addon(im->owner->name);
+            fcitx_utils_launch_configure_tool_for_addon(im->uniqueName);
+        }
+        else {
+            fcitx_utils_launch_configure_tool();
         }
     } else {
         FcitxMenuItem* item = (FcitxMenuItem*) utarray_eltptr(&menu->shell, index);
@@ -509,23 +500,17 @@ boolean MainMenuAction(FcitxUIMenu* menu, int index)
 }
 
 void
-ClassicUIInitWindowAttribute(FcitxClassicUI* classicui, Visual ** vs, Colormap * cmap,
-                             XSetWindowAttributes * attrib,
+ClassicUIInitWindowAttribute(FcitxClassicUI *classicui, Visual **vs,
+                             Colormap *cmap, XSetWindowAttributes *attrib,
                              unsigned long *attribmask, int *depth)
 {
-    FcitxModuleFunctionArg arg;
-    arg.args[0] = vs;
-    arg.args[1] = cmap;
-    arg.args[2] = attrib;
-    arg.args[3] = attribmask;
-    arg.args[4] = depth;
-    InvokeFunction(classicui->owner, FCITX_X11, INITWINDOWATTR, arg);
+    InvokeVaArgs(classicui->owner, FCITX_X11, INITWINDOWATTR,
+                   vs, cmap, attrib, attribmask, depth);
 }
 
 Visual * ClassicUIFindARGBVisual(FcitxClassicUI* classicui)
 {
-    FcitxModuleFunctionArg arg;
-    return InvokeFunction(classicui->owner, FCITX_X11, FINDARGBVISUAL, arg);
+    return InvokeVaArgs(classicui->owner, FCITX_X11, FINDARGBVISUAL);
 }
 
 void ClassicUIMainWindowSizeHint(void* arg, int* x, int* y, int* w, int* h)
@@ -634,7 +619,7 @@ void ResizeSurface(cairo_surface_t** surface, int w, int h)
     cairo_rectangle(c, 0, 0, ow, oh);
     cairo_clip(c);
     cairo_paint(c);
-    cairo_destroy(c);
+    _CAIRO_DESTROY(c);
 
     cairo_surface_destroy(*surface);
 
