@@ -65,6 +65,7 @@ typedef struct _FcitxClassicUI {
     UT_array skinBuf;
     struct _FcitxInstance *owner;
 
+    int fontSize;
     char* font;
     char* menuFont;
     char* strUserLocale;
@@ -81,6 +82,8 @@ typedef struct _FcitxClassicUI {
     FcitxUIMenu mainMenu;
     boolean isSuspend;
     boolean isfallback;
+
+    int dpi;
 } FcitxClassicUI;
 
 void GetScreenSize(FcitxClassicUI* classicui, int* width, int* height);
