@@ -37,6 +37,7 @@
 #include <dirent.h>
 #include <sys/stat.h>
 #include <libintl.h>
+#include <ctype.h>
 
 #include "fcitx/fcitx.h"
 
@@ -78,23 +79,7 @@ int LoadSkinConfig(FcitxSkin* sc, char** skinType)
     int ret = 0;
     if (sc->config.configFile) {
         utarray_done(&sc->skinMainBar.skinPlacement);
-        FcitxConfigFreeConfigFile(sc->config.configFile);
-        free(sc->skinInfo.skinName);
-        free(sc->skinInfo.skinVersion);
-        free(sc->skinInfo.skinAuthor);
-        free(sc->skinInfo.skinDesc);
-        free(sc->skinMainBar.backImg);
-        free(sc->skinMainBar.logo);
-        free(sc->skinMainBar.eng);
-        free(sc->skinMainBar.active);
-        free(sc->skinMainBar.placement);
-        free(sc->skinInputBar.backImg);
-        free(sc->skinInputBar.backArrow);
-        free(sc->skinInputBar.forwardArrow);
-        free(sc->skinTrayIcon.active);
-        free(sc->skinTrayIcon.inactive);
-        free(sc->skinMenu.backImg);
-        free(sc->skinKeyboard.backImg);
+        FcitxConfigFree(&sc->config);
         UnloadImage(sc);
     }
     memset(sc, 0, sizeof(FcitxSkin));
@@ -160,13 +145,14 @@ reload:
 
 SkinImage* LoadImageWithText(FcitxClassicUI* classicui, FcitxSkin* sc, const char* name, const char* text, int w, int h, boolean active)
 {
-    if (!text || strlen(text) == 0)
+    if (!text || *text == '\0')
         return NULL;
 
     UnloadSingleImage(sc, name);
 
-
     int len = fcitx_utf8_char_len(text);
+    if (len == 1 && text[len] && fcitx_utf8_char_len(text + len) == 1)
+        len = 2;
 
     char* iconText = strndup(text, len);
 
@@ -193,12 +179,11 @@ SkinImage* LoadImageWithText(FcitxClassicUI* classicui, FcitxSkin* sc, const cha
         color = sc->skinFont.menuFontColor[1];
 
     int textw, texth;
-    StringSizeStrict(iconText, classicui->font, min, &textw, &texth);
+    StringSizeStrict(iconText, classicui->font, min, false, &textw, &texth);
 
-    int fh = FontHeight(classicui->font, min);
-    OutputString(c, iconText, classicui->font, min, (w - textw) * 0.5, (h - texth) * 0.5 - (fh - texth), &color);
+    OutputString(c, iconText, classicui->font, min, false, (w - textw) * 0.5, 0, &color);
 
-    cairo_destroy(c);
+    _CAIRO_DESTROY(c);
     SkinImage* image = fcitx_utils_malloc0(sizeof(SkinImage));
     image->name = strdup(name);
     image->image = newsurface;
@@ -216,7 +201,7 @@ SkinImage* LoadImageFromTable(SkinImage** imageTable, const char* skinType, cons
     const char* fallbackChainNoFallback[] = { buf };
     const char* fallbackChainPanel[] = { buf, "skin/default" };
     const char* fallbackChainTray[] = { "imicon" };
-    const char* fallbackChainPanelIMIcon[] = { buf, "imicon" };
+    const char* fallbackChainPanelIMIcon[] = { buf, "imicon", "skin/default" };
 
     HASH_FIND_STR(*imageTable, name, image);
     if (image != NULL) {
@@ -552,19 +537,19 @@ void LoadInputMessage(FcitxSkin* sc, InputWindow* inputWindow, const char* font)
     FcitxConfigColor cursorColor = sc->skinInputBar.cursorColor;
 
     if (inputWindow->c_back) {
-        cairo_destroy(inputWindow->c_back);
+        _CAIRO_DESTROY(inputWindow->c_back);
         inputWindow->c_back = NULL;
     }
 
     for (i = 0; i < 7 ; i ++) {
         if (inputWindow->c_font[i]) {
-            cairo_destroy(inputWindow->c_font[i]);
+            _CAIRO_DESTROY(inputWindow->c_font[i]);
             inputWindow->c_font[i] = NULL;
         }
     }
     inputWindow->c_font[7] = NULL;
     if (inputWindow->c_cursor) {
-        cairo_destroy(inputWindow->c_cursor);
+        _CAIRO_DESTROY(inputWindow->c_cursor);
         inputWindow->c_cursor = NULL;
     }
 
@@ -575,7 +560,10 @@ void LoadInputMessage(FcitxSkin* sc, InputWindow* inputWindow, const char* font)
         inputWindow->c_font[i] = cairo_create(inputWindow->cs_input_bar);
         fcitx_cairo_set_color(inputWindow->c_font[i], &sc->skinFont.fontColor[i]);
 #ifndef _ENABLE_PANGO
-        SetFontContext(inputWindow->c_font[i], font, sc->skinFont.fontSize);
+        SetFontContext(inputWindow->c_font[i],
+                       font,
+                       inputWindow->owner->fontSize > 0 ? inputWindow->owner->fontSize : sc->skinFont.fontSize,
+                       dpi);
 #endif
     }
     inputWindow->c_font[7] = inputWindow->c_font[0];
@@ -639,10 +627,13 @@ void DrawInputBar(FcitxSkin* sc, InputWindow* inputWindow, int iCursorPos, Fcitx
         return;
 
     inputWidth = 0;
+    int dpi = sc->skinFont.respectDPI? classicui->dpi : 0;
+    FCITX_UNUSED(dpi);
 #ifdef _ENABLE_PANGO /* special case which only macro unable to handle */
-    SetFontContext(dummy, inputWindow->owner->font, sc->skinFont.fontSize);
+    SetFontContext(dummy, inputWindow->owner->font, inputWindow->owner->fontSize > 0 ? inputWindow->owner->fontSize : sc->skinFont.fontSize, dpi);
 #endif
 
+    int fontHeight = FontHeightWithContext(inputWindow->c_font[0], dpi);
     for (i = 0; i < FcitxMessagesGetMessageCount(msgup) ; i++) {
         char *trans = FcitxInstanceProcessOutputFilter(instance, FcitxMessagesGetMessageString(msgup, i));
         if (trans)
@@ -651,9 +642,12 @@ void DrawInputBar(FcitxSkin* sc, InputWindow* inputWindow, int iCursorPos, Fcitx
             strUp[i] = FcitxMessagesGetMessageString(msgup, i);
         posUpX[i] = sc->skinInputBar.marginLeft + inputWidth;
 
-        StringSizeWithContext(inputWindow->c_font[FcitxMessagesGetMessageType(msgup, i)], strUp[i], &strWidth, &strHeight);
+        StringSizeWithContext(inputWindow->c_font[FcitxMessagesGetMessageType(msgup, i)], dpi, strUp[i], &strWidth, &strHeight);
 
-        posUpY[i] = sc->skinInputBar.marginTop + sc->skinInputBar.iInputPos - strHeight;
+        if (sc->skinFont.respectDPI)
+            posUpY[i] = sc->skinInputBar.marginTop + sc->skinInputBar.iInputPos;
+        else
+            posUpY[i] = sc->skinInputBar.marginTop + sc->skinInputBar.iInputPos - strHeight;
         inputWidth += strWidth;
         if (FcitxInputStateGetShowCursor(input)) {
             int length = strlen(FcitxMessagesGetMessageString(msgup, i));
@@ -664,7 +658,7 @@ void DrawInputBar(FcitxSkin* sc, InputWindow* inputWindow, int iCursorPos, Fcitx
                     strncpy(strTemp, strUp[i], iChar);
                     strTemp[iChar] = '\0';
                     strGBKT = strTemp;
-                    StringSizeWithContext(inputWindow->c_font[FcitxMessagesGetMessageType(msgup, i)], strGBKT, &strWidth, &strHeight);
+                    StringSizeWithContext(inputWindow->c_font[FcitxMessagesGetMessageType(msgup, i)], dpi, strGBKT, &strWidth, &strHeight);
                     cursor_pos = posUpX[i]
                                  + strWidth + 2;
                 }
@@ -680,6 +674,11 @@ void DrawInputBar(FcitxSkin* sc, InputWindow* inputWindow, int iCursorPos, Fcitx
     outputWidth = 0;
     outputHeight = 0;
     int currentX = 0;
+    int offsetY;
+    if (sc->skinFont.respectDPI)
+        offsetY = sc->skinInputBar.marginTop + sc->skinInputBar.iInputPos + fontHeight + (FcitxMessagesGetMessageCount(msgdown) ? sc->skinInputBar.iOutputPos : 0);
+    else
+        offsetY = sc->skinInputBar.marginTop + sc->skinInputBar.iOutputPos - fontHeight;
     for (i = 0; i < FcitxMessagesGetMessageCount(msgdown) ; i++) {
         char *trans = FcitxInstanceProcessOutputFilter(instance, FcitxMessagesGetMessageString(msgdown, i));
         if (trans)
@@ -692,25 +691,25 @@ void DrawInputBar(FcitxSkin* sc, InputWindow* inputWindow, int iCursorPos, Fcitx
                 if (currentX > outputWidth)
                     outputWidth = currentX;
                 if (i != 0) {
-                    outputHeight += sc->skinFont.fontSize + 2;
                     currentX = 0;
                 }
             }
             posDownX[i] = sc->skinInputBar.marginLeft + currentX;
-            StringSizeWithContext(inputWindow->c_font[FcitxMessagesGetMessageType(msgdown, i)], strDown[i], &strWidth, &strHeight);
+            StringSizeWithContext(inputWindow->c_font[FcitxMessagesGetMessageType(msgdown, i)], dpi, strDown[i], &strWidth, &strHeight);
+            if (FcitxMessagesGetMessageType(msgdown, i) == MSG_INDEX && i != 0)
+                outputHeight += fontHeight + 2;
             currentX += strWidth;
-            posDownY[i] =  sc->skinInputBar.marginTop + sc->skinInputBar.iOutputPos + outputHeight - strHeight;
         } else { /* horizontal */
             posDownX[i] = sc->skinInputBar.marginLeft + outputWidth;
-            StringSizeWithContext(inputWindow->c_font[FcitxMessagesGetMessageType(msgdown, i)], strDown[i], &strWidth, &strHeight);
-            posDownY[i] = sc->skinInputBar.marginTop + sc->skinInputBar.iOutputPos - strHeight;
+            StringSizeWithContext(inputWindow->c_font[FcitxMessagesGetMessageType(msgdown, i)], dpi, strDown[i], &strWidth, &strHeight);
             outputWidth += strWidth;
         }
+        posDownY[i] = offsetY + outputHeight;
     }
     if (inputWindow->owner->bVerticalList && currentX > outputWidth)
         outputWidth = currentX;
 
-    newHeight = sc->skinInputBar.marginTop + sc->skinInputBar.iOutputPos + outputHeight + sc->skinInputBar.marginBottom;
+    newHeight = offsetY + outputHeight + sc->skinInputBar.marginBottom + (FcitxMessagesGetMessageCount(msgdown) || !sc->skinFont.respectDPI ? fontHeight : 0);
 
     newWidth = (inputWidth < outputWidth) ? outputWidth : inputWidth;
     newWidth += sc->skinInputBar.marginLeft + sc->skinInputBar.marginRight;
@@ -742,7 +741,7 @@ void DrawInputBar(FcitxSkin* sc, InputWindow* inputWindow, int iCursorPos, Fcitx
                                 sc->skinInputBar.fillV,
                                 sc->skinInputBar.fillH
                                );
-        cairo_destroy(c);
+        _CAIRO_DESTROY(c);
     }
 
     c = cairo_create(inputWindow->cs_input_bar);
@@ -780,27 +779,37 @@ void DrawInputBar(FcitxSkin* sc, InputWindow* inputWindow, int iCursorPos, Fcitx
     }
 
     for (i = 0; i < FcitxMessagesGetMessageCount(msgup) ; i++) {
-        OutputStringWithContext(inputWindow->c_font[FcitxMessagesGetMessageType(msgup, i)], strUp[i], posUpX[i], posUpY[i]);
+        OutputStringWithContext(inputWindow->c_font[FcitxMessagesGetMessageType(msgup, i)], dpi, strUp[i], posUpX[i], posUpY[i]);
         if (strUp[i] != FcitxMessagesGetMessageString(msgup, i))
             free(strUp[i]);
     }
 
     for (i = 0; i < FcitxMessagesGetMessageCount(msgdown) ; i++) {
-        OutputStringWithContext(inputWindow->c_font[FcitxMessagesGetMessageType(msgdown, i)], strDown[i], posDownX[i], posDownY[i]);
+        OutputStringWithContext(inputWindow->c_font[FcitxMessagesGetMessageType(msgdown, i)], dpi, strDown[i], posDownX[i], posDownY[i]);
         if (strDown[i] != FcitxMessagesGetMessageString(msgdown, i))
             free(strDown[i]);
     }
 
-    ResetFontContext();
+    int cursorY1, cursorY2;
+    if (sc->skinFont.respectDPI) {
+        cursorY1 = sc->skinInputBar.marginTop + sc->skinInputBar.iInputPos;
+        cursorY2 = sc->skinInputBar.marginTop + sc->skinInputBar.iInputPos + fontHeight;
+    }
+    else {
+        cursorY1 = sc->skinInputBar.marginTop + sc->skinInputBar.iInputPos - fontHeight - 4;
+        cursorY2 = sc->skinInputBar.marginTop + sc->skinInputBar.iInputPos;
+    }
 
     //画光标
     if (FcitxInputStateGetShowCursor(input)) {
-        cairo_move_to(inputWindow->c_cursor, cursor_pos, sc->skinInputBar.marginTop + sc->skinInputBar.iInputPos);
-        cairo_line_to(inputWindow->c_cursor, cursor_pos, sc->skinInputBar.marginTop + sc->skinInputBar.iInputPos - FontHeightWithContext(inputWindow->c_font[0]) - 4);
+        cairo_move_to(inputWindow->c_cursor, cursor_pos, cursorY1);
+        cairo_line_to(inputWindow->c_cursor, cursor_pos, cursorY2);
         cairo_stroke(inputWindow->c_cursor);
     }
 
-    cairo_destroy(c);
+    ResetFontContext();
+
+    _CAIRO_DESTROY(c);
     FcitxMessagesSetMessageChanged(msgup, false);
     FcitxMessagesSetMessageChanged(msgdown, false);
 }
@@ -983,15 +992,35 @@ void ParsePlacement(UT_array* sps, char* placment)
     utarray_free(array);
 }
 
-SkinImage* GetIMIcon(FcitxInstance* instance, FcitxSkin *sc, const char* fallbackIcon, int flag, boolean fallbackToDefault)
+SkinImage* GetIMIcon(FcitxClassicUI* classicui, FcitxSkin *sc, const char* fallbackIcon, int flag, boolean fallbackToDefault)
 {
-    FcitxIM* im = FcitxInstanceGetCurrentIM(instance);
+    FcitxIM* im = FcitxInstanceGetCurrentIM(classicui->owner);
     char* path;
     if (im->strIconName[0] == '/')
         path = strdup(im->strIconName);
     else
         asprintf(&path, "%s.png", im->strIconName);
-    SkinImage* imicon = LoadImage(sc, path, flag);
+
+    SkinImage* imicon = NULL;
+
+    if (strncmp(im->uniqueName, "fcitx-keyboard-", strlen("fcitx-keyboard-")) == 0) {
+        SkinImage* activeIcon = LoadImage(sc, fallbackIcon, fallbackToDefault);
+        char temp[LANGCODE_LENGTH + 1] = { '\0', };
+        char* iconText = 0;
+        if (*im->langCode) {
+            strncpy(temp, im->langCode, LANGCODE_LENGTH);
+            iconText = temp;
+            iconText[0] = toupper(iconText[0]);
+        }
+        else
+            iconText = im->uniqueName + strlen("fcitx-keyboard-");
+        imicon = LoadImageWithText(classicui, sc, path, iconText,
+                    cairo_image_surface_get_width(activeIcon->image),
+                    cairo_image_surface_get_height(activeIcon->image), true);
+    }
+
+    if (imicon == NULL)
+        imicon = LoadImage(sc, path, flag);
     if (imicon == NULL)
         imicon = LoadImage(sc, fallbackIcon, fallbackToDefault);
     else {
