@@ -81,6 +81,9 @@ const char * kimpanel_introspection_xml =
     "      <arg name=\"label\" direction=\"in\" type=\"a\"/>\n"
     "      <arg name=\"text\" direction=\"in\" type=\"a\"/>\n"
     "    </signal>\n"
+    "    <signal name=\"UpdateLookupTableCursor\">\n"
+    "      <arg name=\"cursor\" direction=\"in\" type=\"i\"/>\n"
+    "    </signal>\n"
     "    <signal name=\"UpdatePreeditCaret\">\n"
     "      <arg name=\"position\" direction=\"in\" type=\"i\"/>\n"
     "    </signal>\n"
@@ -165,8 +168,7 @@ static void KimpanelIntrospectCallback(DBusPendingCall *call, void *data);
 
 #define KIMPANEL_BUFFER_SIZE 4096
 
-FCITX_EXPORT_API
-FcitxUI ui = {
+FCITX_DEFINE_PLUGIN(fcitx_kimpanel_ui, ui, FcitxUI) = {
     KimpanelCreate,
     KimpanelCloseInputWindow,
     KimpanelShowInputWindow,
@@ -189,19 +191,80 @@ FcitxUI ui = {
     NULL
 };
 
-FCITX_EXPORT_API
-int ABI_VERSION = FCITX_ABI_VERSION;
+
+static void SetIMMenu(FcitxIM* pim, char** prop)
+{
+    const char *icon = "";
+    if (strncmp(pim->uniqueName, "fcitx-keyboard-", strlen("fcitx-keyboard-")) != 0)
+        icon = pim->strIconName;
+
+    if (icon[0] == '\0' || icon[0] == '/')
+        asprintf(prop, "/Fcitx/im/%s:%s:%s:%s", pim->uniqueName, pim->strName, icon, pim->strName);
+    else
+        asprintf(prop, "/Fcitx/im/%s:%s:fcitx-%s:%s", pim->uniqueName, pim->strName, icon, pim->strName);
+}
+
+static void SetIMIcon(FcitxInstance* instance, char** prop)
+{
+    char* icon;
+    char* imname;
+    char* description;
+    char temp[LANGCODE_LENGTH + 1] = { '\0', };
+    FCITX_UNUSED(temp);
+    FcitxInputContext* ic = FcitxInstanceGetCurrentIC(instance);
+    if (ic == NULL) {
+        icon = "kbd";
+        imname = _("No input window");
+        description = _("No input window");
+    }
+    else if (FcitxInstanceGetCurrentStatev2(instance) == IS_ACTIVE) {
+        FcitxIM* im = FcitxInstanceGetCurrentIM(instance);
+        if (im) {
+#if 0
+            if (strncmp(im->uniqueName, "fcitx-keyboard-", strlen("fcitx-keyboard-")) == 0) {
+                icon = "";
+                if (*im->langCode) {
+                    strncpy(temp, im->langCode, LANGCODE_LENGTH);
+                    imname = temp;
+                    imname[0] = toupper(imname[0]);
+                }
+                else {
+                    imname = im->uniqueName + strlen("fcitx-keyboard");
+                }
+            }
+            else
+#endif
+            {
+                icon = im->strIconName;
+                imname = im->strName;
+            }
+            description = im->strName;
+        } else {
+            icon = "kbd";
+            imname = _("Disabled");
+            description = _("Input Method Disabled");
+        }
+    } else {
+        icon = "kbd";
+        imname = _("Disabled");
+        description = _("Input Method Disabled");
+    }
+    /* add fcitx- prefix, unless icon name is an absolute path */
+    if (icon[0] == '\0' || icon[0] == '/')
+        asprintf(prop, "/Fcitx/im:%s:%s:%s", imname, icon, description);
+    else
+        asprintf(prop, "/Fcitx/im:%s:fcitx-%s:%s", imname, icon, description);
+}
 
 void* KimpanelCreate(FcitxInstance* instance)
 {
     FcitxKimpanelUI *kimpanel = fcitx_utils_malloc0(sizeof(FcitxKimpanelUI));
-    FcitxModuleFunctionArg arg;
 
     kimpanel->lastCursor = -2;
     kimpanel->version = 1;
     kimpanel->iCursorPos = 0;
     kimpanel->owner = instance;
-    kimpanel->conn = InvokeFunction(instance, FCITX_DBUS, GETCONNECTION, arg);
+    kimpanel->conn = InvokeVaArgs(instance, FCITX_DBUS, GETCONNECTION);
 
     DBusError err;
     dbus_error_init(&err);
@@ -294,30 +357,7 @@ void KimpanelRegisterAllStatus(FcitxKimpanelUI* kimpanel)
     char **prop = fcitx_utils_malloc0(sizeof(char*) * (2 + utarray_len(uistats) + utarray_len(uicompstats)));
     asprintf(&prop[0], "/Fcitx/logo:%s:%s:%s", _("Fcitx"), "fcitx", _("Fcitx"));
 
-    char* icon;
-    char* imname;
-    char* description;
-    if (FcitxInstanceGetCurrentStatev2(instance) == IS_ACTIVE) {
-        FcitxIM* im = FcitxInstanceGetCurrentIM(instance);
-        if (im) {
-            icon = im->strIconName;
-            imname = im->strName;
-            description = im->strName;
-        } else {
-            icon = "kbd";
-            imname = _("Disabled");
-            description = _("Input Method Disabled");
-        }
-    } else {
-        icon = "kbd";
-        imname = _("Disabled");
-        description = _("Input Method Disabled");
-    }
-    /* add fcitx- prefix, unless icon name is an absolute path */
-    if (icon[0] == '/')
-        asprintf(&prop[1], "/Fcitx/im:%s:%s:%s", imname, icon, description);
-    else
-        asprintf(&prop[1], "/Fcitx/im:%s:fcitx-%s:%s", imname, icon, description);
+    SetIMIcon(instance, &prop[1]);
 
     int count = 2;
 
@@ -353,37 +393,8 @@ void KimpanelSetIMStatus(FcitxKimpanelUI* kimpanel)
 {
     FcitxInstance* instance = kimpanel->owner;
     char* status = NULL;
-    char* icon;
-    char* imname;
-    char* description;
-    FcitxInputContext* ic = FcitxInstanceGetCurrentIC(instance);
-    if (ic == NULL) {
-        icon = "kbd";
-        imname = _("No input window");
-        description = _("No input window");
-    }
-    else if (FcitxInstanceGetCurrentStatev2(instance) == IS_ACTIVE) {
-        FcitxIM* im = FcitxInstanceGetCurrentIM(instance);
-        if (im) {
-            icon = im->strIconName;
-            imname = im->strName;
-            description = im->strName;
-        } else {
-            icon = "kbd";
-            imname = _("Disabled");
-            description = _("Input Method Disabled");
-        }
-    } else {
-        icon = "kbd";
-        imname = _("Disabled");
-        description = _("Input Method Disabled");
-    }
-    /* add fcitx- prefix */
-    if (icon[0] == '/')
-        asprintf(&status, "/Fcitx/im:%s:%s:%s", imname, icon, description);
-    else
-        asprintf(&status, "/Fcitx/im:%s:fcitx-%s:%s", imname, icon, description);
 
+    SetIMIcon(instance, &status);
     KimUpdateProperty(kimpanel, status);
     free(status);
 }
@@ -726,9 +737,33 @@ DBusHandlerResult KimpanelDBusFilter(DBusConnection* connection, DBusMessage* ms
             if (strlen(s0) > len) {
                 s0 += len;
                 if (strcmp("logo", s0) == 0) {
-                    FcitxInstanceChangeIMState(instance, FcitxInstanceGetCurrentIC(instance));
+                    size_t len = 3;
+                    char **prop = fcitx_utils_malloc0(len * sizeof(char*));
+                    asprintf(&prop[0], "/Fcitx/logo/toggle:%s::%s", _("Toggle Input Method"), _("Toggle Input Method"));
+                    asprintf(&prop[1], "/Fcitx/logo/configureim:%s:configure:%s", _("Configure Current Input Method"), _("Configure Current Input Method"));
+                    asprintf(&prop[2], "/Fcitx/logo/restart:%s:view-refresh:%s",_("Restart"), _("Restart"));
+                    KimExecMenu(kimpanel, prop, len );
+                    while (len --)
+                        free(prop[len]);
+                    free(prop);
+                } else if (strncmp("logo/", s0, strlen("logo/")) == 0) {
+                    s0 += strlen("logo/");
+                    if (strcmp(s0, "toggle") == 0)
+                        FcitxInstanceChangeIMState(instance, FcitxInstanceGetCurrentIC(instance));
+                    else if (strcmp(s0, "configureim") == 0) {
+                        FcitxIM* im = FcitxInstanceGetCurrentIM(kimpanel->owner);
+                        if (im && im->owner) {
+                            fcitx_utils_launch_configure_tool_for_addon(im->uniqueName);
+                        }
+                        else
+                            fcitx_utils_launch_configure_tool();
+                    }
+                    else if (strcmp(s0, "restart") == 0) {
+                        fcitx_utils_launch_restart();
+                    }
                 } else if (strcmp("keyboard", s0) == 0) {
-                    FcitxInstanceCloseIM(instance, FcitxInstanceGetCurrentIC(instance));
+                    FcitxInstanceCloseIM(instance,
+                                         FcitxInstanceGetCurrentIC(instance));
                 } else if (strncmp("im/", s0, strlen("im/")) == 0) {
                     s0 += strlen("im/");
                     FcitxInstanceSwitchIMByName(instance, s0);
@@ -742,10 +777,7 @@ DBusHandlerResult KimpanelDBusFilter(DBusConnection* connection, DBusMessage* ms
                             pim != NULL;
                             pim = (FcitxIM *) utarray_next(imes, pim)) {
 
-                        if (pim->strIconName[0] == '/')
-                            asprintf(&prop[index], "/Fcitx/im/%s:%s:%s:%s", pim->uniqueName, pim->strName, pim->strIconName, pim->strName);
-                        else
-                            asprintf(&prop[index], "/Fcitx/im/%s:%s:fcitx-%s:%s", pim->uniqueName, pim->strName, pim->strIconName, pim->strName);
+                        SetIMMenu(pim, &prop[index]);
                         index ++;
                     }
                     KimExecMenu(kimpanel, prop, len);
@@ -812,6 +844,10 @@ DBusHandlerResult KimpanelDBusFilter(DBusConnection* connection, DBusMessage* ms
         FcitxLog(DEBUG, "ReloadConfig");
         FcitxInstanceReloadConfig(instance);
         return DBUS_HANDLER_RESULT_HANDLED;
+    } else if (dbus_message_is_signal(msg, "org.kde.impanel", "Restart")) {
+        FcitxLog(DEBUG, "Restart");
+        fcitx_utils_launch_restart();
+        return DBUS_HANDLER_RESULT_HANDLED;
     } else if (dbus_message_is_signal(msg, "org.kde.impanel", "Configure")) {
         FcitxLog(DEBUG, "Configure");
         fcitx_utils_launch_configure_tool();
@@ -888,8 +924,11 @@ void KimExecMenu(FcitxKimpanelUI* kimpanel, char *props[], int n)
     if (n == -1) {
         n = 0;
         while (*(props[n]) != 0) {
+            if (!fcitx_utf8_check_string(props[n]))
+                return;
             n++;
         }
+
     }
 
     int i;
@@ -933,6 +972,8 @@ void KimRegisterProperties(FcitxKimpanelUI* kimpanel, char *props[], int n)
     if (n == -1) {
         n = 0;
         while (*(props[n]) != 0) {
+            if (!fcitx_utf8_check_string(props[n]))
+                return;
             n++;
         }
     }
@@ -975,6 +1016,9 @@ void KimUpdateProperty(FcitxKimpanelUI* kimpanel, char *prop)
         return;
     }
 
+    if (!fcitx_utf8_check_string(prop))
+        return;
+
     // append arguments onto signal
     dbus_message_iter_init_append(msg, &args);
     if (!dbus_message_iter_append_basic(&args, DBUS_TYPE_STRING, &prop)) {
@@ -1007,6 +1051,9 @@ void KimRemoveProperty(FcitxKimpanelUI* kimpanel, char *prop)
         return;
     }
 
+    if (!fcitx_utf8_check_string(prop))
+        return;
+
     // append arguments onto signal
     dbus_message_iter_init_append(msg, &args);
     if (!dbus_message_iter_append_basic(&args, DBUS_TYPE_STRING, &prop)) {
@@ -1186,7 +1233,7 @@ void KimShowLookupTable(FcitxKimpanelUI* kimpanel, boolean toShow)
 
 void KimUpdateLookupTable(FcitxKimpanelUI* kimpanel, char *labels[], int nLabel, char *texts[], int nText, boolean has_prev, boolean has_next)
 {
-
+    int i;
     dbus_uint32_t serial = 0; // unique number to associate replies with requests
     DBusMessage* msg;
     DBusMessageIter args;
@@ -1199,8 +1246,14 @@ void KimUpdateLookupTable(FcitxKimpanelUI* kimpanel, char *labels[], int nLabel,
         FcitxLog(DEBUG, "Message Null");
         return;
     }
-
-    int i;
+    for (i = 0; i < nLabel; i++) {
+        if (!fcitx_utf8_check_string(labels[i]))
+            return;
+    }
+    for (i = 0; i < nText; i++) {
+        if (!fcitx_utf8_check_string(texts[i]))
+            return;
+    }
     DBusMessageIter subLabel;
     DBusMessageIter subText;
     DBusMessageIter subAttrs;
@@ -1327,6 +1380,8 @@ void KimUpdateAux(FcitxKimpanelUI* kimpanel, char *text)
         FcitxLog(DEBUG, "Message Null");
         return;
     }
+    if (!fcitx_utf8_check_string(text))
+        return;
 
     char *attr = "";
 
@@ -1574,11 +1629,9 @@ void KimpanelDestroy(void* arg)
     dbus_connection_unregister_object_path(kimpanel->conn, FCITX_KIMPANEL_PATH);
     dbus_connection_remove_filter(kimpanel->conn, KimpanelDBusFilter, kimpanel);
 
-    DBusError err;
-    dbus_error_init(&err);
     dbus_bus_remove_match(kimpanel->conn,
                           "type='signal',interface='org.kde.impanel'",
-                          &err);
+                          NULL);
     dbus_connection_flush(kimpanel->conn);
 
     dbus_bus_remove_match(kimpanel->conn,
@@ -1586,10 +1639,9 @@ void KimpanelDestroy(void* arg)
                         "interface='" DBUS_INTERFACE_DBUS "',"
                         "path='" DBUS_PATH_DBUS "',"
                         "member='NameOwnerChanged'",
-                        &err);
+                        NULL);
 
     dbus_connection_flush(kimpanel->conn);
-    dbus_error_free(&err);
 
     free(kimpanel->messageUp);
     free(kimpanel->messageDown);
